// Copyright (c) Microsoft Corporation. All rights reserved.

#include "CrashAnalysis.h"

#include <iomanip>
#include <iostream>
#include <memory>
#include <string>
#include <vector>

#include <azure-device-util/Bytes.h>
#include <azure-device-util/FileSystem.h>
#include <azure-device-util/Hash.h>
#include <azure-device-util/Math.h>
#include <azure-device-util/PackageManager.h>
#include <azure-device-util/ScopeExit.h>
#include <azure-device-util/StreamCopier.h>
#include <azure-device-util/String.h>
#include <nonstd/span.hpp>

#include "CrashParameters.h"
#include "ElfArchitectureSpecifics.h"
#include "ElfNoteParsing.h"

namespace Utils = Azure::Device::Utils;

namespace Azure::Device::CrashHandler {

std::ostream& operator<<(std::ostream& out, const ElfHeader& elfHeader)
{
    out << "[Elf Module Header]" << std::endl;
    out << "ProcArch: 0x" << std::hex << elfHeader.e_machine << " "
        << ((elfHeader.e_ident[EI_CLASS] == ELFCLASS32) ? "32-bit" : "64-bit")
        << std::endl;
    out << std::dec << "Program Header offset: " << elfHeader.e_phoff
        << std::endl;
    out << std::dec << "Program Header size: " << elfHeader.e_phentsize
        << std::endl;
    out << std::dec << "Program Header count: " << elfHeader.e_phnum << std::endl;

    return out;
}

static bool
MappedRegionContainsAddress(const MappedRegion& mappedRegion, size_t address)
{
    return mappedRegion.LowAddress <= address &&
        mappedRegion.HighAddress > address;
}

static bool
ProgramHeaderContainsAddress(const ElfProgramHeader& programHeader, size_t address)
{
    return (programHeader.p_vaddr <= address) &&
        (programHeader.p_vaddr + programHeader.p_memsz > address);
}

static CrashAnalysis
CreateAnalysisFromCrashParameters(const CrashParameters& params)
{
    CrashAnalysis analysis;

    if (size_t lastSlash = params.FullPath.find_last_of('/');
        lastSlash != std::string::npos)
    {
        analysis.AppName = params.FullPath.substr(lastSlash + 1);
    }

    if (auto package = Utils::PackageManager::GetFileOwner(params.FullPath))
    {
        analysis.PackageName = package->Name;
        analysis.PackageVersion = package->Version;
        analysis.PackageArch = package->Arch;
    }

    analysis.ExceptionCode = params.Signal;

    return analysis;
}

bool ValidateElfHeader(const ElfHeader& elfHeader, std::ostream& logStream)
{
    if (elfHeader.e_ident[EI_MAG0] != ELFMAG0 ||
        elfHeader.e_ident[EI_MAG1] != ELFMAG1 ||
        elfHeader.e_ident[EI_MAG2] != ELFMAG2 ||
        elfHeader.e_ident[EI_MAG3] != ELFMAG3)
    {
        logStream << "FAILURE: ElfHeader - Incorrect EI_MAG!" << std::endl;
        return false;
    }

    if (elfHeader.e_ident[EI_CLASS] != EXPECTED_ELFCLASS)
    {
        logStream << "FAILURE: ElfHeader - Incorrect EI_CLASS!" << std::endl;
        return false;
    }

    if (elfHeader.e_machine != EXPECTED_E_MACHINE)
    {
        logStream << "FAILURE: ElfHeader - Incorrect E_MACHINE!" << std::endl;
        return false;
    }

    if (elfHeader.e_type != ET_CORE)
    {
        logStream << "FAILURE: ElfHeader - Incorrect E_TYPE!" << std::endl;
        return false;
    }

    if (elfHeader.e_ehsize != sizeof(ElfHeader))
    {
        logStream << "FAILURE: ElfHeader - Incorrect ElfHeader size!"
                  << std::endl;
        return false;
    }

    if (elfHeader.e_phentsize != sizeof(ElfProgramHeader))
    {
        logStream << "FAILURE: ElfHeader - Incorrect ElfProgramHeader size!"
                  << std::endl;
        return false;
    }

    return true;
}

CrashAnalysis CopyAndAnalyzeCrash(
    const CrashParameters& params,
    Utils::StreamCopier& coreStreamCopier,
    std::ostream& logStream)
{
    CrashAnalysis analysis = CreateAnalysisFromCrashParameters(params);

    // Populate module header
    ElfHeader elfHeader;
    coreStreamCopier.CopyNextBytes(Utils::Bytes::ValueToBytes(elfHeader));
    logStream << elfHeader << std::endl;

    if (!ValidateElfHeader(elfHeader, logStream))
    {
        logStream << "FAILURE: Invalid ELF header!" << std::endl;
        return analysis;
    }

    analysis.Is32Bits = elfHeader.e_ident[EI_CLASS] == ELFCLASS32;

    // Populate program headers
    std::vector<ElfProgramHeader> programHeaders{ elfHeader.e_phnum };
    coreStreamCopier.CopyToSpecifiedOffset(elfHeader.e_phoff);
    coreStreamCopier.CopyNextBytes(
        Utils::Bytes::ContainerToWriteableBytes(programHeaders));

    auto noteProgramHeaderItr = std::find_if(
        programHeaders.begin(),
        programHeaders.end(),
        [](const ElfProgramHeader& programHeader) {
            return programHeader.p_type == PT_NOTE;
        });
    if (noteProgramHeaderItr == programHeaders.end())
    {
        logStream << "FAILURE: No note program header found!" << std::endl;
        return analysis;
    }

    logStream
        << "Note program header found: Offset=" << noteProgramHeaderItr->p_offset
        << " Size=" << noteProgramHeaderItr->p_filesz << std::endl;

    // Read PT_NOTE program header and parse its contents
    std::vector<std::byte> notesBytes{ noteProgramHeaderItr->p_filesz };
    coreStreamCopier.CopyToSpecifiedOffset(noteProgramHeaderItr->p_offset);
    coreStreamCopier.CopyNextBytes(notesBytes);

    ElfNotesContents notesContents = ParseElfNotes(notesBytes, logStream);

    // Find the thread state corresponding to the thread blamed in params
    auto blameThreadStateItr = std::find_if(
        notesContents.ThreadStates.begin(),
        notesContents.ThreadStates.end(),
        [&params](const elf_prstatus& threadState) {
            return threadState.pr_pid == params.ThreadId;
        });
    if (blameThreadStateItr == notesContents.ThreadStates.end())
    {
        logStream << "FAILURE: No thread state found for blamed thread!"
                  << std::endl;
        return analysis;
    }

// Get the architecture specific program counter register on the blame thread
#if defined(__arm__) && defined(ANDROID)
    auto userRegs =
        reinterpret_cast<const user_regs*>(&blameThreadStateItr->pr_reg);
#else
    auto userRegs =
        reinterpret_cast<const user_regs_struct*>(&blameThreadStateItr->pr_reg);
#endif

    size_t programCounter = GetProgramCounter(*userRegs);
    if (programCounter == 0)
    {
        logStream
            << "FAILURE: Blamed thread did not have a valid program counter!"
            << std::endl;
        return analysis;
    }

    // Look for the program header that contains the program counter's address
    auto pcProgramHeaderItr = std::find_if(
        programHeaders.begin(),
        programHeaders.end(),
        [programCounter](const ElfProgramHeader& programHeader) {
            return ProgramHeaderContainsAddress(programHeader, programCounter);
        });

    if (pcProgramHeaderItr == programHeaders.end())
    {
        logStream
            << "FAILURE [Non-fatal]: Could not find program header containing program counter address!"
            << std::endl;
    }
    else
    {
        analysis.ModuleOffset = programCounter - pcProgramHeaderItr->p_vaddr;
    }

    // Look for the MappedRegion that contains the program counter's address
    auto pcMappedRegionItr = std::find_if(
        notesContents.MappedRegions.begin(),
        notesContents.MappedRegions.end(),
        [programCounter](const MappedRegion& mappedRegion) {
            return MappedRegionContainsAddress(mappedRegion, programCounter);
        });

    if (pcMappedRegionItr == notesContents.MappedRegions.end())
    {
        logStream
            << "FAILURE [Non-fatal]: Could not find mapped region containing program counter address!"
            << std::endl;
    }
    else
    {
        std::filesystem::path modulePath{ pcMappedRegionItr->ModulePath };
        analysis.ModuleName = modulePath.filename().string();

        // TODO: Populate analysis.ModuleVerison based on SONAME of module
    }

    return analysis;
}
}
