// Copyright (c) Microsoft Corporation. All rights reserved.

#pragma once

#include <algorithm>
#include <cstring>
#include <elf.h>
#include <fstream>
#include <iostream>
#include <map>
#include <sstream>
#include <string>
#include <vector>

#include <sys/procfs.h>
#include <sys/user.h>

//
// Below is needed because the android version of sys/prcfs.h does not incude
// elf_prstatus. See
// https://cs.android.com/android/platform/superproject/+/master:bionic/libc/include/sys/procfs.h;l=1?q=sys%2Fprocf&ss=android%2Fplatform%2Fsuperproject.
// This is copied from
// https://cs.android.com/android/platform/superproject/+/master:prebuilts/gcc/linux-x86/host/x86_64-linux-glibc2.17-4.8/sysroot/usr/include/linux/elfcore.h?q=elf_prstatus&ss=android%2Fplatform%2Fsuperproject&start=1.
//
#ifdef ANDROID
/*
 * Definitions to generate Intel SVR4-like core files.
 * These mostly have the same names as the SVR4 types with "elf_"
 * tacked on the front to prevent clashes with linux definitions,
 * and the typedef forms have been avoided.  This is mostly like
 * the SVR4 structure, but more Linuxy, with things that Linux does
 * not support and which gdb doesn't really use excluded.
 * Fields present but not used are marked with "XXX".
 */
struct elf_prstatus
{
    #if 0
	long	pr_flags;	/* XXX Process flags */
	short	pr_why;		/* XXX Reason for process halt */
	short	pr_what;	/* XXX More detailed reason */
    #endif
    struct elf_siginfo pr_info; /* Info associated with signal */
    short pr_cursig;            /* Current signal */
    unsigned long pr_sigpend;   /* Set of pending signals */
    unsigned long pr_sighold;   /* Set of held signals */
    #if 0
	struct sigaltstack pr_altstack;	/* Alternate stack info */
	struct sigaction pr_action;	/* Signal action for current sig */
    #endif
    pid_t pr_pid;
    pid_t pr_ppid;
    pid_t pr_pgrp;
    pid_t pr_sid;
    struct timeval pr_utime;  /* User time */
    struct timeval pr_stime;  /* System time */
    struct timeval pr_cutime; /* Cumulative user time */
    struct timeval pr_cstime; /* Cumulative system time */
    #if 0
	long	pr_instr;		/* Current instruction */
    #endif
    elf_gregset_t pr_reg; /* GP registers */
    #ifdef CONFIG_BINFMT_ELF_FDPIC
    /* When using FDPIC, the loadmap addresses need to be communicated
     * to GDB in order for GDB to do the necessary relocations.  The
     * fields (below) used to communicate this information are placed
     * immediately after ``pr_reg'', so that the loadmap addresses may
     * be viewed as part of the register set if so desired.
     */
    unsigned long pr_exec_fdpic_loadmap;
    unsigned long pr_interp_fdpic_loadmap;
    #endif
    int pr_fpvalid; /* True if math co-processor being used.  */
};
#endif

namespace Azure::Device::CrashHandler {

#ifdef __x86_64__
    #define EXPECTED_E_MACHINE EM_X86_64
    #define ELF_BITNESS 64

inline size_t GetProgramCounter(const user_regs_struct& userRegs)
{
    return userRegs.rip;
}

#elif __i386__
    #define EXPECTED_E_MACHINE EM_386
    #define ELF_BITNESS 32

inline size_t GetProgramCounter(const user_regs_struct& userRegs)
{
    return userRegs.eip;
}

#elif __aarch64__
    #define EXPECTED_E_MACHINE EM_AARCH64
    #define ELF_BITNESS 64

inline size_t GetProgramCounter(const user_regs_struct& userRegs)
{
    return userRegs.pc;
}

#elif __arm__
    #define EXPECTED_E_MACHINE EM_ARM
    #define ELF_BITNESS 32

    //
    // Below is needed because Android does not have user_regs_struct for arm,
    // but instead they have user_regs. See
    // https://cs.android.com/android/platform/superproject/+/master:bionic/libc/include/sys/user.h?q=user.h
    //
    #ifdef ANDROID
inline size_t GetProgramCounter(const user_regs& userRegs)
{
    return userRegs.uregs[15];
}
    #else
inline size_t GetProgramCounter(const user_regs_struct& userRegs)
{
    return userRegs.uregs[15];
}
    #endif
#else
    #error "Unsupported architecture"
#endif

#define ELF_TYPE_CONCAT(bitness, suffix) Elf##bitness##_##suffix##hdr
#define ELF_TYPE_CONCAT_INDIRECT(bitness, suffix) \
    ELF_TYPE_CONCAT(bitness, suffix)
#define MAKE_ELF_TYPE(suffix) ELF_TYPE_CONCAT_INDIRECT(ELF_BITNESS, suffix)

using ElfHeader = MAKE_ELF_TYPE(E);
using ElfProgramHeader = MAKE_ELF_TYPE(P);
using ElfNoteHeader = MAKE_ELF_TYPE(N);

#define ELFCLASS_CONCAT(bitness) ELFCLASS##bitness
#define ELFCLASS_INDIRECT(bitness) ELFCLASS_CONCAT(bitness)
#define EXPECTED_ELFCLASS ELFCLASS_INDIRECT(ELF_BITNESS)
}
