// Copyright (c) Microsoft Corporation. All rights reserved.

#include "ElfNoteParsing.h"

#include <iomanip>

#include <azure-device-util/Bytes.h>
#include <azure-device-util/Math.h>

namespace Utils = Azure::Device::Utils;

namespace Azure::Device::CrashHandler {

struct NoteParseException : public std::runtime_error
{
    explicit NoteParseException(const std::string& what)
        : std::runtime_error(what)
    {}
};

std::ostream& operator<<(std::ostream& out, const elf_prstatus& prstatus)
{
    out << std::dec;
    out << "[Thread State: " << prstatus.pr_pid << "]" << std::endl;

    out << "Signal: " << prstatus.pr_info.si_signo << std::endl;
    out << "Extra error code: " << prstatus.pr_info.si_code << std::endl;
    out << "Errno: " << prstatus.pr_info.si_errno << std::endl;
    out << "Current signal: " << prstatus.pr_cursig << std::endl;
    out << "Pending signals: " << prstatus.pr_sigpend << std::endl;
    out << "Held signals: " << prstatus.pr_sighold << std::endl;
    out << "Thread ID: " << prstatus.pr_pid << std::endl;
    out << "Process parent ID: " << prstatus.pr_ppid << std::endl;
    out << "Process group ID: " << prstatus.pr_pgrp << std::endl;
    out << "Process session ID: " << prstatus.pr_sid << std::endl;

    auto printTimeVal = [&](std::string_view header, timeval timeVal) {
        out << header << timeVal.tv_sec << "." << timeVal.tv_usec << std::endl;
    };

    printTimeVal("Process user time: ", prstatus.pr_utime);
    printTimeVal("Process system time: ", prstatus.pr_stime);
    printTimeVal("Process cumulative user time: ", prstatus.pr_cutime);
    printTimeVal("Process cumulative system time: ", prstatus.pr_cstime);

    uint32_t regCounter = 0;
    for (auto reg : prstatus.pr_reg)
    {
        out << "Reg" << std::dec << std::setfill('0') << std::setw(2)
            << regCounter << ": 0x" << std::hex << std::setfill('0')
            << std::setw(sizeof(reg) * 2) << reg << std::endl;
        regCounter++;
    }

    return out;
}

std::ostream&
operator<<(std::ostream& out, const std::vector<MappedRegion>& mappedRegions)
{
    out << "[Mapped regions: " << mappedRegions.size() << "]" << std::endl;
    for (const MappedRegion& mappedRegion : mappedRegions)
    {
        out << mappedRegion.LowAddress << "-" << mappedRegion.HighAddress
            << " | " << mappedRegion.ModulePath << std::endl;
    }
    return out;
}

std::vector<MappedRegion>
ParseCoreMappedFile(nonstd::span<const std::byte> mappedFileBytes)
{
    std::vector<MappedRegion> regions;

    // Consume the two header fields for mappings count and page size
    size_t mappingsCount = Utils::Bytes::ConsumeBytes<size_t>(mappedFileBytes);
    regions.resize(mappingsCount);

    Utils::Bytes::ConsumeBytes<size_t>(mappedFileBytes);  // Unused page size
                                                          // value

    // Split the remaining buffer into the static sized mappings
    // and the variable sized string names
    size_t mappingBoundaryByteCount = 3 * sizeof(size_t) * mappingsCount;
    auto mappingBoundariesBytes = mappedFileBytes.first(mappingBoundaryByteCount);
    auto moduleNameBytes = mappedFileBytes.subspan(mappingBoundaryByteCount);

    for (size_t i = 0; i < regions.size(); i++)
    {
        MappedRegion& region = regions[i];

        region.LowAddress =
            Utils::Bytes::ConsumeBytes<size_t>(mappingBoundariesBytes);
        region.HighAddress =
            Utils::Bytes::ConsumeBytes<size_t>(mappingBoundariesBytes);
        Utils::Bytes::ConsumeBytes<size_t>(mappingBoundariesBytes);  // Unused
                                                                     // offset
                                                                     // value
    }

    if (!mappingBoundariesBytes.empty())
    {
        throw NoteParseException(
            "Mapped file entry contained leftover boundary data");
    }

    for (size_t i = 0; i < regions.size(); i++)
    {
        regions[i].ModulePath =
            Utils::Bytes::ConsumeNullTerminatedString(moduleNameBytes);
    }

    return regions;
}

struct ElfNoteHeaderData
{
    std::string_view Name;
    uint32_t Type = 0;
    nonstd::span<const std::byte> Contents;
};

ElfNoteHeaderData
ConsumeElfNoteHeaderData(nonstd::span<const std::byte>& notesBytes)
{
    ElfNoteHeaderData data;

    ElfNoteHeader noteHeader =
        Utils::Bytes::ConsumeBytes<ElfNoteHeader>(notesBytes);

    data.Type = noteHeader.n_type;

    data.Name = (noteHeader.n_namesz == 0) ?
        "CORE" :
        Utils::Bytes::BytesToStringView(notesBytes.first(noteHeader.n_namesz - 1));
    notesBytes = notesBytes.subspan(Utils::Math::AlignUp(noteHeader.n_namesz, 4));

    data.Contents = notesBytes.first(noteHeader.n_descsz);
    notesBytes = notesBytes.subspan(Utils::Math::AlignUp(noteHeader.n_descsz, 4));

    return data;
}

ElfNotesContents
ParseElfNotes(nonstd::span<const std::byte> notesBytes, std::ostream& logStream)
{
    ElfNotesContents contents;

    try
    {
        while (!notesBytes.empty())
        {
            ElfNoteHeaderData headerData = ConsumeElfNoteHeaderData(notesBytes);

            constexpr std::string_view c_coreNoteName = "CORE";
            if (headerData.Name == c_coreNoteName &&
                headerData.Type == NT_PRSTATUS)
            {
                const elf_prstatus& prstatus =
                    Utils::Bytes::BytesToValue<elf_prstatus>(headerData.Contents);

                logStream << std::endl;
                logStream << prstatus << std::endl;

                contents.ThreadStates.push_back(prstatus);
            }
            else if (headerData.Name == c_coreNoteName && headerData.Type == NT_FILE)
            {
                contents.MappedRegions = ParseCoreMappedFile(headerData.Contents);

                logStream << std::endl;
                logStream << contents.MappedRegions << std::endl;
            }
        }
    }
    catch (const NoteParseException& e)
    {
        logStream
            << "NoteParseException, halting further note analysis: " << std::endl
            << "  |  " << e.what() << std::endl;
    }

    return contents;
}
}
