// Copyright (c) Microsoft Corporation. All rights reserved.

#include <chrono>
#include <filesystem>
#include <fstream>
#include <inttypes.h>
#include <string>
#include <optional>

#include <azure-device-util/AndroidLogHelper.h>
#include <azure-device-util/FileSystem.h>
#include <azure-device-util/ScopeExit.h>
#include <azure-device-util/String.h>
#include <azure-device-util/TraceLoggingTypes.h>
#include <boost/process.hpp>
#include <tracelogging/TraceLoggingProvider.h>

#include "Config.h"
#include "CrashAnalysis.h"
#include "CrashParameters.h"
#include "CrashSubmission.h"

// 22731e9c-e31a-4484-98d6-efa23d791456
TRACELOGGING_DEFINE_PROVIDER(
    g_crashReportProvider,
    "Azure.CrashReport.Handler",
    (0x22731e9c, 0xe31a, 0x4484, 0x98, 0xd6, 0xef, 0xa2, 0x3d, 0x79, 0x14, 0x56));

namespace Utils = Azure::Device::Utils;

//
// The crash handler package is mostly concerned with analyzing an ELF core
// dump. To understand the structure of ELF core files, check out these
// resources:
//
// `man core`
// `man elf`
// `eu-readelf -a <program_name|core_dump_file>`
//
// The general structure of the core dump we care about involves:
// 1. A top level ELF header
// 2. A bunch of ProgramHeaders, one of which describes the location of the NOTE
//      data inside the core dump file.
// 3. The NOTE data itself, which is a variable length structure of various note
//      types, each of which has their own internal structure.  The ones we care
//      about are the ones that tell us the state of each thread (register,
//      signals, etc), and the one that tells us the name of the module each
//      virtual address in the address space maps to.
//
// With all that information, we are able to deduce a faulting module and offset
// and then submit the dump to our uploader service, and send appropriate
// telemetry events with the metadata of the crash.
//

namespace Azure::Device::CrashHandler {

/*++
    Registered as our error handling callback with the ErrorHelper from
    azure-device-utils package. Logs to LTTNG.

Parameters:
    message - The message to log.
    file - The file name (in debug builds) which generated the error.
    line - The line number (in debug builds) which generated the error.
    err - The error code generated.
--*/
static void LogCallback(
    std::string_view message,
    std::string_view file,
    int lineNumber,
    std::error_code err)
{
    TraceLoggingWrite(
        g_crashReportProvider,
        "LogError",
        TraceLoggingStringView(file, "FileName"),
        TraceLoggingValue(lineNumber, "Line"),
        TraceLoggingStringView(message, "Message"),
        TraceLoggingErrorCode(err, "ErrorCode"));

    std::cerr << "LogError: ";
    std::cerr << "loc='" << file << "@" << lineNumber << "' ";
    std::cerr << "msg='" << message << "' ";
    std::cerr << "err={msg='" << err.message() << "' ";
    std::cerr << "cat='" << err.category().name() << "' ";
    std::cerr << "val='" << err.value() << "'}" << std::endl;
}

static CrashParameters ParseCommandLineParameters(int32_t argc, char* argv[])
{
    CrashParameters params;

    std::vector<std::function<void(const char* a)>> consumeArguments;
    consumeArguments.push_back(
        [&](const char* a) { params.Pid = Utils::String::ToIntT<pid_t>(a); });

    consumeArguments.push_back([&](const char* a) {
        params.ThreadId = Utils::String::ToIntT<pid_t>(a);
    });

    consumeArguments.push_back([&](const char* a) {
        params.Signal = Utils::String::ToUIntT<uint64_t>(a);
    });

    consumeArguments.push_back([&](const char* a) {
        params.FullPath = a;
        std::replace(params.FullPath.begin(), params.FullPath.end(), '!', '/');
    });

    consumeArguments.push_back([&](const char* a) {
        params.TimestampSeconds = Utils::String::ToUIntT<uint64_t>(a);
    });

    uint32_t expectedArgIndex = 0;

    for (auto i = 1; i < argc; ++i)
    {
        if (Utils::String::ICompare(argv[i], "-onlymetadatafile") == 0)
        {
            params.CollectMetdataOnly = true;
        }
        else
        {
            if (expectedArgIndex < consumeArguments.size())
            {
                ADHS_LOG("%s", argv[i]);
                consumeArguments.at(expectedArgIndex++)(argv[i]);
            }
        }
    }

    return params;
}

static void FinishOutputStreams(
    Utils::StreamCopier& coreStreamCopier,
    std::ofstream& coreOutputStream,
    std::ofstream& logStream)
{
    coreStreamCopier.CopyToEnd();
    coreOutputStream.close();
    logStream.close();
}

std::optional<std::filesystem::path>
FindTombstonePath(const CrashParameters& params)
{
#ifdef AZURE_DEVICE_CRASH_HANDLER_TOMBSTONE_COLLECTION
    ADHS_LOG("Begin searching for tombstone path");
    std::vector<std::filesystem::path> tombstonePaths;

    static const std::string tombstonePath = "/data/tombstones";
    std::filesystem::path tombstoneRoot{ tombstonePath };
    if (std::filesystem::exists(tombstoneRoot) &&
        std::filesystem::is_directory(tombstoneRoot))
    {
        for (auto const& entry :
             std::filesystem::recursive_directory_iterator(tombstoneRoot))
        {
            if (std::filesystem::is_regular_file(entry))
            {
                tombstonePaths.emplace_back(entry.path().filename());
            }
        }
    }

    auto foundPath = tombstonePaths.end();

    for (auto fileIt = tombstonePaths.begin(); fileIt != tombstonePaths.end();
         ++fileIt)
    {
        // Do not check files that ends with .uploaded. Onetrace currently
        // creates these files to indicate an upload has happened.
        if (Utils::String::IEndsWith((*fileIt).string(), ".uploaded"))
        {
            ADHS_LOG(
                "Skipping file with name, %s, because it ends with .uploaded.",
                (*fileIt).string().c_str());

            continue;
        }

        std::filesystem::path path = tombstoneRoot / (*fileIt);

        int64_t size = 0;
        if (FAILED(
                Utils::FileSystem::ReadFileSize(path.string(), size) || size <= 0))
        {
            ADHS_LOG(
                "Failed to read file size of file with name, %s",
                (*fileIt).string().c_str());

            continue;
        }

        std::ifstream file(path.string());
        std::string line;
        while (getline(file, line))
        {
            if (!Utils::String::IStartsWith(line, "pid:"))
            {
                continue;
            }

            // Tombstone file contains a line, "pid: %PID%, tid: %TID%". Using
            // this to identify which tombstone belongs to
            if (!Utils::String::IStartsWith(
                    line,
                    "pid: " + std::to_string(params.Pid) +
                        ", tid: " + std::to_string(params.ThreadId)))
            {
                ADHS_LOG(
                    "Skipping tombstone with name, %s, since it does not match the pid and tid of the crashed process.",
                    (*fileIt).string().c_str());

                break;
            }

            ADHS_LOG(
                "Found tombstone with name, %s, that matches the crashed process.",
                (*fileIt).string().c_str());

            return path;
        }
    }
#else
    UNREFERENCED_PARAMETER(params);
#endif

    return std::nullopt;
}

std::optional<std::filesystem::path> CreateLogcatFileInTempDirectory(
    const CrashParameters& params,
    const std::filesystem::path& tempDir)
{
#ifdef AZURE_DEVICE_CRASH_HANDLER_LOGCAT_COLLECTION
    ADHS_LOG("Creating logcat file");
    std::filesystem::path logcatPath = tempDir / "crashed_process_logcat.txt";
    (void) boost::process::system(
        "/bin/logcat",
        boost::process::args({ "-d",
                               "-f",
                               logcatPath.string(),
                               "--pid=" + std::to_string(params.Pid) }));

    return logcatPath;
#else
    UNREFERENCED_PARAMETER(params);
    UNREFERENCED_PARAMETER(tempDir);
#endif

    return std::nullopt;
}

std::filesystem::path GetQueuePath()
{
    return AZURE_DEVICE_QUEUE_FOLDER;
}

std::filesystem::path GetTempPath()
{
#ifdef AZURE_DEVICE_USE_CUSTOM_TEMP_DIRECTORY
    return AZURE_DEVICE_TEMP_FOLDER;
#else
    return std::filesystem::temp_directory_path();
#endif
}

std::filesystem::path RenameAndGetCoreFilePathBasedOnAbi(
    const std::filesystem::path& tempCorePath,
    bool is32Bits)
{
#ifdef AZURE_DEVICE_USE_ABI_EXTENSION_FOR_CORE
    std::string abiStr = "unknown";
    #if defined(__arm__) || defined(__aarch64__)
    if (is32Bits)
    {
        abiStr = "ARM32";
    }
    else
    {
        abiStr = "ARM64";
    }
    #elif defined(__i386__) || defined(__x86_64__)
    if (is32Bits)
    {
        abiStr = "x86";
    }
    else
    {
        abiStr = "x64";
    }
    #endif

    std::filesystem::path corePath = tempCorePath.parent_path() /
        (tempCorePath.filename().string() + "_" + abiStr);
    std::filesystem::rename(tempCorePath, corePath);

    return corePath;
#else
    UNREFERENCED_PARAMETER(is32Bits);
#endif

    return tempCorePath;
}

bool ProcessCrash(int32_t argc, char* argv[])
{
    std::error_code ec;
    std::string what;

    try
    {
        CrashParameters params = ParseCommandLineParameters(argc, argv);

        TraceLoggingWrite(
            g_crashReportProvider,
            "CoreDumpInvoked",
            TraceLoggingValue(params.Pid, "ProcessID"),
            TraceLoggingValue(params.ThreadId, "ThreadID"),
            TraceLoggingValue(params.Signal, "Signal"),
            TraceLoggingValue(params.FullPath.c_str(), "FullPath"),
            TraceLoggingValue(params.TimestampSeconds, "TimestampSeconds"));

        ADHS_LOG(
            "CoreDumpInvoked with pid: %d, tid: %d signal: %" PRIu64
            ", FullPath: %s, and TimestampSeconds: %" PRIu64,
            params.Pid,
            params.ThreadId,
            params.Signal,
            params.FullPath.c_str(),
            params.TimestampSeconds);

        auto tempDir =
            GetTempPath() /
            std::to_string(
                std::chrono::steady_clock::now().time_since_epoch().count());
        std::filesystem::create_directories(tempDir);

        auto cleanupTempDir = Utils::ScopeExit([&tempDir] {
            try
            {
                std::filesystem::remove_all(tempDir);
            }
            CATCH_LOG();
        });

        Utils::FileSystem::TempOutputFile coreFile{ tempDir / "core.dmp" };
        Utils::FileSystem::TempOutputFile logFile{ tempDir / "log.txt" };

        logFile.GetStream() << params << std::endl;

        Utils::StreamCopier coreStreamCopier{ std::cin, coreFile.GetStream() };

        CrashAnalysis analysis =
            CopyAndAnalyzeCrash(params, coreStreamCopier, logFile.GetStream());
        logFile.GetStream() << std::endl << analysis << std::endl;

        ADHS_LOG("Finished analyzing crash");

        FinishOutputStreams(
            coreStreamCopier, coreFile.GetStream(), logFile.GetStream());
        ADHS_LOG("Finished core dump output stream");

        std::vector<std::filesystem::path> paths;

        if (!params.CollectMetdataOnly)
        {
            if (auto tombstonePath = FindTombstonePath(params))
            {
                std::filesystem::path tombstoneTempPath = tempDir / "core.tomb";
                std::filesystem::copy_file(*tombstonePath, tombstoneTempPath);

                paths.emplace_back(tombstoneTempPath);
            }

            if (auto logcatPath = CreateLogcatFileInTempDirectory(params, tempDir))
            {
                paths.emplace_back(*logcatPath);
            }

            paths.emplace_back(RenameAndGetCoreFilePathBasedOnAbi(
                coreFile.GetPath(), analysis.Is32Bits));
            paths.emplace_back(logFile.GetPath());
        }

        uint64_t reportId =
            SubmitCrashToUploaderService(GetQueuePath(), analysis, paths);

        TraceLoggingWrite(
            g_crashReportProvider,
            "CoreDumpFinished",
            TraceLoggingValue(reportId, "ReportId"),
            TraceLoggingValue(analysis.AppName.c_str(), "AppName"),
            TraceLoggingValue(analysis.AppVersion.c_str(), "AppVersion"),
            TraceLoggingValue(analysis.PackageName.c_str(), "PackageName"),
            TraceLoggingValue(analysis.PackageVersion.c_str(), "PackageVersion"),
            TraceLoggingValue(analysis.PackageArch.c_str(), "PackageArch"),
            TraceLoggingValue(analysis.ModuleName.c_str(), "ModuleName"),
            TraceLoggingValue(analysis.ModuleVersion.c_str(), "ModuleVersion"),
            TraceLoggingValue(analysis.ModuleOffset, "ModuleOffset"),
            TraceLoggingValue(analysis.ExceptionCode, "ExceptionCode"),
            TraceLoggingValue(analysis.StackHash, "StackHash"),
            TraceLoggingValue(coreStreamCopier.GetBytesRead(), "BytesWritten"),
            TraceLoggingValue(true, "Submitted"));

        ADHS_LOG("CoreDumpFinished");

        return true;
    }
    catch (const std::filesystem::filesystem_error& e)
    {
        what = e.what();
        ec = e.code();
    }
    catch (const std::system_error& e)
    {
        what = e.what();
        ec = e.code();
    }
    catch (const std::exception& e)
    {
        what = e.what();
    }
    catch (...)
    {
        what = "Unknown object caught";
    }

    TraceLoggingWrite(
        g_crashReportProvider,
        "CoreDumpFailed",
        TraceLoggingValue(ec.category().name(), "ErrorCategory"),
        TraceLoggingValue(ec.value(), "ErrorValue"),
        TraceLoggingValue(what.c_str(), "ErrorText"));
    ADHS_LOG(
        "CoreDumpFailed with error category: %s, error value: %d, and error text: %s",
        ec.category().name(),
        ec.value(),
        what.c_str());

    return false;
}
}

int main(int32_t argc, char* argv[])
{
    TraceLoggingRegister(g_crashReportProvider);
    Utils::SetResultLoggingCallback(&Azure::Device::CrashHandler::LogCallback);

    uint8_t aId[16];
    lttngh_ActivityIdCreate(aId);
    lttngh_ActivityIdSet(aId);

    TraceLoggingWrite(g_crashReportProvider, "CoreDumpStarted");

    bool result = Azure::Device::CrashHandler::ProcessCrash(argc, argv);

    TraceLoggingUnregister(g_crashReportProvider);

    return !result;
}
