// Copyright (c) Microsoft Corporation. All rights reserved.

#include <filesystem>
#include <fstream>
#include <limits>
#include <random>
#include <stdexcept>
#include <string_view>

#include <azure-device-errorreporting-submitter/Report.h>

#ifdef AZURE_DEVICE_USE_CLOG
    #define CLOG_INLINE_IMPLEMENTATION 1
    #include "Report.cpp.clog.h"
    #include "privacytags.h"
#else
    #include "clog_stubs.h"
#endif

namespace Azure { namespace Device { namespace ErrorReporting {
namespace Submitter {

/*++
    Creates a report that is intended to be sent to the Watson service.

Arguments:
    eventType - The type of report to send.  This type must be properly
registered with the Watson service.
--*/
Report::Report(std::string_view eventType) : m_eventType(eventType)
{
    // Generate a random id for this report.
    std::random_device device;
    std::uniform_int_distribution<size_t> distribution(
        0, std::numeric_limits<size_t>::max());
    m_reportId = distribution(device);
    uuid_generate_random(m_werReportId);
}

/*++
    Adds a bucketing parameter to the report.  Each report can have up to 10
bucketing parameters, which group similar failures together in the Watson
service.

Arguments:
    parameter - The bucketing parameter to add to the report.
--*/
void Report::AddBucketingParameter(std::string_view parameter)
{
    if (m_bucketingParameters.size() > 9)
    {
        throw std::length_error(
            "A maximum of 10 bucketing parameters are allowed per report.");
    }

    if (parameter.empty())
    {
        throw std::invalid_argument("Bucketing parameters cannot be empty string");
    }

    m_bucketingParameters.emplace_back(parameter);
}

/*++
    Registers a file or directory to be included in the report when sent to the
Watson service. When a report is submitted, all registered paths will be
compressed into a zip file.  Each path registered in this manner will be located
in the root directory of the report zip file.  This is equivalent to calling
RegisterPathInSubdirectory(sourcePath, "").

Arguments:
    sourcePath - The path to the file or directory to include in the report.
--*/
void Report::RegisterPath(const std::filesystem::path& sourcePath)
{
    m_registeredPaths.emplace_back(sourcePath, "");
}

/*++
    Registers a file or directory to be included in the report when sent to the
Watson service. When a report is submitted, all registered paths will be
compressed into a zip file.  Each path registered in this manner will be located
in a specified subdirectory of the report zip file.

Arguments:
    sourcePath - The path to the file or directory to include in the report.
    reportSubdirPath - A subdirectory path that specifies where sourcePath
should be placed in the report zip file. An empty string will place sourcePath
at the root of the zip file.
--*/
void Report::RegisterPathInSubdirectory(
    const std::filesystem::path& sourcePath,
    const std::filesystem::path& reportSubdirPath)
{
    m_registeredPaths.emplace_back(sourcePath, reportSubdirPath);
}

/*++
    Stores the Watson metadata of this report (event type and bucketing
parameters) to a file.

Arguments:
    metadataPath - The path on disk to store the report's metadata.
--*/
void Report::WriteMetadataAndEmitLogging(std::filesystem::path& metadataPath)
{
    std::ofstream metadataStream(metadataPath.string().c_str());
    metadataStream
        << R"(<?xml version="1.0" ?><WERReportMetadata><ProblemSignatures>)";
    metadataStream << "<EventType>" << m_eventType << "</EventType>";

    std::string werReportIdString;
    werReportIdString.resize(36);
    uuid_unparse_lower(m_werReportId, &werReportIdString[0]);

    metadataStream << "<WerReportId>" << werReportIdString << "</WerReportId>";

    for (size_t i = 0; i < m_bucketingParameters.size(); i++)
    {
        metadataStream << "<Parameter" << i << ">" << m_bucketingParameters[i]
                       << "</Parameter" << i << ">";
    }

    metadataStream << "</ProblemSignatures></WERReportMetadata>";

    metadataStream.close();

    // Set permissions on report files, such that only root and the owner have
    // any access.
    std::error_code ec;
    std::filesystem::permissions(
        metadataPath,
        std::filesystem::perms::owner_read | std::filesystem::perms::owner_write,
        ec);

    LatteTelemetry(
        SubmitErrorReport,
        "Azure.Device.ErrorReporting",
        PrivacyTags_ProductAndServicePerformance,
        "%{EventType,s} \
         %{WerReportId,s}",
        m_eventType.c_str(),
        werReportIdString.c_str());
}

}}}}
