// Copyright (c) Microsoft Corporation. All rights reserved.

#include <filesystem>
#include <fstream>
#include <iomanip>
#include <map>
#include <memory>
#include <string_view>
#include <zlib.h>

#include <azure-device-errorreporting-submitter/ReportQueue.h>
#include <azure-device-util/ErrorHelper.h>
#include <azure-device-util/FileSystem.h>
#include <azure-device-util/ScopeExit.h>
#include <zipper/zipper.h>

#include "ZipUtils.h"

using namespace Azure::Device::Utils;

namespace Azure { namespace Device { namespace ErrorReporting {
namespace Submitter {

ReportQueue::ReportQueue(const std::filesystem::path& path) : m_path(path)
{
    ReadQueueConfig();
}

/*++
    Opens a report queue, for report submission.  This will throw if
CreateReportQueue has never been called for the given path.  Typically, an
uploader service will create the report queue, and applications will open the
report queue to submit reports.

Arguments:
    path - The location of the target report queue.
--*/
std::unique_ptr<ReportQueue>
ReportQueue::OpenReportQueue(const std::filesystem::path& path)
{
    if (!std::filesystem::exists(path))
    {
        std::string badPath = "Report queue not found" + path.native();
        throw std::system_error(
            std::make_error_code(std::errc::no_such_file_or_directory), badPath);
    }

    return std::make_unique<ReportQueue>(path);
}

/*++
    Creates a report queue to receive report submissions.  Typically, an
uploader service will create the report queue, and applications will open the
report queue to submit reports.

Arguments:
    path - The desired path for the report queue.
    queuePermissions - The permissions to apply for the root directory of the
queue. pendingReportsPermissions - The permissions to apply for the subdirectory
of the queue containing pending reports. submittedReportsPermissions - The
permissions to apply for the subdirectory of the queue containing submitted
reports.
--*/
std::unique_ptr<ReportQueue> ReportQueue::CreateReportQueue(
    const std::filesystem::path& path,
    std::filesystem::perms queuePermissions,
    std::filesystem::perms pendingReportsPermissions,
    std::filesystem::perms submittedReportsPermissions)
{
    // Create directories for the report queue and all pending and submitted
    // data.
    std::filesystem::create_directories(path);
    std::filesystem::permissions(path, queuePermissions);

    std::filesystem::create_directory(path / "pending");
    std::filesystem::permissions(path / "pending", pendingReportsPermissions);

    std::filesystem::create_directory(path / "submitted");
    std::filesystem::permissions(path / "submitted", submittedReportsPermissions);

    return std::make_unique<ReportQueue>(path);
}

/*++
    Creates a report linked to this queue.  No data will be gathered for this
report until Submit is called.

Arguments:
    eventType - The type of report to send.  This type must be properly
registered with the Watson service.
--*/
std::unique_ptr<Report> ReportQueue::CreateReport(std::string_view eventType)
{
    if (eventType.empty())
    {
        throw std::invalid_argument("eventType cannot be empty");
    }

    return std::make_unique<Report>(eventType);
}

/*++
    Reads configuration settings for the report queue.
--*/
void ReportQueue::ReadQueueConfig()
{
    try
    {
        // Read config data from disk.
        std::filesystem::path queueConfigPath = m_path / "queueconfig";
        std::map<std::string, std::string, std::less<>> configData;
        std::error_code result = Utils::FileSystem::ReadDictionaryFromFile(
            queueConfigPath.c_str(), ",", configData);
        if (FAILED(result))
        {
            return;
        }

        // Enumerate all config data, looking for known settings.
        for (const auto& pair : configData)
        {
            std::string key = std::move(pair.first);
            std::string value = std::move(pair.second);

            if (key == "maxQueueSizeBytes")
            {
                m_queueConfig.MaxQueueSizeBytes =
                    strtoul(value.c_str(), nullptr, 0);
            }
        }
    }
    catch (...)
    {}
}

/*++
    Gets the total size of all submitted reports awaiting upload.
--*/
uintmax_t ReportQueue::GetQueueSize()
{
    uintmax_t totalSize = 0;

    for (auto& item :
         std::filesystem::recursive_directory_iterator(GetSubmittedReportsPath()))
    {
        if (std::filesystem::is_regular_file(item.path()))
        {
            totalSize += std::filesystem::file_size(item.path());
        }
    }

    return totalSize;
}

/*++
    Determines whether or not data for a given report is wanted.  In some cases,
such as frequent submissions of a given report type, the report uploader may
choose to ignore a new report submission.
--*/
bool ReportQueue::IsThrottled(Report& report)
{
    // TODO: Unimplemented.
    report.GetId();
    return false;
}

/*++
    Submits a report to the queue for upload.
--*/
void ReportQueue::Submit(Report& report)
{
    if (IsFull())
    {
        throw std::filesystem::filesystem_error(
            "The report queue is full.",
            make_error_code(std::errc::file_too_large));
    }

    // Create the pending report directory and zip file.
    std::filesystem::path pendingReportPath =
        GetPendingReportsPath() / std::to_string(report.GetId());

    std::filesystem::create_directories(pendingReportPath);

    // On either success or failure, we want to ensure that pending data is gone
    // after Submit returns.
    auto cleanupPendingData = Utils::MakeScopeExit([&pendingReportPath]() {
        if (std::filesystem::exists(pendingReportPath))
        {
            std::filesystem::remove_all(pendingReportPath);
        }
    });

    // Create a zip file containing all of the report's registered paths.
    std::filesystem::path zipPath = pendingReportPath / "report.zip";
    zipper::Zipper zipper(zipPath.string());

    for (const auto& pathPair : report.GetRegisteredPaths())
    {
        ZipUtils::AddPathToZip(zipper, pathPair.first, pathPair.second);
    }

    //
    // Add the report metadata to the report directory and zip file.  This will
    // be parsed by the uploader to determine the event type and parameters.
    //
    std::filesystem::path metadataPath =
        pendingReportPath / "WERInternalMetadata.xml";
    report.WriteMetadataAndEmitLogging(metadataPath);
    ZipUtils::AddPathToZip(zipper, metadataPath, "");

    zipper.close();

    // Set permissions on report files, such that only root and the owner have
    // any access.
    std::error_code ec;
    std::filesystem::permissions(
        zipPath,
        std::filesystem::perms::owner_read | std::filesystem::perms::owner_write,
        ec);

    // Get the size of the compressed report data, to see if it fits in the
    // report queue.
    uintmax_t reportSize = std::filesystem::file_size(zipPath) +
        std::filesystem::file_size(metadataPath);

    if (reportSize + GetQueueSize() > m_queueConfig.MaxQueueSizeBytes)
    {
        throw std::filesystem::filesystem_error(
            "Insufficient queue space remaining.",
            make_error_code(std::errc::file_too_large));
    }

    // Now commit the report by moving it to its final location.
    std::filesystem::path submittedReportPath =
        GetSubmittedReportsPath() / std::to_string(report.GetId());
    std::filesystem::rename(pendingReportPath, submittedReportPath);
}

}}}}
