// Copyright (c) Microsoft Corporation. All rights reserved.

#include "INotifyMonitor.h"

#include <errno.h>
#include <string>

#include <sys/eventfd.h>
#include <sys/inotify.h>
#include <sys/poll.h>
#include <sys/unistd.h>

#include "ServiceLoggingProvider.h"

namespace Azure { namespace Device { namespace ErrorReporting {
namespace Uploader {

INotifyMonitor::INotifyMonitor(INotifyMonitorCallbackReceiver* callbackReceiver)
    : m_callbackReceiver(callbackReceiver)
    , m_isStopping(false)
    , m_stopDescriptor(eventfd(0, 0))
{}

INotifyMonitor::~INotifyMonitor()
{
    if (m_monitorDescriptor >= 0)
    {
        for (int watchDescriptor : m_watchDescriptors)
        {
            inotify_rm_watch(m_monitorDescriptor, watchDescriptor);
        }

        close(m_monitorDescriptor);
    }

    if (m_stopDescriptor >= 0)
    {
        close(m_stopDescriptor);
    }
}

/*++
    Adds a file or directory path to monitor for changes.

Parameters:
    path - The path to monitor
--*/
void INotifyMonitor::AddMonitoredPath(const std::string& path)
{
    m_paths.emplace_back(path);
}

/*++
    Continually poll for file or directory changes until a stop call is made
--*/
void INotifyMonitor::Run()
{
    TraceLoggingWrite(g_errorReportingClientProvider, "INotifyMonitor_RunStart");

    // Initialize an inotify instance.
    m_monitorDescriptor = inotify_init();

    if (m_monitorDescriptor < 0)
    {
        TraceLoggingWrite(
            g_errorReportingClientProvider,
            "INotifyMonitor_MonitorDescriptorNotValid",
            TraceLoggingValue(errno, "ErrorCode"));

        return;
    }

    // Configure inotify to inform us when files or subdirectories are created
    // in each registered path.
    for (const auto& path : m_paths)
    {
        int watchDescriptor = inotify_add_watch(
            m_monitorDescriptor, path.c_str(), IN_CREATE | IN_MOVED_TO);

        if (watchDescriptor >= 0)
        {
            TraceLoggingWrite(
                g_errorReportingClientProvider,
                "INotifyMonitor_PathAdded",
                TraceLoggingValue(path.c_str(), "Path"),
                TraceLoggingValue(watchDescriptor, "WatchDescriptor"));

            m_watchDescriptors.emplace_back(watchDescriptor);
        }
        else
        {
            TraceLoggingWrite(
                g_errorReportingClientProvider,
                "INotifyMonitor_FailedToAddWatcher",
                TraceLoggingValue(path.c_str(), "Path"),
                TraceLoggingValue(errno, "ErrorCode"));
        }
    }

    if (m_watchDescriptors.empty())
    {
        TraceLoggingWrite(
            g_errorReportingClientProvider, "INotifyMonitor_NoValidWatchers");

        return;
    }

    // Set up poll descriptors to wake us when inotify has detected a change or
    // when the stop event is fired.
    std::vector<pollfd> pollDescriptors(2);
    pollDescriptors[0].fd = m_stopDescriptor;
    pollDescriptors[0].events = POLLIN;
    pollDescriptors[1].fd = m_monitorDescriptor;
    pollDescriptors[1].events = POLLIN;

    // Poll for changes until we're told to stop.
    while (!m_isStopping)
    {
        int result = poll(pollDescriptors.data(), 2, m_pollIntervalSeconds);

        TraceLoggingWrite(
            g_errorReportingClientProvider,
            "INotifyMonitor_Wake",
            TraceLoggingValue(result, "Result"),
            TraceLoggingValue(pollDescriptors[0].revents, "StopEventMask"),
            TraceLoggingValue(pollDescriptors[1].revents, "MonitorEventMask"));

        // If the stop event was fired, we're done.
        if (pollDescriptors[0].revents != 0)
        {
            TraceLoggingWrite(
                g_errorReportingClientProvider,
                "INotifyMonitor_StopEventReceived");

            break;
        }

        // If the poll failed, retry if we were interrupted.  Otherwise, quit.
        if (result < 0)
        {
            TraceLoggingWrite(
                g_errorReportingClientProvider,
                "INotifyMonitor_PollFailed",
                TraceLoggingValue(errno, "ErrorCode"));

            if (errno == EINTR)
            {
                continue;
            }

            break;
        }

        // The poll succeeded, and we weren't told to stop.  This indicates that
        // a file change or timeout occurred.
        std::vector<INotifyMonitorEvent> monitorEvents;

        if (pollDescriptors[1].revents == 0)
        {
            TraceLoggingWrite(
                g_errorReportingClientProvider, "INotifyMonitor_PollTimeout");

            monitorEvents.emplace_back(0, 0, "Timeout");
        }
        else
        {
            TraceLoggingWrite(
                g_errorReportingClientProvider,
                "INotifyMonitor_FileChangeDetected");

            // Read the change events from inotify.
            std::vector<char> readData(k_maxReadBufferSize);
            size_t length =
                read(pollDescriptors[1].fd, &readData[0], readData.size());

            TraceLoggingWrite(
                g_errorReportingClientProvider,
                "INotifyMonitor_FileChangePayloadSize",
                TraceLoggingValue(length, "Size"));

            // Read each event from the buffer until we've processed the entire
            // payload.
            size_t bytesProcessed = 0;

            while (bytesProcessed < length)
            {
                inotify_event* event =
                    reinterpret_cast<inotify_event*>(&readData[bytesProcessed]);
                monitorEvents.emplace_back(
                    event->mask, event->cookie, event->name);

                bytesProcessed += sizeof(inotify_event) + event->len;

                TraceLoggingWrite(
                    g_errorReportingClientProvider,
                    "INotifyMonitor_EventDetails",
                    TraceLoggingValue(event->wd, "WatchDescriptor"),
                    TraceLoggingValue(event->mask, "EventMask"),
                    TraceLoggingValue(event->cookie, "Cookie"),
                    TraceLoggingValue(event->len, "NameLength"),
                    TraceLoggingString(event->name, "Name"),
                    TraceLoggingValue(bytesProcessed, "BytesProcessed"));
            }
        }

        // Inform the caller with details of the poll result.
        if (!monitorEvents.empty())
        {
            m_callbackReceiver->INotifyMonitorCallback(monitorEvents);
        }
    }

    TraceLoggingWrite(g_errorReportingClientProvider, "INotifyMonitor_RunStop");
}

/*++
    Signals the thread executing Run() to stop.
--*/
void INotifyMonitor::Stop()
{
    m_isStopping = true;

    if (m_stopDescriptor >= 0)
    {
        TraceLoggingWrite(
            g_errorReportingClientProvider, "INotifyMonitor_SignalStop");

        // Write some data to the stop eventfd, to unblock the poll in Run().
        uint64_t stopData = 1;
        ssize_t result = write(m_stopDescriptor, &stopData, sizeof(stopData));

        if (result <= 0)
        {
            TraceLoggingWrite(
                g_errorReportingClientProvider,
                "INotifyMonitor_SignalStopFailed",
                TraceLoggingValue(result, "Result"),
                TraceLoggingValue(errno, "ErrorCode"));
        }
    }
    else
    {
        TraceLoggingWrite(
            g_errorReportingClientProvider,
            "INotifyMonitor_StopDescriptorNotValid");
    }
}
}}}}
