// Copyright (c) Microsoft Corporation. All rights reserved.

#include "KernelDumpProcessor.h"

#include <string>

#include <azure-device-errorreporting-submitter/ReportQueue.h>
#include <azure-device-util/ResultException.h>
#include <sys/utsname.h>

#include "ServiceLoggingProvider.h"

namespace Azure { namespace Device { namespace ErrorReporting {
namespace Uploader {

/*++
    Loops through subdirectories of kernelDumpDirectory.
    If any contains a 'vmcore' file, it submits the entire subdirectory
    in a new Watson report.

    On completion, it will have recursively deleted all subdirectories.
--*/
void KernelDumpProcessor::ProcessDirectory(
    Azure::Device::ErrorReporting::Submitter::ReportQueue& reportQueue,
    std::string_view kernelDumpDirectory) noexcept
{
    constexpr char c_vmCoreFileName[] = "vmcore";

    std::error_code ec;
    for (const auto& subdir : std::filesystem::directory_iterator(
             std::string{ kernelDumpDirectory }, ec))
    {
        try
        {
            if (std::filesystem::exists(subdir.path() / c_vmCoreFileName))
            {
                SubmitKernelDumpReport(reportQueue, subdir);
            }
        }
        CATCH_LOG();

        try
        {
            std::filesystem::remove_all(subdir);
        }
        CATCH_LOG();
    }
}

uint64_t KernelDumpProcessor::SubmitKernelDumpReport(
    Azure::Device::ErrorReporting::Submitter::ReportQueue& queue,
    const std::filesystem::path& dumpInstanceDirectory)
{
    auto report = queue.CreateReport("LinuxAppCrash");

    utsname unameData = {};
    uname(&unameData);

    report->AddBucketingParameter(unameData.sysname);
    report->AddBucketingParameter(unameData.release);
    report->AddBucketingParameter("none");
    report->AddBucketingParameter(unameData.version);

    report->AddBucketingParameter("0");
    report->AddBucketingParameter("1");
    report->AddBucketingParameter("0");

    report->RegisterPath(dumpInstanceDirectory);

    std::cout << "Submitting kernel dump report on: " << dumpInstanceDirectory
              << std::endl;

    TraceLoggingWrite(
        g_errorReportingClientProvider,
        "KernelDumpProcessor_Submit",
        TraceLoggingValue(dumpInstanceDirectory.c_str(), "Directory"),
        TraceLoggingValue(unameData.sysname, "SysName"),
        TraceLoggingValue(unameData.release, "Release"),
        TraceLoggingValue(unameData.version, "Version"));

    queue.Submit(*report);

    uint64_t reportId = report->GetId();

    std::cout << "Submitted kernel dump report successfully! Id = " << reportId
              << std::endl;

    return reportId;
}

}}}}
