// Copyright (c) Microsoft Corporation. All rights reserved.

#include "KernelMessageMonitor.h"

#include <fcntl.h>
#include <iomanip>
#include <time.h>
#include <vector>

#include <azure-device-util/ErrorHelper.h>
#include <azure-device-util/FileSystem.h>
#include <azure-device-util/ResultException.h>
#include <azure-device-util/ScopeExit.h>
#include <azure-device-util/String.h>
#include <sys/eventfd.h>
#include <sys/poll.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/unistd.h>
#include <tracelogging/TraceLoggingProvider.h>

#include "Config.h"
#include "ServiceLoggingProvider.h"

using namespace Azure::Device::Utils;

namespace Azure { namespace Device { namespace ErrorReporting {
namespace Uploader {

KernelMessageMonitor::KernelMessageMonitor()
    : m_isStopping(false), m_stopDescriptor(eventfd(0, 0))
{}

KernelMessageMonitor::~KernelMessageMonitor()
{
    Stop();

    if (m_thread.joinable())
    {
        m_thread.join();
    }

    if (m_stopDescriptor >= 0)
    {
        close(m_stopDescriptor);
    }
}

void KernelMessageMonitor::Run()
{
    m_thread = std::thread(&KernelMessageMonitor::MessageThreadRun, this);
}

/*++
    Begin monitoring kernel message buffer
--*/
void KernelMessageMonitor::MessageThreadRun()
{
    CheckForSentinel();
    CreateOrReplaceSentinel();

    int kmsgHandle = open("/dev/kmsg", O_RDONLY);
    THROW_HR_IF(make_errno_error_code(), kmsgHandle < 0);

    auto closeKmsgScopeExit =
        Utils::MakeScopeExit([kmsgHandle]() { close(kmsgHandle); });

    TraceLoggingWrite(g_errorReportingClientProvider, "KernelMonitorStart");

    uint seqErrorCount = 0;

    std::array<pollfd, 2> pollDescriptors;
    pollDescriptors[0].fd = m_stopDescriptor;
    pollDescriptors[0].events = POLLIN;
    pollDescriptors[1].fd = kmsgHandle;
    pollDescriptors[1].events = POLLIN;

    while (!m_isStopping)
    {
        std::string result;
        std::array<char, 256> buffer;

        if (m_isStopping)
        {
            break;
        }

        int pollRc = poll(pollDescriptors.data(), 2, INT_MAX);
        if (pollRc < 0)
        {
            TraceLoggingWrite(
                g_errorReportingClientProvider,
                "KernelMessageMonitor_PollFailed",
                TraceLoggingValue(errno, "ErrorCode"));

            if (errno == EINTR)
            {
                continue;
            }

            THROW_HR(make_errno_error_code());
        }

        if (pollDescriptors[0].revents != 0)
        {
            TraceLoggingWrite(
                g_errorReportingClientProvider,
                "KernelMessageMonitor_StopEventReceived");

            break;
        }

        // Only thing left is having read data on kmsg buffer
        ssize_t readSize = read(kmsgHandle, buffer.data(), buffer.size() - 1);
        int readError = errno;

        if (readSize > 0)
        {
            // Overwrite new line with terminator
            // TODO: Multipart messages not supported right now
            result.append(buffer.data(), static_cast<size_t>(readSize - 1));
            ProcessMessage(result);
        }

        if (readSize == -1)
        {
            seqErrorCount++;

            // File handle most likely closed, exit
            if (readError == EINTR)
            {
                break;
            }

            // Exit if sequential error count above 3
            if (seqErrorCount > 3)
            {
                break;
            }
        }
        else
        {
            seqErrorCount = 0;
        }
    }

    TraceLoggingWrite(g_errorReportingClientProvider, "KernelMonitorStop");
}

void KernelMessageMonitor::ProcessMessage(std::string_view message)
{
    size_t prefixEnd = message.find(';');
    if (prefixEnd == std::string::npos)
    {
        // Malformed message
        return;
    }

    auto prefix = Utils::String::Split(message.substr(0, prefixEnd), ",");

    if (prefix.size() < 4)
    {
        // Malformed prefix
        return;
    }

    uint64_t temp = Utils::String::ToUIntT<uint64_t>(prefix[0]);

    // Extract priority, first 3 bits
    uint16_t messagePriority = static_cast<uint16_t>(temp & 0x7);
    // Extract facility, other bits
    uint64_t messageFacility =
        static_cast<uint64_t>(temp & ~(static_cast<uint64_t>(0x7)));

    uint64_t messageSeqNum = Utils::String::ToUIntT<uint64_t>(prefix[1]);

    uint64_t messageTimestamp = Utils::String::ToUIntT<uint64_t>(prefix[2]);

    // TODO prefix[3] can indicate multipart message

    std::string_view messageText = "failedtoparse";
    if (prefixEnd + 1 < message.size())
    {
        messageText = message.substr(prefixEnd + 1);
    }

    // Check config for telemetry priority filter
    if (messagePriority < 4)
    {
        TraceLoggingWrite(
            g_errorReportingClientProvider,
            "KernelMessage",
            TraceLoggingValue(messagePriority, "Priority"),
            TraceLoggingValue(messageFacility, "Facility"),
            TraceLoggingValue(messageSeqNum, "SeqNum"),
            TraceLoggingValue(messageTimestamp, "Timestamp"),
            TraceLoggingValue(messageText.data(), "Message"));
    }
}

/*++
    Signals the thread executing Run() to stop.
--*/
void KernelMessageMonitor::Stop()
{
    m_isStopping = true;

    if (m_stopDescriptor >= 0)
    {
        TraceLoggingWrite(
            g_errorReportingClientProvider, "KernelMessageMonitor_SignalStop");

        // Write some data to the stop eventfd, to unblock the poll in Run().
        uint64_t stopData = 1;
        ssize_t result = write(m_stopDescriptor, &stopData, sizeof(stopData));

        if (result <= 0)
        {
            TraceLoggingWrite(
                g_errorReportingClientProvider,
                "KernelMessageMonitor_SignalStopFailed",
                TraceLoggingValue(result, "Result"),
                TraceLoggingValue(errno, "ErrorCode"));
        }

        std::cout << "Writing stop event data\n";
    }
    else
    {
        TraceLoggingWrite(
            g_errorReportingClientProvider,
            "KernelMessageMonitor_StopDescriptorNotValid");
    }

    try
    {
        std::filesystem::remove(GetSentinelPath());
        TraceLoggingWrite(g_errorReportingClientProvider, "KernelSentinelRemoved");
    }
    CATCH_LOG();
}

std::filesystem::path KernelMessageMonitor::GetSentinelPath()
{
    std::filesystem::path configPath =
        std::filesystem::path(AZURE_DEVICE_QUEUE_FOLDER) / "sentinel";
    return configPath;
}

/*++
    Create the Sentinel file, replacing an existing one
--*/
void KernelMessageMonitor::CreateOrReplaceSentinel()
{
    try
    {
        std::filesystem::path SentinelFile = GetSentinelPath();

        // TODO replace with fsync to gurantee disk write
        std::ofstream sentOut(
            SentinelFile.string(), std::ios_base::out | std::ios_base::trunc);
        sentOut << time(nullptr) << std::endl << std::flush;
        sentOut.close();

        TraceLoggingWrite(g_errorReportingClientProvider, "KernelSentinelCreated");
    }
    CATCH_LOG();
}

/*++
    Check for Sentinel, and emit its contents as telemetry
--*/
void KernelMessageMonitor::CheckForSentinel()
{
    try
    {
        std::filesystem::path SentinelFile = GetSentinelPath();
        if (std::filesystem::exists(SentinelFile))
        {
            std::ifstream sentIn(SentinelFile.string(), std::ios_base::in);
            std::string systemTime;

            if (sentIn.peek() != EOF)
            {
                sentIn >> systemTime;
            }

            TraceLoggingWrite(
                g_errorReportingClientProvider,
                "KernelAbnormalShutdown",
                TraceLoggingValue(systemTime.c_str(), "Time"));

            // Write message to journalctl if available
            std::cout << "SENTINEL ABNORMAL SHUTDOWN: " << systemTime
                      << std::endl;
        }
    }
    CATCH_LOG();
}
}}}}
