// Copyright (c) Microsoft Corporation. All rights reserved.

#include <memory>
#include <string_view>
#include <thread>

#include <azure-device-util/AndroidLogHelper.h>
#include <azure-device-util/FileSystem.h>
#include <azure-device-util/ResultException.h>
#include <azure-device-util/ScopeExit.h>
#include <azure-device-util/SignalLoop.h>
#include <azure-device-util/TraceLoggingTypes.h>
#include <azure-device-util/Version.h>
#include <boost/process.hpp>

#ifdef AZURE_DEVICE_ERRORREPORTING_UPLOADER_USE_HTTP_UPLOADER
    #include <cpprest/version.h>
    #include <pplx/threadpool.h>

    #include "HttpErrorReportUploadService.h"
    #include "LocalPolicyManager.h"
#elif AZURE_DEVICE_ERRORREPORTING_UPLOADER_USE_BONDI_UPLOADER

    #include "BondiErrorReportUploadService.h"
    #include "LocalPolicyManager.h"

#else  // AZURE_DEVICE_ERRORREPORTING_UPLOADER_USE_HVSOCKET_UPLOADER
    #include "HvSocketErrorReportUploadService.h"
    #include "RemotePolicyManager.h"
#endif

#ifdef AZURE_DEVICE_ERRORREPORTING_UPLOADER_USE_KERNEL_MESSAGE_MONITOR
    #include "KernelDumpProcessor.h"
    #include "KernelMessageMonitor.h"
#endif

#ifdef AZURE_DEVICE_ERRORREPORTING_USE_CORE_DIRECTORY_MONITORING
    #include "CoreDirectoryMonitorService.h"
#endif

#include "Config.h"
#include "ErrorReportUploadServiceBase.h"
#include "ErrorReportUploadServiceFactory.h"
#include "ReportUploaderLogging.h"
#include "ServiceLoggingProvider.h"

using namespace Azure::Device::ErrorReporting::Uploader;
using namespace Azure::Device::Policy;
using namespace Azure::Device::Utils;

// b59d8b82-fe34-4fc0-8ca7-e5b7c9b3f342
TRACELOGGING_DEFINE_PROVIDER(
    g_errorReportingClientProvider,
    "Azure.Device.ErrorReporting.Client",
    (0xb59d8b82, 0xfe34, 0x4fc0, 0x8c, 0xa7, 0xe5, 0xb7, 0xc9, 0xb3, 0xf3, 0x42));

#ifdef AZURE_DEVICE_ERRORREPORTING_UPLOADER_USE_HTTP_UPLOADER
void InitializeStaticNetworkingThreadpool()
{
    // Set the threadpool to a more reasonable size on cpprest versions that
    // allow it
    #if (CPPREST_VERSION >= (2 * 100000 + 10 * 100 + 8))
    crossplat::threadpool::initialize_with_threads(10);
    #endif
}
#endif

/*++
    Registered as our error handling callback with the ErrorHelper from
    azure-device-utils package. Logs to LTTNG.

Parameters:
    message - The message to log.
    file - The file name (in debug builds) which generated the error.
    line - The line number (in debug builds) which generated the error.
    err - The error code generated.
--*/
static void LogCallback(
    std::string_view message,
    std::string_view file,
    int lineNumber,
    std::error_code err)
{
    TraceLoggingWrite(
        g_errorReportingClientProvider,
        "LogError",
        TraceLoggingStringView(file, "FileName"),
        TraceLoggingValue(lineNumber, "Line"),
        TraceLoggingStringView(message, "Message"),
        TraceLoggingErrorCode(err, "ErrorCode"));

    std::cerr << "LogError: ";
    std::cerr << "loc='" << file << "@" << lineNumber << "' ";
    std::cerr << "msg='" << message << "' ";
    std::cerr << "err={msg='" << err.message() << "' ";
    std::cerr << "cat='" << err.category().name() << "' ";
    std::cerr << "val='" << err.value() << "'}" << std::endl;
}

#ifdef AZURE_DEVICE_ERRORREPORTING_UPLOADER_USE_KERNEL_MESSAGE_MONITOR
void SubmitKernelDumps()
{
    constexpr std::string_view c_kernelDumpPath = "/var/crash";

    KernelDumpProcessor kdumpProcessor;

    auto reportQueuePath = std::filesystem::path(AZURE_DEVICE_QUEUE_FOLDER);
    Azure::Device::ErrorReporting::Submitter::ReportQueue queue{ reportQueuePath };

    kdumpProcessor.ProcessDirectory(queue, c_kernelDumpPath);
}
#endif

#ifdef AZURE_DEVICE_ERRORREPORTING_UPLOADER_REGISTER_CRASH_HANDLER
void RegisterCrashHandler()
{
    std::filesystem::path crashHandlerPath = FileSystem::GetBinaryDirectory();
    crashHandlerPath /= "azure-device-crash-handler";

    std::string registerCrashHandlerCommand;
    registerCrashHandlerCommand += "sysctl -w kernel.core_pattern=\"|";
    registerCrashHandlerCommand += crashHandlerPath.string();
    registerCrashHandlerCommand += " %p %i %s %E %t %c\"";

    // Register crash handler
    system(registerCrashHandlerCommand.c_str());
    system("ulimit -c unlimited");
}
#endif

/*++
    Main routine for the error reporting service.
--*/
int main()
{
    SignalLoop signalLoop{ SIGTERM, SIGINT, SIGHUP };

    std::cout.setf(std::ios::unitbuf);

    try
    {
        // Initialize logging for the service and error reporting libraries.
        TraceLoggingRegister(g_errorReportingClientProvider);
        EnableLogging();

        // Register error callback
        SetResultLoggingCallback(&LogCallback);

        std::string_view version = Version::GetString();
        TraceLoggingWrite(
            g_errorReportingClientProvider,
            "Main_Startup",
            TraceLoggingCountedString(version.data(), version.size(), "Version"));

        ADHS_LOG(
            "Starting azure-device-errorreporting-uploader with version %s",
            version.data());

        std::cout
            << "Staring azure-device-errorreporting-uploaderd with version: "
            << version << std::endl;

#ifdef AZURE_DEVICE_ERRORREPORTING_UPLOADER_REGISTER_CRASH_HANDLER
        RegisterCrashHandler();
#endif

        IPolicyManager* policyManager = nullptr;
        std::unique_ptr<ErrorReportUploadServiceBase> uploadService;
#ifdef AZURE_DEVICE_ERRORREPORTING_UPLOADER_USE_HTTP_UPLOADER
        InitializeStaticNetworkingThreadpool();
        LocalPolicyManager localPolicyManager;
        policyManager = &localPolicyManager;
        uploadService =
            ErrorReportUploadServiceFactory::CreateHttpErrorReportUploadService(
                localPolicyManager);
#elif AZURE_DEVICE_ERRORREPORTING_UPLOADER_USE_BONDI_UPLOADER
        LocalPolicyManager localPolicyManager;
        policyManager = &localPolicyManager;
        uploadService =
            ErrorReportUploadServiceFactory::CreateBondiErrorReportUploadService(
                localPolicyManager);
#else  // AZURE_DEVICE_ERRORREPORTING_UPLOADER_USE_HVSOCKET_UPLOADER
        RemotePolicyManager remotePolicyManager;
        policyManager = &remotePolicyManager;
        uploadService =
            ErrorReportUploadServiceFactory::CreateHvSocketErrorReportUploadService(
                remotePolicyManager);
#endif

#ifdef AZURE_DEVICE_ERRORREPORTING_USE_CORE_DIRECTORY_MONITORING
        static const std::string coreDirectory = "/data/cores";

        // Relabel selinux coredump_file object to our core directory
        boost::process::system(
            "chcon u:object_r:coredump_file:s0 " + coreDirectory);

        CoreDirectoryMonitorService coreDirectoryMonitorService{ coreDirectory,
                                                                 *policyManager };
        std::thread coreDirectoryRunLoopThread{
            &CoreDirectoryMonitorService::Run,
            std::ref(coreDirectoryMonitorService)
        };

        auto joinCoreDirectoryRunLoopThread = MakeScopeExit([&] {
            coreDirectoryMonitorService.Stop();
            coreDirectoryRunLoopThread.join();
        });
#else
        (void) policyManager;
#endif

        std::thread errorReportUploadRunLoopThread{
            &ErrorReportUploadServiceBase::Run, std::ref(*uploadService)
        };
        auto joinErrorReportUploadRunLoopThread = MakeScopeExit([&] {
            uploadService->Stop();
            errorReportUploadRunLoopThread.join();
        });

#ifdef AZURE_DEVICE_ERRORREPORTING_UPLOADER_USE_KERNEL_MESSAGE_MONITOR
        KernelMessageMonitor kernelMonitor;
        kernelMonitor.Run();
        std::thread submitKernelDumpThread{ &SubmitKernelDumps };
        auto joinKernelDumpThread = MakeScopeExit([&] {
            kernelMonitor.Stop();
            submitKernelDumpThread.join();
        });
#endif

        auto shutdownHandler = [&](int signal) {
            std::cout << "Shutting down due to signal " << signal << std::endl;
            ADHS_LOG("Shutting down due to signal %d", signal);
            return false;
        };

        auto reloadHandler = [&](int signal) {
            std::cout << "Reloading configuration because of signal " << signal
                      << std::endl;
            ADHS_LOG("Reloading configuration because of signal %d", signal);

            uploadService->UpdateTelemetryPermission();

            return true;
        };

        signalLoop.RegisterSignalHandler(SIGTERM, shutdownHandler);
        signalLoop.RegisterSignalHandler(SIGINT, shutdownHandler);
        signalLoop.RegisterSignalHandler(SIGHUP, reloadHandler);

        // This will return gracefully on SIGINT or SIGTERM
        signalLoop.Run();
    }
    CATCH_LOG();

    TraceLoggingWrite(g_errorReportingClientProvider, "ServiceStopped");
    ADHS_LOG("Service stopped");

    // Cleanup logging for the service and error reporting libraries.
    TraceLoggingUnregister(g_errorReportingClientProvider);
    DisableLogging();

    return 0;
}
