// Copyright (c) Microsoft Corporation. All rights reserved.

#include "ErrorReportUploadServiceBase.h"

#include <iostream>
#include <map>
#include <memory>
#include <random>
#include <stdlib.h>
#include <thread>

#include <azure-device-util/AndroidLogHelper.h>
#include <azure-device-util/ErrorHelper.h>
#include <azure-device-util/FileSystem.h>
#include <azure-device-util/Identity.h>
#include <azure-device-util/ResultException.h>
#include <azure-device-util/ScopeExit.h>
#include <azure-device-util/UuidBridge.h>
#include <tracelogging/TraceLoggingProvider.h>
#include <uuid/uuid.h>

#include "Config.h"
#include "ServiceLoggingProvider.h"

using namespace Azure::Device::Policy;
using namespace Azure::Device::Utils;

namespace Azure { namespace Device { namespace ErrorReporting {
namespace Uploader {

ErrorReportUploadServiceBase::ErrorReportUploadServiceBase(
    IPolicyManager& policyManager)
    : m_reportQueueWatcher(this), m_policyManager(policyManager)
{
    UpdateTelemetryPermission();

    //
    // Create a report queue in AZURE_DEVICE_QUEUE_FOLDER, if it does not
    // already exist. All users will share one queue. For the top-level queue
    // directory:
    //   Everyone has read/execute access, to determine total queue size.
    //   Only root has write access.
    // For the pending reports directory:
    //   Everyone has full access, to determine total queue size, create report
    //   directories, and delete report directories.
    // For the submitted reports directory:
    //   Everyone has read/write/execute access, to determine total queue size
    //   and create report directories. Only root has access to delete report
    //   directories (sticky_bit).
    //
    std::filesystem::perms queuePermissions = std::filesystem::perms::owner_all |
        std::filesystem::perms::group_read | std::filesystem::perms::group_exec |
        std::filesystem::perms::others_read |
        std::filesystem::perms::others_exec;

    std::filesystem::path queuePath =
        std::filesystem::path(AZURE_DEVICE_QUEUE_FOLDER);

    m_reportQueue = Submitter::ReportQueue::CreateReportQueue(
        queuePath,
        queuePermissions,
        std::filesystem::perms::all | std::filesystem::perms::sticky_bit,
        std::filesystem::perms::all | std::filesystem::perms::sticky_bit);

    // Register for notifications when reports are submitted to the queue.
    m_reportQueueWatcher.AddMonitoredPath(
        m_reportQueue->GetSubmittedReportsPath().string());

    ADHS_LOG(
        "ErrorReportUploadServiceBase is watching path, %s",
        m_reportQueue->GetSubmittedReportsPath().c_str());

    // Get configuration data.  If this fails, use default configuration.
    try
    {
        UploadConfig config = m_uploadConfigManager.GetUploadConfig();
        m_reportQueueWatcher.SetPollIntervalSeconds(
            config.UploadPollIntervalSeconds);
    }
    CATCH_LOG();
}

/*++
    Update telemetry permission
--*/
void ErrorReportUploadServiceBase::UpdateTelemetryPermission()
{
    m_policyManager.Update();
}

/*++
    Callback to execute when changes to the report queue are detected.

Arguments:
    events - Contains details of each change that was detected in a single
    polling interval.
--*/
void ErrorReportUploadServiceBase::INotifyMonitorCallback(
    std::vector<INotifyMonitorEvent> events)
{
    TraceLoggingWrite(
        g_errorReportingClientProvider,
        "ErrorReport_FileSystemChangeCallbackStart");

    // Log details about each event for diagnostic purposes.
    for (const auto& event : events)
    {
        TraceLoggingWrite(
            g_errorReportingClientProvider,
            "ErrorReport_FileSystemChangeEventDetails",
            TraceLoggingValue(event.Name.c_str(), "Name"),
            TraceLoggingValue(event.Mask, "Mask"),
            TraceLoggingValue(event.Cookie, "Cookie"));
    }

    // Process the queue to upload newly submitted reports.
    ProcessReportQueue(m_reportQueue->GetSubmittedReportsPath());

    TraceLoggingWrite(
        g_errorReportingClientProvider,
        "ErrorReport_FileSystemChangeCallbackStop");
}

/*++
    Begin to monitor the report queue for changes.  This will block the calling
    thread until Stop() is called.
--*/
void ErrorReportUploadServiceBase::Run()
{
    m_reportQueueWatcher.Run();
}

/*++
    Stop monitoring the report queue for changes.  This will unblock the thread
    that called Run().
--*/
void ErrorReportUploadServiceBase::Stop()
{
    TraceLoggingWrite(
        g_errorReportingClientProvider, "ErrorReport_UploadServiceStop");
    m_reportQueueWatcher.Stop();
}

/*++
    Attempts to upload all of the submitted reports in a queue.

Arguments:
    queuePath - The path of the submitted report queue to process.
--*/
void ErrorReportUploadServiceBase::ProcessReportQueue(
    const std::filesystem::path& queuePath)
{
    try
    {
        TraceLoggingWrite(
            g_errorReportingClientProvider,
            "ProcessReportQueueStart",
            TraceLoggingValue(queuePath.string().c_str(), "QueuePath"));

        // Upload each submitted report in the queue.
        for (auto& item : std::filesystem::directory_iterator(queuePath))
        {
            //
            // We only expect to find subdirectories for individual reports in
            // the queue path.
            //
            if (!std::filesystem::is_directory(item.path()))
            {
                TraceLoggingWrite(
                    g_errorReportingClientProvider,
                    "UnexpectedItemInQueue",
                    TraceLoggingValue(item.path().string().c_str(), "Path"));

                continue;
            }

            if (m_policyManager.GetTelemetryPermission() >=
                Utils::TelemetryPermission::Required)
            {
                UploadReport(item.path());
            }

            //
            // Delete the report after the upload attempt.  This
            // proof-of-concept code doesn't handle upload errors, archiving
            // reports, etc. That support will come later.
            //
            {
                std::error_code error;
                std::filesystem::remove_all(item.path(), error);

                TraceLoggingWrite(
                    g_errorReportingClientProvider,
                    "DeleteReport",
                    TraceLoggingValue(item.path().c_str(), "ReportPath"),
                    TraceLoggingValue(error.value(), "ErrorCode"));
            }
        }

        TraceLoggingWrite(
            g_errorReportingClientProvider, "ProcessReportQueueStop");
    }
    CATCH_LOG();
}

}}}}