// Copyright (c) Microsoft Corporation. All rights reserved.

#include "ProtocolXmlDocument.h"

#include <iomanip>
#include <map>
#include <memory>
#include <pugixml.hpp>
#include <sstream>
#include <string>

#include "ProtocolArgumentSet.h"

namespace Azure { namespace Device { namespace ErrorReporting {
namespace Uploader {

ProtocolXmlDocument::ProtocolXmlDocument(
    std::shared_ptr<ProtocolArgumentSet> protocolArgs)
    : m_protocolArgs(protocolArgs)
{
    Initialize();
}

/*++
    Creates the basic XML structure for the Watson protocol.
--*/
void ProtocolXmlDocument::Initialize()
{
    // <req><tlm><src><desc><mach>
    pugi::xml_node reqNode = m_xmlDoc.append_child("req");
    reqNode.append_attribute("ver") = "2";

    pugi::xml_node tlmNode = reqNode.append_child("tlm");
    pugi::xml_node srcNode = tlmNode.append_child("src");
    pugi::xml_node descNode = srcNode.append_child("desc");
    pugi::xml_node machNode = descNode.append_child("mach");

    // <req><tlm><src><desc><mach><os>
    AddProtocolSection("os", machNode);

    // <req><tlm><src><desc><mach><hw>
    AddProtocolSection("hw", machNode);

    // <req><tlm><src><desc><mach><ctrl>
    AddProtocolSection("ctrl", machNode);

    // <req><tlm><reqs><req>
    m_reqsNode = tlmNode.append_child("reqs");

    m_reqNode = m_reqsNode.append_child("req");
    m_reqNode.append_attribute("key") = "1";
}

/*++
    Sets the namesapce for a Watson protocol request.

Arguments:
    svc - The service name for the request
    partner - The partner name for the request
    group - The group name for the request
    app - The application name for the request
--*/
void ProtocolXmlDocument::SetNamespace(
    const std::string& svc,
    const std::string& partner,
    const std::string& group,
    const std::string& app)
{
    // <req><tlm><reqs><req><namespace>
    pugi::xml_node namespaceNode = m_reqNode.append_child("namespace");
    namespaceNode.append_attribute("svc") = svc.c_str();
    namespaceNode.append_attribute("ptr") = partner.c_str();
    namespaceNode.append_attribute("gp") = group.c_str();
    namespaceNode.append_attribute("app") = app.c_str();
}

/*++
    Adds a command to a Watson protocol request.

Arguments:
    name - The name of the command to add.
    args - Supplies all protocol argument (key,value) pairs associated with the
    command.
--*/
void ProtocolXmlDocument::AddCommand(
    const std::string& name,
    const std::map<std::string, std::string>& args)
{
    // <req><tlm><reqs><req><cmd>
    pugi::xml_node cmdNode = m_reqNode.append_child("cmd");
    cmdNode.append_attribute("nm") = name.c_str();

    for (const auto& arg : args)
    {
        pugi::xml_node argNode = cmdNode.append_child("arg");
        argNode.append_attribute("nm") = arg.first.c_str();
        argNode.append_attribute("val") = arg.second.c_str();
    }
}

/*++
    Adds metadata about the payload of a Watson protocol request.

Arguments:
    size - The size of the request's payload, in bytes.
--*/
void ProtocolXmlDocument::AddPayload(const std::string& size)
{
    // <req><tlm><reqs><payload>
    pugi::xml_node payloadNode = m_reqsNode.append_child("payload");

    pugi::xml_node payloadSizeArgNode = payloadNode.append_child("arg");
    payloadSizeArgNode.append_attribute("nm") = "size";
    payloadSizeArgNode.append_attribute("val") = size.c_str();

    pugi::xml_node compArgNode = payloadNode.append_child("arg");
    compArgNode.append_attribute("nm") = "comp";
    compArgNode.append_attribute("val") = "zip";
}

/*++
    Adds a section to the Watson protocol request XML document, including all
    arguments registered for that section.

Arguments:
    sectionName - The name of the section to add to the Watson protocol request.
    parentNode - The parent node of the Watson protocol request to which the
    section will be added.
--*/
void ProtocolXmlDocument::AddProtocolSection(
    const std::string& sectionName,
    pugi::xml_node parentNode)
{
    pugi::xml_node sectionNode = parentNode.append_child(sectionName.c_str());
    for (const auto& arg : m_protocolArgs->GetProtocolArgsForSection(sectionName))
    {
        pugi::xml_node argNode = sectionNode.append_child("arg");
        argNode.append_attribute("nm") = arg.first.c_str();
        argNode.append_attribute("val") = arg.second.c_str();
    }
}

/*++
    Serializes the Watson protocol XML document to a string.
--*/
std::string ProtocolXmlDocument::ToProtocolString()
{
    std::stringstream protocolDocumentStream;

    // Reserve 4 bytes to hold the size of the XML document.
    protocolDocumentStream << std::setw(4) << '\0';

    // Save the XML document to a stream, and extract the string to return to
    // the caller.
    m_xmlDoc.save(protocolDocumentStream, "", pugi::format_raw);
    std::string protocolDocumentString = protocolDocumentStream.str();

    // Update the first 4 bytes to represent the actual size of the XML
    // document, in little-endian.
    size_t xmlSize = protocolDocumentString.size() - 4;
    protocolDocumentString[0] = static_cast<unsigned char>(xmlSize & 0xff);
    protocolDocumentString[1] = static_cast<unsigned char>(xmlSize >> 8 & 0xff);
    protocolDocumentString[2] = static_cast<unsigned char>(xmlSize >> 16 & 0xff);
    protocolDocumentString[3] = static_cast<unsigned char>(xmlSize >> 24 & 0xff);

    return protocolDocumentString;
}

}}}}
