// Copyright (c) Microsoft Corporation. All rights reserved.

#include <iostream>
#include <map>
#include <memory>
#include <random>
#include <stdlib.h>
#include <thread>

#include <azure-device-util/ErrorHelper.h>
#include <azure-device-util/FileSystem.h>
#include <azure-device-util/Identity.h>
#include <azure-device-util/ResultException.h>
#include <azure-device-util/ScopeExit.h>
#include <azure-device-util/UuidBridge.h>
#include <tracelogging/TraceLoggingProvider.h>
#include <uuid/uuid.h>

#include "Config.h"
#include "ErrorReportUploadServiceBase.h"
#include "ServiceLoggingProvider.h"

using namespace Azure::Device::Utils;

namespace Azure { namespace Device { namespace ErrorReporting {
namespace Uploader {

/*++
    Reads uploader configuration settings from disk.

Arguments:
    config - Receives the uploader configuration settings.
--*/
void UploadConfigManager::ReadUploadConfig(UploadConfig& config)
{
    try
    {
        // Read config data from disk.
        std::map<std::string, std::string, std::less<>> configData;

        std::filesystem::path configPath =
            std::filesystem::path(AZURE_DEVICE_QUEUE_FOLDER) / "config";

        std::error_code result = Utils::FileSystem::ReadDictionaryFromFile(
            configPath.string(), ",", configData);
        if (FAILED(result))
        {
            TraceLoggingWrite(
                g_errorReportingClientProvider,
                "FailedToReadConfigData",
                TraceLoggingValue(result.value(), "ErrorCode"));

            return;
        }

        // Enumerate all config data, looking for known settings.
        for (const auto& pair : configData)
        {
            std::string key = std::move(pair.first);
            std::string value = std::move(pair.second);

            TraceLoggingWrite(
                g_errorReportingClientProvider,
                "ConfigDataEntry",
                TraceLoggingValue(key.c_str(), "key"),
                TraceLoggingValue(value.c_str(), "value"));

            if (key == "deviceId")
            {
                // TODO 31695739: Use the same device id as
                // azure-device-telemetry
                config.DeviceId = Utils::Identity::GetMachineBasedDeviceId(false);
                if (config.DeviceId.empty())
                {
                    config.DeviceId = std::move(value);
                }
            }
            else if (key == "sampleNumber")
            {
                config.SampleNumber = strtoul(value.c_str(), nullptr, 0);
            }
            else if (key == "proxy")
            {
                config.Proxy = std::move(value);
            }
            else if (key == "disableCertificateValidation" && value == "true")
            {
                config.DisableCertificateValidation = true;
            }
            else if (key == "uploadPollIntervalSeconds")
            {
                config.UploadPollIntervalSeconds =
                    strtoul(value.c_str(), nullptr, 0);
            }
            else
            {
                TraceLoggingWrite(
                    g_errorReportingClientProvider,
                    "UnexpectedConfigKey",
                    TraceLoggingValue(key.c_str(), "key"));
            }
        }
    }
    CATCH_LOG();
}

/*++
    Writes uploader configuration settings to disk.

Arguments:
    config - The active uploader configuration settings.
--*/
void UploadConfigManager::WriteUploadConfig(UploadConfig& config)
{
    try
    {
        TraceLoggingWrite(g_errorReportingClientProvider, "UpdateUploadConfig");

        std::filesystem::path configPath =
            std::filesystem::path(AZURE_DEVICE_QUEUE_FOLDER) / "config";

        std::ofstream configStream(
            configPath.string(), std::ios_base::out | std::ios_base::trunc);

        configStream << "deviceId," << config.DeviceId << "\n"
                     << "sampleNumber," << std::to_string(config.SampleNumber)
                     << "\n";
    }
    CATCH_LOG();
}

/*++
    Returns the active uploader configuration settings.
--*/
UploadConfig UploadConfigManager::GetUploadConfig()
{
    // Read any existing config info from disk.
    UploadConfig config;
    ReadUploadConfig(config);

    // Initialize any required config items that weren't found.
    bool configChanged = false;

    if (config.DeviceId.empty())
    {
        uuid_t uuid;
        uuid_generate_time_safe(uuid);
        config.DeviceId.resize(36);
        uuid_unparse_lower(uuid, &config.DeviceId[0]);

        configChanged = true;

        TraceLoggingWrite(
            g_errorReportingClientProvider,
            "GenerateDeviceId",
            TraceLoggingValue(config.DeviceId.c_str(), "DeviceId"));
    }

    if (config.SampleNumber == 0)
    {
        std::random_device device;
        std::uniform_int_distribution<uint32_t> distribution(1, 100000000);
        config.SampleNumber = distribution(device);

        configChanged = true;

        TraceLoggingWrite(
            g_errorReportingClientProvider,
            "GenerateSampleNumber",
            TraceLoggingValue(config.SampleNumber, "SampleNumber"));
    }

    //
    // If any of the required configuration settings weren't present, write the
    // new configuration to disk.
    //
    if (configChanged)
    {
        WriteUploadConfig(config);
    }

    TraceLoggingWrite(
        g_errorReportingClientProvider,
        "ActiveUploadConfig",
        TraceLoggingValue(config.DeviceId.c_str(), "DeviceId"),
        TraceLoggingValue(config.SampleNumber, "SampleNumber"),
        TraceLoggingValue(config.Proxy.c_str(), "Proxy"),
        TraceLoggingValue(
            config.DisableCertificateValidation, "DisableCertificateValidation"),
        TraceLoggingValue(
            config.UploadPollIntervalSeconds, "UploadPollIntervalSeconds"));

    return config;
}

}}}}
