// Copyright (c) Microsoft Corporation. All rights reserved.

#include "HttpErrorReportUploadService.h"

#include <iostream>
#include <map>
#include <memory>
#include <stdlib.h>
#include <thread>

#include <azure-device-util/ErrorHelper.h>
#include <azure-device-util/FileSystem.h>
#include <azure-device-util/Identity.h>
#include <azure-device-util/ResultException.h>
#include <azure-device-util/ScopeExit.h>
#include <tracelogging/TraceLoggingProvider.h>
#include <uuid/uuid.h>

#include "ServiceLoggingProvider.h"

using namespace Azure::Device::Policy;
using namespace Azure::Device::Utils;

namespace Azure { namespace Device { namespace ErrorReporting {
namespace Uploader {

HttpErrorReportUploadService::HttpErrorReportUploadService(
    LocalPolicyManager& localPolicyManager)
    : ErrorReportUploadServiceBase(localPolicyManager)
{}

HttpErrorReportUploadService::~HttpErrorReportUploadService() {}

/*++
    Adds OS version information to a report.

Arguments:
    report - The report to receive OS version information.
--*/
void HttpErrorReportUploadService::AddOsVersion(HttpReportUploader& report)
{
    try
    {
        std::string osPlatform = "linux";

        // Read release data from /etc/os-release
        std::map<std::string, std::string, std::less<>> releaseData;
        std::error_code result = Utils::FileSystem::ReadDictionaryFromFile(
            "/etc/os-release", "=", releaseData);
        if (FAILED(result))
        {
            TraceLoggingWrite(
                g_errorReportingClientProvider,
                "FailedToReadReleaseData",
                TraceLoggingValue(result.value(), "ErrorCode"));

            return;
        }

        // Enumerate all release data, looking for version or id.
        for (const auto& pair : releaseData)
        {
            std::string key = std::move(pair.first);
            std::string value = std::move(pair.second);

            TraceLoggingWrite(
                g_errorReportingClientProvider,
                "ReleaseDataEntry",
                TraceLoggingValue(key.c_str(), "key"),
                TraceLoggingValue(value.c_str(), "value"));

            if (key == "VERSION")
            {
                //
                // We expect the version string to be formatted like:
                // "major.minor.build description"
                // The string must at least be large enough to contain a pair of
                // quotes and a number.
                //
                if (value.size() < 3)
                {
                    continue;
                }

                std::string version = value.substr(1, value.find(' '));

                std::vector<std::string> versionParts;
                std::istringstream versionStream(version);
                std::string current;
                while (getline(versionStream, current, '.'))
                {
                    versionParts.push_back(current);
                }

                // If we found the expected 3-part version, add it to the report.
                if (versionParts.size() == 3u)
                {
                    report.AddProtocolArg("os", "vermaj", versionParts[0]);
                    report.AddProtocolArg("os", "vermin", versionParts[1]);
                    report.AddProtocolArg("os", "verbld", versionParts[2]);
                    report.AddProtocolArg("os", "vercsdbld", "0");
                }
            }
            else if (key == "ID")
            {
                // We found a more decriptive OS platform.
                // The final value will look like "linux-<ID>"
                osPlatform.push_back('-');
                osPlatform += value;
            }
        }

        // Add the OS platform to the report.
        report.AddProtocolArg("os", "os", osPlatform);
    }
    CATCH_LOG();
}

/*++
    Attempts to upload a submitted report.

Arguments:
    reportPath - The path of the submitted report to upload.
--*/
void HttpErrorReportUploadService::UploadReport(const std::filesystem::path& reportPath)
{
    try
    {
        uuid_t oldActivity, newActivity;
        lttngh_ActivityIdGet(oldActivity);
        lttngh_ActivityIdCreate(newActivity);
        lttngh_ActivityIdSet(newActivity);

        auto resetActivityId = Utils::MakeScopeExit(
            [&oldActivity]() { lttngh_ActivityIdSet(oldActivity); });

        TraceLoggingWrite(
            g_errorReportingClientProvider,
            "UploadReportStart",
            TraceLoggingValue(reportPath.c_str(), "ReportPath"));

        // Create a report uploader for the target path.
        auto reportUploader = std::make_unique<HttpReportUploader>(reportPath);
        auto params = reportUploader->GetBucketingParameters();

        TraceLoggingWrite(
            g_errorReportingClientProvider,
            "ReportDetails",
            TraceLoggingValue(reportUploader->GetReportId(), "ReportId"),
            TraceLoggingValue(reportUploader->GetEventType().c_str(), "EventType"),
            TraceLoggingValue((params.size() > 0 ? params[0].c_str() : ""), "P0"),
            TraceLoggingValue((params.size() > 1 ? params[1].c_str() : ""), "P1"),
            TraceLoggingValue((params.size() > 2 ? params[2].c_str() : ""), "P2"),
            TraceLoggingValue((params.size() > 3 ? params[3].c_str() : ""), "P3"),
            TraceLoggingValue((params.size() > 4 ? params[4].c_str() : ""), "P4"),
            TraceLoggingValue((params.size() > 5 ? params[5].c_str() : ""), "P5"),
            TraceLoggingValue((params.size() > 6 ? params[6].c_str() : ""), "P6"),
            TraceLoggingValue((params.size() > 7 ? params[7].c_str() : ""), "P7"),
            TraceLoggingValue((params.size() > 8 ? params[8].c_str() : ""), "P8"),
            TraceLoggingValue((params.size() > 9 ? params[9].c_str() : ""), "P9"));

        // Upload the report.
        UploadConfig config = m_uploadConfigManager.GetUploadConfig();

        // Get the proxy information for the report from the environment if it exists,
        // otherwise get it from the upload config
        const char* httpsProxy = std::getenv("HTTPS_PROXY");
        if (httpsProxy == nullptr)
        {
            httpsProxy = std::getenv("https_proxy");
        }

        if (httpsProxy == nullptr || !reportUploader->SetProxy(httpsProxy))
        {
            reportUploader->SetProxy(config.Proxy);
        }

        reportUploader->SetValidateCertificates(
            !config.DisableCertificateValidation);
        reportUploader->AddProtocolArg("ctrl", "mid", config.DeviceId);
        reportUploader->AddProtocolArg(
            "ctrl", "sample", std::to_string(config.SampleNumber));

        AddOsVersion(*reportUploader);

        HttpReportUploader::UploadType uploadType =
            (m_policyManager.GetTelemetryPermission() ==
             TelemetryPermission::Optional) ?
            HttpReportUploader::UploadType::DumpUpload :
            HttpReportUploader::UploadType::BucketHitOnly;

        reportUploader->Upload(uploadType);

        // Log whether or not the buket hit and cab were received by the server.
        if (reportUploader->BucketHitReceived())
        {
            TraceLoggingWrite(
                g_errorReportingClientProvider,
                "BucketHitReceived",
                TraceLoggingValue(reportUploader->GetReportId(), "ReportId"),
                TraceLoggingValue(reportUploader->GetBucket().c_str(), "Bucket"),
                TraceLoggingValue(
                    reportUploader->GetBucketTable().c_str(), "BucketTbl"));
        }

        if (reportUploader->CabRequested())
        {
            TraceLoggingWrite(
                g_errorReportingClientProvider,
                "CabRequested",
                TraceLoggingValue(reportUploader->GetReportId(), "ReportId"),
                TraceLoggingValue(reportUploader->GetToken().c_str(), "Token"));
        }

        if (reportUploader->CabReceived())
        {
            TraceLoggingWrite(
                g_errorReportingClientProvider,
                "CabReceived",
                TraceLoggingValue(reportUploader->GetReportId(), "ReportId"),
                TraceLoggingValue(reportUploader->GetCabId().c_str(), "CabGuid"));
        };

        TraceLoggingWrite(
            g_errorReportingClientProvider,
            "UploadReportStop",
            TraceLoggingValue(reportUploader->GetReportId(), "ReportId"),
            TraceLoggingValue(reportPath.c_str(), "ReportPath"));
    }
    CATCH_LOG();

    std::error_code error;
    std::filesystem::remove_all(reportPath, error);
}

}}}}
