// Copyright (c) Microsoft Corporation. All rights reserved.

#include "HttpReportUploader.h"

#include <filesystem>
#include <fstream>
#include <iomanip>
#include <limits>
#include <memory>
#include <pugixml.hpp>
#include <random>
#include <stdexcept>
#include <string>
#include <string_view>
#include <vector>

#include <azure-device-util/String.h>
#include <cpprest/filestream.h>
#include <cpprest/http_client.h>
#include <pplx/pplx.h>

#include "ProtocolXmlDocument.h"
#include "ReportUploaderProvider.h"
#include "XmlUtils.h"

using namespace web;                   // Common features like URIs.
using namespace web::http;             // Common HTTP functionality
using namespace web::http::client;     // HTTP client features
using namespace concurrency::streams;  // Asynchronous streams

namespace Azure { namespace Device { namespace ErrorReporting {
namespace Uploader {

HttpReportUploader::HttpReportUploader(const std::filesystem::path& reportPath)
    : m_reportPath(reportPath)
    , m_protocolArgs(std::make_shared<ProtocolArgumentSet>())
    , m_validateCertificates(true)
{
    m_reportId = Utils::String::ToUIntT<uint64_t>(reportPath.filename().string());
    ParseMetadata();
}

/*++
    Uploads a report to Watson.
--*/
void HttpReportUploader::Upload(UploadType uploadType)
{
    // Ensure that we have the required control arguments for Watson to accept
    // the request.
    if (!m_protocolArgs->IsProtocolArgSet("ctrl", "mid"))
    {
        throw std::invalid_argument(
            "deviceId (ctrl.mid) protocol argument must be set");
    }

    if (!m_protocolArgs->IsProtocolArgSet("ctrl", "sample"))
    {
        throw std::invalid_argument(
            "sampleNumber (ctrl.sample) protocol argument must be set");
    }

    uint32_t sampleNumber = strtoul(
        m_protocolArgs->GetProtocolArg("ctrl", "sample").c_str(), nullptr, 0);
    if (sampleNumber == 0 || sampleNumber > 100000000)
    {
        throw std::invalid_argument(
            "sampleNumber (ctrl.sample) must be in [1,100000000]");
    }

    // Create an http client for Watson requests.
    http_client_config config;
    config.set_validate_certificates(m_validateCertificates);

    if (!m_proxy.empty())
    {
        web_proxy proxy(m_proxy);
        config.set_proxy(proxy);
    }

    http_client client(U("https://nw-umwatson.events.data.microsoft.com"), config);
    uri_builder builder(U("/Telemetry.Request"));

    // Send the 'event' request to Watson.
    http_request request(methods::POST);
    request.set_request_uri(builder.to_uri());
    std::string requestBody = CreateEventRequestMessage();
    request.set_body(requestBody);

    TraceLoggingWrite(
        g_errorReportingUploaderProvider,
        "SendEventRequest",
        TraceLoggingValue(m_proxy.c_str(), "Proxy"),
        TraceLoggingBinary(requestBody.c_str(), 4, "RequestBodySizeBytes"),
        TraceLoggingString(&requestBody[4], "RequestBodyXml"));

    http_response httpResponse;
    httpResponse = client.request(request).get();

    // Parse the response to determine whether or not a cab has been requested.
    ParseEventResponse(httpResponse);

    // If no cab was requested, we're done.
    if (!CabRequested() || uploadType == UploadType::BucketHitOnly)
    {
        return;
    }

    // A cab was requested, so prepare the payload by combining a request XML
    // document with the report's zip file.
    std::filesystem::path zipPath = m_reportPath / "report.zip";
    uintmax_t zipSize = std::filesystem::file_size(zipPath);
    std::string uploadRequestBody = CreateDataUploadRequestMessage(zipSize);

    std::filesystem::path requestMessageDumpPath =
        m_reportPath / "datauploadreq.bin";
    std::ofstream dumpStream(
        requestMessageDumpPath.string().c_str(),
        std::ios_base::out | std::ios::binary);

    // Add the xml document
    dumpStream << uploadRequestBody;

    //
    // Add the payload.  To avoid a file copy, this should eventually be
    // replaced by a custom concurrency stream that combines XML and payload.
    //
    std::ifstream cabStream(
        zipPath.string().c_str(), std::ios_base::in | std::ios::binary);
    cabStream.seekg(0, cabStream.beg);
    dumpStream << cabStream.rdbuf();

    cabStream.close();
    dumpStream.close();

    // Send the 'dataupload' request to Watson.
    http_request uploadRequest(methods::POST);
    uploadRequest.set_request_uri(builder.to_uri());
    uintmax_t contentLength = uploadRequestBody.size() + zipSize;
    uploadRequest.set_body(
        Concurrency::streams::fstream::open_istream(
            requestMessageDumpPath.string().c_str())
            .get(),
        contentLength);

    TraceLoggingWrite(
        g_errorReportingUploaderProvider,
        "SendDataUploadRequest",
        TraceLoggingValue(m_proxy.c_str(), "Proxy"),
        TraceLoggingBinary(uploadRequestBody.c_str(), 4, "RequestBodySizeBytes"),
        TraceLoggingString(&uploadRequestBody[4], "RequestBodyXml"),
        TraceLoggingValue(contentLength, "ContentLength"));

    //
    // Create a second http client, to avoid a crash in cpprestsdk if a proxy
    // like Fiddler is attached. See if we can get rid of this later.
    //
    http_client client2(
        U("https://nw-umwatson.events.data.microsoft.com"), config);
    http_response uploadResponse = client2.request(uploadRequest).get();

    // Parse the response to determine whether or not the cab was received.
    ParseDataUploadResponse(uploadResponse);
}

/*++
    Serializes an 'event' request to a Watson protocol string.
--*/
std::string HttpReportUploader::CreateEventRequestMessage()
{
    ProtocolXmlDocument requestDoc(m_protocolArgs);

    requestDoc.SetNamespace("watson", "generic", "generic", m_eventType);

    std::map<std::string, std::string> commandArgs;
    commandArgs.emplace("cat", "generic");
    commandArgs.emplace("eventtype", m_eventType);

    for (size_t i = 0; i < m_bucketingParameters.size(); i++)
    {
        std::string parameterName = "p" + std::to_string(i + 1);
        commandArgs.emplace(parameterName, m_bucketingParameters[i]);
    }

    requestDoc.AddCommand("event", commandArgs);

    return requestDoc.ToProtocolString();
}

/*++
    Serializes a 'dataupload' request to a Watson protocol string.
--*/
std::string HttpReportUploader::CreateDataUploadRequestMessage(size_t size)
{
    ProtocolXmlDocument requestDoc(m_protocolArgs);

    requestDoc.SetNamespace("watson", "generic", "generic", m_eventType);

    std::string payloadSize = std::to_string(size);
    requestDoc.AddPayload(payloadSize);

    std::map<std::string, std::string> commandArgs;
    commandArgs.emplace("eventtype", m_eventType);
    commandArgs.emplace("bucket", m_bucket);
    commandArgs.emplace("buckettbl", m_buckettbl);
    commandArgs.emplace("size", payloadSize);
    commandArgs.emplace("token", m_token);

    requestDoc.AddCommand("dataupload", commandArgs);

    return requestDoc.ToProtocolString();
}

/*++
    Parses an 'event' response from Watson XML.
--*/
void HttpReportUploader::ParseEventResponse(http_response httpResponse)
{
    // Extract the XML string from the response.
    std::string response = httpResponse.extract_string().get();

    TraceLoggingWrite(
        g_errorReportingUploaderProvider,
        "WatsonEventResponse",
        TraceLoggingValue(httpResponse.status_code(), "StatusCode"),
        TraceLoggingValue(response.c_str(), "Response"));

    // We only expect 200 responses from Watson.
    if (httpResponse.status_code() != status_codes::OK)
    {
        throw http_exception(
            httpResponse.status_code(),
            "The server returned an unexpected status code.");
    }

    // A 200 response with an empty body implies that the request was malformed.
    if (response.empty())
    {
        throw http_exception("The server rejected the request XML.");
    }

    // Parse the response into an XML document.
    pugi::xml_document responseDoc;
    pugi::xml_parse_result result = responseDoc.load_string(response.c_str());
    if (!result)
    {
        throw make_xml_parse_error(result);
    }

    // Look for a 'receipt' command node in the response document.  This node
    // indicates that the bucket hit was received.
    pugi::xpath_node_set receiptArgs =
        responseDoc.select_nodes("/resp/tlm/resps/resp/cmd[@nm='receipt']/arg");
    for (pugi::xpath_node arg : receiptArgs)
    {
        // Extract the bucket and bucket table from the command node.  These may
        // be needed later for a cab upload.
        pugi::xml_node argNode = arg.node();
        std::string name = argNode.attribute("nm").value();

        if (name == "bucket")
        {
            m_bucket = argNode.attribute("val").value();
        }
        else if (name == "buckettbl")
        {
            m_buckettbl = argNode.attribute("val").value();
        }
    }

    // Look for a 'send' command node in the response document.  This node
    // indicates that a cab is requested.
    pugi::xpath_node_set sendArgs =
        responseDoc.select_nodes("/resp/tlm/resps/resp/cmd[@nm='send']/arg");
    for (pugi::xpath_node arg : sendArgs)
    {
        // Extract the token from the command node.  This will be needed later
        // for a cab upload.
        pugi::xml_node argNode = arg.node();
        std::string name = argNode.attribute("nm").value();
        if (name == "token")
        {
            m_token = argNode.attribute("val").value();
        }
    }
}

/*++
    Parses a 'dataupload' response from Watson XML.
--*/
void HttpReportUploader::ParseDataUploadResponse(http_response httpResponse)
{
    // Extract the XML string from the response.
    std::string response = httpResponse.extract_string().get();

    TraceLoggingWrite(
        g_errorReportingUploaderProvider,
        "WatsonDataUploadResponse",
        TraceLoggingValue(httpResponse.status_code(), "StatusCode"),
        TraceLoggingValue(response.c_str(), "Response"));

    // We only expect 200 responses from Watson.
    if (httpResponse.status_code() != status_codes::OK)
    {
        throw http_exception(
            httpResponse.status_code(),
            "The server returned an unexpected status code.");
    }

    // A 200 response with an empty body implies that the request was malformed.
    if (response.empty())
    {
        throw http_exception("The server rejected the request XML.");
    }

    // Parse the response into an XML document.
    pugi::xml_document responseDoc;
    pugi::xml_parse_result result = responseDoc.load_string(response.c_str());
    if (!result)
    {
        throw make_xml_parse_error(result);
    }

    // Look for a 'receipt' command node in the response document.  This node
    // indicates that the cab was received.
    pugi::xpath_node_set receiptArgs =
        responseDoc.select_nodes("/resp/tlm/resps/resp/cmd[@nm='receipt']/arg");
    for (pugi::xpath_node arg : receiptArgs)
    {
        // Extract the cab id from teh receipt node.
        pugi::xml_node argNode = arg.node();
        std::string name = argNode.attribute("nm").value();
        if (name == "cabguid")
        {
            m_cabId = argNode.attribute("val").value();
        }
    }
}

/*++
    Parses report metadata from a submitted report directory.
--*/
void HttpReportUploader::ParseMetadata()
{
    // Load the metadata XML that is written by the report submitter.
    std::filesystem::path metadataPath = m_reportPath / "WERInternalMetadata.xml";

    pugi::xml_document metadataDoc;
    pugi::xml_parse_result result =
        metadataDoc.load_file(metadataPath.string().c_str());
    if (!result)
    {
        throw make_xml_parse_error(result);
    }

    // Extract the event type and bucketing parameters from the metadata XML.
    pugi::xml_node problemSignaturesNode =
        metadataDoc.child("WERReportMetadata").child("ProblemSignatures");

    for (pugi::xml_node sigNode = problemSignaturesNode.first_child(); sigNode;
         sigNode = sigNode.next_sibling())
    {
        std::string_view name = sigNode.name();
        if (name == "EventType")
        {
            m_eventType = sigNode.child_value();
        }
        else if (name.size() == 10 && name.substr(0, 9) == "Parameter")
        {
            if (m_bucketingParameters.size() > 9)
            {
                throw std::length_error(
                    "A maximum of 10 bucketing parameters are allowed per report.");
            }

            m_bucketingParameters.emplace_back(sigNode.child_value());
        }
    }

    // EventType is required.  Bucketing parameters are optional.
    if (m_eventType.empty())
    {
        throw std::invalid_argument("eventType cannot be empty");
    }
}

/*++
    Adds a Watson protocol argument to the report being uploaded.
--*/
void HttpReportUploader::AddProtocolArg(
    std::string_view section,
    std::string_view name,
    std::string_view value)
{
    TraceLoggingWrite(
        g_errorReportingUploaderProvider,
        "AddProtocolArg",
        TraceLoggingCountedString(section.data(), section.size(), "Section"),
        TraceLoggingCountedString(name.data(), name.size(), "Name"),
        TraceLoggingCountedString(value.data(), value.size(), "Value"));

    m_protocolArgs->AddProtocolArg(section, name, value);
}

}}}}
