// Copyright (c) Microsoft Corporation. All rights reserved.

#include <iostream>
#include <string_view>
#include <system_error>

#include <azure-device-health-plugin/plugin-dev.h>
#include <azure-device-util/AndroidLogHelper.h>
#include <azure-device-util/ResultException.h>
#include <azure-device-util/TraceLoggingTypes.h>
#include <tracelogging/TraceLoggingProvider.h>

#ifdef AZURE_DEVICE_HEALTH_USE_PLUGIN_LOG_CENSUS
extern const adh_plugin_definition log_census;
#endif

#ifdef AZURE_DEVICE_HEALTH_USE_PLUGIN_LOG_USB
extern const adh_plugin_definition log_usb;
#endif

#ifdef AZURE_DEVICE_HEALTH_USE_PLUGIN_LOG_PACKAGES
extern const adh_plugin_definition log_packages;
#endif

#ifdef AZURE_DEVICE_HEALTH_USE_PLUGIN_LOG_CONTAINERS
extern const adh_plugin_definition log_containers;
#endif

#ifdef AZURE_DEVICE_HEALTH_USE_PLUGIN_PRODUCE_CONSUME_COOLING
extern const adh_plugin_definition produce_cooling;
extern const adh_plugin_definition consume_cooling;
#endif

#ifdef AZURE_DEVICE_HEALTH_USE_PLUGIN_PRODUCE_CONSUME_PER_PROCESS_CPU
extern const adh_plugin_definition produce_per_process_cpu;
extern const adh_plugin_definition consume_per_process_cpu;
#endif

#ifdef AZURE_DEVICE_HEALTH_USE_PLUGIN_PRODUCE_CONSUME_PER_PROCESS_MEMORY
extern const adh_plugin_definition produce_per_process_memory;
extern const adh_plugin_definition consume_per_process_memory;
#endif

#ifdef AZURE_DEVICE_HEALTH_USE_PLUGIN_PRODUCE_CONSUME_THERMAL
extern const adh_plugin_definition produce_thermal;
extern const adh_plugin_definition consume_thermal;
#endif

#ifdef AZURE_DEVICE_HEALTH_USE_PLUGIN_PRODUCE_CONSUME_TOTAL_CPU
extern const adh_plugin_definition produce_total_cpu;
extern const adh_plugin_definition consume_total_cpu;
#endif

#ifdef AZURE_DEVICE_HEALTH_USE_PLUGIN_PRODUCE_CONSUME_TOTAL_DISK_IO
extern const adh_plugin_definition produce_total_disk_io;
extern const adh_plugin_definition consume_total_disk_io;
#endif

#ifdef AZURE_DEVICE_HEALTH_USE_PLUGIN_PRODUCE_CONSUME_TOTAL_MEMORY
extern const adh_plugin_definition produce_total_memory;
extern const adh_plugin_definition consume_total_memory;
#endif

#ifdef AZURE_DEVICE_HEALTH_USE_PLUGIN_PRODUCE_CONSUME_TOTAL_NETWORK
extern const adh_plugin_definition produce_total_network;
extern const adh_plugin_definition consume_total_network;
#endif

static const adh_plugin_definition* const s_pluginsList[] = {
#ifdef AZURE_DEVICE_HEALTH_USE_PLUGIN_LOG_CENSUS
    &log_census,
#endif

#ifdef AZURE_DEVICE_HEALTH_USE_PLUGIN_LOG_USB
    &log_usb,
#endif

#ifdef AZURE_DEVICE_HEALTH_USE_PLUGIN_LOG_PACKAGES
    &log_packages,
#endif

#ifdef AZURE_DEVICE_HEALTH_USE_PLUGIN_LOG_CONTAINERS
    &log_containers,
#endif

#ifdef AZURE_DEVICE_HEALTH_USE_PLUGIN_PRODUCE_CONSUME_COOLING
    &produce_cooling,
    &consume_cooling,
#endif

#ifdef AZURE_DEVICE_HEALTH_USE_PLUGIN_PRODUCE_CONSUME_PER_PROCESS_CPU
    &produce_per_process_cpu,
    &consume_per_process_cpu,
#endif

#ifdef AZURE_DEVICE_HEALTH_USE_PLUGIN_PRODUCE_CONSUME_PER_PROCESS_MEMORY
    &produce_per_process_memory,
    &consume_per_process_memory,
#endif

#ifdef AZURE_DEVICE_HEALTH_USE_PLUGIN_PRODUCE_CONSUME_THERMAL
    &produce_thermal,
    &consume_thermal,
#endif

#ifdef AZURE_DEVICE_HEALTH_USE_PLUGIN_PRODUCE_CONSUME_TOTAL_CPU
    &produce_total_cpu,
    &consume_total_cpu,
#endif

#ifdef AZURE_DEVICE_HEALTH_USE_PLUGIN_PRODUCE_CONSUME_TOTAL_DISK_IO
    &produce_total_disk_io,
    &consume_total_disk_io,
#endif

#ifdef AZURE_DEVICE_HEALTH_USE_PLUGIN_PRODUCE_CONSUME_TOTAL_MEMORY
    &produce_total_memory,
    &consume_total_memory,
#endif

#ifdef AZURE_DEVICE_HEALTH_USE_PLUGIN_PRODUCE_CONSUME_TOTAL_NETWORK
    &produce_total_network,
    &consume_total_network,
#endif

    nullptr
};

extern "C" __attribute__((visibility("default")))
const adh_plugin_definition* const*
get_adh_plugin_definitions()
{
    return s_pluginsList;
}

TRACELOGGING_DEFINE_PROVIDER(
    g_coreModuleProvider,
    "Azure.DeviceHealth.Plugins.CoreModule",
    (0xf016a36d, 0xb955, 0x754a, 0xbc, 0xf7, 0x79, 0x9b, 0xc1, 0xe9, 0x50, 0x45));

/*++
    Registered as our error handling callback with the ErrorHelper from
    azure-device-utils package. Logs to LTTNG.

Parameters:
    message - The message to log.
    file - The file name (in debug builds) which generated the error.
    line - The line number (in debug builds) which generated the error.
    err - The error code generated.
--*/
static void LogCallback(
    std::string_view message,
    std::string_view file,
    int lineNumber,
    std::error_code err)
{
    TraceLoggingWrite(
        g_coreModuleProvider,
        "LogError",
        TraceLoggingStringView(file, "FileName"),
        TraceLoggingValue(lineNumber, "Line"),
        TraceLoggingStringView(message, "Message"),
        TraceLoggingErrorCode(err, "ErrorCode"));

    ADHS_LOG(
        "LogError: %.*s::%d, %.*s %d",
        static_cast<int>(file.size()),
        file.data(),
        lineNumber,
        static_cast<int>(message.size()),
        message.data(),
        err.value());

    std::cerr << "LogError: ";
    std::cerr << "loc='" << file << "@" << lineNumber << "' ";
    std::cerr << "msg='" << message << "' ";
    std::cerr << "err={msg='" << err.message() << "' ";
    std::cerr << "cat='" << err.category().name() << "' ";
    std::cerr << "val='" << err.value() << "'}" << std::endl;
}

__attribute((constructor)) void SetupCorePluginsModule()
{
    TraceLoggingRegister(g_coreModuleProvider);
    Azure::Device::Utils::SetResultLoggingCallback(&LogCallback);
}

__attribute((destructor)) void TearmdownCorePluginsModule()
{
    TraceLoggingUnregister(g_coreModuleProvider);
}
