// Copyright (c) Microsoft Corporation. All rights reserved.
#include <optional>

#include <azure-device-health-plugin/plugin-dev.h>
#include <azure-device-util/FileSystem.h>
#include <azure-device-util/Process.h>
#include <azure-device-util/ResultException.h>
#include <azure-device-util/String.h>

#include "CpuInfoParser.h"
#include "FirmwareInfoParser.h"
#include "MemInfoParser.h"
#include "NetworkInfo.h"
#include "OEMInfoParser.h"
#include "OSInfoParser.h"
#include "Provider.h"

TRACELOGGING_DEFINE_PROVIDER(
    g_deviceHealthCensusProvider,
    "Census",
    // {F1AE5F35-1318-4F3A-B0B9-67985A85EB70}
    (0xf1ae5f35, 0x1318, 0x4f3a, 0xb0, 0xb9, 0x67, 0x98, 0x5a, 0x85, 0xeb, 0x70));

TRACELOGGING_DEFINE_PROVIDER(
    g_deviceHealthCensusLocalProvider,
    "Azure.DeviceHealth.Debug.Census",
    // {80BECB7E-506E-4BCE-B4B9-F0A0792E7CBB}
    (0x80becb7e, 0x506e, 0x4bce, 0xb4, 0xb9, 0xf0, 0xa0, 0x79, 0x2e, 0x7c, 0xbb));

namespace Azure { namespace Device { namespace Health { namespace Plugin {

class LogCensusPlugin
{
  public:
    bool LogData(const adh_log_data_args* args);

    std::string Override_OEMManufacturerName;
    std::string Override_OEMModelNumber;
    std::string Override_OEMModelBaseBoard;
    std::string Override_OEMModelSystemFamily;
};

}}}}

using namespace Azure::Device::Health::Plugin;
namespace Utils = Azure::Device::Utils;

adh_plugin_instance* log_census_create_plugin_instance(
    const adh_plugin_config_pair* configs,
    size_t configs_count)
{
    try
    {
        TraceLoggingRegister(g_deviceHealthCensusProvider);
        TraceLoggingRegister(g_deviceHealthCensusLocalProvider);

        TraceLoggingWrite(g_deviceHealthCensusLocalProvider, "Create_Initialize");

        auto logCensusPlugin = std::make_unique<LogCensusPlugin>();

        // Handle configs
        for (const adh_plugin_config_pair* config = configs;
             config != (configs + configs_count);
             ++config)
        {
            std::string_view name = config->name;
            std::string_view value = config->value;

            if (Utils::String::ICompare(name, "CensusOEMManufacturerName") == 0)
            {
                logCensusPlugin->Override_OEMManufacturerName = value;
            }
            else if (Utils::String::ICompare(name, "CensusOEMModelNumber") == 0)
            {
                logCensusPlugin->Override_OEMModelNumber = value;
            }
            else if (Utils::String::ICompare(name, "CensusOEMModelBaseBoard") == 0)
            {
                logCensusPlugin->Override_OEMModelBaseBoard = value;
            }
            else if (Utils::String::ICompare(name, "CensusOEMModelSystemFamily") == 0)
            {
                logCensusPlugin->Override_OEMModelSystemFamily = value;
            }
            else
            {
                TraceLoggingWrite(
                    g_deviceHealthCensusLocalProvider,
                    "Create_UnusedConfiguration",
                    TraceLoggingValue(config->name, "Name"),
                    TraceLoggingValue(config->value, "Value"));
            }
        }

        return reinterpret_cast<adh_plugin_instance*>(logCensusPlugin.release());
    }
    CATCH_LOG();
    return nullptr;
}

void log_census_destroy_plugin_instance(adh_plugin_instance* plugin)
{
    LogCensusPlugin* logCensusPlugin = reinterpret_cast<LogCensusPlugin*>(plugin);
    delete logCensusPlugin;

    TraceLoggingWrite(g_deviceHealthCensusLocalProvider, "Uninitialize");

    TraceLoggingUnregister(g_deviceHealthCensusProvider);
    TraceLoggingUnregister(g_deviceHealthCensusLocalProvider);
}

bool log_census_log_data(adh_plugin_instance* plugin, const adh_log_data_args* args)
{
    try
    {
        LogCensusPlugin* logCensusPlugin =
            reinterpret_cast<LogCensusPlugin*>(plugin);
        return logCensusPlugin->LogData(args);
    }
    CATCH_LOG();
    return false;
}

DEFINE_LOGGER_PLUGIN(
    log_census,
    "log-census",
    log_census_create_plugin_instance,
    log_census_destroy_plugin_instance,
    log_census_log_data);

/*++
    Tries to run the binary 'azure-device-health-id' if found on the path.
    If successful, it will take the first line of output and treat that as
    the device id.
--*/
static std::string ConstructHardwareDeviceId()
{
    constexpr std::string_view c_healthIdBinaryName = "azure-device-health-id";
    constexpr uint32_t c_maxDeviceIdChars = 512;

    Utils::Process::CommandResult healthIdCommandResult =
        Utils::Process::RunCommand(c_healthIdBinaryName);

    if (healthIdCommandResult.ExitCode != 0)
    {
        TraceLoggingWrite(
            g_deviceHealthCensusLocalProvider,
            "Census_ConstructHarwareDeviceId_BadExitCode",
            TraceLoggingValue(healthIdCommandResult.ExitCode, "ExitCode"));
        return {};
    }

    if (healthIdCommandResult.Output.empty())
    {
        TraceLoggingWrite(
            g_deviceHealthCensusLocalProvider,
            "Census_ConstructHarwareDeviceId_IllegalOutput");
        return {};
    }

    const std::string& idLine = healthIdCommandResult.Output[0];
    if (idLine.size() > c_maxDeviceIdChars)
    {
        TraceLoggingWrite(
            g_deviceHealthCensusLocalProvider,
            "Census_ConstructHarwareDeviceId_UnreasonableIdLength",
            TraceLoggingValue(idLine.c_str(), "Id"));
        return {};
    }

    TraceLoggingWrite(
        g_deviceHealthCensusLocalProvider,
        "Census_ConstructHarwareDeviceId_HealthIdSucceeded",
        TraceLoggingValue(idLine.c_str(), "Id"));

    return healthIdCommandResult.Output[0];
}

/*++
    Required for plug-ins of type 'log'. This function works almost
    identically to 'Produce' except it does not return any data (via return
    or param). It is used normally for plug-ins which need to log some data
    based on system state to a system or service other than adhd.
--*/
bool LogCensusPlugin::LogData([[maybe_unused]] const adh_log_data_args* args)
{
    TraceLoggingWrite(g_deviceHealthCensusLocalProvider, "LogData");

    CpuInfoParser cpuInfoParser;
    std::optional<CpuInfo> cpuInfo = cpuInfoParser.Parse();
    if (!cpuInfo)
    {
        TraceLoggingWrite(
            g_deviceHealthCensusLocalProvider, "FailedToParseCpuinfo");
    }
    else
    {
        // Keep this consistent with Asimov schema style
        TraceLoggingWrite(
            g_deviceHealthCensusProvider,
            "Processor",
            TraceLoggingValue(cpuInfo->NumCpus, "ProcessorCores"),
            TraceLoggingValue(cpuInfo->CorePerSocket, "ProcessorPhysicalCores"),
            TraceLoggingValue(
                static_cast<int32_t>(cpuInfo->ArchitectureId),
                "ProcessorArchitecture"),
            TraceLoggingValue(cpuInfo->CpuClockSpeed, "ProcessorClockSpeed"),
            TraceLoggingValue(cpuInfo->CpuVendor.c_str(), "ProcessorManufacturer"),
            TraceLoggingValue(cpuInfo->CpuModel.c_str(), "ProcessorModel"),
            TraceLoggingValue(cpuInfo->Sockets, "SocketCount"));

        TraceLoggingWrite(
            g_deviceHealthCensusProvider,
            "VM",
            TraceLoggingValue(
                cpuInfo->HypervisorVendor.empty() ? "0" : "1", "IsVirtualDevice"),
            TraceLoggingValue(
                int(!cpuInfo->HypervisorVendor.empty()), "HyperVisor"),
            TraceLoggingValue(
                cpuInfo->HypervisorVendor.c_str(), "HyperVisorVendor"));
    }

    MemInfo memInfo;
    MemInfoParser memInfoParser;
    if (!memInfoParser.Parse(memInfo))
    {
        TraceLoggingWrite(
            g_deviceHealthCensusLocalProvider, "FailedToParseMeminfo");
    }
    else
    {
        TraceLoggingWrite(
            g_deviceHealthCensusProvider,
            "Memory",
            TraceLoggingValue(memInfo.MemTotalKb, "TotalPhysicalRAM"),
            TraceLoggingValue(memInfo.SwapTotalKb, "TotalSwapSpace"));
    }

    FirmwareInfo firmwareInfo;
    FirmwareInfoParser firmwareInfoParser;
    if (!firmwareInfoParser.Parse(firmwareInfo))
    {
        TraceLoggingWrite(
            g_deviceHealthCensusLocalProvider, "FailedToParseFirmwareinfo");
    }
    else
    {
        TraceLoggingWrite(
            g_deviceHealthCensusProvider,
            "Firmware",
            TraceLoggingValue(
                firmwareInfo.BiosVendor.c_str(), "FirmwareManufacturer"),
            TraceLoggingValue(
                firmwareInfo.BiosReleaseDate.c_str(), "FirmwareReleaseDate"),
            TraceLoggingValue(firmwareInfo.BiosVersion.c_str(), "FirmwareVersion"),
            TraceLoggingValue(
                static_cast<int32_t>(firmwareInfo.FirmwareType), "FirmwareType"));
    }

    NetworkInfo networkInfo;
    std::vector<NetworkInterfaceInfo> netInterfaceInfo;
    if (!networkInfo.GetInfo(netInterfaceInfo))
    {
        TraceLoggingWrite(
            g_deviceHealthCensusLocalProvider, "FailedToGetNetworkInfo");
    }
    else
    {
        for (uint16_t i = 0; i < netInterfaceInfo.size(); i++)
        {
            TraceLoggingWrite(
                g_deviceHealthCensusProvider,
                "Network",
                TraceLoggingValue(
                    netInterfaceInfo[i].InterfaceName.c_str(), "LogicalName"),
                TraceLoggingValue(netInterfaceInfo[i].IsWireless, "Wireless"),
                TraceLoggingValue(netInterfaceInfo[i].Flags, "Flags"),
                TraceLoggingValue(netInterfaceInfo[i].SpeedMbs, "SpeedMbs"),
                TraceLoggingValue(netInterfaceInfo[i].Duplex, "Duplex"),
                TraceLoggingValue(netInterfaceInfo[i].VendorId, "VendorId"),
                TraceLoggingValue(netInterfaceInfo[i].DeviceId, "DeviceId"));
        }
    }

    std::string hardwareDeviceId = ConstructHardwareDeviceId();

    OEMInfo oemInfo;
    OEMInfoParser oemInfoParser;
    if (!oemInfoParser.Parse(oemInfo))
    {
        TraceLoggingWrite(
            g_deviceHealthCensusLocalProvider, "FailedToParseOEMinfo");
    }

    const char* finalOemManufacturerName = Override_OEMManufacturerName.empty() ?
        oemInfo.OEMManufacturerName.c_str() :
        Override_OEMManufacturerName.c_str();

    const char* finalOemModelNumber = Override_OEMModelNumber.empty() ?
        oemInfo.OEMModelNumber.c_str() :
        Override_OEMModelNumber.c_str();

    const char* finalOemModelBaseBoard = Override_OEMModelBaseBoard.empty() ?
        oemInfo.OEMModelBaseBoard.c_str() :
        Override_OEMModelBaseBoard.c_str();

    const char* finalOemModelSystemFamily = Override_OEMModelSystemFamily.empty() ?
        oemInfo.OEMModelSystemFamily.c_str() :
        Override_OEMModelSystemFamily.c_str();

    // Hardware
    TraceLoggingWrite(
        g_deviceHealthCensusProvider,
        "Hardware",
        TraceLoggingValue(hardwareDeviceId.c_str(), "OEMSerialNumber"),
        TraceLoggingValue(finalOemManufacturerName, "OEMManufacturerName"),
        TraceLoggingValue(finalOemModelNumber, "OEMModelNumber"),
        TraceLoggingValue(oemInfo.OEMModelName.c_str(), "OEMModelName"),
        TraceLoggingValue(oemInfo.OEMModelSKU.c_str(), "OEMModelSKU"),
        TraceLoggingValue(finalOemModelBaseBoard, "OEMModelBaseBoard"),
        TraceLoggingValue(
            oemInfo.OEMModelBaseBoardVersion.c_str(), "OEMModelBaseBoardVersion"),
        TraceLoggingValue(
            oemInfo.OEMModelBaseBoardManufacturer.c_str(),
            "OEMModelBaseBoardManufacturer"),
        TraceLoggingValue(finalOemModelSystemFamily, "OEMModelSystemFamily"),
        TraceLoggingValue(
            oemInfo.OEMModelSystemVersion.c_str(), "OEMModelSystemVersion"),
        TraceLoggingValue(oemInfo.ChassisType, "ChassisType"));

    // OS Info
    // Declared here because it consumes values from os-subrelease
    OSInfo osInfo;

    // Read release data from /etc/os-subrelease
    std::map<std::string, std::string, std::less<>> subreleaseData;
    std::error_code result = Utils::FileSystem::ReadDictionaryFromFile(
        "/etc/os-subrelease", "=", subreleaseData);
    if (FAILED(result))
    {
        TraceLoggingWrite(
            g_deviceHealthCensusLocalProvider,
            "FailedToParseOSSubrelease",
            TraceLoggingValue(result.value(), "ErrorCode"));
    }
    else
    {
        // Trim leading/trailing quotation marks
        for (auto&& [key, value] : subreleaseData)
        {
            Utils::String::Trim(value, "\"");
        }

        // Populate OS Info
        osInfo.OSEdition = subreleaseData["ID"];

        // OS SubreleaseInfo
        TraceLoggingWrite(
            g_deviceHealthCensusProvider,
            "OSSubrelease",
            TraceLoggingValue(
                subreleaseData["BUILDER_NAME"].c_str(), "BuilderName"),
            TraceLoggingValue(subreleaseData["BUILD_DATE"].c_str(), "BuildDate"),
            TraceLoggingValue(subreleaseData["ID"].c_str(), "Id"),
            TraceLoggingValue(subreleaseData["VERSION_ID"].c_str(), "VersionId"),
            TraceLoggingValue(subreleaseData["NAME"].c_str(), "Name"),
            TraceLoggingValue(subreleaseData["VERSION"].c_str(), "Version"),
            TraceLoggingValue(subreleaseData["BRANCH"].c_str(), "Branch"),
            TraceLoggingValue(subreleaseData["COMMIT_ID"].c_str(), "CommitId"));
    }

    OSInfoParser osInfoParser;
    osInfoParser.Parse(osInfo);

    // OS
    TraceLoggingWrite(
        g_deviceHealthCensusProvider,
        "OS",
        // OSEdition obtained from /etc/os-subrelease
        TraceLoggingValue(osInfo.OSEdition.c_str(), "OSEdition"),
        TraceLoggingValue(osInfo.KernelRelease.c_str(), "KernelRelease"),
        TraceLoggingValue(osInfo.KernelVersion.c_str(), "KernelVersion"),
        // Secure boot status sent as a string, same as Windows.
        TraceLoggingValue(
            osInfo.IsSecureBootEnabled ? "1" : "0", "IsSecureBootEnabled"));

    return true;
}
