// Copyright (c) Microsoft Corporation. All rights reserved.

#include "CpuInfoParser.h"
#include <optional>

#include <type_traits>

#include <azure-device-util/Process.h>
#include <azure-device-util/String.h>
#include <sys/utsname.h>

#include "Provider.h"

namespace Azure { namespace Device { namespace Health {

std::optional<CpuInfo> CpuInfoParser::Parse()
{
    std::vector<std::string> lscpuLines = GetLscpuOutput();
    return ParseLines(lscpuLines);
}

std::vector<std::string> CpuInfoParser::GetLscpuOutput()
{
    Utils::Process::CommandResult lscpuResult =
        Utils::Process::RunCommand("lscpu");

    if (lscpuResult.ExitCode != 0)
    {
        TraceLoggingWrite(
            g_deviceHealthCensusLocalProvider,
            "CpuInfoParser_LscpuFailed",
            TraceLoggingValue(lscpuResult.ExitCode, "ExitCode"));

        return {};
    }

    return lscpuResult.Output;
}

template<class T>
struct dependent_false : std::false_type
{};

template<class T>
bool ParseLine(std::string_view line, std::string_view header, T& value)
{
    if (!Utils::String::IStartsWith(line, header))
    {
        return false;
    }

    line.remove_prefix(header.size());

    if constexpr (std::is_unsigned_v<T>)
    {
        value = Utils::String::ToUIntT<T>(line);
    }
    else if constexpr (std::is_floating_point_v<T>)
    {
        value = Utils::String::ToFloat<T>(line);
    }
    else if constexpr (std::is_same_v<std::string, T>)
    {
        value = line;
        Utils::String::Trim(value);
    }
    else
    {
        static_assert(
            dependent_false<T>::value, "Only accepts unsigned or string types");
    }

    return true;
}

/*++
    Parses lines of `lscpu` output to populate a CpuInfo.  If not all fields
    can be populated, then this will return an empty optional.

Arguments:
    lscpuLines - Lines of output from lscpu command

Return Value:
    Completely populated CpuInfo, or empty
--*/
std::optional<CpuInfo>
CpuInfoParser::ParseLines(const std::vector<std::string>& lscpuLines)
{
    CpuInfo cpuInfo;

    for (const auto& line : lscpuLines)
    {
        if (ParseLine(line, "Architecture:", cpuInfo.Arch)) {}
        else if (ParseLine(line, "CPU(s):", cpuInfo.NumCpus))
        {}
        else if (ParseLine(line, "Core(s) per socket:", cpuInfo.CorePerSocket))
        {}
        else if (ParseLine(line, "Socket(s):", cpuInfo.Sockets))
        {}
        else if (ParseLine(line, "CPU MHz:", cpuInfo.CpuClockSpeed))
        {}
        else if (ParseLine(line, "Vendor ID:", cpuInfo.CpuVendor))
        {}
        else if (ParseLine(line, "Model name:", cpuInfo.CpuModel))
        {}
        else if (ParseLine(line, "Hypervisor vendor:", cpuInfo.HypervisorVendor))
        {}
    }

    auto parseArch = [](std::string_view Arch) {
        if (Arch.empty())
        {
            return ProcessArchitectureId::UNKNOWN;
        }
        // Catch 64-bit arm CPUs in compat mode and native 64-bit mode
        else if (
            Utils::String::IStartsWith(Arch, "armv8") ||
            Utils::String::IStartsWith(Arch, "aarch64"))
        {
            return ProcessArchitectureId::ARM64;
        }
        // Catch all other 32-bit ARM CPUs
        else if (Utils::String::IStartsWith(Arch, "arm"))
        {
            return ProcessArchitectureId::ARM;
        }
        // Catch x64 machines
        else if (Arch == "x86_64" || Arch == "amd64")
        {
            return ProcessArchitectureId::AMD64;
        }
        // Catch x86 machines
        else if (Arch == "i386" || Arch == "i686")
        {
            return ProcessArchitectureId::INTEL;
        }
        // Catch IA64 machines
        else if (Arch == "ia64")
        {
            return ProcessArchitectureId::IA64;
        }
        else
        {
            return ProcessArchitectureId::UNKNOWN;
        }
    };

    cpuInfo.ArchitectureId = parseArch(cpuInfo.Arch);

    return cpuInfo;
}
}}}
