// Copyright (c) Microsoft Corporation. All rights reserved.

#pragma once

// STL
#include <string>

// Provider definition
#include "Provider.h"

// Utils
#include <azure-device-util/FileSystem.h>
#include <azure-device-util/String.h>

static const char* FirmwareInfoFileList[] = { "bios_vendor",
                                              "bios_date",
                                              "bios_version" };

enum FirmwareFileInfo
{
    BIOS_VENDOR,
    BIOS_DATE,
    BIOS_VERSION,
    NUM_VALUES
};

// Copied from winnt.h
enum FIRMWARE_TYPE
{
    FirmwareTypeUnknown,
    FirmwareTypeBios,
    FirmwareTypeUefi,
    FirmwareTypeMax
};

static_assert(
    FirmwareFileInfo::NUM_VALUES ==
        std::extent<decltype(FirmwareInfoFileList)>::value,
    "Incorrect number of FirmwareFileInfo values.");

struct FirmwareInfo
{
    std::string BiosVendor;
    std::string BiosReleaseDate;
    std::string BiosVersion;
    FIRMWARE_TYPE FirmwareType;
};

namespace Azure { namespace Device { namespace Health {

class FirmwareInfoParser
{
  public:
    FirmwareInfoParser() : _rootPath("/") {}
    FirmwareInfoParser(std::string_view rootPath) : _rootPath(rootPath) {}

    bool Parse(FirmwareInfo& info)
    {
        try
        {
            std::string fwInfoDirPath;
            fwInfoDirPath.append(_rootPath);
            fwInfoDirPath.append("sys/class/dmi/id/");

            std::vector<std::string> fmwInfo;
            for (uint32_t i = 0; i < FirmwareFileInfo::NUM_VALUES; i++)
            {
                std::string filePath = fwInfoDirPath + FirmwareInfoFileList[i];
                std::string content;

                if (FAILED(Utils::FileSystem::ReadUTF8StringFromFile(
                        filePath, content)))
                {
                    continue;
                }

                // Remove newline character at the end & remove space (if
                // any)
                if (content.length() > 0 && content.back() == '\n')
                {
                    content.pop_back();
                }
                Utils::String::Trim(content);
                fmwInfo.push_back(content);
            }

            if (fmwInfo.size() != FirmwareFileInfo::NUM_VALUES)
            {
                return false;
            }

            info.BiosVendor = std::move(fmwInfo[BIOS_VENDOR]);
            info.BiosReleaseDate = std::move(fmwInfo[BIOS_DATE]);
            info.BiosVersion = std::move(fmwInfo[BIOS_VERSION]);

            std::string efiDirPath{ _rootPath };
            efiDirPath.append("sys/firmware/efi");
            info.FirmwareType = std::filesystem::exists(efiDirPath) ?
                FirmwareTypeUefi :
                FirmwareTypeBios;

            return true;
        }
        CATCH_LOG();
        return false;
    }

  private:
    std::string _rootPath;
};

}}}
