// Copyright (c) Microsoft Corporation. All rights reserved.

#pragma once

// STL
#include <string>

// Provider definition
#include "Provider.h"

// Utils
#include <azure-device-util/FileSystem.h>
#include <azure-device-util/String.h>

struct MemInfo
{
  public:
    uint64_t MemTotalKb = 0;   // Total usable memory.
    uint64_t HighTotalKb = 0;  // Total amount of highmem.  Highmem is memory
                               // above ~860MB of physical memory.
    uint64_t LowTotalKb = 0;  // Total amount of lowmem.  Lowmem is memory which
                              // can be used for everything that highmem can be
                              // used for, but it is also available for the
                              // kernel's use for its own data structures.
    uint64_t SwapTotalKb = 0;  // Total amount of swap space available
};

namespace Azure { namespace Device { namespace Health {

class MemInfoParser
{
  public:
    MemInfoParser() : _rootPath("/") {}
    MemInfoParser(std::string_view rootPath) : _rootPath(rootPath) {}

    bool Parse(MemInfo& info)
    {
        try
        {
            std::string memInfoPath;
            memInfoPath.append(_rootPath);
            memInfoPath.append("proc/meminfo");

            std::string memInfoConntent;
            if (FAILED(Utils::FileSystem::ReadUTF8StringFromFile(
                    memInfoPath, memInfoConntent)))
            {
                return false;
            }

            std::vector<std::string> lines =
                Utils::String::Split(memInfoConntent, "\n");
            for (const auto& line : lines)
            {
                if (Utils::String::IStartsWith(line, "MemTotal"))
                {
                    std::vector<std::string> MemTotalKb =
                        Utils::String::Split(line, " ", false);
                    if (MemTotalKb.size() >= 2)
                    {
                        info.MemTotalKb = Utils::String::ToUIntT<uint64_t>(
                            MemTotalKb[1].c_str());
                    }
                }
                else if (Utils::String::IStartsWith(line, "SwapTotal"))
                {
                    std::vector<std::string> SwapTotalKb =
                        Utils::String::Split(line, " ", false);
                    if (SwapTotalKb.size() >= 2)
                    {
                        info.SwapTotalKb =
                            Utils::String::ToUIntT<uint64_t>(SwapTotalKb[1]);
                    }
                }
            }

            return true;
        }
        CATCH_LOG();
        return false;
    }

  private:
    std::string _rootPath;
};

}}}
