// Copyright (c) Microsoft Corporation. All rights reserved.

#pragma once

// String functions
#include <memory>
#include <string.h>
#include <string>

// Provider definition
#include "Provider.h"

// OS System call facade
#include "IOsSystemCallFacade.h"
#include "OsSystemCall.h"

// Utils
#include <azure-device-util/FileSystem.h>
#include <azure-device-util/String.h>

// Symbolic Constants
#include <unistd.h>

// Declarations for inquiring about network interfaces
#include <net/if.h>

// System & other networking headers
#include <linux/ethtool.h>
#include <linux/sockios.h>
#include <linux/wireless.h>
#include <sys/ioctl.h>
#include <sys/socket.h>

// std::extent
#include <type_traits>

struct NetworkInterfaceInfo
{
    std::string InterfaceName;
    uint32_t DeviceId = 0;
    uint32_t VendorId = 0;

    bool IsWireless = false;
    int16_t Flags = -1;     // IFF_UP etc. -1 is unknown
    int64_t SpeedMbs = -1;  // -1 is unknown
    uint8_t Duplex = 0xff;  // DUPLEX_HALF(0x00), DUPLEX_FULL (0x01) or
                            // DUPLEX_UNKNOWN  (0xff)
};

namespace Azure { namespace Device { namespace Health {

class NetworkInfo
{
  public:
    NetworkInfo()
        : _rootPath("/"), _osSystemCall(std::make_unique<OsSystemCall>())
    {}
    NetworkInfo(std::string_view rootPath, std::unique_ptr<IOsSystemCallFacade>&& ossci)
        : _rootPath(rootPath), _osSystemCall(std::move(ossci))
    {}

    bool GetInfo(std::vector<NetworkInterfaceInfo>& info)
    {
        try
        {
            std::vector<std::string> interfaces;
            if (!LoadNetworkInterfaces(interfaces))
            {
                // TraceLoggingWrite(
                //     g_deviceHealthCensusLocalProvider,
                //     "FailedToGetNetworkInterfaceList");
                return false;
            }

            int socketfd = socket(PF_INET, SOCK_DGRAM, 0);

            if (socketfd >= 0)
            {
                ethtool_cmd ecmd;
                ifreq ifr;
                iwreq iwr;
                for (uint32_t i = 0; i < interfaces.size(); i++)
                {
                    NetworkInterfaceInfo interfaceInfo;
                    interfaceInfo.InterfaceName = interfaces[i];

                    std::memset(&ecmd, 0, sizeof(ecmd));
                    std::memset(&ifr, 0, sizeof(ifr));
                    std::memset(&iwr, 0, sizeof(iwr));

                    FAIL_FAST_IF(
                        std::extent<decltype(ifr.ifr_name)>::value <=
                        interfaces[i].size());
                    strcpy(ifr.ifr_name, interfaces[i].c_str());  // NOLINT

                    FAIL_FAST_IF(
                        std::extent<decltype(iwr.ifr_name)>::value <=
                        interfaces[i].size());
                    strcpy(iwr.ifr_name, interfaces[i].c_str());  // NOLINT

                    // Get Interface flags.
                    if (_osSystemCall->os_ioctl(socketfd, SIOCGIFFLAGS, &ifr) == 0)
                    {
                        interfaceInfo.Flags = ifr.ifr_flags;
                    }

                    // Figure out if this is a wireless interface.
                    if (_osSystemCall->os_ioctl(socketfd, SIOCGIWNAME, &iwr) != -1)
                    {
                        interfaceInfo.IsWireless = true;
                    }

                    // Get settings
                    ecmd.cmd = ETHTOOL_GSET;  // Use ETHTOOL_SLINKSETTINGS
                                              // instead.
                    ifr.ifr_data = (caddr_t) &ecmd;
                    if (_osSystemCall->os_ioctl(socketfd, SIOCETHTOOL, &ifr) == 0)
                    {
                        interfaceInfo.SpeedMbs = ethtool_cmd_speed(&ecmd);
                        interfaceInfo.Duplex = ecmd.duplex;
                    }

                    // Grab vendor and device info.
                    interfaceInfo.DeviceId = Utils::String::ToUIntT<uint32_t>(
                        GetDeviceId(interfaces[i]));
                    interfaceInfo.VendorId = Utils::String::ToUIntT<uint32_t>(
                        GetVendorId(interfaces[i]));

                    // Exclude logical and loopback interfaces.
                    if (!(interfaceInfo.Flags & IFF_LOOPBACK) &&
                        !IsLogicalInterface(interfaces[i]))
                    {
                        info.push_back(interfaceInfo);
                    }
                }
            }

            close(socketfd);
            return true;
        }
        CATCH_LOG();
        return false;
    }

  private:
    bool LoadNetworkInterfaces(std::vector<std::string>& interfaces)
    {
        try
        {
            std::string filePath{ _rootPath };
            filePath.append("proc/net/dev");

            std::string content;
            if (FAILED(Utils::FileSystem::ReadUTF8StringFromFile(
                    filePath, content)))
            {
                return false;
            }

            std::vector<std::string> procnetdev =
                Utils::String::Split(content, "\n");

            // We should have more than 2 lines (# of header lines) here.
            if (procnetdev.size() <= 2)
            {
                return false;
            }

            // Remove the first 2 lines, which are the header lines.
            procnetdev.erase(procnetdev.begin(), procnetdev.begin() + 2);

            for (uint32_t i = 0; i < procnetdev.size(); i++)
            {
                if (Utils::String::IContains(procnetdev[i], ":"))
                {
                    auto line = Utils::String::Split(procnetdev[i], ":", false);
                    if (line.size() == 2)
                    {
                        Utils::String::Trim(line[0]);
                        interfaces.push_back(line[0]);
                    }
                }
            }

            return true;
        }
        CATCH_LOG();
        return false;
    }

    bool IsLogicalInterface(std::string_view interfaceName)
    {
        std::string path{ _rootPath };
        path.append(
            "sys/class/net/" + std::string{ interfaceName } + "/device/bridge");
        return std::filesystem::exists(path);
    }

    std::string GetDeviceId(std::string_view interfaceName)
    {
        std::string filePath{ _rootPath };
        filePath.append(
            "sys/class/net/" + std::string{ interfaceName } + "/device/device");
        std::string deviceId;
        if (std::filesystem::exists(filePath))
        {
            std::string content;
            if (Utils::FileSystem::ReadUTF8StringFromFile(filePath, deviceId))
            {
                deviceId = std::move(content);
            }

            // Remove newline character at the end.
            if (deviceId.length() > 0 && deviceId.back() == '\n')
            {
                deviceId.pop_back();
            }
        }

        return deviceId;
    }

    std::string GetVendorId(std::string_view interfaceName)
    {
        std::string filePath{ _rootPath };
        filePath.append(
            "sys/class/net/" + std::string{ interfaceName } + "/device/vendor");
        std::string vendor;
        if (std::filesystem::exists(filePath))
        {
            std::string content;
            if (Utils::FileSystem::ReadUTF8StringFromFile(filePath, vendor))
            {
                vendor = std::move(content);
            }

            // Remove newline character at the end.
            if (vendor.length() > 0 && vendor.back() == '\n')
            {
                vendor.pop_back();
            }
        }

        return vendor;
    }

  private:
    std::string _rootPath;
    std::unique_ptr<IOsSystemCallFacade> _osSystemCall;
};

}}}
