// Copyright (c) Microsoft Corporation. All rights reserved.

#pragma once

// STL
#include <string>

// Provider definition
#include "Provider.h"

// Utils
#include <azure-device-util/FileSystem.h>
#include <azure-device-util/String.h>

static const char* OEMInfoFileList[] = { "sys_vendor",     "product_name",
                                         "product_sku",    "board_name",
                                         "board_version",  "board_vendor",
                                         "product_family", "product_version",
                                         "product_serial", "chassis_type" };

enum OEMFileInfo
{
    SYS_VENDOR,
    PRODUCT_NAME,
    PRODUCT_SKU,
    BOARD_NAME,
    BOARD_VERSION,
    BOARD_VENDOR,
    PRODUCT_FAMILY,
    PRODUCT_VERSION,
    PRODUCT_SERIAL,
    CHASSIS_TYPE,
    OEM_NUM_VALUES
};

static_assert(
    OEMFileInfo::OEM_NUM_VALUES == std::extent<decltype(OEMInfoFileList)>::value,
    "Incorrect number of OEMFileInfo values.");

struct OEMInfo
{
    std::string OEMManufacturerName;
    std::string OEMModelNumber;
    std::string OEMModelName;
    std::string OEMModelSKU;
    std::string OEMModelBaseBoard;
    std::string OEMModelBaseBoardVersion;
    std::string OEMModelBaseBoardManufacturer;
    std::string OEMModelSystemFamily;
    std::string OEMModelSystemVersion;
    std::string OEMSerialNumber;
    int ChassisType;
};

namespace Azure { namespace Device { namespace Health {

class OEMInfoParser
{
  private:
    std::string _rootPath;

  public:
    OEMInfoParser() : _rootPath("/") {}
    OEMInfoParser(std::string_view rootPath) : _rootPath(rootPath) {}

    bool Parse(OEMInfo& info)
    {
        try
        {
            std::string OEMInfoDirPath;
            OEMInfoDirPath.append(_rootPath);
            OEMInfoDirPath.append("sys/class/dmi/id/");

            std::vector<std::string> oemInfo;
            for (uint32_t i = 0; i < OEMFileInfo::OEM_NUM_VALUES; i++)
            {
                std::string filePath = OEMInfoDirPath + OEMInfoFileList[i];
                std::string content;

                if (FAILED(Utils::FileSystem::ReadUTF8StringFromFile(
                        filePath, content)))
                {
                    content.assign(UNKNOWN_STRING);
                }
                else
                {
                    // Remove newline character at the end & remove space (if
                    // any)
                    if (content.length() > 0 && content.back() == '\n')
                    {
                        content.pop_back();
                    }
                    Utils::String::Trim(content);
                }

                oemInfo.push_back(content);
            }

            info.OEMManufacturerName = std::move(oemInfo[SYS_VENDOR]);
            info.OEMModelNumber = std::move(oemInfo[PRODUCT_NAME]);
            info.OEMModelSKU = std::move(oemInfo[PRODUCT_SKU]);
            info.OEMModelBaseBoard = std::move(oemInfo[BOARD_NAME]);
            info.OEMModelBaseBoardVersion = std::move(oemInfo[BOARD_VERSION]);
            info.OEMModelBaseBoardManufacturer = std::move(oemInfo[BOARD_VENDOR]);
            info.OEMModelSystemFamily = std::move(oemInfo[PRODUCT_FAMILY]);
            info.OEMModelSystemVersion = std::move(oemInfo[PRODUCT_VERSION]);
            info.OEMSerialNumber = std::move(oemInfo[PRODUCT_SERIAL]);
            info.ChassisType = Utils::String::ToIntT<int>(oemInfo[CHASSIS_TYPE]);

            info.OEMModelName = info.OEMModelNumber;

            return true;
        }
        CATCH_LOG();
        return false;
    }
};
}}}