// Copyright (c) Microsoft Corporation. All rights reserved.

#pragma once

#include <fstream>
#include <string>

#include <azure-device-util/FileSystem.h>
#include <azure-device-util/String.h>
#include <sys/stat.h>
#include <sys/utsname.h>

struct OSInfo
{
    std::string OSEdition;
    std::string KernelRelease;
    std::string KernelVersion;
    bool IsSecureBootEnabled = false;
};

namespace Azure { namespace Device { namespace Health {

class OSInfoParser
{
  public:
    OSInfoParser() : _rootPath("/"){};
    OSInfoParser(std::string_view rootPath) : _rootPath(rootPath) {}

    bool Parse(OSInfo& osInfo)
    {
        osInfo.IsSecureBootEnabled = GetSecureBootEnabled();

        // Get Kernel Release (aka uname -r)
        struct utsname unameData;
        uname(&unameData);
        osInfo.KernelRelease = std::string(unameData.release);
        osInfo.KernelVersion = std::string(unameData.version);
        return true;
    }

  private:
    bool GetSecureBootEnabled()
    {
        std::string fileName;
        std::filesystem::path efiVarsDirPath = _rootPath;
        efiVarsDirPath /= "sys/firmware/efi/efivars/";

        // If the directory doesn't exist the system is most likely using legacy
        // boot. We assume assume boot is not being used.
        if (!std::filesystem::exists(efiVarsDirPath))
        {
            return false;
        }

        for (const auto& entry :
             std::filesystem::directory_iterator(efiVarsDirPath))
        {
            if (entry.path().has_filename() &&
                Utils::String::IStartsWith(
                    entry.path().filename().string(), "SecureBoot-"))
            {
                fileName = entry.path().string();
                break;
            }
        }

        // If the SecureBoot variable was not found
        if (fileName.empty())
        {
            return false;
        }

        // File stat struct for reading the file's size.
        struct stat results;
        // 4 bytes/chars for the variable attributes.
        char attributes[4];
        // 1 byte/char for the variable data itself.
        char data;

        // Get size of the file
        if (stat(fileName.c_str(), &results) != 0)
        {
            return false;
        }

        // The UEFI standard defines 4 bytes for attributes and the SecureBoot
        // variable as being 1 byte. Therefore the total size of the file must
        // be 5. Otherwise assume an error. mokutil behaves similarly although
        // there is better error recovery in there.
        // https://uefi.org/sites/default/files/resources/UEFI%20Spec%202.8B%20May%202020.pdf
        if (results.st_size != 5)
        {
            return false;
        }

        std::ifstream secureBootEfivar(fileName, std::ios::binary);
        if (!secureBootEfivar.is_open())
        {
            return false;
        }

        secureBootEfivar.read(attributes, 4);
        secureBootEfivar.read(&data, 1);
        secureBootEfivar.close();

        // From the UEFI spec:
        // A value of 1 indicates that platform firmware performs driver and
        // boot application signature verification as specified in UEFI Image
        // Validation during the current boot. A value of 0 indicates that
        // driver and boot application signature verification is not active
        // during the current boot.
        return data == 1;
    }

  private:
    std::string _rootPath;
};

}}}