// Copyright (c) Microsoft Corporation. All rights reserved.

#include <memory>

#include <azure-device-health-plugin/plugin-dev.h>
#include <azure-device-util/Docker.h>
#include <azure-device-util/ResultException.h>
#include <azure-device-util/TraceLoggingTypes.h>
#include <tracelogging/TraceLoggingProvider.h>

TRACELOGGING_DEFINE_PROVIDER(
    g_deviceHealthContainersProvider,
    "Azure.DeviceHealth.SoftwareInventory.Containers",
    (0x0fbfba91, 0xb2af, 0x4bba, 0x83, 0x02, 0xff, 0x0a, 0xf8, 0xff, 0xe3, 0x59));

namespace Azure { namespace Device { namespace Health { namespace Plugin {

class LogContainersPlugin
{
  public:
    bool LogData(const adh_log_data_args* args);
};

}}}}

namespace Utils = Azure::Device::Utils;
using namespace Azure::Device::Health::Plugin;

adh_plugin_instance* log_containers_create_plugin_instance(
    const adh_plugin_config_pair* configs,
    size_t configs_count)
{
    try
    {
        TraceLoggingRegister(g_deviceHealthContainersProvider);
        TraceLoggingWrite(g_deviceHealthContainersProvider, "Initialize");

        auto logContainersPlugin = std::make_unique<LogContainersPlugin>();

        for (const adh_plugin_config_pair* config = configs;
             config != (configs + configs_count);
             ++config)
        {
            TraceLoggingWrite(
                g_deviceHealthContainersProvider,
                "Create_UnusedConfiguration",
                TraceLoggingValue(config->name, "Name"),
                TraceLoggingValue(config->value, "Value"));
        }

        return reinterpret_cast<adh_plugin_instance*>(
            logContainersPlugin.release());
    }
    CATCH_LOG();
    return nullptr;
}

void log_containers_destroy_plugin_instance(adh_plugin_instance* plugin)
{
    LogContainersPlugin* logContainersPlugin =
        reinterpret_cast<LogContainersPlugin*>(plugin);
    delete logContainersPlugin;

    TraceLoggingWrite(g_deviceHealthContainersProvider, "Uninitialize");

    TraceLoggingUnregister(g_deviceHealthContainersProvider);
}

bool log_containers_log_data(adh_plugin_instance* plugin, const adh_log_data_args* args)
{
    try
    {
        LogContainersPlugin* logContainersPlugin =
            reinterpret_cast<LogContainersPlugin*>(plugin);
        return logContainersPlugin->LogData(args);
    }
    CATCH_LOG();
    return false;
}

DEFINE_LOGGER_PLUGIN(
    log_containers,
    "log-containers",
    log_containers_create_plugin_instance,
    log_containers_destroy_plugin_instance,
    log_containers_log_data);

bool LogContainersPlugin::LogData([[maybe_unused]] const adh_log_data_args* args)
{
    TraceLoggingWrite(g_deviceHealthContainersProvider, "LogData");

    auto containers = Utils::Docker::GetAllContainers();

    if (containers.empty())
    {
        TraceLoggingWrite(g_deviceHealthContainersProvider, "NoContainersFound");
        return true;
    }

    Utils::Docker::AddContainerStats(containers);

    for (auto&& container : containers)
    {
        TraceLoggingWrite(
            g_deviceHealthContainersProvider,
            "ContainerFound",
            TraceLoggingValue(args->sample_key, "SampleKey"),
            TraceLoggingValue(container.CreatedTime.c_str(), "CreatedTime"),
            TraceLoggingValue(container.ImageName.c_str(), "ImageName"),
            TraceLoggingValue(container.ImageTag.c_str(), "ImageTag"),
            TraceLoggingValue(container.ImageId.c_str(), "ImageId"),
            TraceLoggingValue(container.Status.c_str(), "Status"),
            TraceLoggingValue(container.StartedTime.c_str(), "StartedTime"),
            TraceLoggingValue(container.FinishedTime.c_str(), "FinishedTime"),
            TraceLoggingValue(container.SizeRw, "SizeRw"),
            TraceLoggingValue(container.SizeRootFs, "SizeRootFs"),
            TraceLoggingValue(container.MemUsage.c_str(), "StatsMemUsage"),
            TraceLoggingValue(container.CpuUsage.c_str(), "StatsCpuUsage"));
    }

    return true;
}
