// Copyright (c) Microsoft Corporation. All rights reserved.

#include <azure-device-health-plugin/plugin-dev.h>
#include <azure-device-util/PackageManager.h>
#include <azure-device-util/ResultException.h>
#include <azure-device-util/TraceLoggingTypes.h>
#include <jsonbuilder/JsonBuilder.h>
#include <jsonbuilder/JsonRenderer.h>
#include <tracelogging/TraceLoggingProvider.h>

TRACELOGGING_DEFINE_PROVIDER(
    g_deviceHealthPackagesProvider,
    "Azure.DeviceHealth.SoftwareInventory.Packages",
    (0x0fbfba91, 0xb2af, 0x4bba, 0x83, 0x02, 0xff, 0x0a, 0xf8, 0xff, 0xe3, 0x59));

namespace Azure { namespace Device { namespace Health { namespace Plugin {

class LogPackagesPlugin
{
  public:
    bool LogData(const adh_log_data_args* args);

  private:
    // Maximum packages to send per event
    static constexpr size_t c_PackageListLimit = 50;
};

}}}}

namespace Utils = Azure::Device::Utils;
using namespace Azure::Device::Health::Plugin;

adh_plugin_instance* log_packages_create_plugin_instance(
    const adh_plugin_config_pair* configs,
    size_t configs_count)
{
    try
    {
        TraceLoggingRegister(g_deviceHealthPackagesProvider);
        TraceLoggingWrite(g_deviceHealthPackagesProvider, "Initialize");

        for (const adh_plugin_config_pair* config = configs;
             config != (configs + configs_count);
             ++config)
        {
            TraceLoggingWrite(
                g_deviceHealthPackagesProvider,
                "Create_UnusedConfiguration",
                TraceLoggingValue(config->name, "Name"),
                TraceLoggingValue(config->value, "Value"));
        }

        return reinterpret_cast<adh_plugin_instance*>(
            std::make_unique<LogPackagesPlugin>().release());
    }
    CATCH_LOG();
    return nullptr;
}

void log_packages_destroy_plugin_instance(adh_plugin_instance* plugin)
{
    LogPackagesPlugin* logPackagesPlugin =
        reinterpret_cast<LogPackagesPlugin*>(plugin);
    delete logPackagesPlugin;

    TraceLoggingWrite(g_deviceHealthPackagesProvider, "Uninitialize");

    TraceLoggingUnregister(g_deviceHealthPackagesProvider);
}

bool log_packages_log_data(adh_plugin_instance* plugin, const adh_log_data_args* args)
{
    try
    {
        LogPackagesPlugin* logPackagesPlugin =
            reinterpret_cast<LogPackagesPlugin*>(plugin);
        return logPackagesPlugin->LogData(args);
    }
    CATCH_LOG();
    return false;
}

DEFINE_LOGGER_PLUGIN(
    log_packages,
    "log-packages",
    log_packages_create_plugin_instance,
    log_packages_destroy_plugin_instance,
    log_packages_log_data);

bool LogPackagesPlugin::LogData([[maybe_unused]] const adh_log_data_args* args)
{
    TraceLoggingWrite(g_deviceHealthPackagesProvider, "LogData");

    if (!Utils::PackageManager::PackageManagerSupported())
    {
        TraceLoggingWrite(g_deviceHealthPackagesProvider, "NoPackageManager");
        return true;
    }

    auto packages = Utils::PackageManager::GetInstalledPackages();

    if (packages.empty())
    {
        TraceLoggingWrite(g_deviceHealthPackagesProvider, "EmptyPackageList");
        return true;
    }

    size_t package_count = packages.size();

    size_t batch_number = 1;
    // Find the minimum amount of events to send all packages found.
    // Calculate integer division and add 1 if there is a remainder.
    size_t batch_total = package_count / c_PackageListLimit +
        ((package_count % c_PackageListLimit) ? 1 : 0);

    // JsonBuilder to accumulate data
    jsonbuilder::JsonBuilder jb;
    jsonbuilder::JsonIterator arrItr;

    // Setup Json renderer
    jsonbuilder::JsonRenderer renderer;
    renderer.Reserve(2048);

    auto prepareJsonBuilder = [&jb, &arrItr]() {
        jb.clear();
        arrItr = jb.push_back(jb.end(), "data", jsonbuilder::JsonArray);
    };

    auto appendPackage =
        [&jb, &arrItr](const Utils::PackageManager::Package& package) {
            auto itr = jb.push_back(arrItr, "", jsonbuilder::JsonArray);
            jb.push_back(itr, "", package.Name);
            jb.push_back(itr, "", package.Version);
            jb.push_back(itr, "", package.Arch);
        };

    // Setup JsonBuilder for the first time
    prepareJsonBuilder();

    for (size_t package_number = 1; package_number <= package_count;
         package_number++)
    {
        appendPackage(packages[package_number - 1]);

        // If the event cargo is full or this is the last package send event
        if ((package_number % c_PackageListLimit == 0) ||
            (package_number == package_count))
        {
            auto data = renderer.Render(jb);
            TraceLoggingWrite(
                g_deviceHealthPackagesProvider,
                "PackagesFound",
                TraceLoggingValue(args->sample_key, "SampleKey"),
                TraceLoggingValue(batch_number++, "BatchNumber"),
                TraceLoggingValue(batch_total, "BatchTotal"),
                TraceLoggingStringView(data, "Packages"));

            // Reset JsonBuilder
            prepareJsonBuilder();
        }
    }

    return true;
}
