// Copyright (c) Microsoft Corporation. All rights reserved.

#include <map>
#include <numeric>
#include <optional>

#include <azure-device-health-plugin-utils/UdevHelpers.h>
#include <azure-device-health-plugin/plugin-dev.h>
#include <azure-device-util/Bytes.h>
#include <azure-device-util/Hash.h>
#include <azure-device-util/ResultException.h>
#include <tracelogging/TraceLoggingProvider.h>

TRACELOGGING_DEFINE_PROVIDER(
    g_deviceHealthUsbProvider,
    "Azure.DeviceHealth.DeviceInventory.Usb",
    (0x0fbeba91, 0xb2af, 0x4bba, 0x83, 0x02, 0x85, 0x0a, 0xf8, 0xff, 0xe3, 0x59));

namespace PluginUtils = Azure::Device::Health::PluginUtils;

namespace Azure { namespace Device { namespace Health { namespace Plugin {

class LogUsbPlugin
{
  public:
    bool LogData(const adh_log_data_args* args);
};

}}}}

struct UsbDevice
{
    std::string SysName;
    std::string VendorID;
    std::string ProductID;
    std::string Serial;
    std::string Product;
};

using namespace Azure::Device::Health::Plugin;

adh_plugin_instance* log_usb_create_plugin_instance(
    const adh_plugin_config_pair* configs,
    size_t configs_count)
{
    try
    {
        TraceLoggingRegister(g_deviceHealthUsbProvider);
        TraceLoggingWrite(g_deviceHealthUsbProvider, "Initialize");

        for (const adh_plugin_config_pair* config = configs;
             config != (configs + configs_count);
             ++config)
        {
            TraceLoggingWrite(
                g_deviceHealthUsbProvider,
                "Create_UnusedConfiguration",
                TraceLoggingValue(config->name, "Name"),
                TraceLoggingValue(config->value, "Value"));
        }

        return reinterpret_cast<adh_plugin_instance*>(
            std::make_unique<LogUsbPlugin>().release());
    }
    CATCH_LOG();
    return nullptr;
}

void log_usb_destroy_plugin_instance(adh_plugin_instance* plugin)
{
    LogUsbPlugin* logUsbPlugin = reinterpret_cast<LogUsbPlugin*>(plugin);
    delete logUsbPlugin;

    TraceLoggingWrite(g_deviceHealthUsbProvider, "Uninitialize");

    TraceLoggingUnregister(g_deviceHealthUsbProvider);
}

bool log_usb_log_data(adh_plugin_instance* plugin, const adh_log_data_args* args)
{
    try
    {
        LogUsbPlugin* logUsbPlugin = reinterpret_cast<LogUsbPlugin*>(plugin);
        return logUsbPlugin->LogData(args);
    }
    CATCH_LOG();
    return false;
}

DEFINE_LOGGER_PLUGIN(
    log_usb,
    "log-usb",
    log_usb_create_plugin_instance,
    log_usb_destroy_plugin_instance,
    log_usb_log_data);

constexpr const char* c_attrIdVendor = "idVendor";
constexpr const char* c_attrIdProduct = "idProduct";
constexpr const char* c_attrSerial = "serial";
constexpr const char* c_attrProduct = "product";

static std::optional<UsbDevice>
TryCreateUsbDeviceFromDeviceEntry(const PluginUtils::UdevDeviceEntry& deviceEntry)
{
    UsbDevice device;
    device.SysName = deviceEntry.SysName;

    if (auto itr = deviceEntry.Attributes.find(c_attrIdVendor);
        itr != deviceEntry.Attributes.end() && !itr->second.empty())
    {
        device.VendorID = itr->second;
    }
    else
    {
        return {};
    }

    if (auto itr = deviceEntry.Attributes.find(c_attrIdProduct);
        itr != deviceEntry.Attributes.end() && !itr->second.empty())
    {
        device.ProductID = itr->second;
    }
    else
    {
        return {};
    }

    // Optional
    if (auto itr = deviceEntry.Attributes.find(c_attrSerial);
        itr != deviceEntry.Attributes.end() && !itr->second.empty())
    {
        device.Serial = itr->second;
    }

    // Optional
    if (auto itr = deviceEntry.Attributes.find(c_attrProduct);
        itr != deviceEntry.Attributes.end() && !itr->second.empty())
    {
        device.Product = itr->second;
    }

    return device;
}

static void LogDeviceDiscovered(uint64_t sampleKey, const UsbDevice& usbDevice)
{
    uint64_t serialSpookyHash = 0;
    if (!usbDevice.Serial.empty())
    {
        auto serialBytes =
            Azure::Device::Utils::Bytes::StringViewToBytes(usbDevice.Serial);

        serialSpookyHash = Azure::Device::Utils::Hash::Spooky64(serialBytes, 0);
    }

    TraceLoggingWrite(
        g_deviceHealthUsbProvider,
        "DeviceDiscovered",
        TraceLoggingValue(sampleKey, "SampleKey"),
        TraceLoggingValue(usbDevice.SysName.c_str(), "SysName"),
        TraceLoggingValue(usbDevice.ProductID.c_str(), "ProductID"),
        TraceLoggingValue(usbDevice.VendorID.c_str(), "VendorID"),
        TraceLoggingValue(usbDevice.Product.c_str(), "Product"),
        TraceLoggingValue(serialSpookyHash, "SerialHash"));
}

bool LogUsbPlugin::LogData([[maybe_unused]] const adh_log_data_args* args)
{
    TraceLoggingWrite(g_deviceHealthUsbProvider, "LogData");

    constexpr std::array<const char*, 4> c_usbDeviceProps = {
        c_attrIdVendor, c_attrIdProduct, c_attrSerial, c_attrProduct
    };

    auto deviceEntries =
        PluginUtils::EnumerateDeviceAttributes("usb", nullptr, c_usbDeviceProps);

    TraceLoggingWrite(
        g_deviceHealthUsbProvider,
        "Produce_FoundDevices",
        TraceLoggingValue(deviceEntries.size(), "DevicesCount"));

    for (const auto& deviceEntry : deviceEntries)
    {
        std::optional<UsbDevice> optUsbDevice =
            TryCreateUsbDeviceFromDeviceEntry(deviceEntry);

        if (optUsbDevice)
        {
            LogDeviceDiscovered(args->sample_key, *optUsbDevice);
        }
    }

    return true;
}
