// Copyright (c) Microsoft Corporation. All rights reserved.

#include <map>
#include <numeric>
#include <optional>

#include <azure-device-health-plugin-utils/MetricRingStorage.h>
#include <azure-device-health-plugin-utils/UdevHelpers.h>
#include <azure-device-health-plugin/plugin-dev.h>
#include <azure-device-util/Math.h>
#include <azure-device-util/ResultException.h>
#include <azure-device-util/String.h>
#include <tracelogging/TraceLoggingProvider.h>

TRACELOGGING_DEFINE_PROVIDER(
    g_coolingProducerProvider,
    "Azure.DeviceHealth.Plugins.ProduceCooling",
    (0x4bcb0b17, 0xe9c6, 0x824f, 0x83, 0xab, 0x18, 0x78, 0xd7, 0x15, 0x25, 0x7c));

TRACELOGGING_DEFINE_PROVIDER(
    g_coolingConsumerProvider,
    "Azure.DeviceHealth.Plugins.ConsumeCooling",
    (0x86b2a53d, 0xf78a, 0x45f6, 0x80, 0x7c, 0x42, 0xdb, 0xc5, 0xf1, 0x64, 0x1f));

namespace Utils = Azure::Device::Utils;
namespace PluginUtils = Azure::Device::Health::PluginUtils;

namespace Azure { namespace Device { namespace Health { namespace Plugin {

struct CoolingDevice
{
    std::string SysName;
    std::string Type;
    int32_t MaxState = 0;
    int32_t CurState = 0;
};

struct CoolingSample
{
    std::vector<CoolingDevice> CoolingDevices;
};

class ProduceCoolingPlugin
{
  public:
    std::unique_ptr<CoolingSample> ProduceSample(uint64_t sampleKey);
};

class ConsumeCoolingPlugin
{
  public:
    bool ConsumeSample(uint64_t sampleKey, const CoolingSample& coolingSample);

  public:
    PluginUtils::MetricRingStorage<CoolingSample> CoolingSamples;
};

}}}}

using namespace Azure::Device::Health::Plugin;

adh_plugin_instance* produce_cooling_create_plugin_instance(
    const adh_plugin_config_pair* configs,
    size_t configs_count)
{
    try
    {
        TraceLoggingRegister(g_coolingProducerProvider);
        TraceLoggingWrite(g_coolingProducerProvider, "Initialize");

        for (const adh_plugin_config_pair* config = configs;
             config != (configs + configs_count);
             ++config)
        {
            TraceLoggingWrite(
                g_coolingProducerProvider,
                "Create_UnusedConfiguration",
                TraceLoggingValue(config->name, "Name"),
                TraceLoggingValue(config->value, "Value"));
        }

        return reinterpret_cast<adh_plugin_instance*>(
            std::make_unique<ProduceCoolingPlugin>().release());
    }
    CATCH_LOG();
    return nullptr;
}

void produce_cooling_destroy_plugin_instance(adh_plugin_instance* plugin)
{
    ProduceCoolingPlugin* produceCoolingPlugin =
        reinterpret_cast<ProduceCoolingPlugin*>(plugin);
    delete produceCoolingPlugin;

    TraceLoggingWrite(g_coolingProducerProvider, "Uninitialize");
    TraceLoggingUnregister(g_coolingProducerProvider);
}

void* produce_cooling_produce_sample(
    adh_plugin_instance* plugin,
    const adh_produce_sample_args* args)
{
    try
    {
        ProduceCoolingPlugin* produceCoolingPlugin =
            reinterpret_cast<ProduceCoolingPlugin*>(plugin);
        return produceCoolingPlugin->ProduceSample(args->sample_key).release();
    }
    CATCH_LOG();
    return nullptr;
}

void produce_cooling_free_sample(
    [[maybe_unused]] adh_plugin_instance* plugin,
    void* sample)
{
    if (sample)
    {
        delete reinterpret_cast<CoolingSample*>(sample);
    }
}

DEFINE_PRODUCER_PLUGIN(
    produce_cooling,
    "produce-cooling",
    produce_cooling_create_plugin_instance,
    produce_cooling_destroy_plugin_instance,
    produce_cooling_produce_sample,
    produce_cooling_free_sample);

static void
LogInvalidIntegerConfiguration(std::string_view name, std::string_view value)
{
    TraceLoggingWrite(
        g_coolingConsumerProvider,
        "Create_InvalidIntegerConfiguration",
        TraceLoggingCountedString(name.data(), name.size(), "Name"),
        TraceLoggingCountedString(value.data(), value.size(), "Value"));
};

adh_plugin_instance* consume_cooling_create_plugin_instance(
    const adh_plugin_config_pair* configs,
    size_t configs_count)
{
    try
    {
        TraceLoggingRegister(g_coolingConsumerProvider);
        TraceLoggingWrite(g_coolingConsumerProvider, "Initialize");

        auto consumeCoolingPlugin = std::make_unique<ConsumeCoolingPlugin>();

        for (const adh_plugin_config_pair* config = configs;
             config != (configs + configs_count);
             ++config)
        {
            std::string_view name = config->name;
            std::string_view value = config->value;

            if (Utils::String::ICompare(name, "CircularBufferSize") == 0)
            {
                size_t val = 0;
                if (SUCCEEDED(Utils::String::ToUIntTStrict(value, val)))
                {
                    consumeCoolingPlugin->CoolingSamples.SetMaxRingSize(val);
                }
                else
                {
                    LogInvalidIntegerConfiguration(name, value);
                }
            }
            else if (Utils::String::ICompare(name, "ReportCadence") == 0)
            {
                size_t val = 0;
                if (SUCCEEDED(Utils::String::ToUIntTStrict(value, val)))
                {
                    consumeCoolingPlugin->CoolingSamples.SetReportFrequency(val);
                }
                else
                {
                    LogInvalidIntegerConfiguration(name, value);
                }
            }
            else
            {
                TraceLoggingWrite(
                    g_coolingConsumerProvider,
                    "Create_UnusedConfiguration",
                    TraceLoggingValue(config->name, "Name"),
                    TraceLoggingValue(config->value, "Value"));
            }
        }

        return reinterpret_cast<adh_plugin_instance*>(
            consumeCoolingPlugin.release());
    }
    CATCH_LOG();
    return nullptr;
}

void consume_cooling_destroy_plugin_instance(adh_plugin_instance* plugin)
{
    ConsumeCoolingPlugin* consumeCoolingPlugin =
        reinterpret_cast<ConsumeCoolingPlugin*>(plugin);
    delete consumeCoolingPlugin;

    TraceLoggingWrite(g_coolingConsumerProvider, "Uninitialize");
    TraceLoggingUnregister(g_coolingConsumerProvider);
}

static const char* s_coolingConsumeTargets[] = { "produce-cooling", nullptr };
const char* const* consume_cooling_list_consume_targets()
{
    return s_coolingConsumeTargets;
}

bool consume_cooling_consume_sample(
    adh_plugin_instance* plugin,
    const adh_consume_sample_args* args)
{
    try
    {
        if (Utils::String::ICompare(args->producer_name, produce_cooling.name) !=
            0)
        {
            TraceLoggingWrite(
                g_coolingConsumerProvider,
                "PassedUnusedSample",
                TraceLoggingValue(args->producer_name, "ProducerName"));
            return false;
        }

        ConsumeCoolingPlugin* consumeCoolingPlugin =
            reinterpret_cast<ConsumeCoolingPlugin*>(plugin);
        return consumeCoolingPlugin->ConsumeSample(
            args->sample_key, *static_cast<CoolingSample*>(args->sample));
    }
    CATCH_LOG();
    return false;
}

DEFINE_CONSUMER_PLUGIN(
    consume_cooling,
    "consume-cooling",
    consume_cooling_create_plugin_instance,
    consume_cooling_destroy_plugin_instance,
    consume_cooling_list_consume_targets,
    consume_cooling_consume_sample);

constexpr const char* c_deviceCurState = "cur_state";
constexpr const char* c_deviceMaxState = "max_state";
constexpr const char* c_deviceType = "type";

std::optional<CoolingDevice>
TryCreateCoolingDeviceFromDeviceEntry(const PluginUtils::UdevDeviceEntry& deviceEntry)
{
    CoolingDevice device;
    device.SysName = deviceEntry.SysName;

    if (auto itr = deviceEntry.Attributes.find(c_deviceType);
        itr != deviceEntry.Attributes.end() && !itr->second.empty())
    {
        device.Type = itr->second;
    }
    else
    {
        return {};
    }

    if (auto itr = deviceEntry.Attributes.find(c_deviceMaxState);
        itr != deviceEntry.Attributes.end() && !itr->second.empty())
    {
        device.MaxState =
            Azure::Device::Utils::String::ToIntT<int32_t>(itr->second);
    }
    else
    {
        return {};
    }

    if (auto itr = deviceEntry.Attributes.find(c_deviceCurState);
        itr != deviceEntry.Attributes.end() && !itr->second.empty())
    {
        device.CurState =
            Azure::Device::Utils::String::ToIntT<int32_t>(itr->second);
    }
    else
    {
        return {};
    }

    return device;
}

std::unique_ptr<CoolingSample>
ProduceCoolingPlugin::ProduceSample(uint64_t sampleKey)
{
    TraceLoggingWrite(
        g_coolingProducerProvider,
        "Produce",
        TraceLoggingValue(sampleKey, "SampleKey"));

    auto coolingSample = std::make_unique<CoolingSample>();

    constexpr std::array<const char*, 3> c_coolingDeviceProps = {
        c_deviceCurState, c_deviceMaxState, c_deviceType
    };

    auto deviceEntries = PluginUtils::EnumerateDeviceAttributes(
        "thermal", "cooling_device*", c_coolingDeviceProps);

    TraceLoggingWrite(
        g_coolingProducerProvider,
        "Produce_FoundDevices",
        TraceLoggingValue(deviceEntries.size(), "DevicesCount"));

    for (const auto& deviceEntry : deviceEntries)
    {
        std::optional<CoolingDevice> optCoolingDevice =
            TryCreateCoolingDeviceFromDeviceEntry(deviceEntry);

        if (optCoolingDevice)
        {
            coolingSample->CoolingDevices.push_back(std::move(*optCoolingDevice));
        }
    }

    return coolingSample;
}

void GenerateReport(
    uint64_t sampleKey,
    const PluginUtils::MetricRingStorage<CoolingSample>& coolingSamples)
{
    struct DeviceKey
    {
        std::string SysName;
        std::string Type;
        int32_t MaxState;
    };

    struct LessDeviceKey
    {
        bool operator()(const DeviceKey& l, const DeviceKey& r) const
        {
            if (l.SysName < r.SysName)
            {
                return true;
            }
            else if (l.SysName > r.SysName)
            {
                return false;
            }

            if (l.Type < r.Type)
            {
                return true;
            }
            else if (l.Type > r.Type)
            {
                return false;
            }

            return l.MaxState < r.MaxState;
        }
    };

    std::map<DeviceKey, std::vector<int64_t>, LessDeviceKey> coolingDeviceToTemps;
    for (const auto& sample : coolingSamples)
    {
        for (const auto& device : sample.CoolingDevices)
        {
            coolingDeviceToTemps[DeviceKey{ device.SysName, device.Type, device.MaxState }]
                .push_back(device.CurState);
        }
    }

    for (auto&& [deviceKey, temps] : coolingDeviceToTemps)
    {
        std::sort(temps.begin(), temps.end());
        nonstd::span<const int64_t> tempsSpan{ temps };

        using namespace Azure::Device::Utils;

        TraceLoggingWrite(
            g_coolingConsumerProvider,
            "CoolingDeviceStats",
            TraceLoggingValue(sampleKey, "SampleKey"),
            TraceLoggingValue(deviceKey.SysName.c_str(), "SysName"),
            TraceLoggingCountedString(
                deviceKey.Type.data(), deviceKey.Type.size(), "DeviceType"),
            TraceLoggingValue(temps.size(), "NumSamples"),
            TraceLoggingValue(deviceKey.MaxState, "MaxState"),
            TraceLoggingValue(Math::Min(tempsSpan), "CurStateMin"),
            TraceLoggingValue(Math::Max(tempsSpan), "CurStateMax"),
            TraceLoggingValue(Math::Mean(tempsSpan), "CurStateMean"),
            TraceLoggingValue(Math::Median(tempsSpan), "CurStateMedian"));
    }
}

bool ConsumeCoolingPlugin::ConsumeSample(uint64_t sampleKey, const CoolingSample& sample)
{
    TraceLoggingWrite(g_coolingConsumerProvider, "Consume");

    CoolingSamples.AddMetric(sample);

    if (CoolingSamples.IsReportReady())
    {
        TraceLoggingWrite(g_coolingConsumerProvider, "Consume_Reporting");

        GenerateReport(sampleKey, CoolingSamples);
    }

    return true;
}
