// Copyright (c) Microsoft Corporation. All rights reserved.

#include <string>
#include <unistd.h>

#include <azure-device-health-plugin-utils/IProcessMetadataCache.h>
#include <azure-device-health-plugin/plugin-dev.h>
#include <azure-device-util/AndroidLogHelper.h>
#include <azure-device-util/String.h>

#include "CpuProcessMetrics.h"
#include "PerProcessMetricStorage.h"
#include "ProcPidStatParser.h"
#include "Provider.h"

TRACELOGGING_DEFINE_PROVIDER(
    g_perProcessCpuProducerProvider,
    "Azure.DeviceHealth.Plugins.ProducePerProcessCpu",
    (0x2eba4200, 0xe237, 0xb348, 0x8e, 0x05, 0x6a, 0x92, 0x60, 0xe1, 0xc6, 0xe3));

TRACELOGGING_DEFINE_PROVIDER(
    g_perProcessCpuConsumerProvider,
    "Azure.DeviceHealth.Plugins.ConsumePerProcessCpu",
    (0x1116584a, 0xdd89, 0x496a, 0x45, 0x45, 0x4d, 0xa7, 0xa5, 0x74, 0x82, 0x51));

namespace Azure { namespace Device { namespace Health { namespace Plugin {

namespace PluginUtils = Azure::Device::Health::PluginUtils;

class ProducePerProcessCpuPlugin
{
  public:
    std::unique_ptr<PerProcessCpuSample> ProduceSample(uint64_t sampleKey);
};

class ConsumePerProcessCpuPlugin
{
  public:
    explicit ConsumePerProcessCpuPlugin(
        std::unique_ptr<PluginUtils::IProcessMetadataCache> processMetadataCache)
        : _processMetadataCache(std::move(processMetadataCache))
    {
        SampleStorage.SetReportCadence(30);
        SampleStorage.SetBufferSize(30);

        _clockTicksPerSecond = sysconf(_SC_CLK_TCK);
    }

    bool
    ConsumeSample(uint64_t sampleKey, const PerProcessCpuSample& perProcessCpuSample);

  public:
    PerProcessMetricStorage SampleStorage;
    uint32_t ReportProcessCount = 5;

  private:
    uint64_t _clockTicksPerSecond = 0;
    std::unique_ptr<PluginUtils::IProcessMetadataCache> _processMetadataCache;
};

}}}}

namespace Utils = Azure::Device::Utils;
using namespace Azure::Device::Health::Plugin;

adh_plugin_instance* produce_per_process_cpu_create_plugin_instance(
    const adh_plugin_config_pair* configs,
    size_t configs_count)
{
    try
    {
        TraceLoggingRegister(g_perProcessCpuProducerProvider);
        TraceLoggingWrite(g_perProcessCpuProducerProvider, "Initialize");
        ADHS_LOG("Initialize PerProcessCpu Plugin");

        for (const adh_plugin_config_pair* config = configs;
             config != (configs + configs_count);
             ++config)
        {
            TraceLoggingWrite(
                g_perProcessCpuProducerProvider,
                "Create_UnusedConfiguration",
                TraceLoggingValue(config->name, "Name"),
                TraceLoggingValue(config->value, "Value"));
        }

        return reinterpret_cast<adh_plugin_instance*>(
            std::make_unique<ProducePerProcessCpuPlugin>().release());
    }
    CATCH_LOG();
    return nullptr;
}

void produce_per_process_cpu_destroy_plugin_instance(adh_plugin_instance* plugin)
{
    ProducePerProcessCpuPlugin* producePerProcessCpuPlugin =
        reinterpret_cast<ProducePerProcessCpuPlugin*>(plugin);
    delete producePerProcessCpuPlugin;

    TraceLoggingWrite(g_perProcessCpuProducerProvider, "Uninitialize");
    TraceLoggingUnregister(g_perProcessCpuProducerProvider);
}

void* produce_per_process_cpu_produce_sample(
    adh_plugin_instance* plugin,
    const adh_produce_sample_args* args)
{
    try
    {
        ProducePerProcessCpuPlugin* producePerProcessCpuPlugin =
            reinterpret_cast<ProducePerProcessCpuPlugin*>(plugin);
        return producePerProcessCpuPlugin->ProduceSample(args->sample_key).release();
    }
    CATCH_LOG();
    return nullptr;
}

void produce_per_process_cpu_free_sample(
    [[maybe_unused]] adh_plugin_instance* plugin,
    void* sample)
{
    if (sample)
    {
        delete reinterpret_cast<PerProcessCpuSample*>(sample);
    }
}

DEFINE_PRODUCER_PLUGIN(
    produce_per_process_cpu,
    "produce-per-process-cpu",
    produce_per_process_cpu_create_plugin_instance,
    produce_per_process_cpu_destroy_plugin_instance,
    produce_per_process_cpu_produce_sample,
    produce_per_process_cpu_free_sample);

static void
LogInvalidIntegerConfiguration(std::string_view name, std::string_view value)
{
    TraceLoggingWrite(
        g_perProcessCpuConsumerProvider,
        "Create_InvalidIntegerConfiguration",
        TraceLoggingCountedString(name.data(), name.size(), "Name"),
        TraceLoggingCountedString(value.data(), value.size(), "Value"));
};

adh_plugin_instance* consume_per_process_cpu_create_plugin_instance(
    const adh_plugin_config_pair* configs,
    size_t configs_count)
{
    try
    {
        TraceLoggingRegister(g_perProcessCpuConsumerProvider);
        TraceLoggingWrite(g_perProcessCpuConsumerProvider, "Initialize");

        auto consumePerProcessCpuPlugin =
            std::make_unique<ConsumePerProcessCpuPlugin>(
                PluginUtils::IProcessMetadataCache::Create());

        for (const adh_plugin_config_pair* config = configs;
             config != (configs + configs_count);
             ++config)
        {
            std::string_view name = config->name;
            std::string_view value = config->value;

            if (Utils::String::ICompare(name, "CircularBufferSize") == 0)
            {
                size_t val = 0;
                if (SUCCEEDED(Utils::String::ToUIntTStrict(value, val)))
                {
                    consumePerProcessCpuPlugin->SampleStorage.SetBufferSize(val);
                }
                else
                {
                    LogInvalidIntegerConfiguration(name, value);
                }
            }
            else if (Utils::String::ICompare(name, "ReportCadence") == 0)
            {
                size_t val = 0;
                if (SUCCEEDED(Utils::String::ToUIntTStrict(value, val)))
                {
                    consumePerProcessCpuPlugin->SampleStorage.SetReportCadence(
                        val);
                }
                else
                {
                    LogInvalidIntegerConfiguration(name, value);
                }
            }
            else if (Utils::String::ICompare(name, "ReportProcessCount") == 0)
            {
                size_t val = 0;
                if (SUCCEEDED(Utils::String::ToUIntTStrict(value, val)))
                {
                    consumePerProcessCpuPlugin->ReportProcessCount = val;
                }
                else
                {
                    LogInvalidIntegerConfiguration(name, value);
                }
            }
            else
            {
                TraceLoggingWrite(
                    g_perProcessCpuConsumerProvider,
                    "Create_UnusedConfiguration",
                    TraceLoggingValue(config->name, "Name"),
                    TraceLoggingValue(config->value, "Value"));
            }
        }

        return reinterpret_cast<adh_plugin_instance*>(
            consumePerProcessCpuPlugin.release());
    }
    CATCH_LOG();
    return nullptr;
}

void consume_per_process_cpu_destroy_plugin_instance(adh_plugin_instance* plugin)
{
    ConsumePerProcessCpuPlugin* consumePerProcessCpuPlugin =
        reinterpret_cast<ConsumePerProcessCpuPlugin*>(plugin);
    delete consumePerProcessCpuPlugin;

    TraceLoggingWrite(g_perProcessCpuConsumerProvider, "Uninitialize");
    TraceLoggingUnregister(g_perProcessCpuConsumerProvider);
}

static const char* s_perProcessCpuConsumeTargets[] = { "produce-per-process-cpu",
                                                       nullptr };
const char* const* consume_per_process_cpu_list_consume_targets()
{
    return s_perProcessCpuConsumeTargets;
}

bool consume_per_process_cpu_consume_sample(
    adh_plugin_instance* plugin,
    const adh_consume_sample_args* args)
{
    try
    {
        if (Utils::String::ICompare(
                args->producer_name, produce_per_process_cpu.name) != 0)
        {
            TraceLoggingWrite(
                g_perProcessCpuConsumerProvider,
                "PassedUnusedSample",
                TraceLoggingValue(args->producer_name, "ProducerName"));
            return false;
        }

        ConsumePerProcessCpuPlugin* consumePerProcessCpuPlugin =
            reinterpret_cast<ConsumePerProcessCpuPlugin*>(plugin);
        return consumePerProcessCpuPlugin->ConsumeSample(
            args->sample_key, *static_cast<PerProcessCpuSample*>(args->sample));
    }
    CATCH_LOG();
    return false;
}

DEFINE_CONSUMER_PLUGIN(
    consume_per_process_cpu,
    "consume-per-process-cpu",
    consume_per_process_cpu_create_plugin_instance,
    consume_per_process_cpu_destroy_plugin_instance,
    consume_per_process_cpu_list_consume_targets,
    consume_per_process_cpu_consume_sample);

std::unique_ptr<PerProcessCpuSample>
ProducePerProcessCpuPlugin::ProduceSample(uint64_t sampleKey)
{
    TraceLoggingWrite(
        g_perProcessCpuProducerProvider,
        "Produce",
        TraceLoggingValue(sampleKey, "SampleKey"));

    auto perProcessCpuSample = std::make_unique<PerProcessCpuSample>();

    std::vector<CpuProcessInfo> processInfos;

    ProcPidStatParser parser;
    if (!parser.Parse(processInfos))
    {
        return nullptr;
    }

    perProcessCpuSample->Processes = std::move(processInfos);

    return perProcessCpuSample;
}

bool ConsumePerProcessCpuPlugin::ConsumeSample(
    uint64_t sampleKey,
    const PerProcessCpuSample& sample)
{
    TraceLoggingWrite(g_perProcessCpuConsumerProvider, "Consume");

    SampleStorage.AddMetric(sampleKey, sample);

    if (SampleStorage.IsReportReady())
    {
        TraceLoggingWrite(g_perProcessCpuConsumerProvider, "Consume_Reporting");

        auto metrics = SampleStorage.GenerateReportMetrics(ReportProcessCount);

        SampleStorage.FireReportMetrics(
            *_processMetadataCache, sampleKey, _clockTicksPerSecond, metrics);
    }

    return true;
}
