// Copyright (c) Microsoft Corporation. All rights reserved.

#include "PerProcessMetricStorage.h"

#include <chrono>
#include <type_traits>
#include <unordered_map>
#include <utility>

#include <azure-device-util/String.h>

#include "Provider.h"

#ifdef AZURE_DEVICE_USE_CLOG
    #define CLOG_INLINE_IMPLEMENTATION 1
    #include "PerProcessMetricStorage.cpp.clog.h"
    #include "privacytags.h"
#else
    #include "clog_stubs.h"
#endif

namespace Azure { namespace Device { namespace Health { namespace Plugin {

namespace PluginUtils = Azure::Device::Health::PluginUtils;

void PerProcessMetricStorage::AddMetric(
    uint64_t sampleKey,
    const PerProcessCpuSample& sample)
{
    //
    // Our very first sample is our delta base and is basically ignored to keep
    // our reporting window about the same.
    //
    if (_oldestReportedEntry != -1)
    {
        _samplesSinceLastReport++;
    }

    //
    // Update oldest valid entry. Note that for our very first sample, we treat
    // the entry as already reported to use it as our delta. Going forward, we
    // will always use the last reported entry as the base for our delta.
    //
    if (_oldestReportedEntry == -1)
    {
        _oldestReportedEntry = 0;
    }
    else if (_currentRingEntry == _oldestReportedEntry)
    {
        IncrementWithWrap(_oldestReportedEntry, _ringMetrics.size());
    }

    _ringMetrics[_currentRingEntry] = std::make_pair(sampleKey, sample.Processes);
    _newestValidEntry = _currentRingEntry;

    IncrementWithWrap(_currentRingEntry, _ringMetrics.size());
}

/*++
    Reports top N processes by CPU usage in total window.

    TODO: We can also use this data to do (high CPU over a longer period and
    high CPU in a single sample)
--*/
PerProcessMetricStorage::UserAndSystemMetrics
PerProcessMetricStorage::GenerateReportMetrics(uint32_t reportProcessCount)
{
    TraceLoggingWrite(g_perProcessCpuConsumerProvider, "ReportMetrics");

    // Calculate total time period covered
    const auto& oldestSample = _ringMetrics[_oldestReportedEntry];
    const auto& newestSample = _ringMetrics[_newestValidEntry];

    auto windowStart = std::chrono::milliseconds(oldestSample.first);
    auto windowEnd = std::chrono::milliseconds(newestSample.first);
    auto totalWindowMs = windowEnd - windowStart;

    // Calculate most expensive processes between oldest and newest sample
    if (oldestSample.second.size() == 0 || newestSample.second.size() == 0)
    {
        // Nothing to report
        // TODO: Fix
        UserAndSystemMetrics metrics;
        metrics.TotalWindowMs = totalWindowMs.count();
        return metrics;
    }

    auto userMetrics = GenerateReportMetricsHelper(
        reportProcessCount,
        oldestSample,
        newestSample,
        CpuProcessState::CPU_PROCESS_STATE_USER);

    auto systemMetrics = GenerateReportMetricsHelper(
        reportProcessCount,
        oldestSample,
        newestSample,
        CpuProcessState::CPU_PROCESS_STATE_SYSTEM);

    _samplesSinceLastReport = 0;

    // Assign oldest reported to the last metric reported
    _oldestReportedEntry = _newestValidEntry;

    return UserAndSystemMetrics{ static_cast<int32_t>(totalWindowMs.count()),
                                 std::move(userMetrics),
                                 std::move(systemMetrics) };
}

std::vector<PerProcessMetricStorage::ReportMetrics>
PerProcessMetricStorage::GenerateReportMetricsHelper(
    uint32_t reportProcessCount,
    const std::pair<unsigned long, std::vector<CpuProcessInfo>>& oldestSample,
    const std::pair<unsigned long, std::vector<CpuProcessInfo>>& newestSample,
    CpuProcessState sampleIndex) const
{
    //
    // Since we know the two vectors are sorted, we can loop over them and
    // advance the indexes based on the process ID we find at the given index.
    // For each matching pair we find, we can create a process_diff_t entry
    // which will have the delta between the two samples. For any item
    // which doesn't fit into a pair (by process ID + name), we will ignore
    // the entry since we cannot calculate a difference.
    //
    size_t oldestSampleIdx = 0;
    size_t newestSampleIdx = 0;

    std::vector<ReportMetrics> result;
    while (true)
    {
        const auto& oldEntry = oldestSample.second.at(oldestSampleIdx);
        const auto& newEntry = newestSample.second.at(newestSampleIdx);

        if (oldEntry.ProcessId == newEntry.ProcessId &&
            Utils::String::ICompare(
                oldEntry.ProcessName, newEntry.ProcessName) == 0)
        {
            result.emplace_back(ReportMetrics{
                oldEntry.ProcessId,
                oldEntry.ProcessName,
                newEntry.Sample[sampleIndex] - oldEntry.Sample[sampleIndex] });

            oldestSampleIdx++;
            newestSampleIdx++;
        }
        else if (oldEntry.ProcessId < newEntry.ProcessId)
        {
            //
            // The process no longer exists, we don't have an end state to
            // compare, so treat it as zero.
            //
            oldestSampleIdx++;
        }
        else if (newEntry.ProcessId < oldEntry.ProcessId)
        {
            //
            // This process didn't exist in the old sample, so count the whole
            // sample amount since our delta base is zero.
            //
            result.emplace_back(ReportMetrics{ newEntry.ProcessId,
                                               newEntry.ProcessName,
                                               newEntry.Sample[sampleIndex] });

            newestSampleIdx++;
        }
        else
        {
            //
            // Mismatching process name, increment both. Treat the newest one as
            // truth.
            //

            result.emplace_back(ReportMetrics{ newEntry.ProcessId,
                                               newEntry.ProcessName,
                                               newEntry.Sample[sampleIndex] });

            oldestSampleIdx++;
            newestSampleIdx++;
        }

        // If either list reaches the end, terminate
        if (oldestSampleIdx == oldestSample.second.size() ||
            newestSampleIdx == newestSample.second.size())
        {
            break;
        }
    }

    // Sort in descending order (biggest first)
    std::sort(
        result.begin(),
        result.end(),
        [](const ReportMetrics& first, const ReportMetrics& second) {
            if (first.ClockTicksOverPeriod > second.ClockTicksOverPeriod)
            {
                return true;
            }

            return false;
        });

    // Trim out lower usage processes
    if (reportProcessCount < result.size())
    {
        result.resize(reportProcessCount);
    }

    return result;
}

void PerProcessMetricStorage::FireReportMetrics(
    PluginUtils::IProcessMetadataCache& processMetadataCache,
    uint64_t sampleKey,
    uint64_t clockTicksPerSecond,
    const UserAndSystemMetrics& metrics) const
{
    // Report top N processes by  time over this period

    auto reportLambda = [sampleKey, clockTicksPerSecond, &processMetadataCache](
                            int32_t totalWindowMs,
                            CpuProcessState sampleIndex,
                            const std::vector<ReportMetrics>& metrics) {
        int rank = 1;
        for (const auto& entry : metrics)
        {
            if (entry.ClockTicksOverPeriod == 0)
            {
                // Since these are sorted in descending order, once we find
                // a zero, we're done
                break;
            }

            auto processProps =
                processMetadataCache.GetProcessPropertiesByPid(entry.ProcessId);

            std::string type =
                (sampleIndex == CpuProcessState::CPU_PROCESS_STATE_USER) ?
                "User" :
                "System";

            // Keep this consistent with Asimov schema style
            TraceLoggingWrite(
                g_perProcessCpuConsumerProvider,
                "ProcessCpuStats",
                TraceLoggingValue(type.c_str(), "Type"),
                TraceLoggingValue(sampleKey, "SampleKey"),
                TraceLoggingValue(rank, "Rank"),
                TraceLoggingValue(totalWindowMs, "TotalWindowDurationMs"),
                TraceLoggingValue(clockTicksPerSecond, "ClockTicksPerSecond"),
                TraceLoggingValue(entry.ProcessId, "ProcessId"),
                TraceLoggingString(entry.ProcessName.data(), "ProcessName"),
                TraceLoggingValue(
                    processProps->PackageName.c_str(), "PackageName"),
                TraceLoggingValue(
                    processProps->PackageVersion.c_str(), "PackageVersion"),
                TraceLoggingValue(
                    processProps->PackageArch.c_str(), "PackageArch"),
                TraceLoggingValue(
                    processProps->SystemdCGroupPath.c_str(), "SystemdCgroupPath"),
                TraceLoggingValue(
                    processProps->DockerInstanceId.c_str(), "DockerInstanceId"),
                TraceLoggingValue(
                    processProps->DockerImageName.c_str(), "DockerImageName"),
                TraceLoggingValue(
                    processProps->DockerImageId.c_str(), "DockerImageId"),
                TraceLoggingValue(entry.ClockTicksOverPeriod, "TotalClockTicks"));

            LatteTelemetry(
                ProcessCpuStats,
                "Azure.DeviceHealth.Plugins.ConsumePerProcessCpu",
                PrivacyTags_ProductAndServicePerformance,
                "%{Type,s} \
                %{SampleKey,d} \
                %{Rank,d} \
                %{TotalWindowDurationMs,lld} \
                %{ClockTicksPerSecond,llu} \
                %{ProcessId,u} \
                %{ProcessName,s} \
                %{PackageName,s} \
                %{PackageVersion,s} \
                %{PackageArch,s} \
                %{SystemdCgroupPath,s} \
                %{DockerInstanceId,s} \
                %{DockerImageName,s} \
                %{DockerImageId,s} \
                %{TotalClockTicks,lld}",
                type.c_str(),
                sampleKey,
                rank,
                totalWindowMs,
                clockTicksPerSecond,
                entry.ProcessId,
                entry.ProcessName.c_str(),
                processProps->PackageName.c_str(),
                processProps->PackageVersion.c_str(),
                processProps->PackageArch.c_str(),
                processProps->SystemdCGroupPath.c_str(),
                processProps->DockerInstanceId.c_str(),
                processProps->DockerImageName.c_str(),
                processProps->DockerImageId.c_str(),
                entry.ClockTicksOverPeriod);

            rank++;
        }
    };

    reportLambda(
        metrics.TotalWindowMs,
        CpuProcessState::CPU_PROCESS_STATE_USER,
        metrics.UserMetrics);
    reportLambda(
        metrics.TotalWindowMs,
        CpuProcessState::CPU_PROCESS_STATE_SYSTEM,
        metrics.SystemMetrics);
}

}}}}
