// Copyright (c) Microsoft Corporation. All rights reserved.

#pragma once

#include <string>

#include <azure-device-util/FileSystem.h>
#include <azure-device-util/String.h>

#include "CpuProcessMetrics.h"

namespace Azure { namespace Device { namespace Health { namespace Plugin {

class ProcPidStatParser
{
  public:
    ProcPidStatParser() : _rootPath("/") {}

    ProcPidStatParser(std::string_view rootPath) : _rootPath(rootPath) {}

    bool Parse(std::vector<CpuProcessInfo>& infos)
    {
        try
        {
            std::string procRoot = _rootPath;
            procRoot.append("proc");

            Utils::FileSystem::DoForEachFileWithPattern(
                procRoot, R"(^\d+$)", [&](const std::filesystem::path& path) {
                    uint32_t processId = atoi(path.filename().c_str());
                    if (processId != 0)
                    {
                        // Get processor sample for this PID
                        {
                            std::stringstream processCpuInfoStream;
                            processCpuInfoStream << procRoot << "/" << processId
                                                 << "/stat";
                            std::string contents;
                            if (FAILED(Utils::FileSystem::ReadUTF8StringFromFile(
                                    processCpuInfoStream.str(), contents)))
                            {
                                // Skip this process
                                return;
                            }

                            infos.push_back(ParseText(contents));
                        }
                    }
                });

            //
            // Sort out-going data by PID so consumers can depend on
            // that as an invariant
            //
            std::sort(infos.begin(), infos.end(), CompareCpuProcessInfo);

            return true;
        }
        CATCH_LOG();

        return false;
    }

  private:
    static bool
    CompareCpuProcessInfo(const CpuProcessInfo& first, const CpuProcessInfo& second)
    {
        return first.ProcessId < second.ProcessId;
    }

    CpuProcessInfo ParseText(std::string_view procPidText)
    {
        CpuProcessInfo info = {};

        if (procPidText.empty())
        {
            // TODO: Deal with invalid samples in some way
            return info;
        }

        //
        // Since executable names can contain spaces and
        // close parens, hand parse the first few elements.
        //
        size_t pidLength = procPidText.find(' ');
        size_t openParenIndex = pidLength + 1;
        size_t closeParenIndex = procPidText.rfind(')');

        if (pidLength == std::string::npos || closeParenIndex == std::string::npos)
        {
            // Give up if we can't find the basic process info
            return info;
        }

        std::string pidString = std::string{ procPidText.data(), 0, pidLength };

        info.ProcessId = atol(pidString.c_str());

        info.ProcessName = std::string{ &procPidText[openParenIndex + 1],
                                        closeParenIndex - openParenIndex - 1 };

        std::string_view splitView;
        if (procPidText.size() > closeParenIndex + 2)
        {
            splitView = procPidText.substr(closeParenIndex + 2);
        }

        std::vector<std::string> entries = Utils::String::Split(splitView, " ");

        constexpr uint32_t k_userTimeIndex = 11;
        constexpr uint32_t k_systemTimeIndex = 12;
        constexpr uint32_t k_childUserTimeIndex = 13;
        constexpr uint32_t k_childSystemTimeIndex = 14;
        constexpr uint32_t k_processStartTime = 19;
        uint32_t entryId = 0;

        for (const auto& entry : entries)
        {
            switch (entryId)
            {
            case k_userTimeIndex:
                info.Sample[CpuProcessState::CPU_PROCESS_STATE_USER] =
                    atol(entry.c_str());
                break;
            case k_systemTimeIndex:
                info.Sample[CpuProcessState::CPU_PROCESS_STATE_SYSTEM] =
                    atol(entry.c_str());
                break;
            case k_childUserTimeIndex:
                info.Sample[CpuProcessState::CPU_PROCESS_STATE_CHILD_USER] =
                    atol(entry.c_str());
                break;
            case k_childSystemTimeIndex:
                info.Sample[CpuProcessState::CPU_PROCESS_STATE_CHILD_SYSTEM] =
                    atol(entry.c_str());
                break;
            case k_processStartTime:
                info.Sample[CpuProcessState::CPU_PROCESS_STATE_START_TIME] =
                    atol(entry.c_str());
                break;
            }

            entryId++;
        }

        // Get the command line for this PID
        {
            std::stringstream commandLineStream;
            commandLineStream << _rootPath << "proc/" << info.ProcessId
                              << "/cmdline";
            std::string procPidText;
            if (SUCCEEDED(Utils::FileSystem::ReadUTF8StringFromFile(
                    commandLineStream.str(), procPidText)))
            {
                info.CommandLine = procPidText;
            }
        }
        return info;
    }

  private:
    std::string _rootPath;
};

}}}}
