// Copyright (c) Microsoft Corporation. All rights reserved.

#include <algorithm>
#include <string>

#include <azure-device-health-plugin-utils/IProcessMetadataCache.h>
#include <azure-device-health-plugin/plugin-dev.h>
#include <azure-device-util/AndroidLogHelper.h>
#include <azure-device-util/FileSystem.h>
#include <azure-device-util/String.h>
#include <tracelogging/TraceLoggingProvider.h>

#ifdef AZURE_DEVICE_USE_CLOG
    #define CLOG_INLINE_IMPLEMENTATION 1
    #include "PerProcessMemoryMain.cpp.clog.h"
    #include "privacytags.h"
#else
    #include "clog_stubs.h"
#endif

#include "MemoryProcessMetrics.h"

TRACELOGGING_DEFINE_PROVIDER(
    g_perProcessMemoryProducerProvider,
    "Azure.DeviceHealth.Plugins.ProducePerProcessMemory",
    (0xaa206621, 0x7449, 0x544b, 0x93, 0xef, 0x74, 0x9b, 0x4d, 0xee, 0x08, 0xa7));

TRACELOGGING_DEFINE_PROVIDER(
    g_perProcessMemoryConsumerProvider,
    "Azure.DeviceHealth.Plugins.ConsumePerProcessMemory",
    (0x4c86e890, 0xd5ca, 0x9448, 0xae, 0xb3, 0xe6, 0x52, 0x4b, 0x15, 0x06, 0x35));

namespace Azure { namespace Device { namespace Health { namespace Plugin {

namespace PluginUtils = Azure::Device::Health::PluginUtils;

class ProducePerProcessMemoryPlugin
{
  public:
    std::unique_ptr<PerProcessMemorySample> ProduceSample(uint64_t sampleKey);
};

class ConsumePerProcessMemoryPlugin
{
  public:
    explicit ConsumePerProcessMemoryPlugin(
        std::unique_ptr<PluginUtils::IProcessMetadataCache> processMetadataCache)
        : _processMetadataCache(std::move(processMetadataCache))
    {}

    bool ConsumeSample(
        uint64_t sampleKey,
        const PerProcessMemorySample& perProcessMemorySample);

  public:
    uint32_t ReportProcessCount = 5;

  private:
    std::unique_ptr<PluginUtils::IProcessMetadataCache> _processMetadataCache;
};

}}}}

namespace Utils = Azure::Device::Utils;
using namespace Azure::Device::Health::Plugin;

adh_plugin_instance* produce_per_process_memory_create_plugin_instance(
    const adh_plugin_config_pair* configs,
    size_t configs_count)
{
    try
    {
        TraceLoggingRegister(g_perProcessMemoryProducerProvider);
        TraceLoggingWrite(g_perProcessMemoryProducerProvider, "Initialize");
        ADHS_LOG("Initialize PerProcessMemory Plugin");

        for (const adh_plugin_config_pair* config = configs;
             config != (configs + configs_count);
             ++config)
        {
            TraceLoggingWrite(
                g_perProcessMemoryProducerProvider,
                "Create_UnusedConfiguration",
                TraceLoggingValue(config->name, "Name"),
                TraceLoggingValue(config->value, "Value"));
        }

        return reinterpret_cast<adh_plugin_instance*>(
            std::make_unique<ProducePerProcessMemoryPlugin>().release());
    }
    CATCH_LOG();
    return nullptr;
}

void produce_per_process_memory_destroy_plugin_instance(adh_plugin_instance* plugin)
{
    ProducePerProcessMemoryPlugin* producePerProcessMemoryPlugin =
        reinterpret_cast<ProducePerProcessMemoryPlugin*>(plugin);
    delete producePerProcessMemoryPlugin;

    TraceLoggingWrite(g_perProcessMemoryProducerProvider, "Uninitialize");
    TraceLoggingUnregister(g_perProcessMemoryProducerProvider);
}

void* produce_per_process_memory_produce_sample(
    adh_plugin_instance* plugin,
    const adh_produce_sample_args* args)
{
    try
    {
        ProducePerProcessMemoryPlugin* producePerProcessMemoryPlugin =
            reinterpret_cast<ProducePerProcessMemoryPlugin*>(plugin);
        return producePerProcessMemoryPlugin->ProduceSample(args->sample_key)
            .release();
    }
    CATCH_LOG();
    return nullptr;
}

void produce_per_process_memory_free_sample(
    [[maybe_unused]] adh_plugin_instance* plugin,
    void* sample)
{
    if (sample)
    {
        delete reinterpret_cast<PerProcessMemorySample*>(sample);
    }
}

DEFINE_PRODUCER_PLUGIN(
    produce_per_process_memory,
    "produce-per-process-memory",
    produce_per_process_memory_create_plugin_instance,
    produce_per_process_memory_destroy_plugin_instance,
    produce_per_process_memory_produce_sample,
    produce_per_process_memory_free_sample);

static void
LogInvalidIntegerConfiguration(std::string_view name, std::string_view value)
{
    TraceLoggingWrite(
        g_perProcessMemoryConsumerProvider,
        "Create_InvalidIntegerConfiguration",
        TraceLoggingCountedString(name.data(), name.size(), "Name"),
        TraceLoggingCountedString(value.data(), value.size(), "Value"));
};

adh_plugin_instance* consume_per_process_memory_create_plugin_instance(
    const adh_plugin_config_pair* configs,
    size_t configs_count)
{
    try
    {
        TraceLoggingRegister(g_perProcessMemoryConsumerProvider);
        TraceLoggingWrite(g_perProcessMemoryConsumerProvider, "Initialize");

        auto consumePerProcessMemoryPlugin =
            std::make_unique<ConsumePerProcessMemoryPlugin>(
                PluginUtils::IProcessMetadataCache::Create());

        for (const adh_plugin_config_pair* config = configs;
             config != (configs + configs_count);
             ++config)
        {
            std::string_view name = config->name;
            std::string_view value = config->value;

            if (Utils::String::ICompare(name, "ReportProcessCount") == 0)
            {
                size_t val = 0;
                if (SUCCEEDED(Utils::String::ToUIntTStrict(value, val)))
                {
                    consumePerProcessMemoryPlugin->ReportProcessCount = val;
                }
                else
                {
                    LogInvalidIntegerConfiguration(name, value);
                }
            }
            else
            {
                TraceLoggingWrite(
                    g_perProcessMemoryConsumerProvider,
                    "Create_UnusedConfiguration",
                    TraceLoggingValue(config->name, "Name"),
                    TraceLoggingValue(config->value, "Value"));
            }
        }

        return reinterpret_cast<adh_plugin_instance*>(
            consumePerProcessMemoryPlugin.release());
    }
    CATCH_LOG();
    return nullptr;
}

void consume_per_process_memory_destroy_plugin_instance(adh_plugin_instance* plugin)
{
    ConsumePerProcessMemoryPlugin* consumePerProcessMemoryPlugin =
        reinterpret_cast<ConsumePerProcessMemoryPlugin*>(plugin);
    delete consumePerProcessMemoryPlugin;

    TraceLoggingWrite(g_perProcessMemoryConsumerProvider, "Uninitialize");
    TraceLoggingUnregister(g_perProcessMemoryConsumerProvider);
}

static const char* s_perProcessMemoryConsumeTargets[] = {
    "produce-per-process-memory",
    nullptr
};
const char* const* consume_per_process_memory_list_consume_targets()
{
    return s_perProcessMemoryConsumeTargets;
}

bool consume_per_process_memory_consume_sample(
    adh_plugin_instance* plugin,
    const adh_consume_sample_args* args)
{
    try
    {
        if (Utils::String::ICompare(
                args->producer_name, produce_per_process_memory.name) != 0)
        {
            TraceLoggingWrite(
                g_perProcessMemoryConsumerProvider,
                "PassedUnusedSample",
                TraceLoggingValue(args->producer_name, "ProducerName"));
            return false;
        }

        ConsumePerProcessMemoryPlugin* consumePerProcessMemoryPlugin =
            reinterpret_cast<ConsumePerProcessMemoryPlugin*>(plugin);
        return consumePerProcessMemoryPlugin->ConsumeSample(
            args->sample_key,
            *static_cast<PerProcessMemorySample*>(args->sample));
    }
    CATCH_LOG();
    return false;
}

DEFINE_CONSUMER_PLUGIN(
    consume_per_process_memory,
    "consume-per-process-memory",
    consume_per_process_memory_create_plugin_instance,
    consume_per_process_memory_destroy_plugin_instance,
    consume_per_process_memory_list_consume_targets,
    consume_per_process_memory_consume_sample);

std::unique_ptr<PerProcessMemorySample>
ProducePerProcessMemoryPlugin::ProduceSample(uint64_t sampleKey)
{
    TraceLoggingWrite(
        g_perProcessMemoryProducerProvider,
        "Produce",
        TraceLoggingValue(sampleKey, "SampleKey"));

    auto perProcessMemorySample = std::make_unique<PerProcessMemorySample>();

    std::vector<MemoryProcessInfo> processInfos;

    Utils::FileSystem::DoForEachFileWithPattern(
        "/proc", R"(^\d+$)", [&](const std::filesystem::path& path) {
            uint32_t processId = atoi(path.filename().c_str());
            if (processId != 0)
            {
                MemoryProcessInfo info = {};

                // Get processor sample for this PID
                {
                    std::stringstream processMemoryInfoStream;
                    processMemoryInfoStream << "/proc/" << processId << "/status";
                    std::string contents;
                    if (FAILED(Utils::FileSystem::ReadUTF8StringFromFile(
                            processMemoryInfoStream.str(), contents)))
                    {
                        // Skip this process
                        return;
                    }

                    info.ProcessId = processId;

                    using namespace std::string_view_literals;

                    // Memory metric names
                    constexpr std::string_view names[MemProcessInfoEnum::NUM_VALUES] = {
                        "VmPeak:\t"sv,  "VmSize:\t"sv,  "VmLck:\t"sv,
                        "VmPin:\t"sv,   "VmHWM:\t"sv,   "VmRSS:\t"sv,
                        "RssAnon:\t"sv, "RssFile:\t"sv, "RssShmem:\t"sv,
                        "VmData:\t"sv,  "VmStk:\t"sv,   "VmExe:\t"sv,
                        "VmLib:\t"sv,   "VmPTE:\t"sv,   "VmPMD:\t"sv,
                        "VmSwap:\t"sv
                    };

                    auto lineSplitContents = Utils::String::Split(contents, "\n");
                    bool firstEntry = true;
                    for (const auto& line : lineSplitContents)
                    {
                        // The first entry is the process name
                        if (firstEntry)
                        {
                            constexpr std::string_view processNamePrefix =
                                "Name:\t"sv;
                            info.ProcessName = &line[processNamePrefix.size()];
                            firstEntry = false;
                            continue;
                        }

                        //
                        // Quick-filter away lines that are non-memory
                        // related. This must stay synchronized with
                        // names array. All VM and Resident Set info
                        // starts with V or R.
                        //
                        if (line.empty() || (line[0] != 'V' && line[0] != 'R'))
                        {
                            continue;
                        }

                        auto takeSampleLambda = [&line, &info, &names](
                                                    uint32_t index) {
                            //
                            // Try to match this line against a given
                            // entry in the names array
                            //
                            if (Utils::String::IStartsWith(line, names[index]))
                            {
                                std::string sampleStr =
                                    std::string{ &line[names[index].size()],
                                                 line.size() -
                                                     names[index].size() };

                                //
                                // The string here has leading spaces
                                // and trailing "KB". We rely on atoll
                                // to strip those for us.
                                //
                                info.Sample[index] = atoll(sampleStr.c_str());
                                return true;
                            }

                            return false;
                        };

                        //
                        // For each line, attempt to match it against
                        // all interesting metrics. Stop if a match is
                        // found.
                        //
                        for (int idx = 0; idx < MemProcessInfoEnum::NUM_VALUES;
                             idx++)
                        {
                            if (takeSampleLambda(idx))
                            {
                                break;
                            }
                        }
                    }

                    processInfos.push_back(std::move(info));
                }
            }
        });

    // Sort out-going data by PID so consumers can depend on that as  an
    // invariant
    std::sort(
        processInfos.begin(),
        processInfos.end(),
        [](const MemoryProcessInfo& first, const MemoryProcessInfo& second) {
            return first.ProcessId < second.ProcessId;
        });

    perProcessMemorySample->Processes = std::move(processInfos);

    return perProcessMemorySample;
}

bool ConsumePerProcessMemoryPlugin::ConsumeSample(
    uint64_t sampleKey,
    const PerProcessMemorySample& sample)
{
    TraceLoggingWrite(g_perProcessMemoryConsumerProvider, "Consume");

    std::vector<MemoryProcessInfo> processInfoCopy{ sample.Processes };

    // Sort in descending order of resident set size (biggest first)
    std::sort(
        processInfoCopy.begin(),
        processInfoCopy.end(),
        [](const MemoryProcessInfo& first, const MemoryProcessInfo& second) {
            if (first.Sample[MEM_PROCESS_INFO_RESIDENT_SIZE_KB] >
                second.Sample[MEM_PROCESS_INFO_RESIDENT_SIZE_KB])
            {
                return true;
            }

            return false;
        });

    // Report top N processes
    if (ReportProcessCount < processInfoCopy.size())
    {
        processInfoCopy.resize(ReportProcessCount);
    }

    //
    // Report top N processes by total resident set size over this
    // period Keep this consistent with Asimov schema style
    //
    int rank = 1;
    for (const auto& entry : processInfoCopy)
    {
        auto processProps =
            _processMetadataCache->GetProcessPropertiesByPid(entry.ProcessId);

        TraceLoggingWrite(
            g_perProcessMemoryConsumerProvider,
            "ProcessMemoryStats",
            TraceLoggingValue(sampleKey, "SampleKey"),
            TraceLoggingValue(rank, "Rank"),
            TraceLoggingValue(entry.ProcessId, "ProcessId"),
            TraceLoggingString(entry.ProcessName.data(), "ProcessName"),
            TraceLoggingValue(processProps->PackageName.c_str(), "PackageName"),
            TraceLoggingValue(
                processProps->PackageVersion.c_str(), "PackageVersion"),
            TraceLoggingValue(processProps->PackageArch.c_str(), "PackageArch"),
            TraceLoggingValue(
                processProps->SystemdCGroupPath.c_str(), "SystemdCgroupPath"),
            TraceLoggingValue(
                processProps->DockerInstanceId.c_str(), "DockerInstanceId"),
            TraceLoggingValue(
                processProps->DockerImageName.c_str(), "DockerImageName"),
            TraceLoggingValue(
                processProps->DockerImageId.c_str(), "DockerImageId"),
            TraceLoggingValue(
                entry.Sample[MEM_PROCESS_INFO_PEAK_VIRTUAL_KB], "PeakVirtualKb"),
            TraceLoggingValue(
                entry.Sample[MEM_PROCESS_INFO_VIRTUAL_SIZE_KB], "VirtualSizeKb"),
            TraceLoggingValue(
                entry.Sample[MEM_PROCESS_INFO_LOCK_SIZE_KB], "LockSizeKb"),
            TraceLoggingValue(
                entry.Sample[MEM_PROCESS_INFO_PIN_SIZE_KB], "PinSizeKb"),
            TraceLoggingValue(
                entry.Sample[MEM_PROCESS_INFO_PEAK_RESIDENT_KB],
                "PeakResidentSetKb"),
            TraceLoggingValue(
                entry.Sample[MEM_PROCESS_INFO_RESIDENT_SIZE_KB],
                "TotalResidentSetKb"),
            TraceLoggingValue(
                entry.Sample[MEM_PROCESS_INFO_RESIDENT_ANON_KB],
                "ResidentSetAnonKb"),
            TraceLoggingValue(
                entry.Sample[MEM_PROCESS_INFO_RESIDENT_FILE_KB],
                "ResidentSetFileKb"),
            TraceLoggingValue(
                entry.Sample[MEM_PROCESS_INFO_RESIDENT_SHARE_KB],
                "ResidentSetShareKb"),
            TraceLoggingValue(
                entry.Sample[MEM_PROCESS_INFO_DATA_KB], "DataSegmentsKb"),
            TraceLoggingValue(
                entry.Sample[MEM_PROCESS_INFO_STACK_KB], "StackSegmentsKb"),
            TraceLoggingValue(
                entry.Sample[MEM_PROCESS_INFO_TEXT_KB], "TextSegmentsKb"),
            TraceLoggingValue(
                entry.Sample[MEM_PROCESS_INFO_SHARED_LIBRARY_KB],
                "SharedLibrarySegmentsKb"),
            TraceLoggingValue(
                entry.Sample[MEM_PROCESS_INFO_PTE_SIZE_KB], "PageTableEntryKb"),
            TraceLoggingValue(
                entry.Sample[MEM_PROCESS_INFO_PMD_SIZE_KB],
                "SecondLevelPageTableEntrySizeKb"),
            TraceLoggingValue(
                entry.Sample[MEM_PROCESS_INFO_VIRTUAL_SWAP_KB], "VirtualSwapKb"));

        LatteTelemetry(
            ProcessMemoryStats,
            "Azure.DeviceHealth.Plugins.ConsumePerProcessMemory",
            PrivacyTags_ProductAndServicePerformance,
            "%{SampleKey,d} \
            %{Rank,d} \
            %{ProcessId,u} \
            %{ProcessName,s} \
            %{PackageName,s} \
            %{PackageVersion,s} \
            %{PackageArch,s} \
            %{SystemdCgroupPath,s} \
            %{DockerInstanceId,s} \
            %{DockerImageName,s} \
            %{DockerImageId,s} \
            %{PeakVirtualKb,lld} \
            %{VirtualSizeKb,lld} \
            %{LockSizeKb,lld} \
            %{PinSizeKb,lld} \
            %{PeakResidentSetKb,lld} \
            %{TotalResidentSetKb,lld} \
            %{ResidentSetAnonKb,lld} \
            %{ResidentSetFileKb,lld} \
            %{ResidentSetShareKb,lld} \
            %{DataSegmentsKb,lld} \
            %{StackSegmentsKb,lld} \
            %{TextSegmentsKb,lld} \
            %{SharedLibrarySegmentsKb,lld} \
            %{PageTableEntryKb,lld} \
            %{SecondLevelPageTableEntrySizeKb,lld} \
            %{VirtualSwapKb,lld}",
            sampleKey,
            rank,
            entry.ProcessId,
            entry.ProcessName.c_str(),
            processProps->PackageName.c_str(),
            processProps->PackageVersion.c_str(),
            processProps->PackageArch.c_str(),
            processProps->SystemdCGroupPath.c_str(),
            processProps->DockerInstanceId.c_str(),
            processProps->DockerImageName.c_str(),
            processProps->DockerImageId.c_str(),
            entry.Sample[MEM_PROCESS_INFO_PEAK_VIRTUAL_KB],
            entry.Sample[MEM_PROCESS_INFO_VIRTUAL_SIZE_KB],
            entry.Sample[MEM_PROCESS_INFO_LOCK_SIZE_KB],
            entry.Sample[MEM_PROCESS_INFO_PIN_SIZE_KB],
            entry.Sample[MEM_PROCESS_INFO_PEAK_RESIDENT_KB],
            entry.Sample[MEM_PROCESS_INFO_RESIDENT_SIZE_KB],
            entry.Sample[MEM_PROCESS_INFO_RESIDENT_ANON_KB],
            entry.Sample[MEM_PROCESS_INFO_RESIDENT_FILE_KB],
            entry.Sample[MEM_PROCESS_INFO_RESIDENT_SHARE_KB],
            entry.Sample[MEM_PROCESS_INFO_DATA_KB],
            entry.Sample[MEM_PROCESS_INFO_STACK_KB],
            entry.Sample[MEM_PROCESS_INFO_TEXT_KB],
            entry.Sample[MEM_PROCESS_INFO_SHARED_LIBRARY_KB],
            entry.Sample[MEM_PROCESS_INFO_PTE_SIZE_KB],
            entry.Sample[MEM_PROCESS_INFO_PMD_SIZE_KB],
            entry.Sample[MEM_PROCESS_INFO_VIRTUAL_SWAP_KB]);

        rank++;
    }

    return true;
}