// Copyright (c) Microsoft Corporation. All rights reserved.

#include <map>
#include <numeric>
#include <optional>

#include <azure-device-health-plugin-utils/MetricRingStorage.h>
#include <azure-device-health-plugin-utils/UdevHelpers.h>
#include <azure-device-health-plugin/plugin-dev.h>
#include <azure-device-util/Math.h>
#include <azure-device-util/String.h>
#include <tracelogging/TraceLoggingProvider.h>

TRACELOGGING_DEFINE_PROVIDER(
    g_thermalProducerProvider,
    "Azure.DeviceHealth.Plugins.ProduceThermal",
    (0xf1f60c77, 0x738f, 0xfc47, 0x9d, 0xcc, 0xa5, 0xe9, 0xe8, 0xf6, 0x2f, 0x68));

TRACELOGGING_DEFINE_PROVIDER(
    g_thermalConsumerProvider,
    "Azure.DeviceHealth.Plugins.ConsumeThermal",
    (0xd778e398, 0xd68b, 0x874d, 0xb3, 0x02, 0x66, 0x92, 0x05, 0xe9, 0x00, 0x39));

namespace Utils = Azure::Device::Utils;
namespace PluginUtils = Azure::Device::Health::PluginUtils;

namespace Azure { namespace Device { namespace Health { namespace Plugin {

struct ThermalZone
{
    std::string SysName;
    std::string Type;
    int32_t MillidegreesCelsius = 0;
};

struct ThermalSample
{
    std::vector<ThermalZone> ThermalZones;
};

class ProduceThermalPlugin
{
  public:
    std::unique_ptr<ThermalSample> ProduceSample(uint64_t sampleKey);
};

class ConsumeThermalPlugin
{
  public:
    bool ConsumeSample(uint64_t sampleKey, const ThermalSample& thermalSample);

  public:
    PluginUtils::MetricRingStorage<ThermalSample> ThermalSamples;
};

}}}}

using namespace Azure::Device::Health::Plugin;

adh_plugin_instance* produce_thermal_create_plugin_instance(
    const adh_plugin_config_pair* configs,
    size_t configs_count)
{
    try
    {
        TraceLoggingRegister(g_thermalProducerProvider);
        TraceLoggingWrite(g_thermalProducerProvider, "Initialize");

        for (const adh_plugin_config_pair* config = configs;
             config != (configs + configs_count);
             ++config)
        {
            TraceLoggingWrite(
                g_thermalProducerProvider,
                "Create_UnusedConfiguration",
                TraceLoggingValue(config->name, "Name"),
                TraceLoggingValue(config->value, "Value"));
        }

        return reinterpret_cast<adh_plugin_instance*>(
            std::make_unique<ProduceThermalPlugin>().release());
    }
    CATCH_LOG();
    return nullptr;
}

void produce_thermal_destroy_plugin_instance(adh_plugin_instance* plugin)
{
    ProduceThermalPlugin* produceThermalPlugin =
        reinterpret_cast<ProduceThermalPlugin*>(plugin);
    delete produceThermalPlugin;

    TraceLoggingWrite(g_thermalProducerProvider, "Uninitialize");
    TraceLoggingUnregister(g_thermalProducerProvider);
}

void* produce_thermal_produce_sample(
    adh_plugin_instance* plugin,
    const adh_produce_sample_args* args)
{
    try
    {
        ProduceThermalPlugin* produceThermalPlugin =
            reinterpret_cast<ProduceThermalPlugin*>(plugin);
        return produceThermalPlugin->ProduceSample(args->sample_key).release();
    }
    CATCH_LOG();
    return nullptr;
}

void produce_thermal_free_sample(
    [[maybe_unused]] adh_plugin_instance* plugin,
    void* sample)
{
    if (sample)
    {
        delete reinterpret_cast<ThermalSample*>(sample);
    }
}

DEFINE_PRODUCER_PLUGIN(
    produce_thermal,
    "produce-thermal",
    produce_thermal_create_plugin_instance,
    produce_thermal_destroy_plugin_instance,
    produce_thermal_produce_sample,
    produce_thermal_free_sample);

static void
LogInvalidIntegerConfiguration(std::string_view name, std::string_view value)
{
    TraceLoggingWrite(
        g_thermalConsumerProvider,
        "Create_InvalidIntegerConfiguration",
        TraceLoggingCountedString(name.data(), name.size(), "Name"),
        TraceLoggingCountedString(value.data(), value.size(), "Value"));
};

adh_plugin_instance* consume_thermal_create_plugin_instance(
    const adh_plugin_config_pair* configs,
    size_t configs_count)
{
    try
    {
        TraceLoggingRegister(g_thermalConsumerProvider);
        TraceLoggingWrite(g_thermalConsumerProvider, "Initialize");

        auto consumeThermalPlugin = std::make_unique<ConsumeThermalPlugin>();

        for (const adh_plugin_config_pair* config = configs;
             config != (configs + configs_count);
             ++config)
        {
            std::string_view name = config->name;
            std::string_view value = config->value;

            if (Utils::String::ICompare(name, "CircularBufferSize") == 0)
            {
                size_t val = 0;
                if (SUCCEEDED(Utils::String::ToUIntTStrict(value, val)))
                {
                    consumeThermalPlugin->ThermalSamples.SetMaxRingSize(val);
                }
                else
                {
                    LogInvalidIntegerConfiguration(name, value);
                }
            }
            else if (Utils::String::ICompare(name, "ReportCadence") == 0)
            {
                size_t val = 0;
                if (SUCCEEDED(Utils::String::ToUIntTStrict(value, val)))
                {
                    consumeThermalPlugin->ThermalSamples.SetReportFrequency(val);
                }
                else
                {
                    LogInvalidIntegerConfiguration(name, value);
                }
            }
            else
            {
                TraceLoggingWrite(
                    g_thermalConsumerProvider,
                    "Create_UnusedConfiguration",
                    TraceLoggingValue(config->name, "Name"),
                    TraceLoggingValue(config->value, "Value"));
            }
        }

        return reinterpret_cast<adh_plugin_instance*>(
            consumeThermalPlugin.release());
    }
    CATCH_LOG();
    return nullptr;
}

void consume_thermal_destroy_plugin_instance(adh_plugin_instance* plugin)
{
    ConsumeThermalPlugin* consumeThermalPlugin =
        reinterpret_cast<ConsumeThermalPlugin*>(plugin);
    delete consumeThermalPlugin;

    TraceLoggingWrite(g_thermalConsumerProvider, "Uninitialize");
    TraceLoggingUnregister(g_thermalConsumerProvider);
}

static const char* s_thermalConsumeTargets[] = { "produce-thermal", nullptr };
const char* const* consume_thermal_list_consume_targets()
{
    return s_thermalConsumeTargets;
}

bool consume_thermal_consume_sample(
    adh_plugin_instance* plugin,
    const adh_consume_sample_args* args)
{
    try
    {
        if (Utils::String::ICompare(args->producer_name, produce_thermal.name) !=
            0)
        {
            TraceLoggingWrite(
                g_thermalConsumerProvider,
                "PassedUnusedSample",
                TraceLoggingValue(args->producer_name, "ProducerName"));
            return false;
        }

        ConsumeThermalPlugin* consumeThermalPlugin =
            reinterpret_cast<ConsumeThermalPlugin*>(plugin);
        return consumeThermalPlugin->ConsumeSample(
            args->sample_key, *static_cast<ThermalSample*>(args->sample));
    }
    CATCH_LOG();
    return false;
}

DEFINE_CONSUMER_PLUGIN(
    consume_thermal,
    "consume-thermal",
    consume_thermal_create_plugin_instance,
    consume_thermal_destroy_plugin_instance,
    consume_thermal_list_consume_targets,
    consume_thermal_consume_sample);

constexpr const char* c_deviceTemp = "temp";
constexpr const char* c_deviceType = "type";

std::optional<ThermalZone>
TryCreateThermalZoneFromDeviceEntry(const PluginUtils::UdevDeviceEntry& deviceEntry)
{
    ThermalZone zone;
    zone.SysName = deviceEntry.SysName;

    if (auto itr = deviceEntry.Attributes.find(c_deviceType);
        itr != deviceEntry.Attributes.end() && !itr->second.empty())
    {
        zone.Type = itr->second;
    }
    else
    {
        return {};
    }

    if (auto itr = deviceEntry.Attributes.find(c_deviceTemp);
        itr != deviceEntry.Attributes.end() && !itr->second.empty())
    {
        zone.MillidegreesCelsius =
            Azure::Device::Utils::String::ToIntT<int32_t>(itr->second);
    }
    else
    {
        return {};
    }

    return zone;
}

std::unique_ptr<ThermalSample>
ProduceThermalPlugin::ProduceSample(uint64_t sampleKey)
{
    TraceLoggingWrite(
        g_thermalProducerProvider,
        "Produce",
        TraceLoggingValue(sampleKey, "SampleKey"));

    auto thermalSample = std::make_unique<ThermalSample>();

    constexpr std::array<const char*, 2> c_thermalZoneProps = { c_deviceTemp,
                                                                c_deviceType };

    auto deviceEntries = PluginUtils::EnumerateDeviceAttributes(
        "thermal", "thermal_zone*", c_thermalZoneProps);

    TraceLoggingWrite(
        g_thermalProducerProvider,
        "Produce_FoundDevices",
        TraceLoggingValue(deviceEntries.size(), "DevicesCount"));

    for (const auto& deviceEntry : deviceEntries)
    {
        std::optional<ThermalZone> optThermalZone =
            TryCreateThermalZoneFromDeviceEntry(deviceEntry);

        if (optThermalZone)
        {
            thermalSample->ThermalZones.push_back(std::move(*optThermalZone));
        }
    }

    return thermalSample;
}

void GenerateReport(
    uint64_t sampleKey,
    const PluginUtils::MetricRingStorage<ThermalSample>& thermalSamples)
{
    struct DeviceKey
    {
        std::string SysName;
        std::string Type;
    };

    struct LessDeviceKey
    {
        bool operator()(const DeviceKey& l, const DeviceKey& r) const
        {
            if (l.SysName < r.SysName)
            {
                return true;
            }
            else if (l.SysName > r.SysName)
            {
                return false;
            }

            return l.Type < r.Type;
        }
    };

    std::map<DeviceKey, std::vector<int64_t>, LessDeviceKey> thermalDeviceToTemps;
    for (const auto& sample : thermalSamples)
    {
        for (const auto& zone : sample.ThermalZones)
        {
            thermalDeviceToTemps[DeviceKey{ zone.SysName, zone.Type }].push_back(
                zone.MillidegreesCelsius);
        }
    }

    for (auto&& [deviceKey, temps] : thermalDeviceToTemps)
    {
        std::sort(temps.begin(), temps.end());
        nonstd::span<const int64_t> tempsSpan{ temps };

        using namespace Azure::Device::Utils;

        TraceLoggingWrite(
            g_thermalConsumerProvider,
            "ThermalZoneStats",
            TraceLoggingValue(sampleKey, "SampleKey"),
            TraceLoggingValue(deviceKey.SysName.c_str(), "SysName"),
            TraceLoggingCountedString(
                deviceKey.Type.data(), deviceKey.Type.size(), "DeviceType"),
            TraceLoggingValue(temps.size(), "NumSamples"),
            TraceLoggingValue(Math::Min(tempsSpan), "Min"),
            TraceLoggingValue(Math::Max(tempsSpan), "Max"),
            TraceLoggingValue(Math::Mean(tempsSpan), "Mean"),
            TraceLoggingValue(Math::Median(tempsSpan), "Median"));
    }
}

bool ConsumeThermalPlugin::ConsumeSample(uint64_t sampleKey, const ThermalSample& sample)
{
    TraceLoggingWrite(g_thermalConsumerProvider, "Consume");

    ThermalSamples.AddMetric(sample);

    if (ThermalSamples.IsReportReady())
    {
        TraceLoggingWrite(g_thermalConsumerProvider, "Consume_Reporting");

        GenerateReport(sampleKey, ThermalSamples);
    }

    return true;
}
