// Copyright (c) Microsoft Corporation. All rights reserved.

#include "CpuMetricStorage.h"

#include <type_traits>
#include <utility>

#include <azure-device-util/AndroidLogHelper.h>
#include <azure-device-util/ResultException.h>

#include "Provider.h"

#ifdef AZURE_DEVICE_USE_CLOG
    #define CLOG_INLINE_IMPLEMENTATION 1
    #include "CpuMetricStorage.cpp.clog.h"
    #include "privacytags.h"
#else
    #include "clog_stubs.h"
#endif

namespace Azure { namespace Device { namespace Health { namespace Plugin {

/*++
    Adds a metric to the storage.

Arguments:
    cpuSet - The sample to add to our storage
--*/
void CpuMetricStorage::AddMetric(uint64_t sampleKey, const TotalCpuSample& sample)
{
    const CpuInfo& cpuInfo = sample.Total;

    TraceLoggingWrite(
        g_totalCpuConsumerProvider,
        "AddMetric",
        TraceLoggingValue(sampleKey, "timestamp"),
        TraceLoggingValue(cpuInfo.Sample[CpuState::CPU_STATE_NICE], "nice_time"),
        TraceLoggingValue(
            cpuInfo.Sample[CpuState::CPU_STATE_SYSTEM], "system_time"),
        TraceLoggingValue(cpuInfo.Sample[CpuState::CPU_STATE_IDLE], "idle_time"),
        TraceLoggingValue(
            cpuInfo.Sample[CpuState::CPU_STATE_IOWAIT], "io_wait_time"),
        TraceLoggingValue(cpuInfo.Sample[CpuState::CPU_STATE_IRQ], "irq_time"),
        TraceLoggingValue(
            cpuInfo.Sample[CpuState::CPU_STATE_SOFTIRQ], "soft_irq_time"),
        TraceLoggingValue(cpuInfo.Sample[CpuState::CPU_STATE_USER], "user_time"),
        TraceLoggingValue(
            cpuInfo.Sample[CpuState::CPU_STATE_STEAL], "steal_time"));

    //
    // Our very first sample is our delta base and is basically ignored to keep
    // our reporting window about the same.
    //
    if (_oldestReportedEntry != -1)
    {
        _samplesSinceLastReport++;
    }

    //
    // Update oldest valid entry. Note that for our very first sample, we treat
    // the entry as already reported to use it as our delta. Going forward, we
    // will always use the last reported entry as the base for our delta.
    //
    if (_oldestReportedEntry == -1)
    {
        _oldestReportedEntry = 0;
    }
    else if (_currentRingEntry == _oldestReportedEntry)
    {
        IncrementWithWrap(_oldestReportedEntry, _ringMetrics.size());
    }

    _ringMetrics[_currentRingEntry] = std::make_pair(sampleKey, cpuInfo);
    _newestValidEntry = _currentRingEntry;

    IncrementWithWrap(_currentRingEntry, _ringMetrics.size());
}

/*++
    Reports a histogram of deciles (0-10, 10-20) as a ratio of active time to
    idle time as well as total times.

Parameters:
    conf - The service configuration given to us by the device health service.
--*/
CpuMetricStorage::ReportMetrics CpuMetricStorage::GenerateReportMetrics()
{
    TraceLoggingWrite(g_totalCpuConsumerProvider, "ReportMetrics");

    try
    {
        // We need at least 2 samples to generate a report
        if (_oldestReportedEntry == -1 || _newestValidEntry == -1 ||
            _oldestReportedEntry == static_cast<int32_t>(_ringMetrics.size()) ||
            _newestValidEntry >= static_cast<int32_t>(_ringMetrics.size()) ||
            _oldestReportedEntry == _newestValidEntry)
        {
            // Logic error in plugin.
            THROW_HR(make_error_code(std::errc::operation_not_permitted));
        }

        // Calculate total time period covered
        const auto& oldestSample = _ringMetrics[_oldestReportedEntry];
        const auto& newestSample = _ringMetrics[_newestValidEntry];

        auto window_start_ms = std::chrono::milliseconds(oldestSample.first);
        auto window_end_ms = std::chrono::milliseconds(newestSample.first);

        auto totalWindowDurationMs = window_end_ms - window_start_ms;

        // Calculate total IDLE vs. ACTIVE CPU time per bucket
        auto const* current_sample_minus_1 =
            &(_ringMetrics[_oldestReportedEntry]);
        int32_t DecileCounts[10] = { 0 };
        std::chrono::milliseconds total_duration(0);

        // Start at oldest reported entry and use that as our base for deltas
        int32_t numEntriesInReport = 0;
        int32_t index = _oldestReportedEntry;
        IncrementWithWrap(index, _ringMetrics.size());

        while (numEntriesInReport < _reportCadence)
        {
            auto const* currentSample = &(_ringMetrics[index]);

            //
            // Each entry is a sum from boot, so we ignore the oldest valid
            // entry and use it as a reference point to start our calculation
            //
            int64_t total_active_ticks = 0;
            {
                auto calc_time_lambda = [&](uint32_t cpu_state_index) -> int64_t {
                    return currentSample->second.Sample[cpu_state_index] -
                        current_sample_minus_1->second.Sample[cpu_state_index];
                };

                int64_t user_ticks = calc_time_lambda(CpuState::CPU_STATE_USER);
                int64_t nice_ticks = calc_time_lambda(CpuState::CPU_STATE_NICE);
                int64_t system_ticks =
                    calc_time_lambda(CpuState::CPU_STATE_SYSTEM);
                int64_t io_ticks = calc_time_lambda(CpuState::CPU_STATE_IOWAIT);
                int64_t irq_ticks = calc_time_lambda(CpuState::CPU_STATE_IRQ);
                int64_t soft_irq_ticks =
                    calc_time_lambda(CpuState::CPU_STATE_SOFTIRQ);
                int64_t steal_ticks = calc_time_lambda(CpuState::CPU_STATE_STEAL);
                int64_t guest_ticks = calc_time_lambda(CpuState::CPU_STATE_GUEST);
                int64_t guest_nice_ticks =
                    calc_time_lambda(CpuState::CPU_STATE_GUEST_NICE);

                total_active_ticks = user_ticks + nice_ticks + system_ticks +
                    io_ticks + irq_ticks + soft_irq_ticks + steal_ticks +
                    guest_ticks + guest_nice_ticks;
            }

            int64_t idle_ticks =
                currentSample->second.Sample[CpuState::CPU_STATE_IDLE] -
                current_sample_minus_1->second.Sample[CpuState::CPU_STATE_IDLE];

            int64_t total_ticks = total_active_ticks + idle_ticks;
            int32_t percent_active_decile =
                ((double) total_active_ticks / total_ticks) * 10;

            // Sanitize data
            if (percent_active_decile >= 10)
            {
                percent_active_decile = 10;
            }
            else if (percent_active_decile < 0)
            {
                percent_active_decile = 0;
            }

            DecileCounts[percent_active_decile]++;

            // Calculate duration covered by this sample
            auto sample_window_start =
                std::chrono::milliseconds(current_sample_minus_1->first);
            auto sample_window_end =
                std::chrono::milliseconds(currentSample->first);

            auto sample_duration_ms = sample_window_end - sample_window_start;
            total_duration += sample_duration_ms;

            // Update sample minus 1
            current_sample_minus_1 = currentSample;

            // Update index
            IncrementWithWrap(index, _ringMetrics.size());

            // Update entries in sample
            numEntriesInReport++;
        }

        // Average time per sample
        uint32_t AverageTimePerSampleMs = static_cast<uint32_t>(
            total_duration.count() / (double) numEntriesInReport);

        // Context-switches in window
        uint64_t ContextSwitchesInWindow = 0;
        if (newestSample.second.ContextSwitches != 0 &&
            oldestSample.second.ContextSwitches != 0)
        {
            ContextSwitchesInWindow = newestSample.second.ContextSwitches -
                oldestSample.second.ContextSwitches;
        }

        // Interesting times over window
        {
            auto calc_total_ticks_lambda =
                [&](uint32_t cpu_state_index) -> int64_t {
                return newestSample.second.Sample[cpu_state_index] -
                    oldestSample.second.Sample[cpu_state_index];
            };

            int64_t TotalUserTicks =
                calc_total_ticks_lambda(CpuState::CPU_STATE_USER);
            int64_t TotalNiceTicks =
                calc_total_ticks_lambda(CpuState::CPU_STATE_NICE);
            int64_t TotalSystemTicks =
                calc_total_ticks_lambda(CpuState::CPU_STATE_SYSTEM);
            int64_t TotalIoTicks =
                calc_total_ticks_lambda(CpuState::CPU_STATE_IOWAIT);
            int64_t TotalIrqTicks =
                calc_total_ticks_lambda(CpuState::CPU_STATE_IRQ);
            int64_t TotalSoftIrqTicks =
                calc_total_ticks_lambda(CpuState::CPU_STATE_SOFTIRQ);
            int64_t TotalStealTicks =
                calc_total_ticks_lambda(CpuState::CPU_STATE_STEAL);
            int64_t TotalGuestTicks =
                calc_total_ticks_lambda(CpuState::CPU_STATE_GUEST);
            int64_t TotalGuestNiceTicks =
                calc_total_ticks_lambda(CpuState::CPU_STATE_GUEST_NICE);

            ReportMetrics metrics;
            metrics.TotalWindowDurationMs = totalWindowDurationMs.count();
            metrics.AverageTimePerSampleMs = AverageTimePerSampleMs;
            metrics.ContextSwitchesInWindow = ContextSwitchesInWindow;
            metrics.TotalUserTicks = TotalUserTicks;
            metrics.TotalNiceTicks = TotalNiceTicks;
            metrics.TotalSystemTicks = TotalSystemTicks;
            metrics.TotalIoTicks = TotalIoTicks;
            metrics.TotalIrqTicks = TotalIrqTicks;
            metrics.TotalSoftIrqTicks = TotalSoftIrqTicks;
            metrics.TotalStealTicks = TotalStealTicks;
            metrics.TotalGuestTicks = TotalGuestTicks;
            metrics.TotalGuestNiceTicks = TotalGuestNiceTicks;
            std::copy(
                DecileCounts,
                DecileCounts + std::extent<decltype(DecileCounts)>::value,
                metrics.DecileCounts);

            _samplesSinceLastReport = 0;

            // Assign oldest reported to the last metric reported
            _oldestReportedEntry = _newestValidEntry;

            return metrics;
        }
    }
    CATCH_FAIL_FAST();
}

void CpuMetricStorage::FireReportMetrics(
    uint64_t clockTicksPerSecond,
    const ReportMetrics& metrics)
{
    // Send report event, keep this style consistent with Asimov schema
    TraceLoggingWrite(
        g_totalCpuConsumerProvider,
        "TotalCpuStats",
        TraceLoggingValue(metrics.TotalWindowDurationMs, "TotalWindowDurationMs"),
        TraceLoggingValue(
            metrics.AverageTimePerSampleMs, "AverageTimePerSampleMs"),
        TraceLoggingValue(clockTicksPerSecond, "ClockTicksPerSecond"),
        TraceLoggingValue(
            metrics.ContextSwitchesInWindow, "ContextSwitchesInWindow"),
        TraceLoggingValue(metrics.TotalUserTicks, "TotalUserClockTicks"),
        TraceLoggingValue(metrics.TotalNiceTicks, "TotalNiceClockTicks"),
        TraceLoggingValue(metrics.TotalSystemTicks, "TotalSystemClockTicks"),
        TraceLoggingValue(metrics.TotalIoTicks, "TotalIoClockTicks"),
        TraceLoggingValue(metrics.TotalIrqTicks, "TotalIrqClockTicks"),
        TraceLoggingValue(metrics.TotalSoftIrqTicks, "TotalSoftIrqClockTicks"),
        TraceLoggingValue(metrics.TotalStealTicks, "TotalStealClockTicks"),
        TraceLoggingValue(metrics.TotalGuestTicks, "TotalGuestClockTicks"),
        TraceLoggingValue(metrics.TotalGuestNiceTicks, "TotalGuestNiceClockTicks"),
        TraceLoggingInt32Array(
            metrics.DecileCounts,
            std::extent<decltype(metrics.DecileCounts)>::value,
            "DecileCounts"));

    LatteTelemetry(
        TotalCpuStats,
        "Azure.DeviceHealth.Plugins.ConsumeTotalCpu",
        PrivacyTags_ProductAndServicePerformance,
        "%{TotalWindowDurationMs,lld} \
        %{AverageTimePerSampleMs,u} \
        %{ClockTicksPerSecond,llu} \
        %{ContextSwitchesInWindow,llu} \
        %{TotalUserClockTicks,lld} \
        %{TotalNiceClockTicks,lld} \
        %{TotalSystemClockTicks,lld} \
        %{TotalIoClockTicks,lld} \
        %{TotalIrqClockTicks,lld} \
        %{TotalSoftIrqClockTicks,lld} \
        %{TotalStealClockTicks,lld} \
        %{TotalGuestClockTicks,lld} \
        %{TotalGuestNiceClockTicks,lld} \
        %{DecileCounts,!INT32ARRAY!}",
        metrics.TotalWindowDurationMs,
        metrics.AverageTimePerSampleMs,
        clockTicksPerSecond,
        metrics.ContextSwitchesInWindow,
        metrics.TotalUserTicks,
        metrics.TotalNiceTicks,
        metrics.TotalSystemTicks,
        metrics.TotalIoTicks,
        metrics.TotalIrqTicks,
        metrics.TotalSoftIrqTicks,
        metrics.TotalStealTicks,
        metrics.TotalGuestTicks,
        metrics.TotalGuestNiceTicks,
        CLOG_ARRAY(
            std::extent<decltype(metrics.DecileCounts)>::value *
                sizeof(decltype(metrics.DecileCounts[0])),
            metrics.DecileCounts));
}

}}}}
