// Copyright (c) Microsoft Corporation. All rights reserved.

#include <mutex>
#include <string>
#include <string_view>
#include <unistd.h>
#include <utility>
#include <vector>

#include <azure-device-health-plugin/plugin-dev.h>
#include <azure-device-util/AndroidLogHelper.h>
#include <azure-device-util/FileSystem.h>
#include <azure-device-util/String.h>

#include "CpuMetricStorage.h"
#include "CpuMetrics.h"
#include "ProcStatParser.h"
#include "Provider.h"

TRACELOGGING_DEFINE_PROVIDER(
    g_totalCpuProducerProvider,
    "Azure.DeviceHealth.Plugins.ProduceTotalCpu",
    (0x2eba4200, 0xe237, 0xb348, 0x8e, 0x05, 0x6a, 0x92, 0x60, 0xe1, 0xc6, 0xe3));

TRACELOGGING_DEFINE_PROVIDER(
    g_totalCpuConsumerProvider,
    "Azure.DeviceHealth.Plugins.ConsumeTotalCpu",
    (0x5b163ef6, 0x67e1, 0xf241, 0x96, 0x11, 0xd9, 0x51, 0x38, 0x07, 0xb7, 0x57));

namespace Azure { namespace Device { namespace Health { namespace Plugin {

class ProduceTotalCpuPlugin
{
  public:
    std::unique_ptr<TotalCpuSample> ProduceSample(uint64_t sampleKey);

  public:
    bool SummaryOnly = true;
};

class ConsumeTotalCpuPlugin
{
  public:
    ConsumeTotalCpuPlugin()
    {
        SampleStorage.SetReportCadence(30);
        SampleStorage.SetBufferSize(30);

        _clockTicksPerSecond = sysconf(_SC_CLK_TCK);
    }

    bool ConsumeSample(uint64_t sampleKey, const TotalCpuSample& totalCpuSample);

  public:
    CpuMetricStorage SampleStorage;

  private:
    uint64_t _clockTicksPerSecond = 0;
};

}}}}

namespace Utils = Azure::Device::Utils;
using namespace Azure::Device::Health::Plugin;

adh_plugin_instance* produce_total_cpu_create_plugin_instance(
    const adh_plugin_config_pair* configs,
    size_t configs_count)
{
    try
    {
        TraceLoggingRegister(g_totalCpuProducerProvider);
        TraceLoggingWrite(g_totalCpuProducerProvider, "Initialize");

        auto produceTotalCpuPlugin = std::make_unique<ProduceTotalCpuPlugin>();

        for (const adh_plugin_config_pair* config = configs;
             config != (configs + configs_count);
             ++config)
        {
            std::string_view name = config->name;
            std::string_view value = config->value;

            if (Utils::String::ICompare(name, "SummaryOnly") == 0)
            {
                if (Utils::String::ICompare(value, "true") == 0)
                {
                    produceTotalCpuPlugin->SummaryOnly = true;
                }
                else if (Utils::String::ICompare(value, "false") == 0)
                {
                    produceTotalCpuPlugin->SummaryOnly = false;
                }
                else
                {
                    TraceLoggingWrite(
                        g_totalCpuProducerProvider,
                        "Create_InvalidBooleanConfiguration",
                        TraceLoggingCountedString(
                            name.data(), name.size(), "Name"),
                        TraceLoggingCountedString(
                            value.data(), value.size(), "Value"));
                }
            }
            else
            {
                TraceLoggingWrite(
                    g_totalCpuProducerProvider,
                    "Create_UnusedConfiguration",
                    TraceLoggingValue(config->name, "Name"),
                    TraceLoggingValue(config->value, "Value"));
            }
        }

        return reinterpret_cast<adh_plugin_instance*>(
            produceTotalCpuPlugin.release());
    }
    CATCH_LOG();
    return nullptr;
}

void produce_total_cpu_destroy_plugin_instance(adh_plugin_instance* plugin)
{
    ProduceTotalCpuPlugin* produceTotalCpuPlugin =
        reinterpret_cast<ProduceTotalCpuPlugin*>(plugin);
    delete produceTotalCpuPlugin;

    TraceLoggingWrite(g_totalCpuProducerProvider, "Uninitialize");
    TraceLoggingUnregister(g_totalCpuProducerProvider);
}

void* produce_total_cpu_produce_sample(
    adh_plugin_instance* plugin,
    const adh_produce_sample_args* args)
{
    try
    {
        ProduceTotalCpuPlugin* produceTotalCpuPlugin =
            reinterpret_cast<ProduceTotalCpuPlugin*>(plugin);
        return produceTotalCpuPlugin->ProduceSample(args->sample_key).release();
    }
    CATCH_LOG();
    return nullptr;
}

void produce_total_cpu_free_sample(
    [[maybe_unused]] adh_plugin_instance* plugin,
    void* sample)
{
    if (sample)
    {
        delete reinterpret_cast<TotalCpuSample*>(sample);
    }
}

DEFINE_PRODUCER_PLUGIN(
    produce_total_cpu,
    "produce-total-cpu",
    produce_total_cpu_create_plugin_instance,
    produce_total_cpu_destroy_plugin_instance,
    produce_total_cpu_produce_sample,
    produce_total_cpu_free_sample);

static void
LogInvalidIntegerConfiguration(std::string_view name, std::string_view value)
{
    TraceLoggingWrite(
        g_totalCpuConsumerProvider,
        "Create_InvalidIntegerConfiguration",
        TraceLoggingCountedString(name.data(), name.size(), "Name"),
        TraceLoggingCountedString(value.data(), value.size(), "Value"));
};

adh_plugin_instance* consume_total_cpu_create_plugin_instance(
    const adh_plugin_config_pair* configs,
    size_t configs_count)
{
    try
    {
        TraceLoggingRegister(g_totalCpuConsumerProvider);
        TraceLoggingWrite(g_totalCpuConsumerProvider, "Initialize");
        ADHS_LOG("Initialize TotalCpu Plugin");

        auto consumeTotalCpuPlugin = std::make_unique<ConsumeTotalCpuPlugin>();

        for (const adh_plugin_config_pair* config = configs;
             config != (configs + configs_count);
             ++config)
        {
            std::string_view name = config->name;
            std::string_view value = config->value;

            if (Utils::String::ICompare(name, "CircularBufferSize") == 0)
            {
                size_t val = 0;
                if (SUCCEEDED(Utils::String::ToUIntTStrict(value, val)))
                {
                    consumeTotalCpuPlugin->SampleStorage.SetBufferSize(val);
                }
                else
                {
                    LogInvalidIntegerConfiguration(name, value);
                }
            }
            else if (Utils::String::ICompare(name, "ReportCadence") == 0)
            {
                size_t val = 0;
                if (SUCCEEDED(Utils::String::ToUIntTStrict(value, val)))
                {
                    consumeTotalCpuPlugin->SampleStorage.SetReportCadence(val);
                }
                else
                {
                    LogInvalidIntegerConfiguration(name, value);
                }
            }
            else
            {
                TraceLoggingWrite(
                    g_totalCpuConsumerProvider,
                    "Create_UnusedConfiguration",
                    TraceLoggingValue(config->name, "Name"),
                    TraceLoggingValue(config->value, "Value"));
            }
        }

        return reinterpret_cast<adh_plugin_instance*>(
            consumeTotalCpuPlugin.release());
    }
    CATCH_LOG();
    return nullptr;
}

void consume_total_cpu_destroy_plugin_instance(adh_plugin_instance* plugin)
{
    ConsumeTotalCpuPlugin* consumeTotalCpuPlugin =
        reinterpret_cast<ConsumeTotalCpuPlugin*>(plugin);
    delete consumeTotalCpuPlugin;

    TraceLoggingWrite(g_totalCpuConsumerProvider, "Uninitialize");
    TraceLoggingUnregister(g_totalCpuConsumerProvider);
}

static const char* s_totalCpuConsumeTargets[] = { "produce-total-cpu", nullptr };
const char* const* consume_total_cpu_list_consume_targets()
{
    return s_totalCpuConsumeTargets;
}

bool consume_total_cpu_consume_sample(
    adh_plugin_instance* plugin,
    const adh_consume_sample_args* args)
{
    try
    {
        if (Utils::String::ICompare(
                args->producer_name, produce_total_cpu.name) != 0)
        {
            TraceLoggingWrite(
                g_totalCpuConsumerProvider,
                "PassedUnusedSample",
                TraceLoggingValue(args->producer_name, "ProducerName"));
            return false;
        }

        ConsumeTotalCpuPlugin* consumeTotalCpuPlugin =
            reinterpret_cast<ConsumeTotalCpuPlugin*>(plugin);
        return consumeTotalCpuPlugin->ConsumeSample(
            args->sample_key, *static_cast<TotalCpuSample*>(args->sample));
    }
    CATCH_LOG();
    return false;
}

DEFINE_CONSUMER_PLUGIN(
    consume_total_cpu,
    "consume-total-cpu",
    consume_total_cpu_create_plugin_instance,
    consume_total_cpu_destroy_plugin_instance,
    consume_total_cpu_list_consume_targets,
    consume_total_cpu_consume_sample);

std::unique_ptr<TotalCpuSample>
ProduceTotalCpuPlugin::ProduceSample(uint64_t sampleKey)
{
    TraceLoggingWrite(
        g_totalCpuProducerProvider,
        "Produce",
        TraceLoggingValue(sampleKey, "SampleKey"));

    auto totalCpuSample = std::make_unique<TotalCpuSample>();

    CpuInfo cpuInfo;

    ProcStatParser parser;
    if (!parser.Parse(cpuInfo, SummaryOnly))
    {
        return nullptr;
    }

    totalCpuSample->Total = std::move(cpuInfo);

    return totalCpuSample;
}

bool ConsumeTotalCpuPlugin::ConsumeSample(
    uint64_t sampleKey,
    const TotalCpuSample& sample)
{
    TraceLoggingWrite(g_totalCpuConsumerProvider, "Consume");

    SampleStorage.AddMetric(sampleKey, sample);

    if (SampleStorage.IsReportReady())
    {
        TraceLoggingWrite(g_totalCpuConsumerProvider, "Consume_Reporting");

        auto metrics = SampleStorage.GenerateReportMetrics();

        SampleStorage.FireReportMetrics(_clockTicksPerSecond, metrics);
    }

    return true;
}
