// Copyright (c) Microsoft Corporation. All rights reserved.

#include "DiskMetricStorage.h"

#include <type_traits>
#include <utility>

#include <azure-device-util/AndroidLogHelper.h>

#include "Provider.h"

#ifdef AZURE_DEVICE_USE_CLOG
    #define CLOG_INLINE_IMPLEMENTATION 1
    #include "DiskMetricStorage.cpp.clog.h"
    #include "privacytags.h"
#else
    #include "clog_stubs.h"
#endif

namespace Azure { namespace Device { namespace Health { namespace Plugin {

/*++
    Adds a metric to the storage and, if necessary, reports to telemetry.

Arguments:
    diskSet - The sample to add to our storage
--*/
void DiskMetricStorage::AddMetric(uint64_t sampleKey, const TotalDiskIoSample& sample)
{
    for (const DiskIoInfo& diskInfo : sample.Disks)
    {
        TraceLoggingWrite(
            g_totalDiskIoConsumerProvider,
            "AddMetric",
            TraceLoggingValue(sampleKey, "SampleKey"),
            TraceLoggingValue(
                diskInfo.Sample[static_cast<uint32_t>(DiskStats::MAJOR_NUMBER)],
                "MajorNumber"),
            TraceLoggingValue(
                diskInfo.Sample[static_cast<uint32_t>(DiskStats::MINOR_NUMBER)],
                "MinorNumber"),
            TraceLoggingValue(diskInfo.DeviceName.c_str(), "DeviceName"),
            TraceLoggingValue(
                diskInfo.Sample[static_cast<uint32_t>(DiskStats::READS_COMPLETED)],
                "ReadsCompleted"),
            TraceLoggingValue(
                diskInfo.Sample[static_cast<uint32_t>(DiskStats::READS_MERGED)],
                "ReadsMerged"),
            TraceLoggingValue(
                diskInfo.Sample[static_cast<uint32_t>(DiskStats::SECTORS_READ)],
                "SectorsRead"),
            TraceLoggingValue(
                diskInfo.Sample[static_cast<uint32_t>(DiskStats::MS_SPENT_READING)],
                "MsSpendReading"),
            TraceLoggingValue(
                diskInfo.Sample[static_cast<uint32_t>(DiskStats::WRITES_COMPLETED)],
                "WritesCompleted"),
            TraceLoggingValue(
                diskInfo.Sample[static_cast<uint32_t>(DiskStats::WRITES_MERGED)],
                "WritesMerged"),
            TraceLoggingValue(
                diskInfo.Sample[static_cast<uint32_t>(DiskStats::SECTORS_WRITTEN)],
                "SectorsWritten"),
            TraceLoggingValue(
                diskInfo.Sample[static_cast<uint32_t>(DiskStats::MS_SPENT_WRITING)],
                "MsSpendWriting"),
            TraceLoggingValue(
                diskInfo.Sample[static_cast<uint32_t>(DiskStats::IOS_IN_PROGRESS)],
                "IosCurrentlyInProgress"),
            TraceLoggingValue(
                diskInfo.Sample[static_cast<uint32_t>(DiskStats::MS_SPENT_DOING_IO)],
                "MsSpendDoingIo"),
            TraceLoggingValue(
                diskInfo.Sample[static_cast<uint32_t>(
                    DiskStats::WEIGHTED_NUM_OF_MS_DOING_IO)],
                "WeightedTimeSpentDoingIoMs"),
            TraceLoggingValue(
                diskInfo.Sample[static_cast<uint32_t>(DiskStats::DISCARDS_COMPLETED)],
                "DiscardsCompleted"),
            TraceLoggingValue(
                diskInfo.Sample[static_cast<long long>(DiskStats::DISCARDS_MERGED)],
                "DiscardsMerged"),
            TraceLoggingValue(
                diskInfo.Sample[static_cast<uint32_t>(DiskStats::SECTORS_DISCARDED)],
                "SectorsDiscarded"),
            TraceLoggingValue(
                diskInfo.Sample[static_cast<long long>(DiskStats::MS_SPENT_DISCARDING)],
                "MsSpendDiscarding"));
    }

    //
    // Our very first sample is our delta base and is basically ignored to keep
    // our reporting window about the same.
    //
    if (_oldestValidEntry != -1)
    {
        _samplesSinceLastReport++;
    }

    if (_oldestValidEntry == -1)
    {
        _oldestValidEntry = 0;
    }
    else if (_currentRingEntry == _oldestValidEntry)
    {
        IncrementWithWrap(
            _oldestValidEntry, static_cast<int32_t>(_ringMetrics.size()));
    }

    _ringMetrics[_currentRingEntry] = std::make_pair(sampleKey, sample);
    _newestValidEntry = _currentRingEntry;

    IncrementWithWrap(
        _currentRingEntry, static_cast<int32_t>(_ringMetrics.size()));
}

/*++
    Reports a histogram of deciles (0-10, 10-20) as a ratio of active time to
    idle time.
--*/
void DiskMetricStorage::ReportMetrics()
{
    TraceLoggingWrite(g_totalDiskIoConsumerProvider, "ReportMetrics");

    const auto& oldestSample = _ringMetrics[_oldestValidEntry];
    const auto& newestSample = _ringMetrics[_newestValidEntry];

    auto windowStart = std::chrono::milliseconds(oldestSample.first);
    auto windowEnd = std::chrono::milliseconds(newestSample.first);

    auto totalWindowDurationMs = windowEnd - windowStart;

    int32_t numPartitions = (int32_t) oldestSample.second.Disks.size();

    for (int32_t diskIndex = 0; diskIndex < numPartitions; diskIndex++)
    {
        auto calcDelta = [&](uint32_t arrayIndex) -> uint64_t {
            return newestSample.second.Disks[diskIndex].Sample[arrayIndex] -
                oldestSample.second.Disks[diskIndex].Sample[arrayIndex];
        };

        long readsCompleted =
            calcDelta(static_cast<uint32_t>(DiskStats::READS_COMPLETED));

        long readsMerged =
            calcDelta(static_cast<uint32_t>(DiskStats::READS_MERGED));

        long sectorsRead =
            calcDelta(static_cast<uint32_t>(DiskStats::SECTORS_READ));

        long msSpentReading =
            calcDelta(static_cast<uint32_t>(DiskStats::MS_SPENT_READING));

        long writesCompleted =
            calcDelta(static_cast<uint32_t>(DiskStats::WRITES_COMPLETED));

        long writesMerged =
            calcDelta(static_cast<uint32_t>(DiskStats::WRITES_MERGED));

        long sectorsWritten =
            calcDelta(static_cast<uint32_t>(DiskStats::SECTORS_WRITTEN));

        long msSpentWriting =
            calcDelta(static_cast<uint32_t>(DiskStats::MS_SPENT_WRITING));

        long iosInprogress =
            calcDelta(static_cast<uint32_t>(DiskStats::IOS_IN_PROGRESS));

        long msSpentDoingIo =
            calcDelta(static_cast<uint32_t>(DiskStats::IOS_IN_PROGRESS));

        long weightedNumOfMsDoingIo = calcDelta(
            static_cast<uint32_t>(DiskStats::WEIGHTED_NUM_OF_MS_DOING_IO));

        long discardsCompleted =
            calcDelta(static_cast<uint32_t>(DiskStats::DISCARDS_COMPLETED));

        long discardsMerged =
            calcDelta(static_cast<uint32_t>(DiskStats::DISCARDS_MERGED));

        long sectorsDiscarded =
            calcDelta(static_cast<uint32_t>(DiskStats::SECTORS_DISCARDED));

        long msSpentDiscarding =
            calcDelta(static_cast<uint32_t>(DiskStats::MS_SPENT_DISCARDING));

        TraceLoggingWrite(
            g_totalDiskIoConsumerProvider,
            "TotalDiskIoStats",
            TraceLoggingValue(
                oldestSample.second.Disks[diskIndex].DeviceName.c_str(),
                "DeviceName"),
            TraceLoggingValue(
                totalWindowDurationMs.count(), "TotalWindowDurationMs"),
            TraceLoggingValue(readsCompleted, "ReadsCompleted"),
            TraceLoggingValue(readsMerged, "ReadsMerged"),
            TraceLoggingValue(sectorsRead, "SectorsRead"),
            TraceLoggingValue(msSpentReading, "MsSpentReading"),
            TraceLoggingValue(writesCompleted, "WritesCompleted"),
            TraceLoggingValue(writesMerged, "WritesMerged"),
            TraceLoggingValue(sectorsWritten, "SectorsWritten"),
            TraceLoggingValue(msSpentWriting, "MsSpentWriting"),
            TraceLoggingValue(iosInprogress, "IosInProgress"),
            TraceLoggingValue(msSpentDoingIo, "MsSpentDoingIo"),
            TraceLoggingValue(weightedNumOfMsDoingIo, "WeightedNumOfMsDoingIo"),
            TraceLoggingValue(discardsCompleted, "DiscardsCompleted"),
            TraceLoggingValue(discardsMerged, "DiscardsMerged"),
            TraceLoggingValue(sectorsDiscarded, "SectorsDiscarded"),
            TraceLoggingValue(msSpentDiscarding, "MsSpentDiscarding"));

        LatteTelemetry(
            TotalDiskIoStats,
            "Azure.DeviceHealth.Plugins.ConsumeTotalDiskIo",
            PrivacyTags_ProductAndServicePerformance,
            "%{DeviceName,s} \
            %{TotalWindowDurationMs,llu} \
            %{ReadsCompleted,d} \
            %{ReadsMerged,d} \
            %{SectorsRead,d} \
            %{MsSpentReading,d} \
            %{WritesCompleted,d} \
            %{WritesMerged,d} \
            %{SectorsWritten,d} \
            %{MsSpentWriting,d} \
            %{IosInProgress,d} \
            %{MsSpentDoingIo,d} \
            %{WeightedNumOfMsDoingIo,d} \
            %{DiscardsCompleted,d} \
            %{DiscardsMerged,d} \
            %{SectorsDiscarded,d} \
            %{MsSpentDiscarding,d}",
            oldestSample.second.Disks[diskIndex].DeviceName.c_str(),
            totalWindowDurationMs.count(),
            readsCompleted,
            readsMerged,
            sectorsRead,
            msSpentReading,
            writesCompleted,
            writesMerged,
            sectorsWritten,
            msSpentWriting,
            iosInprogress,
            msSpentDoingIo,
            weightedNumOfMsDoingIo,
            discardsCompleted,
            discardsMerged,
            sectorsDiscarded,
            msSpentDiscarding);
    }
    _samplesSinceLastReport = 0;
}

}}}}
