// Copyright (c) Microsoft Corporation. All rights reserved.

#include <mutex>
#include <string>
#include <string_view>
#include <utility>
#include <vector>

#include <azure-device-health-plugin/plugin-dev.h>
#include <azure-device-util/AndroidLogHelper.h>
#include <azure-device-util/FileSystem.h>
#include <azure-device-util/String.h>

#include "DiskMetricStorage.h"
#include "DiskMetrics.h"
#include "DiskStatParser.h"
#include "Provider.h"

TRACELOGGING_DEFINE_PROVIDER(
    g_totalDiskIoProducerProvider,
    "Azure.DeviceHealth.Plugins.ProduceTotalDiskIo",
    (0x60d7a51d, 0xe48f, 0x5841, 0x93, 0x1e, 0x94, 0x8c, 0x8b, 0xec, 0xe7, 0x87));

TRACELOGGING_DEFINE_PROVIDER(
    g_totalDiskIoConsumerProvider,
    "Azure.DeviceHealth.Plugins.ConsumeTotalDiskIo",
    (0x390a9142, 0x9b71, 0xa442, 0xad, 0xe0, 0x9f, 0x8d, 0xbf, 0x8c, 0x97, 0x1d));

namespace Azure { namespace Device { namespace Health { namespace Plugin {

class ProduceTotalDiskIoPlugin
{
  public:
    std::unique_ptr<TotalDiskIoSample> ProduceSample(uint64_t sampleKey);

  public:
    bool SummaryOnly = true;
};

class ConsumeTotalDiskIoPlugin
{
  public:
    ConsumeTotalDiskIoPlugin()
    {
        SampleStorage.SetReportCadence(30);
        SampleStorage.SetBufferSize(30);
    }

    bool
    ConsumeSample(uint64_t sampleKey, const TotalDiskIoSample& totalDiskIoSample);

  public:
    DiskMetricStorage SampleStorage;
};

}}}}

namespace Utils = Azure::Device::Utils;
using namespace Azure::Device::Health::Plugin;

adh_plugin_instance* produce_total_disk_io_create_plugin_instance(
    const adh_plugin_config_pair* configs,
    size_t configs_count)
{
    try
    {
        TraceLoggingRegister(g_totalDiskIoProducerProvider);
        TraceLoggingWrite(g_totalDiskIoProducerProvider, "Initialize");
        ADHS_LOG("Initialize TotalDiskIo Plugin");

        auto produceTotalDiskIoPlugin =
            std::make_unique<ProduceTotalDiskIoPlugin>();

        for (const adh_plugin_config_pair* config = configs;
             config != (configs + configs_count);
             ++config)
        {
            std::string_view name = config->name;
            std::string_view value = config->value;

            if (Utils::String::ICompare(name, "SummaryOnly") == 0)
            {
                if (Utils::String::ICompare(value, "true") == 0)
                {
                    produceTotalDiskIoPlugin->SummaryOnly = true;
                }
                else if (Utils::String::ICompare(value, "false") == 0)
                {
                    produceTotalDiskIoPlugin->SummaryOnly = false;
                }
                else
                {
                    TraceLoggingWrite(
                        g_totalDiskIoProducerProvider,
                        "Create_InvalidBooleanConfiguration",
                        TraceLoggingCountedString(
                            name.data(), name.size(), "Name"),
                        TraceLoggingCountedString(
                            value.data(), value.size(), "Value"));
                }
            }
            else
            {
                TraceLoggingWrite(
                    g_totalDiskIoProducerProvider,
                    "Create_UnusedConfiguration",
                    TraceLoggingValue(config->name, "Name"),
                    TraceLoggingValue(config->value, "Value"));
            }
        }

        return reinterpret_cast<adh_plugin_instance*>(
            produceTotalDiskIoPlugin.release());
    }
    CATCH_LOG();
    return nullptr;
}

void produce_total_disk_io_destroy_plugin_instance(adh_plugin_instance* plugin)
{
    ProduceTotalDiskIoPlugin* produceTotalDiskIoPlugin =
        reinterpret_cast<ProduceTotalDiskIoPlugin*>(plugin);
    delete produceTotalDiskIoPlugin;

    TraceLoggingWrite(g_totalDiskIoProducerProvider, "Uninitialize");
    TraceLoggingUnregister(g_totalDiskIoProducerProvider);
}

void* produce_total_disk_io_produce_sample(
    adh_plugin_instance* plugin,
    const adh_produce_sample_args* args)
{
    try
    {
        ProduceTotalDiskIoPlugin* produceTotalDiskIoPlugin =
            reinterpret_cast<ProduceTotalDiskIoPlugin*>(plugin);
        return produceTotalDiskIoPlugin->ProduceSample(args->sample_key).release();
    }
    CATCH_LOG();
    return nullptr;
}

void produce_total_disk_io_free_sample(
    [[maybe_unused]] adh_plugin_instance* plugin,
    void* sample)
{
    if (sample)
    {
        delete reinterpret_cast<TotalDiskIoSample*>(sample);
    }
}

DEFINE_PRODUCER_PLUGIN(
    produce_total_disk_io,
    "produce-total-disk-io",
    produce_total_disk_io_create_plugin_instance,
    produce_total_disk_io_destroy_plugin_instance,
    produce_total_disk_io_produce_sample,
    produce_total_disk_io_free_sample);

static void
LogInvalidIntegerConfiguration(std::string_view name, std::string_view value)
{
    TraceLoggingWrite(
        g_totalDiskIoConsumerProvider,
        "Create_InvalidIntegerConfiguration",
        TraceLoggingCountedString(name.data(), name.size(), "Name"),
        TraceLoggingCountedString(value.data(), value.size(), "Value"));
};

adh_plugin_instance* consume_total_disk_io_create_plugin_instance(
    const adh_plugin_config_pair* configs,
    size_t configs_count)
{
    try
    {
        TraceLoggingRegister(g_totalDiskIoConsumerProvider);
        TraceLoggingWrite(g_totalDiskIoConsumerProvider, "Initialize");

        auto consumeTotalDiskIoPlugin =
            std::make_unique<ConsumeTotalDiskIoPlugin>();

        for (const adh_plugin_config_pair* config = configs;
             config != (configs + configs_count);
             ++config)
        {
            std::string_view name = config->name;
            std::string_view value = config->value;

            if (Utils::String::ICompare(name, "CircularBufferSize") == 0)
            {
                size_t val = 0;
                if (SUCCEEDED(Utils::String::ToUIntTStrict(value, val)))
                {
                    consumeTotalDiskIoPlugin->SampleStorage.SetBufferSize(val);
                }
                else
                {
                    LogInvalidIntegerConfiguration(name, value);
                }
            }
            else if (Utils::String::ICompare(name, "ReportCadence") == 0)
            {
                size_t val = 0;
                if (SUCCEEDED(Utils::String::ToUIntTStrict(value, val)))
                {
                    consumeTotalDiskIoPlugin->SampleStorage.SetReportCadence(val);
                }
                else
                {
                    LogInvalidIntegerConfiguration(name, value);
                }
            }
            else
            {
                TraceLoggingWrite(
                    g_totalDiskIoConsumerProvider,
                    "Create_UnusedConfiguration",
                    TraceLoggingValue(config->name, "Name"),
                    TraceLoggingValue(config->value, "Value"));
            }
        }

        return reinterpret_cast<adh_plugin_instance*>(
            consumeTotalDiskIoPlugin.release());
    }
    CATCH_LOG();
    return nullptr;
}

void consume_total_disk_io_destroy_plugin_instance(adh_plugin_instance* plugin)
{
    ConsumeTotalDiskIoPlugin* consumeTotalDiskIoPlugin =
        reinterpret_cast<ConsumeTotalDiskIoPlugin*>(plugin);
    delete consumeTotalDiskIoPlugin;

    TraceLoggingWrite(g_totalDiskIoConsumerProvider, "Uninitialize");
    TraceLoggingUnregister(g_totalDiskIoConsumerProvider);
}

static const char* s_totalDiskIoConsumeTargets[] = { "produce-total-disk-io",
                                                     nullptr };
const char* const* consume_total_disk_io_list_consume_targets()
{
    return s_totalDiskIoConsumeTargets;
}

bool consume_total_disk_io_consume_sample(
    adh_plugin_instance* plugin,
    const adh_consume_sample_args* args)
{
    try
    {
        if (Utils::String::ICompare(
                args->producer_name, produce_total_disk_io.name) != 0)
        {
            TraceLoggingWrite(
                g_totalDiskIoConsumerProvider,
                "PassedUnusedSample",
                TraceLoggingValue(args->producer_name, "ProducerName"));
            return false;
        }

        ConsumeTotalDiskIoPlugin* consumeTotalDiskIoPlugin =
            reinterpret_cast<ConsumeTotalDiskIoPlugin*>(plugin);
        return consumeTotalDiskIoPlugin->ConsumeSample(
            args->sample_key, *static_cast<TotalDiskIoSample*>(args->sample));
    }
    CATCH_LOG();
    return false;
}

DEFINE_CONSUMER_PLUGIN(
    consume_total_disk_io,
    "consume-total-disk-io",
    consume_total_disk_io_create_plugin_instance,
    consume_total_disk_io_destroy_plugin_instance,
    consume_total_disk_io_list_consume_targets,
    consume_total_disk_io_consume_sample);

std::unique_ptr<TotalDiskIoSample>
ProduceTotalDiskIoPlugin::ProduceSample(uint64_t sampleKey)
{
    TraceLoggingWrite(
        g_totalDiskIoProducerProvider,
        "Produce",
        TraceLoggingValue(sampleKey, "SampleKey"));

    auto totalDiskIoSample = std::make_unique<TotalDiskIoSample>();

    DiskIoInfo diskIoInfo;

    DiskStatParser parser;
    if (!parser.Parse(*totalDiskIoSample, SummaryOnly))
    {
        return nullptr;
    }

    return totalDiskIoSample;
}

bool ConsumeTotalDiskIoPlugin::ConsumeSample(
    uint64_t sampleKey,
    const TotalDiskIoSample& sample)
{
    TraceLoggingWrite(g_totalDiskIoConsumerProvider, "Consume");

    SampleStorage.AddMetric(sampleKey, sample);

    if (SampleStorage.IsReportReady())
    {
        TraceLoggingWrite(g_totalDiskIoConsumerProvider, "Consume_Reporting");

        SampleStorage.ReportMetrics();
    }

    return true;
}
