// Copyright (c) Microsoft Corporation. All rights reserved.

#pragma once

#include <string>
#include <vector>

#include <azure-device-util/FileSystem.h>
#include <azure-device-util/String.h>

#include "MemoryMetrics.h"

namespace Azure { namespace Device { namespace Health { namespace Plugin {

class MemInfoRuntimeParser
{
  public:
    MemInfoRuntimeParser() : _rootPath("/") {}

    MemInfoRuntimeParser(std::string_view rootPath) : _rootPath(rootPath) {}

    bool Parse(TotalMemorySample& sample)
    {
        std::string memInfoPath;
        memInfoPath.append(_rootPath);
        memInfoPath.append("proc/meminfo");
        std::string contents;
        if (FAILED(Utils::FileSystem::ReadUTF8StringFromFile(
                memInfoPath, contents)))
        {
            return false;
        }

        std::vector<std::string> lines = Utils::String::Split(contents, "\n");
        for (const auto& line : lines)
        {
            using namespace std::string_view_literals;

            // Memory metric names
            constexpr std::string_view names[MemInfoState::NUM_VALUES] = {
                "MemTotal:"sv,    "MemFree:"sv,      "Buffers:"sv,
                "Cached:"sv,      "SwapCached:"sv,   "Active:"sv,
                "Inactive:"sv,    "HighTotal:"sv,    "HighFree:"sv,
                "LowTotal:"sv,    "LowFree:"sv,      "SwapTotal:"sv,
                "SwapFree:"sv,    "Dirty:"sv,        "Writeback:"sv,
                "AnonPages:"sv,   "Mapped:"sv,       "Shmem:"sv,
                "CommitLimit:"sv, "Committed_AS:"sv, "VmallocTotal:"sv
            };

            auto takeSampleLambda = [&line, &sample, &names](uint32_t index) {
                //
                // Try to match this line against a given
                // entry in the names array
                //
                if (Utils::String::IStartsWith(line, names[index]))
                {
                    std::string sampleStr =
                        std::string{ &line[names[index].size()],
                                     line.size() - names[index].size() };

                    Utils::String::Trim(sampleStr);

                    //
                    // The string here has leading spaces and
                    // trailing "KB". We rely on atoll to strip
                    // those for us.
                    //
                    // TODO: KB is not guaranteed -- deal with units
                    sample.Sample[index] = atoll(sampleStr.c_str());
                    return true;
                }

                return false;
            };

            //
            // For each line, attempt to match it against all
            // interesting metrics. Stop if a match is found.
            //
            for (int idx = 0; idx < MemInfoState::NUM_VALUES; idx++)
            {
                if (takeSampleLambda(idx))
                {
                    break;
                }
            }
        }

        return true;
    }

  private:
    std::string _rootPath;
};

}}}}
