// Copyright (c) Microsoft Corporation. All rights reserved.

#include "MemoryMetricStorage.h"

#include <type_traits>
#include <utility>

#include <azure-device-util/ResultException.h>

#include "Provider.h"

#ifdef AZURE_DEVICE_USE_CLOG
    #define CLOG_INLINE_IMPLEMENTATION 1
    #include "MemoryMetricStorage.cpp.clog.h"
    #include "privacytags.h"
#else
    #include "clog_stubs.h"
#endif

namespace Azure { namespace Device { namespace Health { namespace Plugin {

/*++
    Adds a metric to the storage.

Arguments:
    memorySet - The sample to add to our storage
--*/
void MemoryMetricStorage::AddMetric(uint64_t sampleKey, const TotalMemorySample& sample)
{
    TraceLoggingWrite(
        g_totalMemoryConsumerProvider,
        "AddMetric",
        TraceLoggingValue(sampleKey, "SampleKey"),
        TraceLoggingValue(
            sample.Sample[MemInfoState::MEMINFO_TOTAL_KB], "TotalKb"),
        TraceLoggingValue(sample.Sample[MemInfoState::MEMINFO_FREE_KB], "FreeKb"),
        TraceLoggingValue(
            sample.Sample[MemInfoState::MEMINFO_BUFFERS_KB], "BuffersKb"),
        TraceLoggingValue(
            sample.Sample[MemInfoState::MEMINFO_CACHED_KB], "CachedKb"),
        TraceLoggingValue(
            sample.Sample[MemInfoState::MEMINFO_SWAPCACHED_KB], "SwapcachedKb"),
        TraceLoggingValue(
            sample.Sample[MemInfoState::MEMINFO_ACTIVE_KB], "ActiveKb"),
        TraceLoggingValue(
            sample.Sample[MemInfoState::MEMINFO_INACTIVE_KB], "InactiveKb"),
        TraceLoggingValue(
            sample.Sample[MemInfoState::MEMINFO_HIGH_TOTAL_KB], "HighTotalKb"),
        TraceLoggingValue(
            sample.Sample[MemInfoState::MEMINFO_HIGH_FREE_KB], "HighFreeKb"),
        TraceLoggingValue(
            sample.Sample[MemInfoState::MEMINFO_LOW_TOTAL_KB], "LowTotalKb"),
        TraceLoggingValue(
            sample.Sample[MemInfoState::MEMINFO_LOW_FREE_KB], "LowFreeKb"),
        TraceLoggingValue(
            sample.Sample[MemInfoState::MEMINFO_SWAP_TOTAL_KB], "SwapTotalKb"),
        TraceLoggingValue(
            sample.Sample[MemInfoState::MEMINFO_SWAP_FREE_KB], "SwapFreeKb"),
        TraceLoggingValue(
            sample.Sample[MemInfoState::MEMINFO_DIRTY_KB], "DirtyKb"),
        TraceLoggingValue(
            sample.Sample[MemInfoState::MEMINFO_WRITEBACK_KB], "WritebackKb"),
        TraceLoggingValue(
            sample.Sample[MemInfoState::MEMINFO_ANON_PAGES_KB], "AnonPagesKb"),
        TraceLoggingValue(
            sample.Sample[MemInfoState::MEMINFO_MAPPED_KB], "MappedKb"),
        TraceLoggingValue(
            sample.Sample[MemInfoState::MEMINFO_SHARED_MEMORY_KB],
            "SharedMemoryKb"),
        TraceLoggingValue(
            sample.Sample[MemInfoState::MEMINFO_COMMIT_LIMIT_KB], "CommitLimitKb"),
        TraceLoggingValue(
            sample.Sample[MemInfoState::MEMINFO_COMMITED_AS_KB], "CommitedAsKb"),
        TraceLoggingValue(
            sample.Sample[MemInfoState::MEMINFO_VMALLOC_TOTAL_KB],
            "VallocTotalKb"));

    if (_oldestValidEntry == -1)
    {
        _oldestValidEntry = 0;
    }
    else if (_currentRingEntry == _oldestValidEntry)
    {
        IncrementWithWrap(_oldestValidEntry, _ringMetrics.size());
    }

    _ringMetrics[_currentRingEntry] = std::make_pair(sampleKey, sample);
    _newestValidEntry = _currentRingEntry;

    _samplesSinceLastReport++;

    IncrementWithWrap(_currentRingEntry, _ringMetrics.size());
}

/*++
    Reports a histogram of deciles (0-10, 10-20) as a ratio of active time to
    idle time as well as total times.

Parameters:
    conf - The service configuration given to us by the device health service.
--*/
MemoryMetricStorage::ReportMetrics MemoryMetricStorage::GenerateReportMetrics()
{
    TraceLoggingWrite(g_totalMemoryConsumerProvider, "ReportMetrics");

    try
    {
        // We need at least 2 samples to generate a report
        if (_oldestValidEntry == -1 || _newestValidEntry == -1 ||
            _oldestValidEntry >= static_cast<int32_t>(_ringMetrics.size()) ||
            _newestValidEntry >= static_cast<int32_t>(_ringMetrics.size()) ||
            _oldestValidEntry == _newestValidEntry)
        {
            // Logic error in plugin.
            THROW_HR(make_error_code(std::errc::operation_not_permitted));
        }

        // Calculate total time period covered
        const auto& oldestSample = _ringMetrics[_oldestValidEntry];
        const auto& newestSample = _ringMetrics[_newestValidEntry];

        auto windowStartMs = std::chrono::milliseconds(oldestSample.first);
        auto windowEndMs = std::chrono::milliseconds(newestSample.first);

        auto totalWindowDurationMs = windowEndMs - windowStartMs;

        // Calculate total free memory per bucket
        int32_t DecileCounts[10] = { 0 };

        int32_t numEntriesInReport = 0;
        int32_t index = _oldestValidEntry;

        while (numEntriesInReport <= _reportCadence)
        {
            auto const* currentSample = &(_ringMetrics[index]);

            int32_t percentTotalMemUsedDecile =
                ((double) currentSample->second.Sample[MemInfoState::MEMINFO_FREE_KB] /
                 currentSample->second.Sample[MemInfoState::MEMINFO_TOTAL_KB]) *
                10;

            // Sanitize data
            if (percentTotalMemUsedDecile >= 10)
            {
                percentTotalMemUsedDecile = 10;
            }
            else if (percentTotalMemUsedDecile < 0)
            {
                percentTotalMemUsedDecile = 0;
            }

            DecileCounts[percentTotalMemUsedDecile]++;

            // Update index
            IncrementWithWrap(index, _ringMetrics.size());

            // Update entries in sample
            numEntriesInReport++;
        }

        ReportMetrics metrics;
        metrics.TotalWindowDurationMs = totalWindowDurationMs.count();
        metrics.MemTotalKb =
            newestSample.second.Sample[MemInfoState::MEMINFO_TOTAL_KB];
        metrics.MemFreeKb =
            newestSample.second.Sample[MemInfoState::MEMINFO_FREE_KB];
        metrics.BuffersKb =
            newestSample.second.Sample[MemInfoState::MEMINFO_BUFFERS_KB];
        metrics.CachedKb =
            newestSample.second.Sample[MemInfoState::MEMINFO_CACHED_KB];
        metrics.SwapCachedKb =
            newestSample.second.Sample[MemInfoState::MEMINFO_SWAPCACHED_KB];
        metrics.ActiveKb =
            newestSample.second.Sample[MemInfoState::MEMINFO_ACTIVE_KB];
        metrics.InactiveKb =
            newestSample.second.Sample[MemInfoState::MEMINFO_INACTIVE_KB];
        metrics.HighTotalKb =
            newestSample.second.Sample[MemInfoState::MEMINFO_HIGH_TOTAL_KB];
        metrics.HighFreeKb =
            newestSample.second.Sample[MemInfoState::MEMINFO_HIGH_FREE_KB];
        metrics.LowTotalKb =
            newestSample.second.Sample[MemInfoState::MEMINFO_LOW_TOTAL_KB];
        metrics.LowFreeKb =
            newestSample.second.Sample[MemInfoState::MEMINFO_LOW_FREE_KB];
        metrics.SwapTotalKb =
            newestSample.second.Sample[MemInfoState::MEMINFO_SWAP_TOTAL_KB];
        metrics.SwapFreeKb =
            newestSample.second.Sample[MemInfoState::MEMINFO_SWAP_FREE_KB];
        metrics.DirtyKb =
            newestSample.second.Sample[MemInfoState::MEMINFO_DIRTY_KB];
        metrics.WritebackKb =
            newestSample.second.Sample[MemInfoState::MEMINFO_WRITEBACK_KB];
        metrics.AnonPagesKb =
            newestSample.second.Sample[MemInfoState::MEMINFO_ANON_PAGES_KB];
        metrics.MappedKb =
            newestSample.second.Sample[MemInfoState::MEMINFO_MAPPED_KB];
        metrics.SharedMemoryKb =
            newestSample.second.Sample[MemInfoState::MEMINFO_SHARED_MEMORY_KB];
        metrics.CommitLimitKb =
            newestSample.second.Sample[MemInfoState::MEMINFO_COMMIT_LIMIT_KB];
        metrics.CommitedAsKb =
            newestSample.second.Sample[MemInfoState::MEMINFO_COMMITED_AS_KB];
        metrics.VmallocTotalKb =
            newestSample.second.Sample[MemInfoState::MEMINFO_VMALLOC_TOTAL_KB];
        std::copy(
            DecileCounts,
            DecileCounts + std::extent<decltype(DecileCounts)>::value,
            metrics.TotalMemoryDecileCounts);

        _samplesSinceLastReport = 0;

        return metrics;
    }
    CATCH_FAIL_FAST();
}

void MemoryMetricStorage::FireReportMetrics(const ReportMetrics& metrics)
{
    // Send report event, keep this style consistent with Asimov schema
    TraceLoggingWrite(
        g_totalMemoryConsumerProvider,
        "TotalMemoryStats",
        TraceLoggingValue(metrics.TotalWindowDurationMs, "TotalWindowDurationMs"),
        TraceLoggingValue(metrics.MemTotalKb, "MemoryTotalKb"),
        TraceLoggingValue(metrics.MemFreeKb, "MemoryFreeKb"),
        TraceLoggingValue(metrics.BuffersKb, "BuffersKb"),
        TraceLoggingValue(metrics.CachedKb, "CachedKb"),
        TraceLoggingValue(metrics.SwapCachedKb, "SwapCachedKb"),
        TraceLoggingValue(metrics.ActiveKb, "ActiveKb"),
        TraceLoggingValue(metrics.InactiveKb, "InactiveKb"),
        TraceLoggingValue(metrics.HighTotalKb, "HighTotalKb"),
        TraceLoggingValue(metrics.HighFreeKb, "HighFreeKb"),
        TraceLoggingValue(metrics.LowTotalKb, "LowTotalKb"),
        TraceLoggingValue(metrics.LowFreeKb, "LowFreeKb"),
        TraceLoggingValue(metrics.SwapTotalKb, "SwapTotalKb"),
        TraceLoggingValue(metrics.SwapFreeKb, "SwapFreeKb"),
        TraceLoggingValue(metrics.DirtyKb, "DirtyKb"),
        TraceLoggingValue(metrics.WritebackKb, "WritebackKb"),
        TraceLoggingValue(metrics.AnonPagesKb, "AnonPagesKb"),
        TraceLoggingValue(metrics.MappedKb, "MappedKb"),
        TraceLoggingValue(metrics.SharedMemoryKb, "SharedMemoryKb"),
        TraceLoggingValue(metrics.CommitLimitKb, "CommitLimitKb"),
        TraceLoggingValue(metrics.CommitedAsKb, "CommitedAsKb"),
        TraceLoggingValue(metrics.VmallocTotalKb, "VmAllocKb"),
        TraceLoggingInt32Array(
            metrics.TotalMemoryDecileCounts,
            std::extent<decltype(metrics.TotalMemoryDecileCounts)>::value,
            "DecileCounts"));

    LatteTelemetry(
        TotalMemoryStats,
        "Azure.DeviceHealth.Plugins.ConsumeTotalMemory",
        PrivacyTags_ProductAndServicePerformance,
        "%{TotalWindowDurationMs,lld} \
        %{MemoryTotalKb,u} \
        %{MemoryFreeKb,u} \
        %{BuffersKb,u} \
        %{CachedKb,u} \
        %{SwapCachedKb,u} \
        %{ActiveKb,u} \
        %{InactiveKb,u} \
        %{HighTotalKb,u} \
        %{HighFreeKb,u} \
        %{LowTotalKb,u} \
        %{LowFreeKb,u} \
        %{SwapTotalKb,u} \
        %{SwapFreeKb,u} \
        %{DirtyKb,u} \
        %{WritebackKb,u} \
        %{AnonPagesKb,u} \
        %{MappedKb,u} \
        %{SharedMemoryKb,u} \
        %{CommitLimitKb,u} \
        %{CommitedAsKb,u} \
        %{VmAllocKb,u} \
        %{DecileCounts,!INT32ARRAY!}",
        metrics.TotalWindowDurationMs,
        metrics.MemTotalKb,
        metrics.MemFreeKb,
        metrics.BuffersKb,
        metrics.CachedKb,
        metrics.SwapCachedKb,
        metrics.ActiveKb,
        metrics.InactiveKb,
        metrics.HighTotalKb,
        metrics.HighFreeKb,
        metrics.LowTotalKb,
        metrics.LowFreeKb,
        metrics.SwapTotalKb,
        metrics.SwapFreeKb,
        metrics.DirtyKb,
        metrics.WritebackKb,
        metrics.AnonPagesKb,
        metrics.MappedKb,
        metrics.SharedMemoryKb,
        metrics.CommitLimitKb,
        metrics.CommitedAsKb,
        metrics.VmallocTotalKb,
        CLOG_ARRAY(
            std::extent<decltype(metrics.TotalMemoryDecileCounts)>::value *
                sizeof(decltype(metrics.TotalMemoryDecileCounts[0])),
            metrics.TotalMemoryDecileCounts));
}

}}}}
