// Copyright (c) Microsoft Corporation. All rights reserved.

#include <mutex>
#include <string>
#include <string_view>
#include <unistd.h>
#include <utility>
#include <vector>

#include <azure-device-health-plugin/plugin-dev.h>
#include <azure-device-util/AndroidLogHelper.h>
#include <azure-device-util/FileSystem.h>
#include <azure-device-util/String.h>

#include "MemInfoRuntimeParser.h"
#include "MemoryMetricStorage.h"
#include "MemoryMetrics.h"
#include "Provider.h"

TRACELOGGING_DEFINE_PROVIDER(
    g_totalMemoryProducerProvider,
    "Azure.DeviceHealth.Plugins.ProduceTotalMemory",
    (0x8cdf4f54, 0x0a75, 0x1a4c, 0xa2, 0x5a, 0x03, 0xdb, 0x1b, 0xae, 0x95, 0x81));

TRACELOGGING_DEFINE_PROVIDER(
    g_totalMemoryConsumerProvider,
    "Azure.DeviceHealth.Plugins.ConsumeTotalMemory",
    (0x43686241, 0x1d26, 0xfe4d, 0xbb, 0x9c, 0x24, 0xd5, 0x14, 0x0f, 0x12, 0x9a));

namespace Azure { namespace Device { namespace Health { namespace Plugin {

class ProduceTotalMemoryPlugin
{
  public:
    std::unique_ptr<TotalMemorySample> ProduceSample(uint64_t sampleKey);

  public:
    bool SummaryOnly = true;
};

class ConsumeTotalMemoryPlugin
{
  public:
    ConsumeTotalMemoryPlugin()
    {
        SampleStorage.SetReportCadence(30);
        SampleStorage.SetBufferSize(30);

        _clockTicksPerSecond = sysconf(_SC_CLK_TCK);
    }

    bool
    ConsumeSample(uint64_t sampleKey, const TotalMemorySample& totalMemorySample);

  public:
    MemoryMetricStorage SampleStorage;

  private:
    uint64_t _clockTicksPerSecond = 0;
};

}}}}

namespace Utils = Azure::Device::Utils;
using namespace Azure::Device::Health::Plugin;

adh_plugin_instance* produce_total_memory_create_plugin_instance(
    const adh_plugin_config_pair* configs,
    size_t configs_count)
{
    try
    {
        TraceLoggingRegister(g_totalMemoryProducerProvider);
        TraceLoggingWrite(g_totalMemoryProducerProvider, "Initialize");
        ADHS_LOG("Initialize TotalMemory Plugin");

        auto produceTotalMemoryPlugin =
            std::make_unique<ProduceTotalMemoryPlugin>();

        for (const adh_plugin_config_pair* config = configs;
             config != (configs + configs_count);
             ++config)
        {
            std::string_view name = config->name;
            std::string_view value = config->value;

            if (Utils::String::ICompare(name, "SummaryOnly") == 0)
            {
                if (Utils::String::ICompare(value, "true") == 0)
                {
                    produceTotalMemoryPlugin->SummaryOnly = true;
                }
                else if (Utils::String::ICompare(value, "false") == 0)
                {
                    produceTotalMemoryPlugin->SummaryOnly = false;
                }
                else
                {
                    TraceLoggingWrite(
                        g_totalMemoryProducerProvider,
                        "Create_InvalidBooleanConfiguration",
                        TraceLoggingCountedString(
                            name.data(), name.size(), "Name"),
                        TraceLoggingCountedString(
                            value.data(), value.size(), "Value"));
                }
            }
            else
            {
                TraceLoggingWrite(
                    g_totalMemoryProducerProvider,
                    "Create_UnusedConfiguration",
                    TraceLoggingValue(config->name, "Name"),
                    TraceLoggingValue(config->value, "Value"));
            }
        }

        return reinterpret_cast<adh_plugin_instance*>(
            produceTotalMemoryPlugin.release());
    }
    CATCH_LOG();
    return nullptr;
}

void produce_total_memory_destroy_plugin_instance(adh_plugin_instance* plugin)
{
    ProduceTotalMemoryPlugin* produceTotalMemoryPlugin =
        reinterpret_cast<ProduceTotalMemoryPlugin*>(plugin);
    delete produceTotalMemoryPlugin;

    TraceLoggingWrite(g_totalMemoryProducerProvider, "Uninitialize");
    TraceLoggingUnregister(g_totalMemoryProducerProvider);
}

void* produce_total_memory_produce_sample(
    adh_plugin_instance* plugin,
    const adh_produce_sample_args* args)
{
    try
    {
        ProduceTotalMemoryPlugin* produceTotalMemoryPlugin =
            reinterpret_cast<ProduceTotalMemoryPlugin*>(plugin);
        return produceTotalMemoryPlugin->ProduceSample(args->sample_key).release();
    }
    CATCH_LOG();
    return nullptr;
}

void produce_total_memory_free_sample(
    [[maybe_unused]] adh_plugin_instance* plugin,
    void* sample)
{
    if (sample)
    {
        delete reinterpret_cast<TotalMemorySample*>(sample);
    }
}

DEFINE_PRODUCER_PLUGIN(
    produce_total_memory,
    "produce-total-memory",
    produce_total_memory_create_plugin_instance,
    produce_total_memory_destroy_plugin_instance,
    produce_total_memory_produce_sample,
    produce_total_memory_free_sample);

static void
LogInvalidIntegerConfiguration(std::string_view name, std::string_view value)
{
    TraceLoggingWrite(
        g_totalMemoryConsumerProvider,
        "Create_InvalidIntegerConfiguration",
        TraceLoggingCountedString(name.data(), name.size(), "Name"),
        TraceLoggingCountedString(value.data(), value.size(), "Value"));
};

adh_plugin_instance* consume_total_memory_create_plugin_instance(
    const adh_plugin_config_pair* configs,
    size_t configs_count)
{
    try
    {
        TraceLoggingRegister(g_totalMemoryConsumerProvider);
        TraceLoggingWrite(g_totalMemoryConsumerProvider, "Initialize");

        auto consumeTotalMemoryPlugin =
            std::make_unique<ConsumeTotalMemoryPlugin>();

        for (const adh_plugin_config_pair* config = configs;
             config != (configs + configs_count);
             ++config)
        {
            std::string_view name = config->name;
            std::string_view value = config->value;

            if (Utils::String::ICompare(name, "CircularBufferSize") == 0)
            {
                size_t val = 0;
                if (SUCCEEDED(Utils::String::ToUIntTStrict(value, val)))
                {
                    consumeTotalMemoryPlugin->SampleStorage.SetBufferSize(val);
                }
                else
                {
                    LogInvalidIntegerConfiguration(name, value);
                }
            }
            else if (Utils::String::ICompare(name, "ReportCadence") == 0)
            {
                size_t val = 0;
                if (SUCCEEDED(Utils::String::ToUIntTStrict(value, val)))
                {
                    consumeTotalMemoryPlugin->SampleStorage.SetReportCadence(val);
                }
                else
                {
                    LogInvalidIntegerConfiguration(name, value);
                }
            }
            else
            {
                TraceLoggingWrite(
                    g_totalMemoryConsumerProvider,
                    "Create_UnusedConfiguration",
                    TraceLoggingValue(config->name, "Name"),
                    TraceLoggingValue(config->value, "Value"));
            }
        }

        return reinterpret_cast<adh_plugin_instance*>(
            consumeTotalMemoryPlugin.release());
    }
    CATCH_LOG();
    return nullptr;
}

void consume_total_memory_destroy_plugin_instance(adh_plugin_instance* plugin)
{
    ConsumeTotalMemoryPlugin* consumeTotalMemoryPlugin =
        reinterpret_cast<ConsumeTotalMemoryPlugin*>(plugin);
    delete consumeTotalMemoryPlugin;

    TraceLoggingWrite(g_totalMemoryConsumerProvider, "Uninitialize");
    TraceLoggingUnregister(g_totalMemoryConsumerProvider);
}

static const char* s_totalMemoryConsumeTargets[] = { "produce-total-memory",
                                                     nullptr };
const char* const* consume_total_memory_list_consume_targets()
{
    return s_totalMemoryConsumeTargets;
}

bool consume_total_memory_consume_sample(
    adh_plugin_instance* plugin,
    const adh_consume_sample_args* args)
{
    try
    {
        if (Utils::String::ICompare(
                args->producer_name, produce_total_memory.name) != 0)
        {
            TraceLoggingWrite(
                g_totalMemoryConsumerProvider,
                "PassedUnusedSample",
                TraceLoggingValue(args->producer_name, "ProducerName"));
            return false;
        }

        ConsumeTotalMemoryPlugin* consumeTotalMemoryPlugin =
            reinterpret_cast<ConsumeTotalMemoryPlugin*>(plugin);
        return consumeTotalMemoryPlugin->ConsumeSample(
            args->sample_key, *static_cast<TotalMemorySample*>(args->sample));
    }
    CATCH_LOG();
    return false;
}

DEFINE_CONSUMER_PLUGIN(
    consume_total_memory,
    "consume-total-memory",
    consume_total_memory_create_plugin_instance,
    consume_total_memory_destroy_plugin_instance,
    consume_total_memory_list_consume_targets,
    consume_total_memory_consume_sample);

std::unique_ptr<TotalMemorySample>
ProduceTotalMemoryPlugin::ProduceSample(uint64_t sampleKey)
{
    TraceLoggingWrite(
        g_totalMemoryProducerProvider,
        "Produce",
        TraceLoggingValue(sampleKey, "SampleKey"));

    auto totalMemorySample = std::make_unique<TotalMemorySample>();

    MemInfoRuntimeParser parser;
    if (!parser.Parse(*totalMemorySample))
    {
        return nullptr;
    }

    return totalMemorySample;
}

bool ConsumeTotalMemoryPlugin::ConsumeSample(
    uint64_t sampleKey,
    const TotalMemorySample& sample)
{
    TraceLoggingWrite(g_totalMemoryConsumerProvider, "Consume");

    SampleStorage.AddMetric(sampleKey, sample);

    if (SampleStorage.IsReportReady())
    {
        TraceLoggingWrite(g_totalMemoryConsumerProvider, "Consume_Reporting");

        auto metrics = SampleStorage.GenerateReportMetrics();

        SampleStorage.FireReportMetrics(metrics);
    }

    return true;
}

// extern "C" __attribute__((visibility("default"))) bool
// Produce(const ServiceConfiguration& conf, Sample& sample)
// {
//     (void) conf;

//     try
//     {
//         TraceLoggingWrite(
//             g_deviceHealthTotalMemoryProducerConsumerProvider, "Produce");

//         // Get processor sample
//         MemoryInfo info;
//         meminfo_runtime_parser_t parser;
//         if (!parser.parse(info))
//         {
//             return false;
//         }

//         sample.SampleData = make_unique_void(new
//         MemoryInfo(std::move(info))); return true;
//     }
//     CATCH_LOG();

//     return false;
// }

// extern "C" __attribute__((visibility("default"))) bool
// Consume(const ServiceConfiguration& conf, const Sample& sample)
// {
//     (void) conf;

//     try
//     {
//         TraceLoggingWrite(
//             g_deviceHealthTotalMemoryProducerConsumerProvider, "Consume");

//         // Save the memory info into our circular buffer.
//         g_storage.AddMetric(sample);

//         // Report if needed
//         if (g_storage.IsReportReady())
//         {
//             auto metrics = g_storage.GenerateReportMetrics();
//             g_storage.FireReportMetrics(metrics);
//         }

//         return true;
//     }
//     CATCH_LOG();

//     return false;
// }