// Copyright (c) Microsoft Corporation. All rights reserved.

#include "NetworkMetricStorage.h"

#include <type_traits>
#include <utility>

#include "Provider.h"

#ifdef AZURE_DEVICE_USE_CLOG
    #define CLOG_INLINE_IMPLEMENTATION 1
    #include "NetworkMetricStorage.cpp.clog.h"
    #include "privacytags.h"
#else
    #include "clog_stubs.h"
#endif

namespace Azure { namespace Device { namespace Health { namespace Plugin {

/*++
    Adds a metric to the storage and, if necessary, reports to telemetry.

Arguments:
    networkSet - The sample to add to our storage
--*/
void NetworkMetricStorage::AddMetric(
    uint64_t sampleKey,
    const TotalNetworkSample& sample)
{
    for (const NetworkInterfaceInfo& networkInfo : sample.Interfaces)
    {
        TraceLoggingWrite(
            g_totalNetworkConsumerProvider,
            "AddMetric",
            TraceLoggingValue(networkInfo.InterfaceName.c_str(), "DeviceName"),
            TraceLoggingValue(networkInfo.Sample[BYTES_RECEIVED], "BytesReceived"),
            TraceLoggingValue(
                networkInfo.Sample[BYTES_TRANSMITTED], "BytesTransmitted"),
            TraceLoggingValue(
                networkInfo.Sample[PACKETS_RECEVED], "PacketsReceived"),
            TraceLoggingValue(
                networkInfo.Sample[PACKETS_TRANSMITTED], "PacketsTransmitted"),
            TraceLoggingValue(
                networkInfo.Sample[PACKETS_RECEIVED_DROPPED],
                "PacketsReceivedAndDropped"),
            TraceLoggingValue(
                networkInfo.Sample[PACKETS_TRANSMITTED_DROPPED],
                "PacketsDroppedDuringTransmission"),
            TraceLoggingValue(
                networkInfo.Sample[MULTICAST_PACKETS], "MulticastPackets"),
            TraceLoggingValue(
                networkInfo.Sample[PACKET_COLLISIONS], "PacketsWithCollisions"));
    }

    //
    // Our very first sample is our delta base and is basically ignored to keep
    // our reporting window about the same.
    //
    if (_oldestValidEntry != -1)
    {
        _samplesSinceLastReport++;
    }

    if (_oldestValidEntry == -1)
    {
        _oldestValidEntry = 0;
    }
    else if (_currentRingEntry == _oldestValidEntry)
    {
        IncrementWithWrap(
            _oldestValidEntry, static_cast<int32_t>(_ringMetrics.size()));
    }

    _ringMetrics[_currentRingEntry] = std::make_pair(sampleKey, sample);
    _newestValidEntry = _currentRingEntry;

    IncrementWithWrap(
        _currentRingEntry, static_cast<int32_t>(_ringMetrics.size()));
}

/*++
    Reports a delta of network statistics between _newestValidEntry &&
_oldestValidEntry
--*/
void NetworkMetricStorage::ReportMetrics()
{
    TraceLoggingWrite(g_totalNetworkConsumerProvider, "ReportMetrics");

    const auto& oldestSample = _ringMetrics[_oldestValidEntry];
    const auto& newestSample = _ringMetrics[_newestValidEntry];

    auto windowStart = std::chrono::milliseconds(oldestSample.first);
    auto windowEnd = std::chrono::milliseconds(newestSample.first);

    auto totalWindowDurationMs = windowEnd - windowStart;

    int32_t numPartitions = (int32_t) oldestSample.second.Interfaces.size();

    for (int32_t interfaceIndex = 0; interfaceIndex < numPartitions;
         interfaceIndex++)
    {
        auto calcDelta = [&](uint32_t arrayIndex) -> long long {
            return newestSample.second.Interfaces[interfaceIndex].Sample[arrayIndex] -
                oldestSample.second.Interfaces[interfaceIndex].Sample[arrayIndex];
        };

        long bytesReceived =
            calcDelta(static_cast<uint32_t>(NetworkStats::BYTES_RECEIVED));

        long bytesTransmitted =
            calcDelta(static_cast<uint32_t>(NetworkStats::BYTES_TRANSMITTED));

        long packetsReceived =
            calcDelta(static_cast<uint32_t>(NetworkStats::PACKETS_RECEVED));

        long packetsTranismitted =
            calcDelta(static_cast<uint32_t>(NetworkStats::PACKETS_TRANSMITTED));

        long packetsReceivedDropped = calcDelta(
            static_cast<uint32_t>(NetworkStats::PACKETS_RECEIVED_DROPPED));

        long packetsDroppedDuringTransmission = calcDelta(
            static_cast<uint32_t>(NetworkStats::PACKETS_TRANSMITTED_DROPPED));

        long multicastPackets =
            calcDelta(static_cast<uint32_t>(NetworkStats::MULTICAST_PACKETS));

        long collisionsObserved =
            calcDelta(static_cast<uint32_t>(NetworkStats::PACKET_COLLISIONS));

        TraceLoggingWrite(
            g_totalNetworkConsumerProvider,
            "TotalNetworkStats",
            TraceLoggingValue(
                oldestSample.second.Interfaces[interfaceIndex].InterfaceName.c_str(),
                "DeviceName"),
            TraceLoggingValue(
                totalWindowDurationMs.count(), "TotalWindowDurationMs"),
            TraceLoggingValue(bytesReceived, "BytesReceived"),
            TraceLoggingValue(bytesTransmitted, "BytesTransmitted"),
            TraceLoggingValue(packetsReceived, "PacketsReceived"),
            TraceLoggingValue(packetsTranismitted, "PacketsTransmitted"),
            TraceLoggingValue(packetsReceivedDropped, "PacketsReceivedAndDropped"),
            TraceLoggingValue(
                packetsDroppedDuringTransmission,
                "PacketsDroppedDuringTransmission"),
            TraceLoggingValue(multicastPackets, "MulticastPackets"),
            TraceLoggingValue(collisionsObserved, "PacketsWithCollisions"));

        LatteTelemetry(
            TotalNetworkStats,
            "Azure.DeviceHealth.Plugins.ConsumeTotalNetwork",
            PrivacyTags_ProductAndServicePerformance,
            "%{DeviceName,s} \
            %{TotalWindowDurationMs,lld} \
            %{BytesReceived,ld} \
            %{BytesTransmitted,ld} \
            %{PacketsReceived,ld} \
            %{PacketsTransmitted,ld} \
            %{PacketsReceivedAndDropped,ld} \
            %{PacketsDroppedDuringTransmission,ld} \
            %{MulticastPackets,ld} \
            %{PacketsWithCollisions,ld}",
            oldestSample.second.Interfaces[interfaceIndex].InterfaceName.c_str(),
            totalWindowDurationMs.count(),
            bytesReceived,
            bytesTransmitted,
            packetsReceived,
            packetsTranismitted,
            packetsReceivedDropped,
            packetsDroppedDuringTransmission,
            multicastPackets,
            collisionsObserved);
    }
    _samplesSinceLastReport = 0;
}

}}}}
