// Copyright (c) Microsoft Corporation. All rights reserved.

#include <mutex>
#include <string>
#include <string_view>
#include <unistd.h>
#include <utility>
#include <vector>

#include <azure-device-health-plugin/plugin-dev.h>
#include <azure-device-util/AndroidLogHelper.h>
#include <azure-device-util/FileSystem.h>
#include <azure-device-util/String.h>

#include "NetworkMetricStorage.h"
#include "NetworkMetrics.h"
#include "NetworkStatParser.h"
#include "Provider.h"

TRACELOGGING_DEFINE_PROVIDER(
    g_totalNetworkProducerProvider,
    "Azure.DeviceHealth.Plugins.ProduceTotalNetwork",
    (0xe714f6ab, 0x1fb5, 0x764d, 0xbb, 0xb0, 0xb4, 0x04, 0xa8, 0xce, 0x8d, 0xfb));

TRACELOGGING_DEFINE_PROVIDER(
    g_totalNetworkConsumerProvider,
    "Azure.DeviceHealth.Plugins.ConsumeTotalNetwork",
    (0x4f208d31, 0x1efb, 0x7143, 0x93, 0xbd, 0x6d, 0xd0, 0x05, 0x72, 0xf8, 0x23));

namespace Azure { namespace Device { namespace Health { namespace Plugin {

class ProduceTotalNetworkPlugin
{
  public:
    std::unique_ptr<TotalNetworkSample> ProduceSample(uint64_t sampleKey);

  public:
    bool SummaryOnly = true;
};

class ConsumeTotalNetworkPlugin
{
  public:
    ConsumeTotalNetworkPlugin()
    {
        SampleStorage.SetReportCadence(30);
        SampleStorage.SetBufferSize(30);

        _clockTicksPerSecond = sysconf(_SC_CLK_TCK);
    }

    bool
    ConsumeSample(uint64_t sampleKey, const TotalNetworkSample& totalNetworkSample);

  public:
    NetworkMetricStorage SampleStorage;

  private:
    uint64_t _clockTicksPerSecond = 0;
};

}}}}

namespace Utils = Azure::Device::Utils;
using namespace Azure::Device::Health::Plugin;

adh_plugin_instance* produce_total_network_create_plugin_instance(
    const adh_plugin_config_pair* configs,
    size_t configs_count)
{
    try
    {
        TraceLoggingRegister(g_totalNetworkProducerProvider);
        TraceLoggingWrite(g_totalNetworkProducerProvider, "Initialize");
        ADHS_LOG("Initialize TotalNetwork Plugin");

        auto produceTotalNetworkPlugin =
            std::make_unique<ProduceTotalNetworkPlugin>();

        for (const adh_plugin_config_pair* config = configs;
             config != (configs + configs_count);
             ++config)
        {
            std::string_view name = config->name;
            std::string_view value = config->value;

            if (Utils::String::ICompare(name, "SummaryOnly") == 0)
            {
                if (Utils::String::ICompare(value, "true") == 0)
                {
                    produceTotalNetworkPlugin->SummaryOnly = true;
                }
                else if (Utils::String::ICompare(value, "false") == 0)
                {
                    produceTotalNetworkPlugin->SummaryOnly = false;
                }
                else
                {
                    TraceLoggingWrite(
                        g_totalNetworkProducerProvider,
                        "Create_InvalidBooleanConfiguration",
                        TraceLoggingCountedString(
                            name.data(), name.size(), "Name"),
                        TraceLoggingCountedString(
                            value.data(), value.size(), "Value"));
                }
            }
            else
            {
                TraceLoggingWrite(
                    g_totalNetworkProducerProvider,
                    "Create_UnusedConfiguration",
                    TraceLoggingValue(config->name, "Name"),
                    TraceLoggingValue(config->value, "Value"));
            }
        }

        return reinterpret_cast<adh_plugin_instance*>(
            produceTotalNetworkPlugin.release());
    }
    CATCH_LOG();
    return nullptr;
}

void produce_total_network_destroy_plugin_instance(adh_plugin_instance* plugin)
{
    ProduceTotalNetworkPlugin* produceTotalNetworkPlugin =
        reinterpret_cast<ProduceTotalNetworkPlugin*>(plugin);
    delete produceTotalNetworkPlugin;

    TraceLoggingWrite(g_totalNetworkProducerProvider, "Uninitialize");
    TraceLoggingUnregister(g_totalNetworkProducerProvider);
}

void* produce_total_network_produce_sample(
    adh_plugin_instance* plugin,
    const adh_produce_sample_args* args)
{
    try
    {
        ProduceTotalNetworkPlugin* produceTotalNetworkPlugin =
            reinterpret_cast<ProduceTotalNetworkPlugin*>(plugin);
        return produceTotalNetworkPlugin->ProduceSample(args->sample_key).release();
    }
    CATCH_LOG();
    return nullptr;
}

void produce_total_network_free_sample(
    [[maybe_unused]] adh_plugin_instance* plugin,
    void* sample)
{
    if (sample)
    {
        delete reinterpret_cast<TotalNetworkSample*>(sample);
    }
}

DEFINE_PRODUCER_PLUGIN(
    produce_total_network,
    "produce-total-network",
    produce_total_network_create_plugin_instance,
    produce_total_network_destroy_plugin_instance,
    produce_total_network_produce_sample,
    produce_total_network_free_sample);

static void
LogInvalidIntegerConfiguration(std::string_view name, std::string_view value)
{
    TraceLoggingWrite(
        g_totalNetworkConsumerProvider,
        "Create_InvalidIntegerConfiguration",
        TraceLoggingCountedString(name.data(), name.size(), "Name"),
        TraceLoggingCountedString(value.data(), value.size(), "Value"));
};

adh_plugin_instance* consume_total_network_create_plugin_instance(
    const adh_plugin_config_pair* configs,
    size_t configs_count)
{
    try
    {
        TraceLoggingRegister(g_totalNetworkConsumerProvider);
        TraceLoggingWrite(g_totalNetworkConsumerProvider, "Initialize");

        auto consumeTotalNetworkPlugin =
            std::make_unique<ConsumeTotalNetworkPlugin>();

        for (const adh_plugin_config_pair* config = configs;
             config != (configs + configs_count);
             ++config)
        {
            std::string_view name = config->name;
            std::string_view value = config->value;

            if (Utils::String::ICompare(name, "CircularBufferSize") == 0)
            {
                size_t val = 0;
                if (SUCCEEDED(Utils::String::ToUIntTStrict(value, val)))
                {
                    consumeTotalNetworkPlugin->SampleStorage.SetBufferSize(val);
                }
                else
                {
                    LogInvalidIntegerConfiguration(name, value);
                }
            }
            else if (Utils::String::ICompare(name, "ReportCadence") == 0)
            {
                size_t val = 0;
                if (SUCCEEDED(Utils::String::ToUIntTStrict(value, val)))
                {
                    consumeTotalNetworkPlugin->SampleStorage.SetReportCadence(val);
                }
                else
                {
                    LogInvalidIntegerConfiguration(name, value);
                }
            }
            else
            {
                TraceLoggingWrite(
                    g_totalNetworkConsumerProvider,
                    "Create_UnusedConfiguration",
                    TraceLoggingValue(config->name, "Name"),
                    TraceLoggingValue(config->value, "Value"));
            }
        }

        return reinterpret_cast<adh_plugin_instance*>(
            consumeTotalNetworkPlugin.release());
    }
    CATCH_LOG();
    return nullptr;
}

void consume_total_network_destroy_plugin_instance(adh_plugin_instance* plugin)
{
    ConsumeTotalNetworkPlugin* consumeTotalNetworkPlugin =
        reinterpret_cast<ConsumeTotalNetworkPlugin*>(plugin);
    delete consumeTotalNetworkPlugin;

    TraceLoggingWrite(g_totalNetworkConsumerProvider, "Uninitialize");
    TraceLoggingUnregister(g_totalNetworkConsumerProvider);
}

static const char* s_totalNetworkConsumeTargets[] = { "produce-total-network",
                                                      nullptr };
const char* const* consume_total_network_list_consume_targets()
{
    return s_totalNetworkConsumeTargets;
}

bool consume_total_network_consume_sample(
    adh_plugin_instance* plugin,
    const adh_consume_sample_args* args)
{
    try
    {
        if (Utils::String::ICompare(
                args->producer_name, produce_total_network.name) != 0)
        {
            TraceLoggingWrite(
                g_totalNetworkConsumerProvider,
                "PassedUnusedSample",
                TraceLoggingValue(args->producer_name, "ProducerName"));
            return false;
        }

        ConsumeTotalNetworkPlugin* consumeTotalNetworkPlugin =
            reinterpret_cast<ConsumeTotalNetworkPlugin*>(plugin);
        return consumeTotalNetworkPlugin->ConsumeSample(
            args->sample_key, *static_cast<TotalNetworkSample*>(args->sample));
    }
    CATCH_LOG();
    return false;
}

DEFINE_CONSUMER_PLUGIN(
    consume_total_network,
    "consume-total-network",
    consume_total_network_create_plugin_instance,
    consume_total_network_destroy_plugin_instance,
    consume_total_network_list_consume_targets,
    consume_total_network_consume_sample);

std::unique_ptr<TotalNetworkSample>
ProduceTotalNetworkPlugin::ProduceSample(uint64_t sampleKey)
{
    TraceLoggingWrite(
        g_totalNetworkProducerProvider,
        "Produce",
        TraceLoggingValue(sampleKey, "SampleKey"));

    auto totalNetworkSample = std::make_unique<TotalNetworkSample>();

    NetworkStatParser parser;
    if (!parser.Parse(*totalNetworkSample))
    {
        return nullptr;
    }

    return totalNetworkSample;
}

bool ConsumeTotalNetworkPlugin::ConsumeSample(
    uint64_t sampleKey,
    const TotalNetworkSample& sample)
{
    TraceLoggingWrite(g_totalNetworkConsumerProvider, "Consume");

    SampleStorage.AddMetric(sampleKey, sample);

    if (SampleStorage.IsReportReady())
    {
        TraceLoggingWrite(g_totalNetworkConsumerProvider, "Consume_Reporting");

        SampleStorage.ReportMetrics();
    }

    return true;
}
