// Copyright (c) Microsoft Corporation. All rights reserved.

#pragma once

#include <list>

namespace Azure { namespace Device { namespace Health { namespace PluginUtils {

template<class MetricType>
class MetricRingStorage
{
  public:
    using const_iterator = typename std::list<MetricType>::const_iterator;

    MetricRingStorage() = default;

    MetricRingStorage(size_t maxRingSize, size_t reportFrequency)
        : _maxRingSize(maxRingSize), _reportFrequency(reportFrequency)
    {}

    size_t GetMaxRingSize() const { return _maxRingSize; }
    size_t GetReportFrequency() const { return _reportFrequency; }

    void SetMaxRingSize(size_t maxRingSize) { _maxRingSize = maxRingSize; }
    void SetReportFrequency(size_t reportFrequency)
    {
        _reportFrequency = reportFrequency;
    }

    void AddMetric(const MetricType& metric)
    {
        // Increment _currentReportCount to track next report window
        if (_currentReportCount == _reportFrequency)
        {
            _currentReportCount = 1;
        }
        else
        {
            _currentReportCount++;
        }

        // Add new element to end of the list
        _metrics.emplace_back(metric);

        // Remove oldest element from the front of the list to stay under max
        if (_metrics.size() > _maxRingSize)
        {
            _metrics.pop_front();
        }
    }

    bool IsReportReady() const
    {
        return _currentReportCount == _reportFrequency;
    }

    const_iterator begin() const { return _metrics.begin(); }
    const_iterator end() const { return _metrics.end(); }

  private:
    size_t _maxRingSize = 30;
    size_t _reportFrequency = 30;

    size_t _currentReportCount = 0;
    std::list<MetricType> _metrics;
};

}}}}
