// Copyright (c) Microsoft Corporation. All rights reserved.

#include "ProcessMetadataCache.h"

#include <azure-device-util/FileSystem.h>
#include <azure-device-util/PackageManager.h>
#include <azure-device-util/Process.h>
#include <azure-device-util/String.h>

namespace Azure { namespace Device { namespace Health { namespace PluginUtils {

std::unique_ptr<IProcessMetadataCache> IProcessMetadataCache::Create()
{
    return std::make_unique<ProcessMetadataCache>();
}

std::shared_ptr<const ProcessProperties>
ProcessMetadataCache::GetProcessPropertiesByPid(uint32_t processId)
{
    // See if we already have the pid in our map, return
    if (auto props = _processLru.Get(processId))
    {
        return props;
    }

    auto props = std::make_shared<ProcessProperties>();

    //
    // Android doesn't support querying data about containers
    // and does not have a Linux style package manager to query
    // nor systemd. In this case, return empty properties.
    //
#ifndef ANDROID
    // Parse up /proc/<PID>/cgroup for 'name=systemd' entry
    props->SystemdCGroupPath = LookupSystemdCGroupPath(processId);

    constexpr std::string_view dockerPrefix = "/docker/";
    if (Utils::String::IStartsWith(props->SystemdCGroupPath, dockerPrefix))
    {
        props->DockerInstanceId =
            props->SystemdCGroupPath.substr(dockerPrefix.size());

        auto dockerProperties = LookupDockerProperties(props->DockerInstanceId);

        constexpr std::string_view sha256Prefix = "sha256:";
        if (Utils::String::IStartsWith(dockerProperties.ImageId, sha256Prefix))
        {
            props->DockerImageId =
                dockerProperties.ImageId.substr(sha256Prefix.size());
        }

        props->DockerImageName = dockerProperties.ImageName;
    }

    // Try to find owning package
    auto optOwner = Utils::PackageManager::GetProcessOwnerByPid(processId);
    if (optOwner)
    {
        props->PackageName = optOwner->Name;
        props->PackageVersion = optOwner->Version;
        props->PackageArch = optOwner->Arch;
    }
#endif

    // Add new info to map
    _processLru.Insert(processId, props);

    // return
    return props;
}

std::string ProcessMetadataCache::LookupSystemdCGroupPath(uint32_t processId)
{
    std::filesystem::path cgroupFilePath = "/proc";
    cgroupFilePath /= std::to_string(processId);
    cgroupFilePath /= "cgroup";

    std::ifstream cgroupFileStream{ cgroupFilePath };
    std::string line;
    while (std::getline(cgroupFileStream, line))
    {
        std::vector<std::string> splitPieces = Utils::String::Split(line, ":");
        if (splitPieces.size() == 3 && splitPieces[1] == "name=systemd")
        {
            return splitPieces[2];
        }
    }

    return {};
}

ProcessMetadataCache::DockerProperties
ProcessMetadataCache::LookupDockerProperties(std::string_view dockerObjectId)
{
    Utils::Process::CommandResult dockerResult = Utils::Process::RunCommand(
        "docker",
        { "inspect",
          "--format='{{println .Config.Image}}{{.Image}}'",
          std::string{ dockerObjectId } });

    if (dockerResult.ExitCode != 0 || dockerResult.Output.size() != 2)
    {
        return {};
    }

    DockerProperties props;
    props.ImageName = std::move(dockerResult.Output[0]);
    props.ImageId = std::move(dockerResult.Output[1]);

    return props;
}

}}}}
