// Copyright (c) Microsoft Corporation. All rights reserved.

#include <libudev.h>
#include <memory>

#include <azure-device-health-plugin-utils/UdevHelpers.h>
#include <azure-device-util/ResultException.h>

namespace Azure { namespace Device { namespace Health { namespace PluginUtils {

struct UdevClose
{
    void operator()(udev* u) { udev_unref(u); }
};
using UdevHandle = std::unique_ptr<udev, UdevClose>;

struct UdevEnumerateClose
{
    void operator()(udev_enumerate* e) { udev_enumerate_unref(e); }
};
using UdevEnumerateHandle = std::unique_ptr<udev_enumerate, UdevEnumerateClose>;

struct UdevDeviceClose
{
    void operator()(udev_device* e) { udev_device_unref(e); }
};
using UdevDeviceHandle = std::unique_ptr<udev_device, UdevDeviceClose>;

std::vector<UdevDeviceEntry> EnumerateDeviceAttributes(
    const char* subsystemName,
    const char* sysNamePattern,
    nonstd::span<const char* const> attributeNames)
{
    UdevHandle udev{ udev_new() };
    if (!udev)
    {
        THROW_HR(Utils::make_errno_error_code());
    }

    UdevEnumerateHandle enumerate{ udev_enumerate_new(udev.get()) };
    if (!enumerate)
    {
        THROW_HR(Utils::make_errno_error_code());
    }

    if (subsystemName)
    {
        udev_enumerate_add_match_subsystem(enumerate.get(), subsystemName);
    }

    if (sysNamePattern)
    {
        udev_enumerate_add_match_sysname(enumerate.get(), sysNamePattern);
    }

    udev_enumerate_scan_devices(enumerate.get());

    std::vector<UdevDeviceEntry> deviceEntries;

    udev_list_entry* device = nullptr;
    udev_list_entry_foreach(device, udev_enumerate_get_list_entry(enumerate.get()))
    {
        // If any individual device encounters an error, just skip it.
        const char* path = udev_list_entry_get_name(device);
        UdevDeviceHandle dev{ udev_device_new_from_syspath(udev.get(), path) };
        if (!dev)
        {
            continue;
        }

        const char* sysName = udev_device_get_sysname(dev.get());
        if (sysName == nullptr)
        {
            continue;
        }

        UdevDeviceEntry deviceEntry;
        deviceEntry.SysName = sysName;

        for (auto attrName : attributeNames)
        {
            // Attributes are allowed to come back empty, just use empty string
            const char* attrValue =
                udev_device_get_sysattr_value(dev.get(), attrName);

            deviceEntry.Attributes[attrName] =
                attrValue ? std::string{ attrValue } : std::string{};
        }

        deviceEntries.push_back(std::move(deviceEntry));
    }

    return deviceEntries;
}

}}}}