// Copyright (c) Microsoft Corporation. All rights reserved.

#include <algorithm>
#include <atomic>
#include <chrono>
#include <fstream>
#include <iostream>
#include <limits>
#include <math.h>
#include <mutex>
#include <string>
#include <thread>
#include <unistd.h>
#include <utility>
#include <vector>

#include <azure-device-util/AndroidLogHelper.h>
#include <azure-device-util/FileSystem.h>
#include <azure-device-util/ResultException.h>
#include <azure-device-util/SignalLoop.h>
#include <azure-device-util/String.h>
#include <azure-device-util/Toml.h>
#include <azure-device-util/TraceLoggingTypes.h>

#include "ErrorDefinition.h"
#include "PluginManager.h"
#include "Provider.h"

TRACELOGGING_DEFINE_PROVIDER(
    g_deviceHealthDaemonProvider,
    "Azure.DeviceHealth.Daemon",
    (0x8c4c65d7, 0x3210, 0x4c65, 0xac, 0xc3, 0xd9, 0x20, 0x46, 0xc5, 0xcd, 0x89));

namespace Utils = Azure::Device::Utils;
using namespace Azure::Device::Health;

/*++
    Registered as our error handling callback with the ErrorHelper from
    azure-device-utils package. Logs to LTTNG.

Parameters
    message - The message to log.
    file - The file name (in debug builds) which generated the error.
    line - The line number (in debug builds) which generated the error.
    err - The error code generated.
--*/
static void LogCallback(
    std::string_view message,
    std::string_view file,
    int lineNumber,
    std::error_code err)
{
    TraceLoggingWrite(
        g_deviceHealthDaemonProvider,
        "LogError",
        TraceLoggingStringView(file, "FileName"),
        TraceLoggingValue(lineNumber, "Line"),
        TraceLoggingStringView(message, "Message"),
        TraceLoggingErrorCode(err, "ErrorCode"));

    ADHS_LOG(
        "LogError: %.*s::%d, %.*s %d",
        static_cast<int>(file.size()),
        file.data(),
        lineNumber,
        static_cast<int>(message.size()),
        message.data(),
        err.value());

    std::cerr << "LogError: ";
    std::cerr << "loc='" << file << "@" << lineNumber << "' ";
    std::cerr << "msg='" << message << "' ";
    std::cerr << "err={msg='" << err.message() << "' ";
    std::cerr << "cat='" << err.category().name() << "' ";
    std::cerr << "val='" << err.value() << "'}" << std::endl;
}

/*++
    Performs initialization steps for the health service such as loading
    our configuration and loading plug-ins.

Parameters:
    properties - Service properties which will be used during the runtime of
        the service.
    subsystems - The subsystems singleton.
--*/
static void Initialize(PluginManager& pluginManager)
{
    //
    // Read service properties. Prefer a local configuration file to the one
    // under etc
    //
    std::string configPath = "./adhd_local.conf";
    if (access(configPath.c_str(), F_OK) == -1)
    {
        std::filesystem::path confFilePath =
            Utils::FileSystem::GetConfDirectory();
        confFilePath /= "azure-device-healthd/adhd.conf";
        configPath = confFilePath.string();
    }

    std::filesystem::path libraryPath = Utils::FileSystem::GetLibraryDirectory();
    libraryPath /= std::string{ PluginManager::c_pluginRelativeDirectory };

    pluginManager.DiscoverPluginsInDirectory(libraryPath.string());

    const auto data = toml::parse(configPath);
    const auto servicePropertiesTable = toml::find<toml::value>(data, "service");

    uint32_t quantumMs =
        toml::find<uint32_t>(servicePropertiesTable, "QuantumMs");
    pluginManager.SetQuantumDuration(std::chrono::milliseconds{ quantumMs });

    TraceLoggingWrite(
        g_deviceHealthDaemonProvider,
        "service_main.Initialize",
        TraceLoggingValue(configPath.c_str(), "config_path"),
        TraceLoggingValue(quantumMs, "QuantumSizeMs"));

    // Load plug-ins
    if (data.contains("plugin"))
    {
        const auto pluginsTable =
            toml::get<std::vector<toml::value>>(data.at("plugin"));
        pluginManager.CreateInstancesFromConfig(pluginsTable);
    }
}

int main()
{
    // Register tracelogging
    TraceLoggingRegister(g_deviceHealthDaemonProvider);
    TraceLoggingWrite(g_deviceHealthDaemonProvider, "service_main.main");

    // Register error callback
    Utils::SetResultLoggingCallback(&LogCallback);

    Utils::SignalLoop signalLoop{ SIGTERM, SIGINT };

    PluginManager pluginManager;
    Initialize(pluginManager);

    std::thread runLoopThread{ &PluginManager::RunLoop, std::ref(pluginManager) };

    auto shutdownHandler = [&](int signal) {
        std::cout << "Shutting down due to signal " << signal << std::endl;
        return false;
    };

    signalLoop.RegisterSignalHandler(SIGTERM, shutdownHandler);
    signalLoop.RegisterSignalHandler(SIGINT, shutdownHandler);

    // This will return gracefully on SIGINT or SIGTERM
    signalLoop.Run();

    pluginManager.Stop();
    runLoopThread.join();

    // Unregister tracelogging
    TraceLoggingUnregister(g_deviceHealthDaemonProvider);

    return 0;
}