// Copyright (c) Microsoft Corporation. All rights reserved.

#include "PluginFactory.h"

#include <azure-device-health-plugin/plugin-dev.h>
#include <azure-device-util/ResultException.h>

#include "ErrorHelper.h"
#include "PluginInstance.h"
#include "PluginModule.h"
#include "Provider.h"

namespace Azure { namespace Device { namespace Health {

std::shared_ptr<PluginFactory> PluginFactory::Create(
    std::shared_ptr<PluginModule> module,
    const adh_plugin_definition* pluginDef)
{
    return std::shared_ptr<PluginFactory>{ new PluginFactory{ module,
                                                              pluginDef } };
}

PluginFactory::PluginFactory(
    std::shared_ptr<PluginModule> module,
    const adh_plugin_definition* pluginDef)
    : _module(module), _pluginDefinition(pluginDef)
{}

std::string_view PluginFactory::GetName() const
{
    return _pluginDefinition->name;
}

std::vector<std::string_view> PluginFactory::GetConsumeTargets() const
{
    auto consumeTargetsList = _pluginDefinition->list_consume_targets_fn();
    if (!consumeTargetsList || !*consumeTargetsList)
    {
        TraceLoggingWrite(
            g_deviceHealthDaemonProvider,
            "PluginFactory_GetConsumeTargets_NullOrEmptyList",
            TraceLoggingValue(_pluginDefinition->name, "PluginName"));
        return {};
    };

    std::vector<std::string_view> targets;
    for (auto entry = consumeTargetsList; *entry != nullptr; ++entry)
    {
        targets.emplace_back(*entry);
    }

    return targets;
}

bool PluginFactory::IsProducer() const
{
    return _pluginDefinition->produce_sample_fn != nullptr &&
        _pluginDefinition->free_sample_fn != nullptr;
}

bool PluginFactory::IsConsumer() const
{
    return _pluginDefinition->consume_sample_fn != nullptr &&
        _pluginDefinition->list_consume_targets_fn != nullptr;
}

bool PluginFactory::IsLogger() const
{
    return _pluginDefinition->log_data_fn != nullptr;
}

adh_plugin_instance*
PluginFactory::CreateInstance(const std::map<std::string, std::string>& configMap)
{
    std::vector<adh_plugin_config_pair> configArg;
    for (const auto& config : configMap)
    {
        configArg.push_back(adh_plugin_config_pair{ config.first.c_str(),
                                                    config.second.c_str() });
    }

    adh_plugin_instance* instance = _pluginDefinition->create_plugin_instance_fn(
        configArg.data(), configArg.size());
    THROW_HR_IF(HealthError::E_PLUGIN_FAILED_TO_CREATE_INSTANCE, !instance);

    return instance;
}

}}}
