// Copyright (c) Microsoft Corporation. All rights reserved.

#include "PluginInstance.h"

#include "PluginFactory.h"

namespace Azure { namespace Device { namespace Health {

PluginInstance::PluginInstance(
    std::shared_ptr<PluginFactory> factory,
    adh_plugin_instance* instance)
    : _factory(factory), _instance(instance)
{}

PluginInstance::~PluginInstance()
{
    _factory->GetDefinition()->destroy_plugin_instance_fn(_instance);
}

std::string_view PluginInstance::GetName() const
{
    return _factory->GetName();
}

SamplePtr ProducerPluginInstance::ProduceSample(uint64_t sampleKey)
{
    adh_produce_sample_args produceSampleArgs;
    produceSampleArgs.sample_key = sampleKey;

    void* sample = _factory->GetDefinition()->produce_sample_fn(
        _instance, &produceSampleArgs);

    return SamplePtr{ sample, [this](void* sample) {
                         if (sample)
                             _factory->GetDefinition()->free_sample_fn(
                                 _instance, sample);
                     } };
}

bool ConsumerPluginInstance::ConsumeSample(
    uint64_t sampleKey,
    std::string_view producerName,
    SamplePtr sample)
{
    std::string producerNameNullTerminated{ producerName };

    adh_consume_sample_args consumeSampleArgs;
    consumeSampleArgs.sample_key = sampleKey;
    consumeSampleArgs.producer_name = producerNameNullTerminated.c_str();
    consumeSampleArgs.sample = sample.get();

    return _factory->GetDefinition()->consume_sample_fn(
        _instance, &consumeSampleArgs);
}

bool LoggerPluginInstance::LogData(uint64_t sampleKey)
{
    adh_log_data_args logDataArgs;
    logDataArgs.sample_key = sampleKey;

    return _factory->GetDefinition()->log_data_fn(_instance, &logDataArgs);
}

bool ProducerPluginInstance::ShouldProduceData()
{
    return _quantumCounter.IncrementAndCheckForMax();
}

bool ConsumerPluginInstance::ShouldConsumeProducer(std::string_view producerName)
{
    const auto& consumeTargets = _factory->GetConsumeTargets();
    return std::find(consumeTargets.begin(), consumeTargets.end(), producerName) !=
        consumeTargets.end();
}

bool LoggerPluginInstance::ShouldLogData()
{
    return _quantumCounter.IncrementAndCheckForMax();
}

}}}
