// Copyright (c) Microsoft Corporation. All rights reserved.

#include "PluginManager.h"

#include <map>
#include <thread>
#include <optional>

#include <azure-device-util/AndroidLogHelper.h>
#include <azure-device-util/ErrorHelper.h>
#include <azure-device-util/FileSystem.h>
#include <azure-device-util/ResultException.h>
#include <azure-device-util/String.h>
#include <azure-device-util/Toml.h>

#include "ErrorHelper.h"
#include "PluginFactory.h"
#include "PluginInstance.h"
#include "PluginModule.h"
#include "Provider.h"

namespace Azure { namespace Device { namespace Health {

void PluginManager::DiscoverPluginsInDirectory(std::string_view pluginDirectory)
{
    ADHS_LOG(
        "Discover plugins in directory. %.*s",
        static_cast<int>(pluginDirectory.size()),
        pluginDirectory.data());

    TraceLoggingWrite(
        g_deviceHealthDaemonProvider,
        "PluginManager_DiscoverPluginsInDirectory",
        TraceLoggingCountedString(
            pluginDirectory.data(), pluginDirectory.size(), "Directory"));

    Utils::FileSystem::DoForEachFileWithPattern(
        pluginDirectory, ".*", [this](const std::filesystem::path& filePath) {
            TraceLoggingWrite(
                g_deviceHealthDaemonProvider,
                "PluginManager_DiscoverPluginsInFile",
                TraceLoggingValue(filePath.c_str(), "File"));
            DiscoverPluginsInFile(filePath.c_str());
        });
}

void PluginManager::DiscoverPluginsInFile(const char* filePath)
{
    try
    {
        auto pluginModule = PluginModule::Create(filePath);
        auto pluginFactories = pluginModule->GetPluginFactories();

        for (const auto& factory : pluginFactories)
        {
            auto [itr, inserted] = _pluginFactories.emplace(
                std::string{ factory->GetName() }, factory);

            if (!inserted)
            {
                TraceLoggingWrite(
                    g_deviceHealthDaemonProvider,
                    "PluginManger_DiscoverPluginsInFile_IgnoredDuplicateNamedPlugin",
                    TraceLoggingCountedString(
                        factory->GetName().data(),
                        factory->GetName().size(),
                        "PluginName"));
                LOG_HR(HealthError::E_IGNORED_DUPLICATE_PLUGIN_NAME);
            }
        }
    }
    CATCH_LOG();
}

void PluginManager::CreateInstancesFromConfig(
    const std::vector<toml::value>& pluginsConfig)
{
    PluginNameToFactoryAndConfigMap pluginFactoryConfigMap =
        CreatePluginFactoryConfigMap(pluginsConfig);

    CreatePluginInstances(pluginFactoryConfigMap);

    // PluginInstance refer back to the Factory -> Module that they need,
    // so this will unload modules not referenced by the config file
    _pluginFactories.clear();
}

void PluginManager::RunLoop()
{
    if (_producerPlugins.empty() && _consumerPlugins.empty() &&
        _loggerPlugins.empty())
    {
        TraceLoggingWrite(
            g_deviceHealthDaemonProvider,
            "PluginManager_RunLoop_NoPluginsLoaded");

        return;
    }

    while (!_stop)
    {
        RunIteration();

        std::unique_lock lock{ _stopMutex };
        _stopCondition.wait_for(
            lock, _quantumDuration, [this]() { return _stop.load(); });
    }
}

void PluginManager::Stop()
{
    {
        std::lock_guard lock{ _stopMutex };
        _stop = true;
    }

    _stopCondition.notify_one();
}

void PluginManager::RunIteration()
{
    uint64_t sampleKey =
        std::chrono::duration_cast<std::chrono::nanoseconds>(
            std::chrono::high_resolution_clock::now().time_since_epoch())
            .count();

    uuid_t aId;
    lttngh_ActivityIdCreate(aId);
    lttngh_ActivityIdSet(aId);

    // Give each log plugin a chance to log
    for (auto& logger : _loggerPlugins)
    {
        if (logger->ShouldLogData())
        {
            logger->LogData(sampleKey);
        }
    }

    // Generate data from each producer, and feed it to appropriate consumers
    for (auto& producer : _producerPlugins)
    {
        if (producer->ShouldProduceData())
        {
            SamplePtr sample = producer->ProduceSample(sampleKey);
            if (sample)
            {
                for (auto& consumer : _consumerPlugins)
                {
                    if (consumer->ShouldConsumeProducer(producer->GetName()))
                    {
                        consumer->ConsumeSample(
                            sampleKey, producer->GetName(), sample);
                    }
                }
            }
        }
    }
}

PluginManager::PluginNameToFactoryAndConfigMap
PluginManager::CreatePluginFactoryConfigMap(
    const std::vector<toml::value>& pluginsConfig)
{
    const std::string c_nameProperty = "Name";

    PluginNameToFactoryAndConfigMap configuredPluginFactories;

    // Loop through all configuration entries, mapping all named plugins
    // to their factory and configuration table.
    for (const auto& pluginConfig : pluginsConfig)
    {
        try
        {
            std::string name =
                toml::find<std::string>(pluginConfig, c_nameProperty);
            if (name.empty())
            {
                TraceLoggingWrite(
                    g_deviceHealthDaemonProvider,
                    "PluginManager_SkippingUnnamedPluginConfig");
                continue;
            }

            auto pluginFactoryItr = _pluginFactories.find(name);
            if (pluginFactoryItr == _pluginFactories.end())
            {
                TraceLoggingWrite(
                    g_deviceHealthDaemonProvider,
                    "PluginManager_PluginNotFoundForName",
                    TraceLoggingValue(name.c_str(), "Name"));
                continue;
            }

            configuredPluginFactories[name] =
                std::make_pair(pluginFactoryItr->second, &pluginConfig);
        }
        CATCH_LOG();
    }

    // Rules for adding to the set of plugins to be instantiated:
    // 1. Add all Log plugins
    // 2. For each Consume plugins, locate all dependent Producers:
    //    a. If they are all found, and dependencies and the Consume plugin
    //    b. If they are not all found, add none.
    //
    // Anything left over will not be added, because it is either an unsatisifed
    // consumer or a needless consumer

    PluginNameToFactoryAndConfigMap acceptedPluginEntries;

    // Pass 1: Add all Log plugins
    for (auto&& [name, factoryTablePair] : configuredPluginFactories)
    {
        if (factoryTablePair.first->IsLogger())
        {
            acceptedPluginEntries.emplace(name, factoryTablePair);
        }
    }

    // Pass 2: For each Consume plugin, add Produce plugins
    for (auto&& [name, factoryTablePair] : configuredPluginFactories)
    {
        std::shared_ptr<PluginFactory> factory = factoryTablePair.first;
        if (!factory->IsConsumer())
        {
            continue;
        }

        PluginNameToFactoryAndConfigMap proposedProducerEntries;

        bool failedToLocateProducer = false;
        for (auto producerName : factory->GetConsumeTargets())
        {
            // Look in the accepted plugins first to see if another
            // plugin put it there already
            {
                auto acceptedPluginItr = acceptedPluginEntries.find(producerName);
                if (acceptedPluginItr != acceptedPluginEntries.end())
                {
                    // No need to add, it's already there
                    continue;
                }
            }

            // Next, look in the configured map, and add to proposed if found
            {
                auto configedProducerItr =
                    configuredPluginFactories.find(producerName);
                if (configedProducerItr != configuredPluginFactories.end())
                {
                    proposedProducerEntries.emplace(
                        configedProducerItr->first, configedProducerItr->second);
                    continue;
                }
            }

            // Last, look in the plugin factories to see if we know about
            // this plugin at all and can create it will no config
            {
                auto defaultProducerItr = _pluginFactories.find(producerName);
                if (defaultProducerItr != _pluginFactories.end())
                {
                    proposedProducerEntries.emplace(
                        defaultProducerItr->first,
                        std::make_pair(defaultProducerItr->second, nullptr));
                    continue;
                }
            }

            failedToLocateProducer = true;

            TraceLoggingWrite(
                g_deviceHealthDaemonProvider,
                "PluginManager_FailedToLocateProducer",
                TraceLoggingCountedString(
                    factory->GetName().data(),
                    factory->GetName().size(),
                    "ConsumerName"),
                TraceLoggingCountedString(
                    producerName.data(), producerName.size(), "ProducerName"));

            break;
        }

        if (failedToLocateProducer)
        {
            continue;
        }

        // We found all the producer plugins, add the ones in the proposed
        // list to accepted, along with the consumer
        acceptedPluginEntries.insert(
            proposedProducerEntries.begin(), proposedProducerEntries.end());
        acceptedPluginEntries.emplace(name, factoryTablePair);
    }

    // PASS 3: Log all Producers in config that aren't in accepted as useless
    for (auto&& [name, factoryTablePair] : configuredPluginFactories)
    {
        std::shared_ptr<PluginFactory> factory = factoryTablePair.first;
        if (factory->IsProducer() &&
            acceptedPluginEntries.find(name) == acceptedPluginEntries.end())
        {
            TraceLoggingWrite(
                g_deviceHealthDaemonProvider,
                "PluginManager_ConfiguredProducerNotUsed",
                TraceLoggingValue(name.c_str(), "ProducerName"));
        }
    }

    return acceptedPluginEntries;
}

void PluginManager::CreatePluginInstances(
    const PluginNameToFactoryAndConfigMap& acceptedPluginFactories)
{
    // FUTURE: This could do some smart dependency management to optimize:
    //  1. Don't create Consumer if any dependency Producer failed to create
    //  2. Don't create Producer if all dependent Consumers failed to create
    // Create all the defined plugins

    const std::string c_pollQuantumName{ "PollQuantum" };
    const std::string c_fireOnBootName{ "FireOnBoot" };

    for (auto&& [name, factoryTablePair] : acceptedPluginFactories)
    {
        try
        {
            auto& [factory, pluginConfig] = factoryTablePair;

            std::optional<std::uint64_t> pollQuantumOpt;
            std::map<std::string, std::string> configMap;
            if (pluginConfig != nullptr)
            {
                if (pluginConfig->contains("configuration"))
                {
                    configMap = toml::find<std::map<std::string, std::string>>(
                        *pluginConfig, "configuration");
                }
                if (pluginConfig->contains("PollQuantum"))
                {
                    pollQuantumOpt =
                        toml::find<uint64_t>(*pluginConfig, "PollQuantum");
                }
            }

            adh_plugin_instance* adhInstance = factory->CreateInstance(configMap);
            if (factory->IsProducer())
            {
                auto instance = std::make_shared<ProducerPluginInstance>(
                    factory, adhInstance);
                if (pollQuantumOpt)
                {
                    instance->SetPollQuantum(*pollQuantumOpt);
                }
                _producerPlugins.push_back(instance);
            }
            else if (factory->IsConsumer())
            {
                auto instance = std::make_shared<ConsumerPluginInstance>(
                    factory, adhInstance);
                _consumerPlugins.push_back(instance);
            }
            else if (factory->IsLogger())
            {
                auto instance =
                    std::make_shared<LoggerPluginInstance>(factory, adhInstance);
                if (pollQuantumOpt)
                {
                    instance->SetPollQuantum(*pollQuantumOpt);
                }
                _loggerPlugins.push_back(instance);
            }
            else
            {
                FAIL_FAST();
            }
        }
        catch (const std::system_error& ex)
        {
            ADHS_LOG("Failed to create plugin instance.");

            TraceLoggingWrite(
                g_deviceHealthDaemonProvider,
                "PluginManager_FailedToCreatePluginInstance",
                TraceLoggingValue(name.c_str(), "PluginName"),
                TraceLoggingValue(ex.code().category().name(), "ErrorCat"),
                TraceLoggingValue(ex.code().value(), "ErrorVal"));

            LOG_HR(ex.code());
        }
        CATCH_LOG();
    }
}
}}}
