// Copyright (c) Microsoft Corporation. All rights reserved.

#pragma once

#include <atomic>
#include <condition_variable>
#include <map>
#include <string>
#include <vector>

#include <azure-device-util/Toml.h>

namespace Azure { namespace Device { namespace Health {

class PluginFactory;
class ProducerPluginInstance;
class ConsumerPluginInstance;
class LoggerPluginInstance;

class PluginManager
{
  public:
    void DiscoverPluginsInDirectory(std::string_view pluginDirectory);
    void CreateInstancesFromConfig(const std::vector<toml::value>& pluginsConfig);
    void RunLoop();
    void Stop();

    void SetQuantumDuration(std::chrono::milliseconds quantumDuration)
    {
        _quantumDuration = quantumDuration;
    }

  private:
    void DiscoverPluginsInFile(const char* filePath);

    using PluginNameToFactoryAndConfigMap = std::map<
        std::string,
        std::pair<std::shared_ptr<PluginFactory>, const toml::value*>,
        std::less<>>;

    PluginNameToFactoryAndConfigMap
    CreatePluginFactoryConfigMap(const std::vector<toml::value>& pluginsConfig);

    void CreatePluginInstances(
        const PluginNameToFactoryAndConfigMap& acceptedPluginFactories);

    void RunIteration();

  public:
    static constexpr std::string_view c_pluginRelativeDirectory =
        "azure-device-healthd/plugins";

  private:
    std::atomic_bool _stop = false;
    std::mutex _stopMutex;
    std::condition_variable _stopCondition;

    std::chrono::milliseconds _quantumDuration = std::chrono::minutes{ 1 };
    std::map<std::string, std::shared_ptr<PluginFactory>, std::less<>> _pluginFactories;

    std::vector<std::shared_ptr<ProducerPluginInstance>> _producerPlugins;
    std::vector<std::shared_ptr<ConsumerPluginInstance>> _consumerPlugins;
    std::vector<std::shared_ptr<LoggerPluginInstance>> _loggerPlugins;
};

}}}