// Copyright (c) Microsoft Corporation. All rights reserved.

#include "PluginModule.h"

#include <azure-device-util/ResultException.h>

#include "ErrorHelper.h"
#include "PluginFactory.h"
#include "Provider.h"

namespace Azure { namespace Device { namespace Health {

std::shared_ptr<PluginModule> PluginModule::Create(std::string_view path)
{
    return std::shared_ptr<PluginModule>{ new PluginModule{ path } };
}

PluginModule::PluginModule(std::string_view path) : _path(path)
{
    // FUTURE: Remove RTLD_NODELETE when we drop support for LTTNG < 2.11.
    // LTTNG < 2.11 has a bug where the LTTNG code still references memory in
    // dlclose'd modules, causing segfaults.

    _dlModule.reset(dlopen(_path.c_str(), RTLD_LAZY | RTLD_NODELETE));
    if (!_dlModule)
    {
        char* dlError = dlerror();

        TraceLoggingWrite(
            g_deviceHealthDaemonProvider,
            "PluginModule.Construct_DlOpenFailed",
            TraceLoggingValue(_path.c_str(), "Path"),
            TraceLoggingValue(dlError ? dlError : "", "DlErrorText"));

        THROW_HR(HealthError::E_FAILED_TO_LOAD_PLUGIN_SO);
    }

    _getPluginsFunc = reinterpret_cast<GetPluginsFuncType>(
        dlsym(_dlModule.get(), "get_adh_plugin_definitions"));
    if (!_getPluginsFunc)
    {
        char* dlError = dlerror();

        TraceLoggingWrite(
            g_deviceHealthDaemonProvider,
            "PluginModule.Construct_DlSymFailed",
            TraceLoggingValue(_path.c_str(), "Path"),
            TraceLoggingValue(dlError ? dlError : "", "DlErrorText"));

        THROW_HR(HealthError::E_FAILED_TO_LOAD_PLUGIN_EXPORTED_FUNCTION);
    }

    TraceLoggingWrite(
        g_deviceHealthDaemonProvider,
        "PluginModule.Construct_Succeeded",
        TraceLoggingValue(_path.c_str(), "Path"));
}

std::vector<std::shared_ptr<PluginFactory>> PluginModule::GetPluginFactories()
{
    std::vector<std::shared_ptr<PluginFactory>> definitions;

    auto pluginPointerList = _getPluginsFunc();
    if (!pluginPointerList || !*pluginPointerList)
    {
        TraceLoggingWrite(
            g_deviceHealthDaemonProvider,
            "PluginModule.GetPluginFactories_NullOrEmptyList",
            TraceLoggingValue(_path.c_str(), "Path"));
        return definitions;
    }

    uint32_t pluginCount = 0;
    for (auto entry = pluginPointerList; *entry != nullptr; ++entry, ++pluginCount)
    {
        definitions.emplace_back(
            PluginFactory::Create(shared_from_this(), *entry));
    }

    TraceLoggingWrite(
        g_deviceHealthDaemonProvider,
        "PluginModule.GetPluginFactories_CreatedFactories",
        TraceLoggingValue(_path.c_str(), "Path"),
        TraceLoggingValue(pluginCount, "PluginCount"));

    return definitions;
}

}}}