// Copyright (c) Microsoft Corporation. All rights reserved.

#pragma once

#include <cstddef>
#include <memory>
#include <mutex>
#include <string>
#include <string_view>

#include <jsonbuilder/JsonBuilder.h>

#include "CommonDefs.h"

namespace Azure { namespace Device { namespace Telemetry {

using namespace std::string_view_literals;

class AsimovSharedPartAState;
class IEventStore;
class LttngSession;

//
// Base class for all events that will eventually be translated into Asimov JSON
// events. It contains logic for adding all common part A fields and rendering
// out to the final JSON string.
//
class AsimovEvent
{
  public:
    static constexpr uint32_t k_defaultJsonRendererReserveSize = 2048;

  public:
    AsimovEvent() {}

    virtual ~AsimovEvent() {}

    std::string Render(AsimovSharedPartAState& partAState);

    std::string Render(
        AsimovSharedPartAState& partAState,
        IEventStore& eventStore,
        std::unique_lock<std::recursive_mutex>& eventStoreLock);

    std::string_view GetName() const;
    std::chrono::system_clock::time_point GetTimestamp() const;

    const jsonbuilder::JsonBuilder& GetJsonBuilder() const
    {
        return _jsonBuilder;
    }

    jsonbuilder::JsonBuilder& GetJsonBuilderUnsafe() { return _jsonBuilder; }

    TriggerPersistence GetPersistence() const { return _persistence; }
    TriggerLatency GetLatency() const { return _latency; }

    TriggerTelemetryCategory GetTelemetryCategory() const
    {
        return _telemetryCategory;
    }

    TelemetryPermission GetMinimumNeededTelemetryPermission() const
    {
        return _minTelemetryPermission;
    }

    TelemetryPermission GetTelemetryPermissionAtCreation() const
    {
        return _telemetryPermissionAtCreation;
    }
    void SetTelemetryPermissionAtCreation(TelemetryPermission telemetryPermission)
    {
        _telemetryPermissionAtCreation = telemetryPermission;
    }

    uint64_t GetKeywords() { return _keywords; }

    jsonbuilder::JsonConstIterator CreateOrGetExtensionObject(
        jsonbuilder::JsonConstIterator parent,
        std::string_view extensionName);

    void AddEpochAndSeq(
        AsimovSharedPartAState& partAState,
        jsonbuilder::JsonConstIterator dest);

    static std::unique_ptr<AsimovEvent> CreateFromLttngConsumer(
        jsonbuilder::JsonBuilder&& builder,
        TelemetryPermission minTelemetryPermission);

    static std::unique_ptr<AsimovEvent>
    CreateFromHwBinderConsumer(jsonbuilder::JsonBuilder&& builder);

    static std::unique_ptr<AsimovEvent>
    CreateFromLogcatConsumer(jsonbuilder::JsonBuilder&& builder);

  protected:
    jsonbuilder::JsonBuilder _jsonBuilder;
    TriggerPersistence _persistence = TriggerPersistence::Normal;
    TriggerLatency _latency = TriggerLatency::Normal;
    TelemetryPermission _minTelemetryPermission = TelemetryPermission::Optional;
    TelemetryPermission _telemetryPermissionAtCreation =
        TelemetryPermission::Optional;
    uint64_t _keywords = 0;

    // Currently, the default here is CriticalData for Mariner
    TriggerTelemetryCategory _telemetryCategory =
        TriggerTelemetryCategory::CriticalData;

  private:
    static void InitializePartA(
        jsonbuilder::JsonBuilder& jb,
        jsonbuilder::JsonConstIterator& extItr,
        jsonbuilder::JsonConstIterator& utcItr,
        jsonbuilder::JsonConstIterator& metadataItr);

    void AddTopLevelPartAFields(
        jsonbuilder::JsonConstIterator partAItr,
        std::string_view tenantToken);

    void AddExtFields(
        AsimovSharedPartAState& partAState,
        jsonbuilder::JsonConstIterator extItr,
        jsonbuilder::JsonConstIterator utcItr);

    void AddOsFields(
        AsimovSharedPartAState& partAState,
        jsonbuilder::JsonConstIterator dest);

    static std::byte CalculateLatencyByte(TriggerLatency latency);

    void SplicePartAIntoJsonRoot();

    void AddEventFlags(jsonbuilder::JsonConstIterator utcItr);

    static void ExtractTemporaryFieldsInJsonBuilder(
        jsonbuilder::JsonBuilder& builder,
        AsimovEvent& event,
        jsonbuilder::JsonConstIterator& dataItr,
        jsonbuilder::JsonConstIterator& extItr,
        jsonbuilder::JsonConstIterator& metadataItr);
};

}}}
