// Copyright (c) Microsoft Corporation. All rights reserved.

#pragma once

#include <map>
#include <memory>
#include <set>
#include <string_view>

#include <azure-device-util/ConversionFactors.h>
#include <azure-device-util/DeflateStream.h>

#include "AsimovConstantEventBuffer.h"

namespace Azure { namespace Device { namespace Telemetry {

class IAsimovEventBuffer
{
  public:
    virtual ~IAsimovEventBuffer(){};

    virtual bool AddEvent(std::string_view event) = 0;

    virtual AsimovConstantEventBuffer Commit() const = 0;

    virtual bool IsEmpty() const = 0;

    virtual void Clear() noexcept = 0;

    virtual void StoreTenantToken(std::string&& tenantToken) = 0;
};

class IAsimovEventBufferFactory
{
  public:
    virtual ~IAsimovEventBufferFactory(){};

    virtual std::shared_ptr<IAsimovEventBuffer>
    Create(size_t maxBytes, uint32_t maxEvents) = 0;
};

class AsimovEventBufferFactory : public IAsimovEventBufferFactory
{
  public:
    std::shared_ptr<IAsimovEventBuffer>
    Create(size_t maxBytes, uint32_t maxEvents) override;
};

//
// This class holds an array of JSON serialized Asimov events.  It can fill
// up by:
//  1. Hitting an overall byte limit
//  2. Hitting a limit on the number of total events
//  3. A timer occuring to snap the buffer for upload
//
class AsimovEventBuffer : public IAsimovEventBuffer
{
  public:
    AsimovEventBuffer(size_t maxBytes, uint32_t maxEvents);

    bool AddEvent(std::string_view event) override;

    AsimovConstantEventBuffer Commit() const override;

    bool IsEmpty() const override;

    void Clear() noexcept override;

    void StoreTenantToken(std::string&& tenantToken) override
    {
        _tenantTokenSet.emplace(std::move(tenantToken));
    }

    uint32_t GetNumEvents() const { return _numEvents; }

  public:
    static uint32_t GetNormalEventBufferMaxEvents()
    {
        return k_maxEventsPerNormalBufferDefault;
    }

    static uint32_t GetCriticalEventBufferMaxEvents()
    {
        return k_maxEventsPerCriticalBufferDefault;
    }

    static uint32_t GetEventBufferMaxBytes()
    {
        return k_maxBytesPerBufferDefault;
    }

  public:
    static constexpr uint32_t k_oneCollectorProtocol_eventCountPerBufferMax =
        1500;
    static constexpr uint32_t k_oneCollectorProtocol_bytesPerBufferMax =
        4 * ConversionFactors::mbToBytes;
    static constexpr uint32_t k_oneCollectorProtocol_bytesPerEventMax = 2097119;

    static constexpr uint32_t k_maxEventsPerNormalBufferDefault = 1000;
    static constexpr uint32_t k_maxEventsPerNormalBufferAbsoluteMax =
        1 * k_maxEventsPerNormalBufferDefault;
    static constexpr uint32_t k_maxEventsPerCriticalBufferDefault = 500;
    static constexpr uint32_t k_maxEventsPerCriticalBufferAbsoluteMax =
        1 * k_maxEventsPerCriticalBufferDefault;
    static constexpr uint32_t k_maxBytesPerBufferDefault =
        1 * ConversionFactors::mbToBytes;
    static constexpr uint32_t k_maxBytesPerBufferAbsoluteMax =
        2 * k_maxBytesPerBufferDefault;
    static constexpr uint32_t k_maxEventSizeInBytes =
        64 * ConversionFactors::kbToBytes;

    static_assert(
        k_maxEventsPerNormalBufferAbsoluteMax <=
            k_oneCollectorProtocol_eventCountPerBufferMax,
        "OneCollector requirement");
    static_assert(
        k_maxEventsPerCriticalBufferAbsoluteMax <=
            k_oneCollectorProtocol_eventCountPerBufferMax,
        "OneCollector requirement");
    static_assert(
        k_maxBytesPerBufferDefault <= k_oneCollectorProtocol_bytesPerBufferMax,
        "OneCollector requirement");
    static_assert(
        k_maxEventSizeInBytes <= k_oneCollectorProtocol_bytesPerEventMax,
        "OneCollector requirement");

  private:
    bool ShouldUseStreamingCompression(size_t newEventSize);
    bool IsAtEventCapacity() const;
    bool IsAtByteCapacity() const;

    std::string GenerateApiKeyHeader() const;

  private:
    static const std::string k_eventSeparator;

  private:
    const uint32_t _maxEvents;
    const size_t _maxBytes;

    uint32_t _numEvents;

    std::set<std::string> _tenantTokenSet;

    // Revelant structures while streaming mode is false
    std::string _json;

    // Relevant structures while streaming mode is true
    std::unique_ptr<Utils::DeflateStream> _deflateStream;
};

}}}
