// Copyright (c) Microsoft Corporation. All rights reserved.

#pragma once

#include <atomic>
#include <filesystem>
#include <mutex>
#include <string>
#include <string_view>

#include <azure-device-util/IConfig.h>

namespace Azure { namespace Device { namespace Telemetry {

using namespace std::string_view_literals;

//
// This class holds state shared among all Asimov events for population in Part
// A.
//
class AsimovSharedPartAState
{
  public:
    AsimovSharedPartAState(Utils::IConfig& config);
    void ReloadConfiguration(Utils::IConfig& config);

    std::string_view GetDeviceId() const;
    std::string_view GetTenantToken() const { return _tenantToken; }

    std::string_view GetDevicePlatform() const { return _devicePlatform; }
    std::string_view GetDeviceMake() const { return _deviceMake; }
    std::string_view GetDeviceModel() const { return _deviceModel; }
    std::string_view GetOsVersion() const { return _osReleaseData.Version; }
    std::string_view GetOsName() const { return _osReleaseData.Name; }
    std::string_view GetOsBuild() const { return _osReleaseData.Build; }

    std::string_view GetEpoch() const { return _epochString; }
    uint64_t GetEventCount() const { return _eventCount; }
    double GetDeviceSampleRate() const
    {
        return (_deviceSampleNumber * 100.0) / c_maxDeviceSampleNumber;
    }

    uint64_t IncrementEventCount() { return ++_eventCount; }

    bool IsTestMachine() const { return _isTestMachine; }
    bool IsMsftInternal() const { return _isMsftInternal; }

    struct OsReleaseData
    {
        std::string Name = "Linux";
        std::string Version = "Unknown";
        std::string Build = "Unknown";
    };

    void SetOsReleaseData(OsReleaseData&& data)
    {
        _osReleaseData = std::move(data);
    }

  public:
    static OsReleaseData ReadOsReleaseData();

  private:
    struct PersistedData
    {
        std::string LocalId;
        uint32_t LaunchCount = 0;
        uint64_t DeviceSampleNumber = 0;
    };

    static PersistedData LoadPersistedData();
    void SetMembersBasedOnLoadedData(const PersistedData& persistedData);

    static void SavePersistedData(const PersistedData& persistedData);

    static std::filesystem::path GetPersistedDataFilePath();

    void GenerateEpochString(uint32_t launchCount);

  private:
    static constexpr std::string_view c_localIdKey = "localId"sv;
    static constexpr std::string_view c_launchCountKey = "launchCount"sv;
    static constexpr std::string_view c_deviceSampleNumberKey =
        "deviceSampleNumber"sv;

    static constexpr uint64_t c_maxDeviceSampleNumber = 10'000'000;

  private:
    OsReleaseData _osReleaseData;

    std::string _devicePlatform;
    std::string _deviceMake;
    std::string _deviceModel;

    std::string _machineBasedDeviceId;

    std::string _asimovGeneratedDeviceId;
    std::string _tenantToken;

    uint64_t _deviceSampleNumber;

    uint32_t _launchCount = 0;
    std::string _epochString;

    uint64_t _eventCount = 0;

    bool _isTestMachine = false;
    bool _isMsftInternal = false;
};

}}}
