// Copyright (c) Microsoft Corporation. All rights reserved.

#pragma once

#include <filesystem>
#include <memory>
#include <mutex>
#include <string>
#include <string_view>
#include <thread>

#include <boost/asio/io_service.hpp>
#include <boost/asio/steady_timer.hpp>

#include "CommonDefs.h"

namespace Azure { namespace Device { namespace Telemetry {

struct HeartBeatCounters;
class IEventSerializer;

class IHeartBeat
{
  public:
    virtual ~IHeartBeat() {}

    virtual void Fire() = 0;

    virtual void RecordNetworkAttempt() = 0;
    virtual void RecordNetworkFailure(std::error_code networkResult) = 0;

    virtual void RecordHttpResponseFailure(uint32_t httpStatusCode) = 0;
    virtual void RecordHttpAcceptedAndRejectedEvents(
        uint32_t acceptedEvents,
        uint32_t rejectedEvents,
        uint32_t compressedBytesUploaded) = 0;

    virtual void
    RecordHttpRepeatedUploadFailureDroppedEvents(uint32_t eventsLost) = 0;

    virtual void RecordDbDropped(
        uint32_t eventCount,
        TriggerPersistence persistence,
        bool isBecauseFull) = 0;

    virtual void RecordDbReset(uint32_t eventStoreSize) = 0;
    virtual void RecordDbEnteringCriticalOverflow(uint32_t bytes) = 0;

    virtual void RecordDbEventPersisted() = 0;

    virtual void
    RecordDbEventSizeOffender(std::string_view eventSizeOffender) = 0;

    virtual void RecordContainerBufferFullDropped() = 0;

    virtual void RecordEventsForwardedToHost() = 0;

    virtual void RecordHostConnectionTimeouts() = 0;

    virtual void RecordHostConnectionErrors(std::error_code error) = 0;

    virtual void SetContainerHeartbeatFields(
        uint64_t HalRegisterReceiverCount,
        uint64_t halConnectionFailures,
        uint64_t halBufferFullDrops,
        uint64_t halBufferOtherDrops,
        uint64_t proxyConnectionFailures,
        uint64_t proxyConnectionSuccess,
        uint64_t proxyMissingCallbackDrops) = 0;
};

class HeartBeat : public IHeartBeat
{
  public:
    HeartBeat(std::string_view name, IEventSerializer& serializer);
    ~HeartBeat();

    void StartTimer();
    void StopTimer();

    void Fire() override;

    // IHeartBeat
    void RecordNetworkAttempt() override;
    void RecordNetworkFailure(std::error_code networkResult) override;

    void RecordHttpResponseFailure(uint32_t httpStatusCode) override;
    void RecordHttpAcceptedAndRejectedEvents(
        uint32_t acceptedEvents,
        uint32_t rejectedEvents,
        uint32_t compressedBytesUploaded) override;

    void
    RecordHttpRepeatedUploadFailureDroppedEvents(uint32_t eventsLost) override;

    void RecordDbDropped(
        uint32_t eventCount,
        TriggerPersistence persistence,
        bool isBecauseFull) override;

    void RecordDbReset(uint32_t eventStoreSize) override;
    void RecordDbEnteringCriticalOverflow(uint32_t bytes) override;

    void RecordDbEventPersisted() override;

    void RecordDbEventSizeOffender(std::string_view eventSizeOffender) override;

    void RecordContainerBufferFullDropped() override;
    void RecordEventsForwardedToHost() override;
    void RecordHostConnectionTimeouts() override;
    void RecordHostConnectionErrors(std::error_code error) override;

    void SetContainerHeartbeatFields(
        uint64_t HalRegisterReceiverCount,
        uint64_t halConnectionFailures,
        uint64_t halBufferFullDrops,
        uint64_t halBufferOtherDrops,
        uint64_t proxyConnectionFailures,
        uint64_t proxyConnectionSuccess,
        uint64_t proxyMissingCallbackDrops) override;

  private:
    void ArmTimer();

    bool LogSynthetic(
        const HeartBeatCounters& counters,
        std::chrono::system_clock::time_point& lastHeartbeatTimestamp);

    static HeartBeatCounters LoadPersistedData();
    void SetMembersBasedOnLoadedData(const HeartBeatCounters& persistedData);

    static void SavePersistedData(const HeartBeatCounters& persistedData);

    static std::filesystem::path GetPersistedDataFilePath();

  private:
    std::string _name;
    IEventSerializer& _serializer;

    std::recursive_mutex _countersMutex;
    std::unique_ptr<HeartBeatCounters> _counters;

    // Members needed for timer
    boost::asio::io_service _timerIoService;
    std::thread _timerThread;

    boost::asio::steady_timer _timer{ _timerIoService };
    static constexpr std::chrono::minutes c_timerPeriod{ 30 };
};

}}}
