// Copyright (c) Microsoft Corporation. All rights reserved.

#pragma once

#include <cstddef>
#include <map>
#include <memory>
#include <mutex>
#include <string>
#include <vector>

#include "CommonDefs.h"

namespace Azure { namespace Device { namespace Telemetry {

struct UploadData
{
    TriggerPersistence Persistence = TriggerPersistence::Normal;
    TriggerLatency Latency = TriggerLatency::Normal;
    std::string Payload;
};

struct ReadResult
{
    std::string ApiKeyHttpHeader;
    std::vector<std::byte> Payload;
    uint32_t SequenceNumber = 0;
    uint32_t EventCount = 0;
    uint32_t UploadFailCount = 0;
};

// Defines Interface of EventStore Type
class IEventStore
{
  public:
    virtual ~IEventStore(){};

    virtual std::unique_ptr<ReadResult>
    GetDataByLatency(TriggerLatency latency) = 0;

    virtual bool AddData(const UploadData& uploadData) = 0;

    virtual void
    DeleteData(uint32_t bufferSeqNum, bool commitTransaction = true) = 0;

    virtual void IncrementUploadFailCount(const ReadResult& readResult) = 0;

    virtual void FlushToFile() = 0;

    virtual void FlushPreCompressBuffers(TriggerLatency latency) = 0;

    virtual double GetPercentageFull() const = 0;

    virtual bool HasRemainingData(
        const TriggerPersistence* persistence,
        const TriggerLatency* latency) = 0;

    virtual bool ArePreCompressBuffersEmpty(TriggerLatency latency) = 0;

    virtual void StoreTenantToken(
        TriggerPersistence persistence,
        TriggerLatency latency,
        std::string&& tenantToken) = 0;

    virtual std::recursive_mutex& GetEventStoreMutex() const = 0;

    virtual uint32_t GetEventStoreSizeBytes() const = 0;

    virtual void SetEventStoreSizeBytes(uint32_t bytes) = 0;

  private:
    virtual std::vector<ReadResult> GetAndDeleteAllData() = 0;
};

}}}
