// Copyright (c) Microsoft Corporation. All rights reserved.

#include "Journal.hpp"
#include <optional>

#include <azure-device-util/ResultException.h>

namespace Azure { namespace Device { namespace Telemetry {

Journal::Journal(std::chrono::milliseconds pollInterval)
    : _pollInterval(pollInterval)
{
    InitJournal();
}

Journal::~Journal()
{
    sd_journal_close(sdj);
}

void Journal::StopCollecting()
{
    _stopCollecting = true;
}

std::optional<MessageEvent> Journal::GetNextEvent()
{
    // Number of read entries (0 or 1)
    // Might be negative in case of an error.
    int res = 0;
    int errCount = 0;
    do
    {
        // Try to advance read pointer by one entry
        // r will contain the number of entries read (0 or 1)
        res = sd_journal_next(sdj);
        // If r == 0, the pointer is at the end and didn't move
        if (res == 0)
        {
            WaitForNextEvent();  // Wait out JOURNAL_NOPs
            if (_stopCollecting)
            {
                // Break immediately after
                // break flag
                return {};
            }
        }
        else if (res < 0)
        {
            // Severe error, try to recover up to 5 times.
            if (errCount++ < 5)
            {
                continue;
            }
            else
            {
                throw std::runtime_error(
                    "Advancing the journal read pointer failed 5 times in a row.");
            }
        }
        // If res is not 1 (i.e. one entry received) go back and try again
    } while (res != 1);

    MessageEvent event{ LoadFields(), std::chrono::system_clock::now() };

    return event;
}

MessageFields Journal::LoadFields()
{
    const void* data;
    size_t length;
    MessageFields fields;
    SD_JOURNAL_FOREACH_DATA(sdj, data, length)
    {
        std::string line(static_cast<const char*>(data), length);

        size_t delimPos = line.find("=");
        if (delimPos == std::string::npos || delimPos == 0)
        {
            continue;
        }

        std::string key = line.substr(0, delimPos);
        std::string value = line.substr(delimPos + 1);
        fields.emplace(key, value);
    }

    return fields;
}

void Journal::WaitForNextEvent()
{
    auto timeout_usec =
        std::chrono::duration_cast<std::chrono::microseconds>(_pollInterval).count();
    int res;
    do
    {
        res = sd_journal_wait(sdj, timeout_usec);
    } while (res == SD_JOURNAL_NOP && !_stopCollecting);

    if (res < 0)
    {
        throw std::runtime_error("Waiting for journald entried failed.");
    }
}

void Journal::InitJournal()
{
    int res = 0;

    // Open journal connection
    res = sd_journal_open(&sdj, SD_JOURNAL_LOCAL_ONLY);
    FAIL_FAST_IF(res < 0);

    // Generate internal file descriptor
    res = sd_journal_get_fd(sdj);
    FAIL_FAST_IF(res < 0);

    // Go to the end of the journal
    res = sd_journal_seek_tail(sdj);
    FAIL_FAST_IF(res < 0);

    // Clear record queue
    res = sd_journal_previous(sdj);
    FAIL_FAST_IF(res < 0);

    // Add a very restrictive match so nothing
    sd_journal_add_match(sdj, "PRIORITY=0", 0);
}

void Journal::AddSource(std::string_view filter)
{
    // Insert an OR in the match list
    sd_journal_add_disjunction(sdj);
    // Insert the provided match as-is, return code
    if (int r = sd_journal_add_match(sdj, filter.data(), filter.size()) != 0)
    {
        throw std::runtime_error(
            "Could not add journald source \"" + std::string{ filter } +
            "\": " + strerror(-r));
    }
}

void Journal::AddSourceList(const SourceList& sourceList)
{
    sd_journal_flush_matches(sdj);
    sd_journal_add_match(sdj, "PRIORITY=0", 0);

    for (std::string_view filter : sourceList)
    {
        AddSource(filter);
    }
}

}}}