// Copyright (c) Microsoft Corporation. All rights reserved.

#include "JournalDConsumer.hpp"

#include <iostream>
#include <vector>

#include <azure-device-util/FileSystem.h>
#include <azure-device-util/String.h>

#include "ConsumerConfig.hpp"
#include "Router.hpp"

using namespace Azure::Device::Policy;

namespace Azure { namespace Device { namespace Telemetry {

JournalDConsumer::JournalDConsumer(
    IEventSerializer& serializer,
    IPolicyManager& policyManager,
    std::chrono::milliseconds _pollInterval)
    : _serializer(serializer)
    , _policyManager(policyManager)
    , _journal(_pollInterval)
{}

void JournalDConsumer::StartConsuming()
{
    _stopCollecting = false;
    _consumerThread = std::thread(&JournalDConsumer::ConsumeTask, this);
}

void JournalDConsumer::ConsumeTask()
{
    std::filesystem::path journaldCollectorConfigPath =
        Utils::FileSystem::GetConfDirectory();
    journaldCollectorConfigPath /= "azure-device-telemetryd/events.conf.d";

    // Check if path exists
    if (!std::filesystem::exists(journaldCollectorConfigPath.c_str()))
    {
        // Nothing to collect
        TraceLoggingWrite(g_tlgProv, "JournalD_NoConfigDirectory");
        return;
    }

    JournalDConsumerConfig config(_policyManager);

    // Extract event and config data from files in the directory
    for (const auto& entry :
         std::filesystem::directory_iterator(journaldCollectorConfigPath.c_str()))
    {
        try
        {
            const auto data = toml::parse(entry.path().string());
            config.AddConfigFile(data);
        }
        catch (std::exception& ex)
        {
            std::cout << "JournalDConsumer: Failed to parse file "
                      << entry.path() << std::endl
                      << ex.what() << std::endl;

            TraceLoggingWrite(
                g_tlgProv,
                "JournalD_ParseFailed",
                TraceLoggingString(entry.path().c_str(), "file"),
                TraceLoggingString(ex.what(), "what"));
        }
    }

    if (!config.HasCollectionConfigured())
    {
        TraceLoggingWrite(g_tlgProv, "JournalD_NothingToCollect");
        return;
    }

    // Apply filters to journal
    try
    {
        _journal.AddSourceList(config.GetJournalSourceList());
    }
    catch (std::exception& ex)
    {
        // Severe error. Shouldn't happen normaly as the creation of journald
        // match strings is abstracted from the user. Only an extremely
        // malformed source would trigger this.
        std::cout << "JournalDConsumer: Could not add journal source! "
                  << std::endl
                  << ex.what() << std::endl;

        TraceLoggingWrite(
            g_tlgProv,
            "JournalD_JournalDSourceFailed",
            TraceLoggingString(ex.what(), "what"));

        FAIL_FAST();
    }

    // Create a router with the serializer and settings
    Router router(_serializer, config);

    std::cout << "JournalDConsumer: Starting thread!" << std::endl;
    while (!_stopCollecting)
    {
        auto optEvent = _journal.GetNextEvent();
        if (optEvent)
        {
            router.RouteMessageEvent(*optEvent);
        }
    }
    std::cout << "JournalDConsumer: Finished thread!" << std::endl;
}

void JournalDConsumer::ReloadConfiguration()
{
    StopConsuming();
    WaitForCompletion();
    StartConsuming();
}

void JournalDConsumer::StopConsuming()
{
    _stopCollecting = true;
    _journal.StopCollecting();
}

void JournalDConsumer::WaitForCompletion()
{
    _consumerThread.join();
}

}}}