// Copyright (c) Microsoft Corporation. All rights reserved.

#include "JournalDConsumerConfig.hpp"
#include <optional>

#include <azure-device-util/AndroidLogHelper.h>
#include <azure-device-util/Permission.h>

using namespace Azure::Device::Policy;

namespace Azure { namespace Device { namespace Telemetry {

JournalDConsumerConfig::JournalDConsumerConfig(IPolicyManager& policyManager)
    : _consumerConfig(policyManager)
{}

bool JournalDConsumerConfig::ReadJournalFilter(
    const toml::value& data,
    std::string fieldName,
    size_t maxChars)
{
    if (!data.contains(fieldName))
    {
        return false;
    }

    const auto values = toml::find<std::vector<std::string>>(data, fieldName);

    if (values.empty())
    {
        return false;
    }

    for (const auto& value : values)
    {
        _journalSources.emplace(fieldName + "=" + value.substr(0, maxChars));
    }

    return true;
}

void JournalDConsumerConfig::AddConfigFile(const toml::value& data)
{
    const auto journaldConfig = toml::find<toml::value>(data, "JournaldConfig");

    bool sourceRegistered = false;
    sourceRegistered |= ReadJournalFilter(journaldConfig, "SYSLOG_IDENTIFIER");
    sourceRegistered |= ReadJournalFilter(journaldConfig, "_SYSTEMD_UNIT");
    sourceRegistered |= ReadJournalFilter(journaldConfig, "_COMM", c_CommMaxSize);
    sourceRegistered |= ReadJournalFilter(journaldConfig, "_EXE");

    if (!sourceRegistered)
    {
        throw JournalDConfigurationException{
            "File must include at least one source in [JournaldConfig]"
        };
    }

    _consumerConfig.AddConfigFile(data);
}

bool JournalDConsumerConfig::HasCollectionConfigured()
{
    return !_journalSources.empty() && _consumerConfig.HasCollectionConfigured();
}

const SourceList& JournalDConsumerConfig::GetJournalSourceList() const
{
    return _journalSources;
}

const StaticParserRuleSet& JournalDConsumerConfig::GetStaticParserRuleSet() const
{
    return _consumerConfig.GetStaticParserRuleSet();
};

void JournalDConsumerConfig::Reset()
{
    _consumerConfig.Reset();
    _journalSources.clear();
}

std::optional<TelemetryPermission>
JournalDConsumerConfig::GetTelemetryPermissionFromJsonBuilder(
    jsonbuilder::JsonBuilder& jsonBuilder)
{
    return _consumerConfig.GetTelemetryPermissionFromJsonBuilder(jsonBuilder);
}

}}}