// Copyright (c) Microsoft Corporation. All rights reserved.

#include "EventConfigurationParser.h"

#include <azure-device-util/FileSystem.h>
#include <azure-device-util/ResultException.h>
#include <azure-device-util/String.h>

#include "TlgProvider.h"

namespace Azure { namespace Device { namespace Telemetry {

EventConfigurationParser::EventConfigurationParser(std::string_view confFolder)
    : _confFolder(confFolder)
{}

/*++
    Call ParseFile for every .conf file in _confFolder, calling the passed
    callback on completion.

Arguments:
    onFileParseCallback - File parse completion callback to pass to ParseFile
--*/
void EventConfigurationParser::Parse(const ConfFileParseCallback& onFileParseCallback)
{
    Utils::FileSystem::DoForEachFileWithPattern(
        _confFolder,
        ".*\\.conf",
        [this, &onFileParseCallback](const std::filesystem::path& fullPath) {
            ParseFile(onFileParseCallback, fullPath);
        });
}

/*++
    Call ParseFile for every .conf file in _confFolder, calling the passed
    callback on completion.

Arguments:
    onFileParseCallback - File parse completion callback to pass to ParseFile
    fullPath            - Path to file to parse
--*/
void EventConfigurationParser::ParseFile(
    const ConfFileParseCallback& onFileParseCallback,
    const std::filesystem::path& fullPath)
{
    TraceLoggingWrite(
        g_tlgProv,
        "EventConfigurationParser_ParseFile_Start",
        TraceLoggingValue(fullPath.c_str(), "FullPath"));

    std::string fileContents;
    THROW_IF_FAILED(Utils::FileSystem::ReadUTF8StringFromFile(
        fullPath.c_str(), fileContents));

    TraceLoggingWrite(
        g_tlgProv,
        "EventConfigurationParser_ParseFile_Contents",
        TraceLoggingValue(fullPath.c_str(), "FullPath"),
        TraceLoggingValue(fileContents.c_str(), "Contents"));

    std::vector<std::string> fileLines = Utils::String::Split(fileContents, "\n");

    ListeningRuleSet listeningRuleSet;
    TelemetryPermission currentTelemetryPermissionState =
        TelemetryPermission::Optional;

    for (const std::string& line : fileLines)
    {
        if (line == "Required:")
        {
            currentTelemetryPermissionState = TelemetryPermission::Required;
        }
        else if (line == "Optional:")
        {
            currentTelemetryPermissionState = TelemetryPermission::Optional;
        }
        else if (line.find('*') != std::string::npos)
        {
            TraceLoggingWrite(
                g_tlgProv,
                "EventConfigurationParser_ParseFile_RejectedWildcardLine",
                TraceLoggingValue(fullPath.c_str(), "FullPath"),
                TraceLoggingValue(line.c_str(), "Line"));
        }
        else if (line.find(':') == std::string::npos)
        {
            TraceLoggingWrite(
                g_tlgProv,
                "EventConfigurationParser_ParseFile_RejectedMissingProviderEventSeparator",
                TraceLoggingValue(fullPath.c_str(), "FullPath"),
                TraceLoggingValue(line.c_str(), "Line"));
        }
        else if (!line.empty())
        {
            TraceLoggingWrite(
                g_tlgProv,
                "EventConfigurationParser_ParseFile_AcceptingLine",
                TraceLoggingValue(fullPath.c_str(), "FullPath"),
                TraceLoggingValue(line.c_str(), "Line"));

            listeningRuleSet[currentTelemetryPermissionState].insert(line);
        }
    }

    onFileParseCallback(fullPath.c_str(), std::move(listeningRuleSet));
}

}}}
