// Copyright (c) Microsoft Corporation. All rights reserved.

#include "LttngEventConsumer.h"
#include <optional>

#include <unistd.h>

#include <azure-device-util/FileSystem.h>
#include <azure-device-util/Json.h>
#include <azure-device-util/ResultException.h>
#include <lttng-consume/LttngConsumer.h>
#include <nlohmann/json.hpp>

#include "EventConfigurationParser.h"
#include "IPolicyManager.h"
#include "LttngApi.h"
#include "LttngSession.h"
#include "TelemetryError.h"
#include "TlgProvider.h"

using namespace Azure::Device::Policy;

namespace Utils = Azure::Device::Utils;
namespace Azure { namespace Device { namespace Telemetry {

LttngEventConsumer::LttngEventConsumer(
    IEventSerializer& serializer,
    IPolicyManager& policyManager,
    std::chrono::milliseconds pollInterval)
    : _serializer(serializer)
    , _policyManager(policyManager)
    , _pollInterval(pollInterval)
{}

LttngEventConsumer::~LttngEventConsumer() = default;

void LttngEventConsumer::StartConsuming()
{
    constexpr std::string_view c_sessionName = "azure-device-telemetry-"
                                               "session"sv;

    constexpr std::chrono::milliseconds c_singleStepWait{ 500 };

    char hostnameBuf[256];
    gethostname(hostnameBuf, 256);

    auto lttngApi = std::make_unique<LttngApi>();

    _session = std::make_unique<LttngSession>(
        std::move(lttngApi), c_sessionName, hostnameBuf, c_singleStepWait, 10);

    std::filesystem::path telemetryEventConfigurationDir =
        Utils::FileSystem::GetConfDirectory();
    telemetryEventConfigurationDir /= "azure-device-telemetryd";

    EventConfigurationParser eventConfigParser{
        telemetryEventConfigurationDir.string()
    };

    _session->SetTelemetryPermission(_policyManager.GetTelemetryPermission());

    eventConfigParser.Parse(
        [this](std::string_view fullPath, ListeningRuleSet&& listeningRuleSet) {
            _session->AddRuleSet(fullPath, std::move(listeningRuleSet));
        });

    _session->EnableDotNetEventSourceEvents();
    _session->Start();

    std::string connectionString =
        LttngSession::MakeConnectionString(c_sessionName, hostnameBuf);

    _consumer = std::make_unique<LttngConsume::LttngConsumer>(
        connectionString, _pollInterval);

    _consumerThread = std::thread(&LttngEventConsumer::ConsumeTask, this);
}

void LttngEventConsumer::ReloadConfiguration()
{
    _session->SetTelemetryPermission(_policyManager.GetTelemetryPermission());
}

void LttngEventConsumer::StopConsuming()
{
    _consumer->StopConsuming();
}

void LttngEventConsumer::WaitForCompletion()
{
    _consumerThread.join();
}

/*++
    Thread function to block on receiving events from lttng-consume, and
    handling them.
--*/
void LttngEventConsumer::ConsumeTask()
{
    try
    {
        // This will block until a SIGTERM handler calls StopConsuming(), or
        // the consumer exits for some other reason, like someone killing
        // the session or it encountering an error.
        _consumer->StartConsuming([&](jsonbuilder::JsonBuilder&& builder) {
            try
            {
                OnNewJsonBuilder(_serializer, *_session, std::move(builder));
            }
            CATCH_LOG();
        });
    }
    CATCH_LOG();
}

/*++
    Convert a jsonbuilder received from lttng-consume into an AsimovEvent, then
    serialize it if it's allowed at the current permission level.

    This will perform additional fixes to EventSource-style events to make
    them more Common Schema friendly.

Arguments:
    builder - jsonbuilder structure from lttng-consume
--*/
void LttngEventConsumer::OnNewJsonBuilder(
    IEventSerializer& serializer,
    LttngSession& session,
    jsonbuilder::JsonBuilder&& builder)
{
    std::string lttngName{ GetCleanedUpLttngName(builder) };
    if (lttngName == LttngSession::c_eventSourceLttngName)
    {
        lttngName = RemoveEventSourceEnvelope(builder);
    }

    std::optional<TelemetryPermission> optMinTelemetryPermission =
        session.GetMinimumRequiredTelemetryPermissionForEvent(lttngName);
    if (!optMinTelemetryPermission)
    {
        TraceLoggingWrite(
            g_tlgProv,
            "LttngEventConsumer_EventDiscarded_NoPermissionDefinitionFound",
            TraceLoggingStringView(lttngName, "EventName"));

        return;
    }

    std::unique_ptr<AsimovEvent> event = AsimovEvent::CreateFromLttngConsumer(
        std::move(builder), *optMinTelemetryPermission);

    serializer.WriteEvent(std::move(event));
}

/*++
    Lookup the lttngName from the jsonBuilder.  This is disctinct from 'name',
    as it is still using ':' as the separator between provider name and event
    name.

    This will also chop off any characters including and after the first ';'.
    This is to solve the problem where TraceLogging inserts ';k;' style suffixes
    to the event name as pseudo-keywords for filtering.
--*/
std::string_view
LttngEventConsumer::GetCleanedUpLttngName(const jsonbuilder::JsonBuilder& builder)
{
    auto lttngNameItr = builder.find("metadata", "lttngName");
    FAIL_FAST_IF(lttngNameItr == builder.end());

    std::string_view lttngNameView;
    FAIL_FAST_IF(!lttngNameItr->ConvertTo(lttngNameView));

    // Chop off any trailing keyword information inserted into event name
    // by TraceLogging
    if (auto semiColonIndex = lttngNameView.find(';');
        semiColonIndex != std::string::npos)
    {
        lttngNameView = lttngNameView.substr(0, semiColonIndex);
    }

    return lttngNameView;
}

/*++
    Mutates builder to:
     1. Change DotNETRuntime:EventSource name to MyEventSource:MyEventName
     2. Use the Payload field's escaped JSON as the real Part C for the event.

    On entry, 'data' for DotNETRuntime:EventSource has these fields:
     * EventID           - ignored
     * EventName         - event name
     * EventSourceName   - provider name
     * Payload           - escaped JSON object containing the real fields

    In the end, we only want the embedded fields in Payload to be in 'data'.

Arguments:
    builder - jsonbuilder structure from lttng-consume, heavily mutated during
                this method

Return Value:
    LTTNG colon-separated new name of the event
--*/
std::string
LttngEventConsumer::RemoveEventSourceEnvelope(jsonbuilder::JsonBuilder& builder)
{
    auto dataItr = builder.find("data");

    auto getRequiredString = [dataItr, &builder](std::string_view fieldName) {
        auto itr = builder.find(dataItr, fieldName);

        THROW_HR_IF(
            TelemetryError::E_ILLEGAL_DOT_NET_EVENT_SOURCE_EVENT,
            itr == builder.end());

        std::string_view stringView;
        THROW_HR_IF(
            TelemetryError::E_ILLEGAL_DOT_NET_EVENT_SOURCE_EVENT,
            !itr->ConvertTo(stringView));

        return stringView;
    };

    // Replace name based on EventSourceName/EventName fields
    std::string_view eventSourceName = getRequiredString("EventSourceName");
    std::string_view eventName = getRequiredString("EventName");

    auto buildEventName = [](std::string_view providerName,
                             std::string_view eventName,
                             char separator) {
        std::string name;
        name += providerName;
        name += separator;
        name += eventName;
        return name;
    };

    std::string asimovEventName = buildEventName(eventSourceName, eventName, '.');
    std::string lttngName = buildEventName(eventSourceName, eventName, ':');

    auto nameItr = builder.find("name");
    THROW_HR_IF(
        TelemetryError::E_ILLEGAL_DOT_NET_EVENT_SOURCE_EVENT,
        nameItr == builder.end());

    builder.erase(nameItr);
    builder.push_back(builder.root(), "name", asimovEventName);

    // Get Payload from the original structure
    std::string payload{ getRequiredString("Payload") };

    // Blow away all data section fields and insert anything parsed from Payload
    // into a clean new data object
    builder.erase(dataItr);
    dataItr = builder.push_back(builder.root(), "data", jsonbuilder::JsonObject);

    if (!payload.empty())
    {
        AddParsedPayloadToData(builder, dataItr, payload);
    }

    return lttngName;
}

/*++
    Parses bizarrely-escaped JSON string from DotNETRuntime:EventSource's
    Payload field.

    Then adds those elements into the fresh new 'data' section for this event.

Arguments:
    builder     - jsonbuilder to add data to
    dataItr     - iterator to 'data' to insert at
    payloadStr  - escapedJson string form EventSource Payload field

Return Value:
    Legal, parseable JSON string
--*/
void LttngEventConsumer::AddParsedPayloadToData(
    jsonbuilder::JsonBuilder& builder,
    jsonbuilder::JsonBuilder::iterator dataItr,
    std::string_view payloadStr)
{
    std::string unescapedJson = UnescapeJson(payloadStr);
    nlohmann::json parsedPayload = nlohmann::json::parse(unescapedJson);

    for (const auto& [key, value] : parsedPayload.items())
    {
        Utils::Json::InsertJsonValueIntoJsonBuilder(builder, dataItr, key, value);
    }
}

/*++
    Removes a level of escaping from the supplied string.

Arguments:
    escapedJson - JSON string with an extra level of JSON escaping

Return Value:
    Legal, parseable JSON string
--*/
std::string LttngEventConsumer::UnescapeJson(std::string_view escapedJson)
{
    std::string output;
    output.reserve(escapedJson.size());

    bool escaping = false;
    for (char c : escapedJson)
    {
        if (escaping)
        {
            switch (c)
            {
            case 'b':
                output += '\b';
                break;
            case 'f':
                output += '\f';
                break;
            case 'n':
                output += '\n';
                break;
            case 'r':
                output += '\r';
                break;
            case 't':
                output += '\t';
                break;
            default:  // This covers backslash and quote as well
                output += c;
                break;
            }

            escaping = false;
        }
        else
        {
            if (c == '\\')
            {
                escaping = true;
            }
            else
            {
                output += c;
            }
        }
    }

    return output;
}
}}}