// Copyright (c) Microsoft Corporation. All rights reserved.

#include "LttngSession.h"

#include <cstring>
#include <thread>
#include <optional>
#include <unistd.h>

#include <azure-device-util/ErrorHelper.h>
#include <azure-device-util/Process.h>
#include <azure-device-util/ResultException.h>
#include <lttng/lttng-error.h>
#include <lttng/lttng.h>
#include <lttng/session.h>

#include "LttngApi.h"
#include "LttngSessionLogging.h"
#include "TelemetryError.h"

namespace Azure { namespace Device { namespace Telemetry {

/*++
    This creates the connection string for a live LTTNG session running
    on the local machine.

Arguments:
    sessionName - Name of LTTNG session to connect to

Return Value:
    Formatted net:// string to pass to LTTNG APIs
--*/
std::string LttngSession::MakeConnectionString(
    std::string_view sessionName,
    std::string_view hostName)
{
    std::string result = "net://localhost/host/";

    result += hostName;
    result += "/";
    result.append(sessionName.data(), sessionName.size());

    return result;
}

/*++
    Create a fresh session and enable necessary context information.

Arguments:
    sessionName - Name of session to manage
--*/
LttngSession::LttngSession(
    std::unique_ptr<ILttngApi> lttngApi,
    std::string_view sessionName,
    std::string_view hostName,
    std::chrono::milliseconds retryInterval,
    uint32_t retriesAllowed)
    : _lttngApi(std::move(lttngApi)), _sessionName(sessionName)
{
    const char* name = _sessionName.c_str();
    std::string connectionString = MakeConnectionString(sessionName, hostName);

    // Try to destroy the session name just in case, to ensure it's available
    int destroyErr = _lttngApi->DestroySession(name);
    LttngSessionLogging::Create_DestroyName(name, destroyErr);

    // First, try to create the session directly.
    int lttngErr =
        _lttngApi->CreateSessionLive(name, connectionString.c_str(), 1000000);
    if (lttngErr < 0)
    {
        LttngSessionLogging::Create_FirstAttemptFailed(
            name, connectionString, lttngErr);

        uint32_t retries = 0;
        do
        {
            std::cout << "Backing off due to LTTNG start error: " << lttngErr
                      << std::endl;
            retries++;
            std::this_thread::sleep_for(retryInterval);

            // Then try to create the session again.
            lttngErr = _lttngApi->CreateSessionLive(
                name, connectionString.c_str(), 1000000);

            LttngSessionLogging::Create_Retry(name, connectionString, lttngErr);

        } while (lttngErr == -LTTNG_ERR_NO_SESSIOND && retries < retriesAllowed);

        if (lttngErr < 0)
        {
            if (lttngErr == -LTTNG_ERR_NO_SESSIOND)
            {
                std::cout
                    << "Could not connect to lttng-sessiond.  Was it started before running azure-device-telemetryd?"
                    << std::endl;
            }

            LttngSessionLogging::Create_TotalFailure(
                name, connectionString, lttngErr);

            THROW_HR(make_lttng_error_code(lttngErr));
        }
    }

    lttng_domain domain = lttng_domain();
    domain.type = LTTNG_DOMAIN_UST;
    domain.buf_type = LTTNG_BUFFER_PER_UID;

    lttng_handle* handle = _lttngApi->CreateHandle(name, &domain);
    if (!handle)
    {
        LttngSessionLogging::Create_HandleFailed(name);

        THROW_HR(make_error_code(std::errc::bad_file_descriptor));
    }

    _lttngHandle = handle;

    constexpr lttng_event_context_type contextsToAdd[] = {
        LTTNG_EVENT_CONTEXT_PROCNAME, LTTNG_EVENT_CONTEXT_VPID
    };

    for (auto contextType : contextsToAdd)
    {
        lttng_event_context context = lttng_event_context();
        context.ctx = contextType;
        int contextErr =
            _lttngApi->AddContext(handle, &context, nullptr, nullptr);
        if (contextErr < 0)
        {
            if (contextErr == -LTTNG_ERR_RELAYD_CONNECT_FAIL)
            {
                std::cout
                    << "Could not connect to lttng-relayd.  Was it started before running azure-device-telemetryd?"
                    << std::endl;
            }

            LttngSessionLogging::Create_AddContextFailed(
                name, static_cast<int32_t>(contextType), contextErr);

            THROW_HR(make_lttng_error_code(contextErr));
        }
    }

    lttng_channel* channel = _lttngApi->CreateChannel();

    static_assert(
        std::extent<decltype(channel->name)>::value > c_channelName.size());
    memcpy(channel->name, c_channelName.data(), c_channelName.size());

    channel->attr.subbuf_size = 256 * 1024;
    channel->attr.num_subbuf = 4;

    channel->attr.tracefile_size = channel->attr.subbuf_size;
    channel->attr.tracefile_count = 2;

    int enableChannelError = _lttngApi->EnableChannel(_lttngHandle, channel);
    if (enableChannelError < 0)
    {
        LttngSessionLogging::Create_EnableChannelFailed(name, enableChannelError);

        THROW_HR(make_lttng_error_code(enableChannelError));
    }
}

/*++
    Enables DotNETRuntime:EventSource static event that acts as a wrapper for
    all usage of EventSource in dotnet client applicaitons.

    LttngEventConsumer has special logic to 'unwrap' this envelope event and
    create dynamically named events like TraceLogging produces.
--*/
void LttngSession::EnableDotNetEventSourceEvents()
{
    EnableEvents({ std::string{ c_eventSourceLttngName } });
}

/*++
    Loop over rules and enable those that are relevant to current
TelemetryPermission. Also saves the rules with the fileName so rules can be
re-evaluated if TelemetryPermission changes at runtime.

Arguments:
    fileName    - Name of file rules came from
    ruleSet     - Set of event enablement strings and TelemetryPermission they
apply to
--*/
void LttngSession::AddRuleSet(std::string_view fileName, ListeningRuleSet&& ruleSet)
{
    // TODO: Use try_emplace, test result.  Or do a find first.  If find turns
    // up something, then we need to calculate the diff and call disable_event
    // for removed things, leave alone the intersection, and add the new things
    auto itrInsertedPair =
        _fileNameToRuleSets.emplace(fileName, std::move(ruleSet));
    if (!itrInsertedPair.second)
    {
        THROW_HR(make_error_code(std::errc::invalid_argument));
    }

    LttngSessionLogging::AddRuleSet(fileName);

    for (const auto& listeningRule : itrInsertedPair.first->second)
    {
        const auto& telemetryPermission = listeningRule.first;
        const auto& rules = listeningRule.second;

        if (telemetryPermission <= _telemetryPermission)
        {
            EnableEvents(rules);
        }
    }
}

void LttngSession::SetTelemetryPermission(TelemetryPermission telemetryPermission)
{
    TelemetryPermission oldPermission = _telemetryPermission;
    _telemetryPermission = telemetryPermission;

    // Nothing to do if there was no change
    if (_telemetryPermission == oldPermission)
    {
        return;
    }

    for (const auto& [fileName, listeningRuleSet] : _fileNameToRuleSets)
    {
        for (const auto& [rulePermission, rules] : listeningRuleSet)
        {
            // Previously denied rule is now allowed, do an enable
            if (oldPermission < rulePermission &&
                rulePermission <= _telemetryPermission)
            {
                EnableEvents(rules);
            }
            // Previously allowed rule is now denied, do a disable
            else if (_telemetryPermission < rulePermission && rulePermission <= oldPermission)
            {
                DisableEvents(rules);
            }
            // No change in whether the rule is allowed, leave it alone
        }
    }
}

/*++
    Start LTTNG session.
--*/
void LttngSession::Start()
{
    int lttngErr = _lttngApi->StartTracing(_sessionName.c_str());

    LttngSessionLogging::Start(_sessionName, lttngErr);

    if (lttngErr < 0)
    {
        THROW_HR(make_lttng_error_code(lttngErr));
    }
}

/*++
    Stop LTTNG session.
--*/
void LttngSession::Stop()
{
    int lttngErr = _lttngApi->StopTracing(_sessionName.c_str());

    LttngSessionLogging::Stop(_sessionName, lttngErr);

    if (lttngErr < 0)
    {
        THROW_HR(make_lttng_error_code(lttngErr));
    }
}

/*++
    Determines the minimum TelemetryPermission that needs to be enabled
    for this event to be allowed.

Arguments:
    eventName - Name of event to turn on

Return Value:
    Required   - Allowed at Required+
    Optional   - Allowed at Optional only
    Empty      - No rule found
--*/
std::optional<TelemetryPermission>
LttngSession::GetMinimumRequiredTelemetryPermissionForEvent(std::string_view eventName)
{
    for (const auto& fileNameRuleSetPair : _fileNameToRuleSets)
    {
        for (const auto& telemetryPermissionRuleSetPair :
             fileNameRuleSetPair.second)
        {
            if (telemetryPermissionRuleSetPair.second.find(eventName) !=
                telemetryPermissionRuleSetPair.second.end())
            {
                return telemetryPermissionRuleSetPair.first;
            }
        }
    }

    return {};
}

/*++
    Loop over event enablement rules turning each of them on in the
    current session.

Arguments:
    eventsToEnable - Set of event rule strings to turn on
--*/
void LttngSession::EnableEvents(
    const std::set<std::string, std::less<>>& eventsToEnable)
{
    std::string channelName{ c_channelName };
    for (const auto& eventRule : eventsToEnable)
    {
        lttng_event event = lttng_event();
        event.type = LTTNG_EVENT_TRACEPOINT;

        // Need one extra space for trailing * we will tack on to deal with
        // TraceLogging adding ';k;' type keyword information to the event of
        // all event names.
        if (eventRule.size() >= std::extent<decltype(event.name)>::value - 1)
        {
            LOG_HR(TelemetryError::E_EVENT_RULE_TOO_LONG);
            continue;
        }

        strcpy(event.name, eventRule.c_str());  // NOLINT
        strcat(event.name, "*");                // NOLINT

        int lttngErr =
            _lttngApi->EnableEvent(_lttngHandle, &event, channelName.c_str());

        LttngSessionLogging::EventEnabled(
            _sessionName,
            reinterpret_cast<size_t>(_lttngHandle),
            event.name,
            lttngErr);

        if (lttngErr < 0)
        {
            LOG_HR(make_lttng_error_code(lttngErr));
        }
    }
}

void LttngSession::DisableEvents(
    const std::set<std::string, std::less<>>& eventsToDisable)
{
    std::string channelName{ c_channelName };
    for (const auto& eventRule : eventsToDisable)
    {
        std::string rule = eventRule;
        rule += "*";

        int lttngErr = _lttngApi->DisableEvent(
            _lttngHandle, rule.c_str(), channelName.c_str());

        LttngSessionLogging::EventDisabled(
            _sessionName, reinterpret_cast<size_t>(_lttngHandle), rule, lttngErr);

        if (lttngErr < 0)
        {
            LOG_HR(make_lttng_error_code(lttngErr));
        }
    }
}

LttngSession::~LttngSession()
{
    if (_lttngHandle != nullptr)
    {
        _lttngApi->DestroyHandle(_lttngHandle);
    }

    if (!_sessionName.empty())
    {
        _lttngApi->DestroySession(_sessionName.c_str());
    }
}

}}}