// Copyright (c) Microsoft Corporation. All rights reserved.

#include "ConsumerConfig.hpp"
#include <optional>

#include <azure-device-util/Permission.h>

using namespace Azure::Device::Policy;

namespace Azure { namespace Device { namespace Telemetry {

ConsumerConfig::ConsumerConfig(IPolicyManager& policyManager)
    : _policyManager(policyManager)
{}

bool ConsumerConfig::RegisterEventFromToml(
    const toml::value& eventDefinition,
    EventType type)
{
    std::string name;
    switch (type)
    {
    case EventType::Tagged:
        name = toml::find<std::string>(eventDefinition, "name");
        break;

    case EventType::Static:
        name = toml::find<std::string>(eventDefinition, "provider") + ":" +
            toml::find<std::string>(eventDefinition, "event");
        break;
    }

    // Look for level setting and convert to TelemetryPermission.
    // If not present assume Optional.
    TelemetryPermission telemetryPermission = eventDefinition.contains("level") ?
        Utils::Permission::ConvertStringToTelemetryPermission(
            toml::find<std::string_view>(eventDefinition, "level")) :
        TelemetryPermission::Optional;

    if (telemetryPermission <= _policyManager.GetTelemetryPermission())
    {
        _eventList.emplace(name, telemetryPermission);
        return true;
    }
    return false;
}

void ConsumerConfig::AddConfigFile(const toml::value& data)
{
    if (data.contains("StaticEvents"))
    {
        const auto staticEvents =
            toml::find<std::vector<toml::value>>(data, "StaticEvents");
        for (const auto& table : staticEvents)
        {
            // If event is registered successufully, meaning the ADHS's
            // telemetry permission allows for this event to be sent, add it to
            // the static rule list. Otherwise don't even look for it.
            if (RegisterEventFromToml(table, EventType::Static))
            {
                _staticRules.push_back(StaticParserRule::FromTomlTable(table));
            }
        }
    }

    if (data.contains("TaggedEvents"))
    {
        const auto taggedEvents =
            toml::find<std::vector<toml::value>>(data, "TaggedEvents");

        for (const auto& table : taggedEvents)
        {
            RegisterEventFromToml(table, EventType::Tagged);
        }
    }
}

bool ConsumerConfig::HasCollectionConfigured()
{
    return !_eventList.empty();
}

const StaticParserRuleSet& ConsumerConfig::GetStaticParserRuleSet() const
{
    return _staticRules;
};

void ConsumerConfig::Reset()
{
    _eventList.clear();
    _staticRules.clear();
}

std::optional<TelemetryPermission>
ConsumerConfig::GetTelemetryPermissionFromJsonBuilder(
    jsonbuilder::JsonBuilder& jsonBuilder)
{
    auto lttngNameItr = jsonBuilder.find("metadata", "lttngName");
    std::string_view eventName = lttngNameItr->GetUnchecked<std::string_view>();

    if (auto itr = _eventList.find(eventName); itr != _eventList.end())
    {
        return itr->second;
    }

    return {};
}

}}}