// Copyright (c) Microsoft Corporation. All rights reserved.

#include "DynamicParser.hpp"
#include <optional>

#include <azure-device-util/Json.h>
#include <azure-device-util/String.h>
#include <nlohmann/json.hpp>

#include "TlgProvider.h"

using json = nlohmann::json;

namespace Azure { namespace Device { namespace Telemetry {

DynamicParser::DynamicParser()
    : _regexTagStart(c_EventRegexStart, std::regex::extended | std::regex::optimize)
    , _regexTagEnd(c_EventRegexEnd, std::regex::extended | std::regex::optimize)
    , _regexKeyValue(c_KeyValueRegex, std::regex::extended | std::regex::optimize)
{}

std::optional<jsonbuilder::JsonBuilder>
DynamicParser::ParseEvent(const MessageEvent& event)
{
    std::string_view message = event.Message();
    std::match_results<std::string_view::const_iterator> base_match;
    switch (event.GetTagLocation())
    {
    case MessageEvent::TagLocation::Start:
        if (std::regex_match(
                message.begin(), message.end(), base_match, _regexTagStart))
        {
            return BuildEventJsonBuilder(
                base_match[static_cast<int>(RegexStartGroups::Provider)].str(),  // Provider
                base_match[static_cast<int>(RegexStartGroups::Name)].str(),  // Name
                base_match[static_cast<int>(RegexStartGroups::Mode)].str(),  // Content
                base_match[static_cast<int>(RegexStartGroups::Content)].str()  // Mode
            );
        }
        break;
    case MessageEvent::TagLocation::End:
        if (std::regex_match(
                message.begin(), message.end(), base_match, _regexTagEnd))
        {
            return BuildEventJsonBuilder(
                base_match[static_cast<int>(RegexEndGroups::Provider)].str(),  // Provider
                base_match[static_cast<int>(RegexEndGroups::Name)].str(),  // Name
                base_match[static_cast<int>(RegexEndGroups::Mode)].str(),  // Content
                base_match[static_cast<int>(RegexEndGroups::Content)].str()  // Mode
            );
        }
        break;
    case MessageEvent::TagLocation::None:
        return {};
    }

    return {};
}

std::optional<jsonbuilder::JsonBuilder> DynamicParser::BuildEventJsonBuilder(
    const std::string& providerName,
    const std::string& eventName,
    std::string_view mode,
    std::string_view content)
{
    jsonbuilder::JsonBuilder jb;
    auto metaDataItr =
        jb.push_back(jb.end(), "metadata", jsonbuilder::JsonObject);
    jb.push_back(metaDataItr, "lttngName", (providerName + ":" + eventName));

    jb.push_back(jb.end(), "name", (providerName + "." + eventName));
    jb.push_back(jb.end(), "time", std::chrono::system_clock::now());

    auto dataItr = jb.push_back(jb.end(), "data", jsonbuilder::JsonObject);

    if (!ParseData(mode, content, jb, dataItr))
    {
        return {};
    }

    return jb;
}

bool DynamicParser::ParseData(
    std::string_view mode,
    std::string_view content,
    jsonbuilder::JsonBuilder& jb,
    jsonbuilder::JsonIterator itr)
{
    if (mode.empty())
    {
        jb.push_back(itr, "message", content);
        return true;
    }
    else if (!Utils::String::ICompare("KV", mode))
    {
        return ParseKeyValue(content, jb, itr);
    }
    else if (!Utils::String::ICompare("Json", mode))
    {
        return ParseJson(content, jb, itr);
    }
    else
    {
        return false;
    }
}

bool DynamicParser::ParseKeyValue(
    std::string_view content,
    jsonbuilder::JsonBuilder& jb,
    jsonbuilder::JsonIterator itr)
{
    auto matches_begin = std::regex_iterator<std::string_view::const_iterator>(
        content.begin(), content.end(), _regexKeyValue);
    auto matches_end = std::regex_iterator<std::string_view::const_iterator>();

    for (std::regex_iterator<std::string_view::const_iterator> i = matches_begin;
         i != matches_end;
         i++)
    {
        std::match_results<std::string_view::const_iterator> kv_matches = *i;
        jb.push_back(itr, kv_matches[1].str(), kv_matches[2].str());
    }

    return true;
}

bool DynamicParser::ParseJson(
    std::string_view content,
    jsonbuilder::JsonBuilder& jb,
    jsonbuilder::JsonIterator itr)
{
    try
    {
        auto jsonStart = content.find_first_of('{');
        auto jsonLength = content.find_last_of('}') - jsonStart + 1;
        auto j = json::parse(content.substr(jsonStart, jsonLength));
        if (j.type() != nlohmann::json::value_t::object)
        {
            return false;
        }

        for (const auto& [key, value] : j.items())
        {
            Azure::Device::Utils::Json::InsertJsonValueIntoJsonBuilder(
                jb, itr, key, value);
        }
        return true;
    }
    catch (std::exception& ex)
    {
        std::cout << "JournalDConsumer: Could not parse JSON event!" << std::endl
                  << ex.what() << std::endl;

        TraceLoggingWrite(
            g_tlgProv,
            "JournalD_JsonParsingFailed",
            TraceLoggingString(ex.what(), "what"));

        return false;
    }
}

}}}