// Copyright (c) Microsoft Corporation. All rights reserved.

#include "StaticParserRule.hpp"
#include <optional>

#include <azure-device-util/AndroidLogHelper.h>

namespace Azure { namespace Device { namespace Telemetry {

StaticParserRule::StaticParserRule(
    std::string providerName,
    std::string eventName,
    std::optional<uint32_t> isCoreData,
    std::optional<uint32_t> privacyTags,
    std::optional<uint64_t> keyword,
    std::string regexStr)
    : _providerName(std::move(providerName))
    , _eventName(std::move(eventName))
    , _isCoreData(isCoreData)
    , _privacyTags(privacyTags)
    , _keyword(keyword)
    , _regex(std::move(regexStr), std::regex::optimize)
{}

StaticParserRule StaticParserRule::FromTomlTable(const toml::value& table)
{
    // Provider and event name
    auto provider = toml::find<std::string>(table, "provider");
    auto event = toml::find<std::string>(table, "event");

    std::optional<uint32_t> isCoreData;
    if (table.contains("iscoredata"))
    {
        isCoreData = toml::find<uint32_t>(table, "iscoredata");
    }

    std::optional<uint32_t> privacyTags;
    if (table.contains("privacytags"))
    {
        isCoreData = toml::find<uint32_t>(table, "privacytags");
    }

    std::optional<uint64_t> keyword;
    if (table.contains("keyword"))
    {
        keyword = toml::find<uint32_t>(table, "keyword");
    }

    // Regex for matching messages
    auto regex = toml::find<std::string>(table, "regex");

    StaticParserRule rule{
        std::move(provider), std::move(event), isCoreData,
        privacyTags,         keyword,          std::move(regex)
    };

    // Part C fields
    if (table.contains("groups"))
    {
        auto groups = toml::find<std::vector<toml::value>>(table, "groups");
        rule._groups.reserve(groups.size());

        for (const auto& table : groups)
        {
            uint32_t index = toml::find<uint32_t>(table, "index");
            std::string name = toml::find<std::string>(table, "name");
            rule._groups.push_back(std::make_pair(index, name));
        }
    }

    // Journal filters
    if (table.contains("rules"))
    {
        const auto rules =
            toml::find<std::map<std::string, std::string>>(table, "rules");

        rule._fields = std::vector<std::pair<std::string, std::string>>(
            rules.begin(), rules.end());
    }

    return rule;
}

bool StaticParserRule::CheckFilter(const MessageEvent& event) const
{
    for (const auto& field : _fields)
    {
        // Get journal field value
        // Fail if the journal entry has no such field
        auto eventVal = event.Field(field.first);
        if (!eventVal)
        {
            return false;
        }

        // Check if the value is empty or doesn't match the rule's value
        if (field.second != *eventVal)
        {
            return false;
        }
    }

    return true;
}

std::optional<jsonbuilder::JsonBuilder>
StaticParserRule::MatchEvent(const MessageEvent& event) const
{
    const std::string_view message = event.Message();
    std::match_results<std::string_view::const_iterator> base_match;
    if (std::regex_match(message.begin(), message.end(), base_match, _regex))
    {
        jsonbuilder::JsonBuilder jb;
        auto metaDataItr =
            jb.push_back(jb.end(), "metadata", jsonbuilder::JsonObject);
        jb.push_back(metaDataItr, "lttngName", _providerName + ":" + _eventName);
        jb.push_back(jb.end(), "name", _providerName + "." + _eventName);
        jb.push_back(jb.end(), "time", event.GetPartATime());

        auto dataItr = jb.push_back(jb.end(), "data", jsonbuilder::JsonObject);
        jb.push_back(dataItr, "message", message);
        if (_keyword)
        {
            jb.push_back(dataItr, "_ms_Keywords", *_keyword);
        }

        if (_isCoreData)
        {
            jb.push_back(dataItr, "_ms_IsCoreData", *_isCoreData);
        }

        if (_privacyTags)
        {
            jb.push_back(dataItr, "_ms_PrivacyTags", *_privacyTags);
        }

        for (const auto& group : _groups)
        {
            jb.push_back(dataItr, group.second, base_match[group.first].str());
        }

        return jb;
    }

    return {};
}

}}}
