// Copyright (c) Microsoft Corporation. All rights reserved.

#include "AsimovEvent.h"

#include <bitset>
#include <memory>

#include <azure-device-util/AndroidLogHelper.h>
#include <azure-device-util/Bytes.h>
#include <azure-device-util/ResultException.h>
#include <azure-device-util/String.h>
#include <azure-device-util/Version.h>
#include <jsonbuilder/JsonRenderer.h>

#include "AsimovSharedPartAState.h"
#include "IEventStore.h"
#include "TelemetryError.h"

using namespace jsonbuilder;

namespace Azure { namespace Device { namespace Telemetry {

/*++
    Renders out the AsimovEvent to a UTF-8 JSON string

Arguments:
    partAState - Helper that holds non-event state needed for serialization
                 (epoch, seqNum, etc)
    eventStore - Event store the event will be persisted to
                 (necessary for correctly choosing ticket LUIDs)
    eventStoreLock - Lock for eventStore.GetMutex(), probably deferred.
                     Will become acquired.

Return Value:
    UTF-8 JSON string
--*/
std::string AsimovEvent::Render(
    AsimovSharedPartAState& partAState,
    IEventStore& eventStore,
    std::unique_lock<std::recursive_mutex>& eventStoreLock)
{
    {
        auto partAItr = _jsonBuilder.find(_jsonBuilder.root(), "a");
        FAIL_FAST_IF(partAItr == _jsonBuilder.end());

        AddTopLevelPartAFields(partAItr, partAState.GetTenantToken());

        auto extItr = CreateOrGetExtensionObject(partAItr, "ext");
        auto utcItr = CreateOrGetExtensionObject(extItr, "utc");

        AddExtFields(partAState, extItr, utcItr);

        // Begin work that requires eventstore synchronization
        eventStoreLock.lock();

        //
        // Add the epoch and seqnum after locking the event store, so that
        // events can't be rendered and written to the event store with
        // different sequence number ordering.
        //
        AddEpochAndSeq(partAState, utcItr);

        // The full tenant token is moved to a set which is sent as a header on
        // upload (apikey).
        // Do this after taking the storagebuffer lock
        eventStore.StoreTenantToken(
            _persistence, _latency, std::string{ partAState.GetTenantToken() });
    }

    SplicePartAIntoJsonRoot();

    JsonRenderer renderer;
    renderer.Reserve(k_defaultJsonRendererReserveSize);
    std::string_view result = renderer.Render(_jsonBuilder);
    return std::string(result.data(), result.size());
}

std::string AsimovEvent::Render(AsimovSharedPartAState& partAState)
{
    auto partAItr = _jsonBuilder.find(_jsonBuilder.root(), "a");
    FAIL_FAST_IF(partAItr == _jsonBuilder.end());

    AddTopLevelPartAFields(partAItr, partAState.GetTenantToken());

    auto extItr = CreateOrGetExtensionObject(partAItr, "ext");
    auto utcItr = CreateOrGetExtensionObject(extItr, "utc");

    AddExtFields(partAState, extItr, utcItr);
    AddEpochAndSeq(partAState, utcItr);

    SplicePartAIntoJsonRoot();

    JsonRenderer renderer;
    renderer.Reserve(k_defaultJsonRendererReserveSize);
    std::string_view result = renderer.Render(_jsonBuilder);
    return std::string(result.data(), result.size());
}

/*++
    Get timestamp of event.  Does not work after fields have been spliced and
    moved.
--*/
std::chrono::system_clock::time_point AsimovEvent::GetTimestamp() const
{
    JsonConstIterator itr = _jsonBuilder.find("a", "time");
    return itr->GetUnchecked<std::chrono::system_clock::time_point>();
}

/*++
    Get full name of event.  Does not work after fields have been spliced and
    moved.
--*/
std::string_view AsimovEvent::GetName() const
{
    auto nameItr = _jsonBuilder.find("a", "name");
    FAIL_FAST_IF(nameItr == _jsonBuilder.end());

    std::string_view name;
    FAIL_FAST_IF(!nameItr->ConvertTo(name));

    return name;
}

/*++
    Adds more data to the 'a' JsonBuilder subobject

Arguments:
    partAItr - Iterator to part a{} json object
--*/
void AsimovEvent::AddTopLevelPartAFields(
    JsonConstIterator partAItr,
    std::string_view tenantToken)
{
    FAIL_FAST_IF(tenantToken.empty());

    auto tenantIdEnd = tenantToken.find('-');
    FAIL_FAST_IF(tenantIdEnd == std::string::npos);

    std::string prefixedTenantId = "o:";
    prefixedTenantId += tenantToken.substr(0, tenantIdEnd);

    // Add only the o:GUID section of the tenant token to the part A.
    _jsonBuilder.push_back(partAItr, "iKey", prefixedTenantId);
}

/*++
    Adds eventFlags field.

Arguments:
    utcItr - Iterator to utc{} json object
--*/
void AsimovEvent::AddEventFlags(JsonConstIterator utcItr)
{
    std::byte persistenceByte{ static_cast<uint8_t>(
        static_cast<int>(_persistence) + 1) };

    std::byte latencyByte = CalculateLatencyByte(_latency);

    constexpr std::byte c_zeroByte{ 0 };

    const std::byte partABytes[8] = { persistenceByte, latencyByte, c_zeroByte,
                                      c_zeroByte,      c_zeroByte,  c_zeroByte,
                                      c_zeroByte,      c_zeroByte };

    nonstd::span<const std::byte> partABytesSpan{ partABytes };

    uint64_t partAflags = Utils::Bytes::BytesToValue<uint64_t>(partABytesSpan);

    _jsonBuilder.push_back(utcItr, "eventFlags", partAflags);
}

/*++
    Adds 'epoch' and 'seq' values.

Arguments:
    partAState    - Helper that holds non-event state needed for serialization
                    (epoch, seqNum, etc)
    utcItr        - Iterator to utc{} json object
--*/
void AsimovEvent::AddEpochAndSeq(
    AsimovSharedPartAState& partAState,
    JsonConstIterator utcItr)
{
    // push_back fields into PartA.  ver, name, timestamp are already present
    // from event decode time.
    std::string_view epoch = partAState.GetEpoch();
    _jsonBuilder.push_back(utcItr, "epoch", epoch);

    // sequence id - a monotonically incrementing counter for this instance
    // of the uploader
    uint64_t sequenceNumber = partAState.IncrementEventCount();
    _jsonBuilder.push_back(utcItr, "seq", sequenceNumber);
}

/*++
    Adds all extensions and their fields into the object.

Arguments:
    partAState    - Helper that holds non-event state needed for serialization
                    (epoch, seqNum, etc)
    extItr      - Iterator to the ext{} json object
    utcItr      - Iterator to the utc{} json object
--*/
void AsimovEvent::AddExtFields(
    AsimovSharedPartAState& partAState,
    JsonConstIterator extItr,
    JsonConstIterator utcItr)
{
    AddOsFields(partAState, extItr);

    {
        // Create Device Extension
        auto deviceItr = CreateOrGetExtensionObject(extItr, "device");

        // Device { localId } (aka SQM ID)
        std::string_view deviceId = partAState.GetDeviceId();
        _jsonBuilder.push_back(deviceItr, "localId", deviceId);

        //
        // DeviceClass
        // This is actually the device Platform data, but needs to be called
        // DeviceClass because reasons. Some people apparently really want it
        // called deviceClass.
        //
        _jsonBuilder.push_back(
            deviceItr, "deviceClass", partAState.GetDevicePlatform());
    }

    // Create protocol extension
    {
        auto protocolItr = CreateOrGetExtensionObject(extItr, "protocol");
        _jsonBuilder.push_back(protocolItr, "devMake", partAState.GetDeviceMake());
        _jsonBuilder.push_back(
            protocolItr, "devModel", partAState.GetDeviceModel());

        // This bit signifies participation in AzureDeviceBuilder
        uint32_t msp = 0b100;
        _jsonBuilder.push_back(protocolItr, "msp", msp);
    }

    // Create sdk extension
    {
        std::string versionString = "ADHS-Linux-None-";
        versionString += Utils::Version::GetString();

        auto sdkItr = CreateOrGetExtensionObject(extItr, "sdk");
        _jsonBuilder.push_back(sdkItr, "ver", versionString);
    }

    AddEventFlags(utcItr);

    {
        // Bits 16-23 (0-indexed)
        constexpr bool collectedDueToOverride = false;
        constexpr bool isOnBattery = false;
        constexpr bool isScreenOff = false;
        constexpr bool isInputIdle = false;
        bool isMeasures =
            (_telemetryCategory == TriggerTelemetryCategory::Measures);
        bool isTelemetry =
            (_telemetryCategory == TriggerTelemetryCategory::Telemetry);

        std::bitset<8> bitFlagSet1;
        bitFlagSet1.set(0, partAState.IsTestMachine());
        bitFlagSet1.set(1, partAState.IsMsftInternal());
        bitFlagSet1.set(2, collectedDueToOverride);
        bitFlagSet1.set(3, isOnBattery);
        bitFlagSet1.set(4, isScreenOff);
        bitFlagSet1.set(5, isInputIdle);
        bitFlagSet1.set(6, isMeasures);
        bitFlagSet1.set(7, isTelemetry);

        // Bits 24-31 (0-indexed)
        constexpr bool isFileBacked = false;
        bool isCoreData =
            (_minTelemetryPermission <= TelemetryPermission::Required);
        std::bitset<2> telPermissionLevelBitset{ static_cast<uint8_t>(
            _telemetryPermissionAtCreation) };
        bool isCriticalData =
            (_telemetryCategory == TriggerTelemetryCategory::CriticalData);

        std::bitset<8> bitFlagSet2;
        bitFlagSet2.set(0, isFileBacked);
        bitFlagSet2.set(1, isCoreData);
        bitFlagSet2.set(2, telPermissionLevelBitset.test(0));
        bitFlagSet2.set(3, telPermissionLevelBitset.test(1));
        bitFlagSet2.set(4, isCriticalData);

        constexpr std::byte c_zeroByte{ 0 };
        const std::byte utcFlagBytes[8] = {
            c_zeroByte,
            c_zeroByte,
            Utils::Bytes::BitsetToByte(bitFlagSet1),
            Utils::Bytes::BitsetToByte(bitFlagSet2),
            c_zeroByte,
            c_zeroByte,
            c_zeroByte,
            c_zeroByte
        };

        nonstd::span<const std::byte> utcFlagBytesSpan{ utcFlagBytes };

        uint64_t utcFlags =
            Utils::Bytes::BytesToValue<uint64_t>(utcFlagBytesSpan);

        _jsonBuilder.push_back(utcItr, "flags", utcFlags);
    }
}

/*++
    Adds more data to the 'a' JsonBuilder subobject based on data from
    partAState.

Arguments:
    partAState    - Helper that holds non-event state needed for serialization
                    (epoch, seqNum, etc)
    extItr        - Iterator to ext{} json object
--*/
void AsimovEvent::AddOsFields(
    AsimovSharedPartAState& partAState,
    JsonConstIterator extItr)
{
    auto osItr = CreateOrGetExtensionObject(extItr, "os");

    _jsonBuilder.push_back(osItr, "name", partAState.GetOsName());
    _jsonBuilder.push_back(osItr, "ver", partAState.GetOsVersion());
    _jsonBuilder.push_back(osItr, "build", partAState.GetOsBuild());
}

/*++
    Creates a new JsonObject under a parent iterator, if an object with the
    given name does not already exist. Optionally adds a version field to the
    new object.

Arguments:
    parent - The parent JsonObject that should contain the new object.
    extensionName - The name of the new object.

Return Value:
    Iterator to a new object or existing object with given name
--*/
JsonConstIterator AsimovEvent::CreateOrGetExtensionObject(
    JsonConstIterator parent,
    std::string_view extensionName)
{
    JsonConstIterator itr = _jsonBuilder.find(parent, extensionName);
    if (itr == _jsonBuilder.end())
    {
        itr = _jsonBuilder.push_back(parent, extensionName, JsonObject);
        FAIL_FAST_IF(itr == _jsonBuilder.end());
    }

    return itr;
}

/*++
    Maps TriggerLatency values to the bitmask value server expects to represent
    each such latency.

Arguments:
    latency - Latency value to be translated

Return Value:
    byte value expected by server for the given latency
--*/
std::byte AsimovEvent::CalculateLatencyByte(TriggerLatency latency)
{
    switch (latency)
    {
    case TriggerLatency::Normal:
        return std::byte{ 1 };

    case TriggerLatency::Realtime:
        return std::byte{ 2 };

    default:
        FAIL_FAST();
    }
}

/*++
    Move some elements of the JsonBuilder from what is convenient to deal
with in the intermediate to what is absolutely required by the CommonSchema
spec.
--*/
void AsimovEvent::SplicePartAIntoJsonRoot()
{
    auto partAItr = _jsonBuilder.find(_jsonBuilder.root(), "a");
    FAIL_FAST_IF(partAItr == _jsonBuilder.end());

    auto oldExtItr = _jsonBuilder.find(partAItr, "ext");
    FAIL_FAST_IF(partAItr == _jsonBuilder.end());

    auto newExtItr =
        _jsonBuilder.push_front(_jsonBuilder.root(), "ext", JsonObject);

    // Take the children out of root.a.ext and front-insert them into
    // root.ext, then delete the empty root.a.ext object
    _jsonBuilder.splice_front(oldExtItr, newExtItr);
    _jsonBuilder.erase(oldExtItr);

    // Take the children out of root.a and front-insert them into the root,
    // then delete the empty root.a object
    _jsonBuilder.splice_front(partAItr, _jsonBuilder.root());
    _jsonBuilder.erase(partAItr);
}

void AsimovEvent::InitializePartA(
    jsonbuilder::JsonBuilder& jb,
    jsonbuilder::JsonConstIterator& extItr,
    jsonbuilder::JsonConstIterator& utcItr,
    jsonbuilder::JsonConstIterator& metadataItr)
{
    auto partAItr = jb.push_back(jb.root(), "a", JsonObject);

    auto nameItr = jb.find("name");
    auto timestampItr = jb.find("time");
    auto timestamp =
        timestampItr->GetUnchecked<std::chrono::system_clock::time_point>();

    jb.push_back(partAItr, "ver", "4.0");
    jb.push_back(partAItr, "name", nameItr->GetUnchecked<std::string_view>());
    jb.push_back(partAItr, "time", timestamp);

    // Now clean up the top level name/timestamp values
    jb.erase(nameItr);
    jb.erase(timestampItr);

    extItr = jb.push_back(partAItr, "ext", JsonObject);
    utcItr = jb.push_back(extItr, "utc", JsonObject);
    metadataItr = jb.push_back(extItr, "metadata", JsonObject);
}

void AsimovEvent::ExtractTemporaryFieldsInJsonBuilder(
    jsonbuilder::JsonBuilder& builder,
    AsimovEvent& event,
    jsonbuilder::JsonConstIterator& dataItr,
    jsonbuilder::JsonConstIterator& extItr,
    jsonbuilder::JsonConstIterator& metadataItr)
{
    auto keywordsItr = builder.find(dataItr, "_ms_Keywords");
    if (keywordsItr != builder.end())
    {
        event._keywords = keywordsItr->GetUnchecked<uint64_t>();
        if (event._keywords == TriggerTelemetryKeyword::Telemetry)
        {
            event._telemetryCategory = TriggerTelemetryCategory::Telemetry;
        }
        else if (event._keywords == TriggerTelemetryKeyword::Measures)
        {
            event._telemetryCategory = TriggerTelemetryCategory::Measures;
        }
        else if (event._keywords == TriggerTelemetryKeyword::CriticalData)
        {
            event._telemetryCategory = TriggerTelemetryCategory::CriticalData;
        }
        else
        {
            // This isn't a telemetry event, but can be sent to relog locally
            event._telemetryCategory = TriggerTelemetryCategory::None;
        }

        builder.erase(keywordsItr);
    }

    auto isCoreDataItr = builder.find(dataItr, "_ms_IsCoreData");
    if (isCoreDataItr != builder.end())
    {
        uint32_t isCoreData = isCoreDataItr->GetUnchecked<uint32_t>();
        if (isCoreData >= 1)
        {
            // Core data can flow at Required permission
            event._minTelemetryPermission = TelemetryPermission::Required;

            // Core data events must always be critical data keyword
            event._telemetryCategory = TriggerTelemetryCategory::CriticalData;

            // For forwarding to ETW, include the ETW keyword that represents
            // critical data as well
            event._keywords = 0x0000800000000000L;
        }

        builder.erase(isCoreDataItr);
    }

    uint64_t privacyTags = 0;
    auto privacyTagItr = builder.find(dataItr, "_ms_PrivacyTags");
    if (privacyTagItr != builder.end())
    {
        privacyTags = privacyTagItr->GetUnchecked<uint64_t>();
        builder.erase(privacyTagItr);
    }

    std::string processName;
    auto processNameItr = builder.find(dataItr, "_ms_ProcessName");
    if (processNameItr != builder.end())
    {
        processName = processNameItr->GetUnchecked<std::string_view>();
        builder.erase(processNameItr);
    }

    builder.push_back(extItr, "container", JsonObject);

    jsonbuilder::JsonConstIterator appItr =
        builder.push_back(extItr, "app", JsonObject);

    if (privacyTags != 0)
    {
        builder.push_back(metadataItr, "privTags", privacyTags);
    }

    //
    // Since we only have the process name available, this is a 'Linux'
    // style app identifier with prefix "L:"
    //
    if (!processName.empty())
    {
        std::string appId = "L:";
        appId.append(processName);
        builder.push_back(appItr, "id", appId);
    }
}

/*++
    Factory function to create an AsimovEvent from the JsonBuilder emitted
    by our CLOG-reader.

    This will erase unneeded metadata sections of the builder and move
    items around to be suitable for Common Schema consumption.
--*/
std::unique_ptr<AsimovEvent>
AsimovEvent::CreateFromHwBinderConsumer(jsonbuilder::JsonBuilder&& builder)
{
    auto event = std::make_unique<AsimovEvent>();

    //
    // Read some metadata properties out of the metadata section to
    // establish core properties of this event
    //
    event->_jsonBuilder = std::move(builder);
    auto& jb = event->_jsonBuilder;

    auto dataItr = jb.find("data");
    if (dataItr == jb.end())
    {
        THROW_HR(std::make_error_code(std::errc::invalid_argument));
    }

    jsonbuilder::JsonConstIterator extItr;
    jsonbuilder::JsonConstIterator utcItr;
    jsonbuilder::JsonConstIterator metadataItr;
    InitializePartA(jb, extItr, utcItr, metadataItr);
    ExtractTemporaryFieldsInJsonBuilder(jb, *event, dataItr, extItr, metadataItr);

    return event;
}

std::unique_ptr<AsimovEvent>
AsimovEvent::CreateFromLogcatConsumer(jsonbuilder::JsonBuilder&& builder)
{
    auto event = std::make_unique<AsimovEvent>();

    auto eraseIfFound = [&builder](std::string_view name) {
        auto itr = builder.find(name);
        if (itr != builder.end())
        {
            builder.erase(itr);
        }
    };

    eraseIfFound("metadata");

    //
    // Read some metadata properties out of the metadata section to
    // establish core properties of this event
    //
    auto dataItr = builder.find("data");
    if (dataItr == builder.end())
    {
        THROW_HR(std::make_error_code(std::errc::invalid_argument));
    }

    event->_jsonBuilder = std::move(builder);
    auto& jb = event->_jsonBuilder;

    jsonbuilder::JsonConstIterator extItr;
    jsonbuilder::JsonConstIterator utcItr;
    jsonbuilder::JsonConstIterator metadataItr;
    InitializePartA(jb, extItr, utcItr, metadataItr);
    ExtractTemporaryFieldsInJsonBuilder(jb, *event, dataItr, extItr, metadataItr);

    return event;
}

/*++
    Factory function to create an AsimovEvent from the JsonBuilder emitted
    by lttng-consume.

    This will erase unneeded metadata sections of the builder and move
    items around to be suitable for Common Schema consumption.
--*/
std::unique_ptr<AsimovEvent> AsimovEvent::CreateFromLttngConsumer(
    jsonbuilder::JsonBuilder&& builder,
    TelemetryPermission minTelemetryPermission)
{
    auto event = std::make_unique<AsimovEvent>();

    auto eraseIfFound = [&builder](std::string_view name) {
        auto itr = builder.find(name);
        if (itr != builder.end())
        {
            builder.erase(itr);
        }
    };

    eraseIfFound("packetContext");
    eraseIfFound("eventHeader");
    eraseIfFound("streamEventContext");
    eraseIfFound("eventContext");
    eraseIfFound("metadata");

    auto getActivityId = [&builder](std::string_view name) -> std::string {
        std::string activityIdStr;

        auto activityIdItr = builder.find("data", name);
        if (activityIdItr != builder.end())
        {
            if (activityIdItr->Type() == jsonbuilder::JsonType::JsonArray)
            {
                uuid_t activityId;
                uint32_t bytesWritten = 0;
                auto bytesItr = activityIdItr.begin();
                auto endBytesItr = activityIdItr.end();
                while (bytesItr != endBytesItr &&
                       bytesWritten < sizeof(activityId))
                {
                    if (!bytesItr->ConvertTo(activityId[bytesWritten]))
                    {
                        break;
                    }

                    ++bytesItr;
                    ++bytesWritten;
                }

                if (bytesItr == endBytesItr && bytesWritten == sizeof(activityId))
                {
                    activityIdStr =
                        Utils::String::StringFromGuid(activityId, false);
                }
            }
            builder.erase(activityIdItr);
        }

        return activityIdStr;
    };

    std::string activityId = getActivityId("_ms_ActivityId");
    std::string relatedActivityId = getActivityId("_ms_RelatedActivityId");

    uint64_t privacyTags = 0;
    auto privacyTagItr = builder.find("data", "_ms_PrivacyTags");
    if (privacyTagItr != builder.end())
    {
        privacyTags = privacyTagItr->GetUnchecked<uint64_t>();
    }

    event->_minTelemetryPermission = minTelemetryPermission;

    event->_jsonBuilder = std::move(builder);
    auto& jb = event->_jsonBuilder;

    jsonbuilder::JsonConstIterator extItr;
    jsonbuilder::JsonConstIterator utcItr;
    jsonbuilder::JsonConstIterator metadataItr;
    InitializePartA(jb, extItr, utcItr, metadataItr);

    if (!activityId.empty())
    {
        jb.push_back(utcItr, "aId", activityId);
    }

    if (!relatedActivityId.empty())
    {
        jb.push_back(utcItr, "raId", relatedActivityId);
    }

    if (privacyTags != 0)
    {
        jb.push_back(metadataItr, "privTags", privacyTags);
    }

    return event;
}
}}}
