// Copyright (c) Microsoft Corporation. All rights reserved.

#include "AsimovEventBuffer.h"

#include <atomic>
#include <iomanip>
#include <memory>
#include <string_view>

#include <azure-device-util/Bytes.h>
#include <azure-device-util/FileSystem.h>
#include <azure-device-util/String.h>
#include <azure-device-util/Time.h>

#include "TelemetryError.h"
#include "TlgProvider.h"

namespace Azure { namespace Device { namespace Telemetry {

const std::string AsimovEventBuffer::k_eventSeparator = "\r\n";

std::shared_ptr<IAsimovEventBuffer>
AsimovEventBufferFactory::Create(size_t maxBytes, uint32_t maxEvents)
{
    return std::make_shared<AsimovEventBuffer>(maxBytes, maxEvents);
}

/*++
    AsimovEventBuffer constructor - Creates a new empty buffer

Arguments:
    maxBytes    -   Maximum bytes this event buffer will hold.
    maxEvents   -   Maximum number of events this event buffer will hold.
--*/
AsimovEventBuffer::AsimovEventBuffer(size_t maxBytes, uint32_t maxEvents)
    : _maxEvents(maxEvents), _maxBytes(maxBytes)
{
    AsimovEventBuffer::Clear();
}

/*++
    Adds an event to the buffer.

    Throws if adds are attempted after previous invocation returned false to
    indicate that no space remains, and the buffer has not been reset.

Arguments:
    event - UTF-8 payload string to store in the internal array

Return Value:
    true - If there is still room for more events / bytes in this buffer AFTER
           the event is added
    false - If this is the last event that can be added to this buffer.
--*/
bool AsimovEventBuffer::AddEvent(std::string_view event)
{
    THROW_HR_IF(make_error_code(std::errc::no_buffer_space), IsAtEventCapacity());
    THROW_HR_IF(make_error_code(std::errc::no_buffer_space), IsAtByteCapacity());

    //
    // Add this event to the buffer. We intentionally keep our max buffer size
    // and max events smaller than the servers's so that if we go over our size
    // slightly, the server will still accept the upload.
    //
    if (ShouldUseStreamingCompression(event.size() + k_eventSeparator.size()))
    {
        _deflateStream->AddData(Utils::Bytes::StringViewToBytes(event));
        _deflateStream->AddData(Utils::Bytes::StringViewToBytes(k_eventSeparator));
    }
    else
    {
        _json.append(event.data(), event.size());
        _json.append(k_eventSeparator);
    }

    _numEvents++;

    return !IsAtEventCapacity() && !IsAtByteCapacity();
}

/*++
    Creates a AsimovConstantEventBuffer from the contents of AsimovEventBuffer.

    The internal IJsonArray is serialized out to a string, which is then
    converted to UTF-8 to construct the AsimovConstantEventBuffer, which does
    further compression using the deflate algorithm.

    NOTE: Throws if this AsimovEventBuffer is empty at the beginning of the call

Return Value:
    On success, this function returns the AsimovConstantEventBuffer created.
--*/
AsimovConstantEventBuffer AsimovEventBuffer::Commit() const
{
    THROW_HR_IF(make_error_code(std::errc::operation_not_permitted), IsEmpty());

    std::vector<std::byte> compressedJson;

    if (_deflateStream.get() == nullptr)
    {
        Utils::DeflateStream deflateStream;

        deflateStream.AddData(Utils::Bytes::StringViewToBytes(_json));

        compressedJson = deflateStream.Finish();
    }
    else
    {
        compressedJson = _deflateStream->Finish();
    }

    return AsimovConstantEventBuffer{ GenerateApiKeyHeader(),
                                      std::move(compressedJson),
                                      _numEvents };
}

/*++
    Returns whether the event buffer is empty or not

Return Value:
    true - event buffer is empty
    false - event buffer is not empty
--*/
bool AsimovEventBuffer::IsEmpty() const
{
    return _json.empty() && _deflateStream.get() == nullptr;
}

/*++
    Resets the event buffer to a clean, empty state.
--*/
void AsimovEventBuffer::Clear() noexcept
{
    _numEvents = 0;
    _json.clear();
    _deflateStream.reset();
    _tenantTokenSet.clear();
}

/*++
    Creates a comma-separated list of Tenant Tokens stored in the buffer

Return Value:
    The generated apikey header
--*/
std::string AsimovEventBuffer::GenerateApiKeyHeader() const
{
    std::string result;

    auto tokenIt = _tenantTokenSet.begin();

    if (tokenIt != _tenantTokenSet.end())
    {
        result += *tokenIt;

        while (++tokenIt != _tenantTokenSet.end())
        {
            result += ",";
            result += *tokenIt;
        }
    }

    return result;
}

/*++
    Decides whether to use streaming compression or not and handles the
    transition from storing uncompressed data to using streaming compression.

    The transition happens when the memory cost of the uncompressed data exceeds
    the memory cost of the compressor.

Arguments:
    newEventSize - size in bytes of the new event that to add to the buffer

Return Value:
    true  - Currently using streaming compression, or switched to it with this
            new event.
    false - Not using streaming compression.
--*/
bool AsimovEventBuffer::ShouldUseStreamingCompression(size_t newEventSize)
{
    if (_deflateStream.get() != nullptr)
    {
        return true;
    }

    size_t newNeededSize = _json.size() + newEventSize;
    if (newNeededSize > Utils::DeflateStream::GetCompressorMemoryCost())
    {
        TraceLoggingWrite(
            g_tlgProv,
            "AsimovEventBuffer_SwitchToStreaming",
            TraceLoggingValue(newNeededSize, "NewNeededSize"));

        _deflateStream = std::make_unique<Utils::DeflateStream>();

        _deflateStream->AddData(Utils::Bytes::StringViewToBytes(_json));
        _json.clear();
        return true;
    }

    return false;
}

/*++
    Returns whether or not the buffer is full based on event capacity.
--*/
bool AsimovEventBuffer::IsAtEventCapacity() const
{
    return _numEvents >= _maxEvents;
}

/*++
    Returns whether or not the buffer is full based on byte capacity.  This can
    happen in two ways:

    1. We are not using streaming compression, and our uncompressed data has
       exceeded the byte limit. Note here that we are exceeding the byte limit,
       but theoretically only by a little, and compression should bring the
       final, compressed value down below the limit.

    2. We are using streaming compression, and there is not room for another
       max-sized event.  This is a pretty conservative estimate.  It is going to
       cut off when the current output bytes + size of the compression window +
       maximum uncompressed event size is larger than the byte limit.
--*/
bool AsimovEventBuffer::IsAtByteCapacity() const
{
    if (_deflateStream.get() == nullptr)
    {
        return _json.size() >= _maxBytes;
    }
    else
    {
        return (_deflateStream->GetMaxCurrentOutputSize() +
                k_maxEventSizeInBytes) >= _maxBytes;
    }
}

}}}
