// Copyright (c) Microsoft Corporation. All rights reserved.

#include "AsimovSharedPartAState.h"

#include <iomanip>
#include <sstream>

#include <azure-device-util/ConversionFactors.h>
#include <azure-device-util/FileSystem.h>
#include <azure-device-util/Identity.h>
#include <azure-device-util/Math.h>
#include <azure-device-util/Process.h>
#include <azure-device-util/String.h>
#include <nlohmann/json.hpp>

#include "AppState.h"
#include "ConfigPaths.h"
#include "TlgProvider.h"

namespace Azure { namespace Device { namespace Telemetry {

constexpr std::string_view AsimovSharedPartAState::c_localIdKey;
constexpr std::string_view AsimovSharedPartAState::c_launchCountKey;

/*++
    Loads persisted data from file for LocalId/LaunchCount.
    Sets defaults/generates for any fields that were not loaded.
    Then writes those fields back to a file.

    Also populates other OEM information based on the AzureDeviceManifest
    file which should be build-time constant.
--*/
AsimovSharedPartAState::AsimovSharedPartAState(Utils::IConfig& config)
{
    _devicePlatform = "Linux";

    ReloadConfiguration(config);

    PersistedData dataToLoad = LoadPersistedData();
    SetMembersBasedOnLoadedData(dataToLoad);

    PersistedData dataToSave;
    dataToSave.LocalId = _asimovGeneratedDeviceId;
    dataToSave.LaunchCount = _launchCount;
    dataToSave.DeviceSampleNumber = _deviceSampleNumber;

    SavePersistedData(dataToSave);

    // For now this is 1 + 0 + random number.
    GenerateEpochString(_launchCount);

    _machineBasedDeviceId = Utils::Identity::GetMachineBasedDeviceId();
    _osReleaseData = ReadOsReleaseData();

    TraceLoggingWrite(
        g_tlgProv,
        "AsimovSharedPartAState_FinalizedMembers",
        TraceLoggingValue(_machineBasedDeviceId.c_str(), "MachineBasedLocalId"),
        TraceLoggingValue(
            _asimovGeneratedDeviceId.c_str(), "AsimovGeneratedLocalId"),
        TraceLoggingValue(_launchCount, "LaunchCount"),
        TraceLoggingValue(_deviceMake.c_str(), "DeviceMake"),
        TraceLoggingValue(_deviceModel.c_str(), "DeviceModel"),
        TraceLoggingValue(_deviceSampleNumber, "DeviceSampleNumber"),
        TraceLoggingValue(_osReleaseData.Name.c_str(), "OsName"),
        TraceLoggingValue(_osReleaseData.Version.c_str(), "OsVersion"),
        TraceLoggingValue(_isTestMachine, "IsTestMachine"),
        TraceLoggingValue(_isMsftInternal, "IsMsftInternal"));
}

void AsimovSharedPartAState::ReloadConfiguration(Utils::IConfig& config)
{
    std::string subClass = config.GetString(Config::Telemetry_DeviceClass, "");

    _devicePlatform += (subClass.empty() ? "" : ".") + subClass;
    _deviceMake = config.GetString(Config::Telemetry_DeviceMake, "Unknown");
    _deviceModel = config.GetString(Config::Telemetry_DeviceModel, "Unknown");

    _tenantToken = config.GetString(Config::Telemetry_TenantToken, "");

    _isTestMachine = Utils::FileSystem::FileExists("/etc/test-machine");
    if (_isTestMachine)
    {
        std::cout << "Test machine mode ON." << std::endl;
    }

    _isMsftInternal = Utils::FileSystem::FileExists("/etc/msft-machine");
    if (_isMsftInternal)
    {
        std::cout << "MSFT internal mode ON." << std::endl;
    }
}

std::string_view AsimovSharedPartAState::GetDeviceId() const
{
    return _machineBasedDeviceId.empty() ? _asimovGeneratedDeviceId :
                                           _machineBasedDeviceId;
}

/*++
    Gets the full file path where persisted PartAState data is saved.
--*/
std::filesystem::path AsimovSharedPartAState::GetPersistedDataFilePath()
{
    return AppState::GetRuntimeDataDirectory() / "partAState.json";
}

/*++
    Loads data from the persisted file into the PersistedData struct.

    Any value that is not valid and present in the file will keep its default
    constructed value.
--*/
AsimovSharedPartAState::PersistedData AsimovSharedPartAState::LoadPersistedData()
{
    PersistedData data;

    std::string jsonString;
    std::error_code readPersistedDataResult =
        Utils::FileSystem::ReadUTF8StringFromFile(
            GetPersistedDataFilePath().string(), jsonString);

    if (FAILED(readPersistedDataResult))
    {
        TraceLoggingWrite(
            g_tlgProv,
            "AsimovSharedPartAState_FailedToReadPersistedDataFile",
            TraceLoggingErrorCode(readPersistedDataResult, "ErrorCode"));

        return data;
    }

    auto jsonVal = nlohmann::json::parse(jsonString, nullptr, false);

    // launchCount
    {
        auto itr = jsonVal.find(c_launchCountKey);
        if (itr != jsonVal.end())
        {
            if (itr->is_number_unsigned())
            {
                data.LaunchCount = *itr;
            }
        }
    }

    // localId
    {
        auto itr = jsonVal.find(c_localIdKey);
        if (itr != jsonVal.end())
        {
            if (itr->is_string())
            {
                std::string jsonLocalId = *itr;
                std::string_view jsonLocalIdView = jsonLocalId;

                // Verify it is of the form "s:UUID_STRING"
                if (jsonLocalIdView.size() == 38 &&
                    jsonLocalIdView.find(Utils::Identity::c_localIdPrefix) == 0)
                {
                    char* uuidView =
                        &jsonLocalId[Utils::Identity::c_localIdPrefix.size()];

                    uuid_t parse;
                    if (uuid_parse(uuidView, parse) == 0)
                    {
                        data.LocalId = jsonLocalId;
                    }
                }
            }
        }
    }

    // deviceSampleNumber
    {
        auto itr = jsonVal.find(c_deviceSampleNumberKey);
        if (itr != jsonVal.end())
        {
            if (itr->is_number_unsigned())
            {
                data.DeviceSampleNumber = *itr;
            }
        }
    }

    return data;
}

/*++
    Sets members based on values in persistedData struct.
    If the values are illegal, correct defaults are used.

Arguments:
    persistedData - PersistedData with value to store in members.
--*/
void AsimovSharedPartAState::SetMembersBasedOnLoadedData(
    const AsimovSharedPartAState::PersistedData& persistedData)
{
    if (persistedData.LaunchCount != 0)
    {
        _launchCount = persistedData.LaunchCount + 1;
    }
    else
    {
        _launchCount = 1;

        TraceLoggingWrite(
            g_tlgProv,
            "AsimovSharedPartAState_SetMemberDefault_LaunchCount",
            TraceLoggingValue(_launchCount, "LaunchCount"));
    }

    if (!persistedData.LocalId.empty())
    {
        _asimovGeneratedDeviceId = persistedData.LocalId;
    }
    else
    {
        _asimovGeneratedDeviceId = "s:" + Utils::String::GenerateGuidString(true);

        TraceLoggingWrite(
            g_tlgProv,
            "AsimovSharedPartAState_SetMemberDefault_DeviceId",
            TraceLoggingValue(_asimovGeneratedDeviceId.c_str(), "DeviceId"));
    }

    if (persistedData.DeviceSampleNumber > 0 &&
        persistedData.DeviceSampleNumber <= c_maxDeviceSampleNumber)
    {
        _deviceSampleNumber = persistedData.DeviceSampleNumber;
    }
    else
    {
        _deviceSampleNumber =
            Utils::Math::GenerateUniformDistributionRandomIntInRange<uint64_t>(
                1, c_maxDeviceSampleNumber);

        TraceLoggingWrite(
            g_tlgProv,
            "AsimovSharedPartAState_SetMemberDefault_DeviceSampleNumber",
            TraceLoggingValue(_deviceSampleNumber, "DeviceSampleNumber"));
    }
}

/*++
    Writes persistedData members to a JSON string and then to a file to be
    loaded on the next boot.

Arguments:
    persistedData - PersistedData to store in file.
--*/
void AsimovSharedPartAState::SavePersistedData(
    const AsimovSharedPartAState::PersistedData& persistedData)
{
    nlohmann::json json;
    json[std::string{ c_launchCountKey }] = persistedData.LaunchCount;
    json[std::string{ c_localIdKey }] = persistedData.LocalId;
    json[std::string{ c_deviceSampleNumberKey }] =
        persistedData.DeviceSampleNumber;

    std::string jsonString = json.dump();

    THROW_IF_FAILED(Utils::FileSystem::WriteUTF8StringToFile(
        GetPersistedDataFilePath().string(), jsonString));
}

/*++
    Generates the epoch string

    Bottom 4 digits:  random number
    5th digit:        0
    Higher digits:    launch count

    Example:
        BootCount = 27, RandomNumber = 8924
        Epoch: 2708924

    To regenerate component strings from output string:
        int  LaunchCount  = EpochNum / 100000;
        int  DataType     = (EpochNum / 10000) % 10;
        int  RandomNumber = EpochNum % 10000;

Arguments:
    launchCount  - Count of times service has booted
--*/
void AsimovSharedPartAState::GenerateEpochString(uint32_t launchCount)
{
    // randomEpochComponent should be a random 4-digit number, [0, 9999]
    uint32_t randomEpochComponent =
        Utils::Math::GenerateUniformDistributionRandomIntInRange(0U, 9999U);

    std::stringstream epochStringStream;
    epochStringStream << launchCount << "0" << std::setfill('0') << std::setw(4)
                      << randomEpochComponent;

    _epochString = epochStringStream.str();
}

AsimovSharedPartAState::OsReleaseData AsimovSharedPartAState::ReadOsReleaseData()
{
    OsReleaseData osRelease;

    // Read release data from /etc/os-release
    std::map<std::string, std::string, std::less<>> releaseData;
    std::error_code result = Utils::FileSystem::ReadDictionaryFromFile(
        "/etc/os-release", "=", releaseData);
    if (FAILED(result))
    {
        TraceLoggingWrite(
            g_tlgProv,
            "AsimovSharedPartAState_LoadOsReleaseData_FailedToReadOsRelease",
            TraceLoggingValue(result.value(), "ErrorCode"));

        return osRelease;
    }

    for (auto&& [key, value] : releaseData)
    {
        if (key == "ID")
        {
            Utils::String::Trim(value, "\"");
            osRelease.Name = std::move(value);
        }
        else if (key == "VERSION")
        {
            Utils::String::Trim(value, "\"");
            osRelease.Version = std::move(value);
        }
    }

    // Read subrelease data from /etc/os-subrelease
    std::map<std::string, std::string, std::less<>> subreleaseData;
    result = Utils::FileSystem::ReadDictionaryFromFile(
        "/etc/os-subrelease", "=", subreleaseData);
    if (FAILED(result))
    {
        TraceLoggingWrite(
            g_tlgProv,
            "AsimovSharedPartAState_LoadOsReleaseData_FailedToReadOsSubrelease",
            TraceLoggingValue(result.value(), "ErrorCode"));
    }
    else
    {
        std::string subreleaseId = subreleaseData["ID"];
        Utils::String::Trim(subreleaseId, "\"");
        std::string subreleaseVersionId = subreleaseData["VERSION_ID"];
        Utils::String::Trim(subreleaseVersionId, "\"");
        osRelease.Build = subreleaseId + "|" + subreleaseVersionId;
    }

    return osRelease;
}

}}}
