// Copyright (c) Microsoft Corporation. All rights reserved.

#include "HeartBeat.h"

#include <memory>

#include <azure-device-util/FileSystem.h>
#include <azure-device-util/ResultException.h>
#include <azure-device-util/Time.h>
#include <nlohmann/json.hpp>

#include "AppState.h"
#include "AsimovSyntheticEvent.h"
#include "IEventSerializer.h"
#include "TelemetryError.h"
#include "TlgProvider.h"

namespace Azure { namespace Device { namespace Telemetry {

constexpr std::chrono::minutes HeartBeat::c_timerPeriod;

using namespace std::string_view_literals;

struct HeartBeatCountersBase
{
#define DECLARE_HEARTBEAT_FIELD(type, name) \
    type name;                              \
    static constexpr std::string_view name##Name = #name##sv;

    struct IngestionType
    {
        // # of events dropped due to decoding failure
        DECLARE_HEARTBEAT_FIELD(uint32_t, DecodingDroppedCounter);
    };

    struct UploaderType
    {
        // **** Fields relating to network connection ****

        // # of http connections attempted against server
        DECLARE_HEARTBEAT_FIELD(uint32_t, NetworkAttempts);

        // # of http connections against server that failed
        DECLARE_HEARTBEAT_FIELD(uint32_t, NetworkErrorCounter);

        // std::error_code of last failed http connection
        DECLARE_HEARTBEAT_FIELD(uint32_t, NetworkLastError);

        // # of server connection failures that resulted in a timeout
        DECLARE_HEARTBEAT_FIELD(uint32_t, NetworkTimeoutCounter);

        // **** Fields relating to server response ****

        // # of server requests that were valid but had >=1 rejected event
        DECLARE_HEARTBEAT_FIELD(uint32_t, HttpResponsesWithRejectedEvents);

        // # of times server responded with a non 2XX/400 status code
        DECLARE_HEARTBEAT_FIELD(uint32_t, HttpResponseFailures);

        // # of server http requests that returned a 4xx status code
        DECLARE_HEARTBEAT_FIELD(uint32_t, HttpFailures4xx);

        // # of server http requests that returned a 5xx status code
        DECLARE_HEARTBEAT_FIELD(uint32_t, HttpFailures5xx);

        // # of events uploaded to server
        DECLARE_HEARTBEAT_FIELD(uint32_t, HttpEventsUploaded);

        // # of events dropped in upload buffer
        DECLARE_HEARTBEAT_FIELD(uint32_t, HttpEventsRejectedCounter);

        // Compressed number of bytes uploaded to server
        DECLARE_HEARTBEAT_FIELD(uint64_t, HttpCompressedBytesUploaded);

        // # of events dropped due to repeated upload failure
        DECLARE_HEARTBEAT_FIELD(uint32_t, HttpRepeatedUploadFailureDropped);
    };

    struct StorageType
    {
        // # of events dropped by event store
        DECLARE_HEARTBEAT_FIELD(uint32_t, DbDroppedCounter);

        // # of events dropped by event store due to full storage (throttling)
        DECLARE_HEARTBEAT_FIELD(uint32_t, DbDroppedFullCounter);

        // # of events dropped by event store due to failures other than full
        // storage (non-throttling)
        DECLARE_HEARTBEAT_FIELD(uint32_t, DbDroppedFailureCounter);

        // # of critical events dropped by event store
        DECLARE_HEARTBEAT_FIELD(uint32_t, DbCriticalDroppedCounter);

        // # of times an event store was reset for the lifetime of machine
        DECLARE_HEARTBEAT_FIELD(uint32_t, DbLifetimeResetCounter);

        // # of times an event store was reset
        DECLARE_HEARTBEAT_FIELD(uint32_t, DbResetCounter);

        // # of times we have entered critical overflow mode
        DECLARE_HEARTBEAT_FIELD(uint32_t, DbCriticalOverflowEntersCounter);

        // sum of bytes dropped from normal buffers due to entering critical
        // overflow
        DECLARE_HEARTBEAT_FIELD(uint32_t, DbEnteringCriticalOverflowDroppedCounter);

        // # of critical-sample events dropped by the event store
        DECLARE_HEARTBEAT_FIELD(uint32_t, DbCriticalDataDroppedCounter);

        // sum of sizes (in bytes) of all event stores reset
        DECLARE_HEARTBEAT_FIELD(uint64_t, DbResetSizeSum);

        // # of events persisted
        DECLARE_HEARTBEAT_FIELD(uint32_t, DbEventsPersistedCounter);
    };

    struct ContainerType
    {
        // # of events dropped due to container buffer full
        DECLARE_HEARTBEAT_FIELD(uint32_t, ContainerBufferFullDropCounter);

        // # of events forwarded to the host
        DECLARE_HEARTBEAT_FIELD(uint32_t, EventsForwardedToHost);

        // # of timeouts when connecting between host and agent
        DECLARE_HEARTBEAT_FIELD(uint32_t, HostConnectionTimeoutsCounter);

        // # of non-timeout errors encounterd in the host/agent socket transport
        // channel
        DECLARE_HEARTBEAT_FIELD(uint32_t, HostConnectionErrorsCounter);

        // Last host connection error code
        DECLARE_HEARTBEAT_FIELD(uint32_t, LastHostConnectionError);

        // Total number of times a reciever was registered with the proxy
        DECLARE_HEARTBEAT_FIELD(uint64_t, HalRegisterReceiverCount);

        // Total number of times a receiver failed to connect
        DECLARE_HEARTBEAT_FIELD(uint64_t, HalConnectionFailures);

        // Total number of events dropped due to buffer full
        DECLARE_HEARTBEAT_FIELD(uint64_t, HalBufferFullDrops);

        // Total number of events dropped due to other issues (e.g. exceptions)
        DECLARE_HEARTBEAT_FIELD(uint64_t, HalBufferOtherDrops);

        // Total number of failures to connect from Proxy API
        DECLARE_HEARTBEAT_FIELD(uint64_t, ProxyConnectionFailures);

        // Total number of successful connections from Proxy API
        DECLARE_HEARTBEAT_FIELD(uint64_t, ProxyConnectionSuccess);

        // Total number of events dropped due to missing callback in Proxy API
        DECLARE_HEARTBEAT_FIELD(uint64_t, ProxyMissingCallbackDrops);
    };

    // monotonically increasing heartbeat counter
    DECLARE_HEARTBEAT_FIELD(uint32_t, HeartBeatSequenceNumber);

    IngestionType Ingestion;
    UploaderType Uploader;
    StorageType Storage;
    ContainerType Container;
};

constexpr std::string_view HeartBeatCountersBase::IngestionType::DecodingDroppedCounterName;

constexpr std::string_view HeartBeatCountersBase::UploaderType::NetworkAttemptsName;
constexpr std::string_view HeartBeatCountersBase::UploaderType::NetworkErrorCounterName;
constexpr std::string_view HeartBeatCountersBase::UploaderType::NetworkLastErrorName;
constexpr std::string_view HeartBeatCountersBase::UploaderType::NetworkTimeoutCounterName;
constexpr std::string_view
    HeartBeatCountersBase::UploaderType::HttpResponsesWithRejectedEventsName;
constexpr std::string_view HeartBeatCountersBase::UploaderType::HttpResponseFailuresName;
constexpr std::string_view HeartBeatCountersBase::UploaderType::HttpFailures4xxName;
constexpr std::string_view HeartBeatCountersBase::UploaderType::HttpFailures5xxName;
constexpr std::string_view HeartBeatCountersBase::UploaderType::HttpEventsUploadedName;
constexpr std::string_view
    HeartBeatCountersBase::UploaderType::HttpEventsRejectedCounterName;
constexpr std::string_view
    HeartBeatCountersBase::UploaderType::HttpCompressedBytesUploadedName;
constexpr std::string_view
    HeartBeatCountersBase::UploaderType::HttpRepeatedUploadFailureDroppedName;

constexpr std::string_view HeartBeatCountersBase::StorageType::DbDroppedCounterName;
constexpr std::string_view HeartBeatCountersBase::StorageType::DbDroppedFullCounterName;
constexpr std::string_view HeartBeatCountersBase::StorageType::DbDroppedFailureCounterName;
constexpr std::string_view HeartBeatCountersBase::StorageType::DbCriticalDroppedCounterName;
constexpr std::string_view HeartBeatCountersBase::StorageType::DbLifetimeResetCounterName;
constexpr std::string_view HeartBeatCountersBase::StorageType::DbResetCounterName;
constexpr std::string_view
    HeartBeatCountersBase::StorageType::DbCriticalOverflowEntersCounterName;
constexpr std::string_view
    HeartBeatCountersBase::StorageType::DbEnteringCriticalOverflowDroppedCounterName;
constexpr std::string_view
    HeartBeatCountersBase::StorageType::DbCriticalDataDroppedCounterName;
constexpr std::string_view HeartBeatCountersBase::StorageType::DbResetSizeSumName;
constexpr std::string_view HeartBeatCountersBase::StorageType::DbEventsPersistedCounterName;

constexpr std::string_view
    HeartBeatCountersBase::ContainerType::ContainerBufferFullDropCounterName;
constexpr std::string_view HeartBeatCountersBase::ContainerType::EventsForwardedToHostName;
constexpr std::string_view
    HeartBeatCountersBase::ContainerType::HostConnectionTimeoutsCounterName;
constexpr std::string_view
    HeartBeatCountersBase::ContainerType::HostConnectionErrorsCounterName;
constexpr std::string_view
    HeartBeatCountersBase::ContainerType::LastHostConnectionErrorName;
constexpr std::string_view
    HeartBeatCountersBase::ContainerType::HalRegisterReceiverCountName;
constexpr std::string_view HeartBeatCountersBase::ContainerType::HalConnectionFailuresName;
constexpr std::string_view HeartBeatCountersBase::ContainerType::HalBufferFullDropsName;
constexpr std::string_view HeartBeatCountersBase::ContainerType::HalBufferOtherDropsName;
constexpr std::string_view
    HeartBeatCountersBase::ContainerType::ProxyConnectionFailuresName;
constexpr std::string_view HeartBeatCountersBase::ContainerType::ProxyConnectionSuccessName;
constexpr std::string_view
    HeartBeatCountersBase::ContainerType::ProxyMissingCallbackDropsName;

constexpr std::string_view HeartBeatCountersBase::HeartBeatSequenceNumberName;

//
// Defines the collection of heartbeat counters collected by this class,
// and passed to the uploader.
//
// N.B. HeartBeatCounters cannot be zero'd, but HeartBeatCountersBase can
// Use 'm_heartBeat = HBC();' to clear.
//
struct HeartBeatCounters : public HeartBeatCountersBase
{
  public:
    // These fields cannot go in the base because we ZeroMemory on those

    // last upload time, as a system_clock::time_point
    DECLARE_HEARTBEAT_FIELD(std::chrono::system_clock::time_point, LastUploadTime);

    // Name of last event to exceed the allowed single-upload event payload size
    DECLARE_HEARTBEAT_FIELD(std::string, LastEventSizeOffender);

    // System Uptime in seconds from /proc/uptime
    DECLARE_HEARTBEAT_FIELD(uint64_t, SystemUptimeSec);

    HeartBeatCounters() noexcept { Clear(); }

    HeartBeatCounters(const HeartBeatCounters& other) noexcept
    {
        *this = other;
    }

    HeartBeatCounters& operator=(const HeartBeatCounters& other) noexcept
    {
        std::memcpy(
            static_cast<HeartBeatCountersBase*>(this),
            static_cast<const HeartBeatCountersBase*>(&other),
            sizeof(HeartBeatCountersBase));

        LastUploadTime = other.LastUploadTime;
        LastEventSizeOffender = other.LastEventSizeOffender;

        return *this;
    }

    void Clear() noexcept
    {
        std::memset(
            static_cast<HeartBeatCountersBase*>(this),
            0,
            sizeof(HeartBeatCountersBase));

        LastUploadTime = std::chrono::system_clock::time_point{};
        LastEventSizeOffender.clear();
    }
};

constexpr std::string_view HeartBeatCounters::LastUploadTimeName;
constexpr std::string_view HeartBeatCounters::LastEventSizeOffenderName;

#undef DECLARE_HEARTBEAT_FIELD

/*++
    Constructor. Reads stored values from registry, default values if
    not existing
--*/
HeartBeat::HeartBeat(std::string_view name, IEventSerializer& serializer)
    : _name(name)
    , _serializer(serializer)
    , _counters(std::make_unique<HeartBeatCounters>())
{
    HeartBeatCounters dataToLoad = LoadPersistedData();
    SetMembersBasedOnLoadedData(dataToLoad);
}

void HeartBeat::StartTimer()
{
    _timerThread = std::thread{ [this] {
        // work is created to prevent run() from returning immediately
        boost::asio::io_service::work work{ _timerIoService };
        _timerIoService.run();
        std::cout << "HeartBeatThread exiting!" << std::endl;
    } };

    Fire();
    ArmTimer();
}

void HeartBeat::ArmTimer()
{
    _timer.expires_from_now(c_timerPeriod);
    _timer.async_wait([this](const boost::system::error_code& ec) {
        if (!ec)
        {
            Fire();
            ArmTimer();
        }
        else
        {
            TraceLoggingWrite(
                g_tlgProv,
                "HeartBeat_SetTimer_TimerError",
                TraceLoggingErrorCode(ec, "ErrorCode"));

            std::cout << "HeartBeat Timer error: " << ec;
        }
    });
}

void HeartBeat::StopTimer()
{
    if (!_timerIoService.stopped())
    {
        _timerIoService.stop();
    }

    if (_timerThread.joinable())
    {
        _timerThread.join();
    }
}

/*++
    Returns a HeartBeatCounters filled in from parsed JSON values from persisted
    file.
--*/
HeartBeatCounters HeartBeat::LoadPersistedData()
{
    HeartBeatCounters counters;

    std::string jsonString;
    std::error_code readJsonFile = Utils::FileSystem::ReadUTF8StringFromFile(
        GetPersistedDataFilePath().string(), jsonString);

    if (FAILED(readJsonFile))
    {
        TraceLoggingWrite(
            g_tlgProv,
            "HeartBeat_LoadPersistedData",
            TraceLoggingErrorCode(readJsonFile, "ErrorCode"));

        return counters;
    }

    auto jsonVal = nlohmann::json::parse(jsonString, nullptr, false);

#define READ_JSON_UINT(fieldName)                          \
    {                                                      \
        auto itr = jsonVal.find(counters.fieldName##Name); \
        if (itr != jsonVal.end())                          \
        {                                                  \
            if (itr->is_number_unsigned())                 \
            {                                              \
                counters.fieldName = *itr;                 \
            }                                              \
        }                                                  \
    }

    READ_JSON_UINT(Ingestion.DecodingDroppedCounter);
    READ_JSON_UINT(Uploader.NetworkAttempts);
    READ_JSON_UINT(Uploader.NetworkErrorCounter);
    READ_JSON_UINT(Uploader.NetworkLastError);
    READ_JSON_UINT(Uploader.NetworkTimeoutCounter);
    READ_JSON_UINT(Uploader.HttpResponsesWithRejectedEvents);
    READ_JSON_UINT(Uploader.HttpResponseFailures);
    READ_JSON_UINT(Uploader.HttpFailures4xx);
    READ_JSON_UINT(Uploader.HttpFailures5xx);
    READ_JSON_UINT(Uploader.HttpEventsUploaded);
    READ_JSON_UINT(Uploader.HttpEventsRejectedCounter);
    READ_JSON_UINT(Uploader.HttpCompressedBytesUploaded);
    READ_JSON_UINT(Uploader.HttpRepeatedUploadFailureDropped);
    READ_JSON_UINT(Storage.DbDroppedCounter);
    READ_JSON_UINT(Storage.DbDroppedFullCounter);
    READ_JSON_UINT(Storage.DbDroppedFailureCounter);
    READ_JSON_UINT(Storage.DbCriticalDroppedCounter);
    READ_JSON_UINT(Storage.DbLifetimeResetCounter);
    READ_JSON_UINT(Storage.DbResetCounter);
    READ_JSON_UINT(Storage.DbCriticalOverflowEntersCounter);
    READ_JSON_UINT(Storage.DbEnteringCriticalOverflowDroppedCounter);
    READ_JSON_UINT(Storage.DbCriticalDataDroppedCounter);
    READ_JSON_UINT(Storage.DbResetSizeSum);
    READ_JSON_UINT(Storage.DbEventsPersistedCounter);
    READ_JSON_UINT(Container.ContainerBufferFullDropCounter);
    READ_JSON_UINT(Container.EventsForwardedToHost);
    READ_JSON_UINT(Container.HostConnectionTimeoutsCounter);
    READ_JSON_UINT(Container.HostConnectionErrorsCounter);
    READ_JSON_UINT(Container.LastHostConnectionError);
    READ_JSON_UINT(Container.HalRegisterReceiverCount);
    READ_JSON_UINT(Container.HalConnectionFailures);
    READ_JSON_UINT(Container.HalBufferFullDrops);
    READ_JSON_UINT(Container.HalBufferOtherDrops);
    READ_JSON_UINT(Container.ProxyConnectionFailures);
    READ_JSON_UINT(Container.ProxyConnectionSuccess);
    READ_JSON_UINT(Container.ProxyMissingCallbackDrops);
    READ_JSON_UINT(HeartBeatSequenceNumber);

#undef READ_JSON_UINT

    // This value is specially deserialized
    {
        auto itr = jsonVal.find(counters.LastUploadTimeName);
        if (itr != jsonVal.end())
        {
            if (itr->is_number_unsigned())
            {
                std::chrono::system_clock::duration durSinceEpoch{ *itr };

                counters.LastUploadTime =
                    std::chrono::system_clock::time_point{ durSinceEpoch };
            }
        }
    }

    return counters;
}

void HeartBeat::SetMembersBasedOnLoadedData(const HeartBeatCounters& persistedData)
{
    *_counters = persistedData;
}

void HeartBeat::SavePersistedData(const HeartBeatCounters& persistedData)
{
    nlohmann::json json;

#define WRITE_JSON_UINT(fieldName)                       \
    json[std::string{ persistedData.fieldName##Name }] = \
        persistedData.fieldName;

    WRITE_JSON_UINT(Ingestion.DecodingDroppedCounter);
    WRITE_JSON_UINT(Uploader.NetworkAttempts);
    WRITE_JSON_UINT(Uploader.NetworkErrorCounter);
    WRITE_JSON_UINT(Uploader.NetworkLastError);
    WRITE_JSON_UINT(Uploader.NetworkTimeoutCounter);
    WRITE_JSON_UINT(Uploader.HttpResponsesWithRejectedEvents);
    WRITE_JSON_UINT(Uploader.HttpResponseFailures);
    WRITE_JSON_UINT(Uploader.HttpFailures4xx);
    WRITE_JSON_UINT(Uploader.HttpFailures5xx);
    WRITE_JSON_UINT(Uploader.HttpEventsUploaded);
    WRITE_JSON_UINT(Uploader.HttpEventsRejectedCounter);
    WRITE_JSON_UINT(Uploader.HttpCompressedBytesUploaded);
    WRITE_JSON_UINT(Uploader.HttpRepeatedUploadFailureDropped);
    WRITE_JSON_UINT(Storage.DbDroppedCounter);
    WRITE_JSON_UINT(Storage.DbDroppedFullCounter);
    WRITE_JSON_UINT(Storage.DbDroppedFailureCounter);
    WRITE_JSON_UINT(Storage.DbCriticalDroppedCounter);
    WRITE_JSON_UINT(Storage.DbLifetimeResetCounter);
    WRITE_JSON_UINT(Storage.DbResetCounter);
    WRITE_JSON_UINT(Storage.DbCriticalOverflowEntersCounter);
    WRITE_JSON_UINT(Storage.DbEnteringCriticalOverflowDroppedCounter);
    WRITE_JSON_UINT(Storage.DbCriticalDataDroppedCounter);
    WRITE_JSON_UINT(Storage.DbResetSizeSum);
    WRITE_JSON_UINT(Storage.DbEventsPersistedCounter);
    WRITE_JSON_UINT(Container.ContainerBufferFullDropCounter);
    WRITE_JSON_UINT(Container.EventsForwardedToHost);
    WRITE_JSON_UINT(Container.HostConnectionTimeoutsCounter);
    WRITE_JSON_UINT(Container.HostConnectionErrorsCounter);
    WRITE_JSON_UINT(Container.LastHostConnectionError);
    WRITE_JSON_UINT(Container.HalRegisterReceiverCount);
    WRITE_JSON_UINT(Container.HalConnectionFailures);
    WRITE_JSON_UINT(Container.HalBufferFullDrops);
    WRITE_JSON_UINT(Container.HalBufferOtherDrops);
    WRITE_JSON_UINT(Container.ProxyConnectionFailures);
    WRITE_JSON_UINT(Container.ProxyConnectionSuccess);
    WRITE_JSON_UINT(Container.ProxyMissingCallbackDrops);
    WRITE_JSON_UINT(HeartBeatSequenceNumber);

#undef WRITE_JSON_UINT

    // This value is specially serialized
    json[std::string{ persistedData.LastUploadTimeName }] =
        persistedData.LastUploadTime.time_since_epoch().count();

    std::string jsonString = json.dump();

    THROW_IF_FAILED(Utils::FileSystem::WriteUTF8StringToFile(
        GetPersistedDataFilePath().string(), jsonString));
}

std::filesystem::path HeartBeat::GetPersistedDataFilePath()
{
    return AppState::GetRuntimeDataDirectory() / "heartBeat.json";
}

/*++
    Destructor. Stores counters in registry.
--*/
HeartBeat::~HeartBeat()
{
    StopTimer();
    SavePersistedData(*_counters);
}

/*++
    Injects a heart beat event in the upload queue and resets the counters for
    next heartbeat.
--*/
void HeartBeat::Fire()
{
    TraceLoggingWrite(g_tlgProv, "HeartBeat_Fire");

    //
    // Sample the counters under a lock. Drop the lock before uploading
    // to avoid deadlocking with the event writer thread.
    //
    std::unique_ptr<HeartBeatCounters> sample;
    {
        std::unique_lock<std::recursive_mutex> lock(_countersMutex);

        //
        // Allow the serializer to add it's own heartbeat information
        // on-demand (useful for polling other systems that aren't part of ADHS)
        //
        _serializer.PollHeartbeatStatistics();

        sample = std::make_unique<HeartBeatCounters>(*_counters);
        _counters->Clear();
    }

    uint32_t newSequenceNumber = sample.get()->HeartBeatSequenceNumber + 1;

    // Read current system uptime and update struct
    sample.get()->SystemUptimeSec = Utils::Time::GetSystemUptime().count();

    std::chrono::system_clock::time_point lastHeartbeatTimestamp;
    bool logSucceeded = LogSynthetic(*sample, lastHeartbeatTimestamp);

    {
        std::unique_lock<std::recursive_mutex> lock(_countersMutex);
        if (logSucceeded)
        {
            // Reset counters for next event, update upload time, update
            // sequence number
            _counters->LastUploadTime = lastHeartbeatTimestamp;
            _counters->HeartBeatSequenceNumber = newSequenceNumber;
            _counters->Storage.DbLifetimeResetCounter =
                sample.get()->Storage.DbLifetimeResetCounter;
        }
        else
        {
            // Failed firing the heartbeat, merge sample with current values
            _counters->Ingestion.DecodingDroppedCounter +=
                sample.get()->Ingestion.DecodingDroppedCounter;

            _counters->Uploader.NetworkAttempts +=
                sample.get()->Uploader.NetworkAttempts;
            _counters->Uploader.NetworkErrorCounter +=
                sample.get()->Uploader.NetworkErrorCounter;
            _counters->Uploader.NetworkTimeoutCounter +=
                sample.get()->Uploader.NetworkTimeoutCounter;
            _counters->Uploader.NetworkLastError =
                _counters->Uploader.NetworkLastError == 0 ?
                sample.get()->Uploader.NetworkLastError :
                _counters->Uploader.NetworkLastError;

            _counters->Uploader.HttpResponsesWithRejectedEvents +=
                sample.get()->Uploader.HttpResponsesWithRejectedEvents;
            _counters->Uploader.HttpResponseFailures +=
                sample.get()->Uploader.HttpResponseFailures;
            _counters->Uploader.HttpFailures4xx +=
                sample.get()->Uploader.HttpFailures4xx;
            _counters->Uploader.HttpFailures5xx +=
                sample.get()->Uploader.HttpFailures5xx;
            _counters->Uploader.HttpEventsUploaded +=
                sample.get()->Uploader.HttpEventsUploaded;
            _counters->Uploader.HttpEventsRejectedCounter +=
                sample.get()->Uploader.HttpEventsRejectedCounter;
            _counters->Uploader.HttpCompressedBytesUploaded +=
                sample.get()->Uploader.HttpCompressedBytesUploaded;
            _counters->Uploader.HttpRepeatedUploadFailureDropped +=
                sample.get()->Uploader.HttpRepeatedUploadFailureDropped;

            _counters->Storage.DbDroppedCounter +=
                sample.get()->Storage.DbDroppedCounter;
            _counters->Storage.DbDroppedFullCounter +=
                sample.get()->Storage.DbDroppedFullCounter;
            _counters->Storage.DbDroppedFailureCounter +=
                sample.get()->Storage.DbDroppedFailureCounter;
            _counters->Storage.DbCriticalDroppedCounter +=
                sample.get()->Storage.DbCriticalDroppedCounter;
            _counters->Storage.DbLifetimeResetCounter =
                sample.get()->Storage.DbLifetimeResetCounter;
            _counters->Storage.DbResetCounter +=
                sample.get()->Storage.DbResetCounter;
            _counters->Storage.DbCriticalOverflowEntersCounter +=
                sample.get()->Storage.DbCriticalOverflowEntersCounter;
            _counters->Storage.DbEnteringCriticalOverflowDroppedCounter +=
                sample.get()->Storage.DbEnteringCriticalOverflowDroppedCounter;
            _counters->Storage.DbCriticalDataDroppedCounter +=
                sample.get()->Storage.DbCriticalDataDroppedCounter;
            _counters->Storage.DbResetSizeSum +=
                sample.get()->Storage.DbResetSizeSum;
            _counters->Storage.DbEventsPersistedCounter =
                sample.get()->Storage.DbEventsPersistedCounter;
            _counters->Container.ContainerBufferFullDropCounter +=
                sample.get()->Container.ContainerBufferFullDropCounter;
            _counters->Container.EventsForwardedToHost +=
                sample.get()->Container.EventsForwardedToHost;
            _counters->Container.HostConnectionTimeoutsCounter +=
                sample.get()->Container.HostConnectionTimeoutsCounter;
            _counters->Container.HostConnectionErrorsCounter +=
                sample.get()->Container.HostConnectionErrorsCounter;
            _counters->Container.LastHostConnectionError =
                sample.get()->Container.LastHostConnectionError;
            _counters->Container.HalRegisterReceiverCount =
                sample.get()->Container.HalRegisterReceiverCount;
            _counters->Container.HalConnectionFailures =
                sample.get()->Container.HalConnectionFailures;
            _counters->Container.HalBufferFullDrops =
                sample.get()->Container.HalBufferFullDrops;
            _counters->Container.HalBufferOtherDrops =
                sample.get()->Container.HalBufferOtherDrops;
            _counters->Container.ProxyConnectionFailures =
                sample.get()->Container.ProxyConnectionFailures;
            _counters->Container.ProxyConnectionSuccess =
                sample.get()->Container.ProxyConnectionSuccess;
            _counters->Container.ProxyMissingCallbackDrops =
                sample.get()->Container.ProxyMissingCallbackDrops;
            _counters->LastEventSizeOffender =
                _counters->LastEventSizeOffender.empty() ?
                sample.get()->LastEventSizeOffender :
                _counters->LastEventSizeOffender;

            _counters->LastUploadTime = sample.get()->LastUploadTime;
        }

        // Save current state
        SavePersistedData(*_counters);
    }
}

/*++
    Logs the HeartBeat synthetic event

Arguments:
    counters - Rest of the counters to actually log.
    lastHeartbeatTimestamp - Receives the timestamp of the logged heartbeat
synthetic event.
--*/
bool HeartBeat::LogSynthetic(
    const HeartBeatCounters& counters,
    std::chrono::system_clock::time_point& lastHeartbeatTimestamp)
{
    try
    {
        std::string eventName = "HeartBeat_";
        eventName += _name;

        auto event = std::make_unique<AsimovSyntheticEvent>(
            eventName,
            TriggerPersistence::Critical,
            TriggerLatency::Normal,
            TelemetryPermission::Required,
            PrivacyTags_ProductAndServicePerformance,
            TriggerTelemetryCategory::CriticalData);

        lastHeartbeatTimestamp = event->GetTimestamp();

#define ADD_PARTC_FIELD(fieldName) \
    event->AddPartCField(counters.fieldName##Name, counters.fieldName);

        ADD_PARTC_FIELD(Ingestion.DecodingDroppedCounter);
        ADD_PARTC_FIELD(Uploader.NetworkAttempts);
        ADD_PARTC_FIELD(Uploader.NetworkErrorCounter);
        ADD_PARTC_FIELD(Uploader.NetworkLastError);
        ADD_PARTC_FIELD(Uploader.NetworkTimeoutCounter);
        ADD_PARTC_FIELD(Uploader.HttpResponsesWithRejectedEvents);
        ADD_PARTC_FIELD(Uploader.HttpResponseFailures);
        ADD_PARTC_FIELD(Uploader.HttpFailures4xx);
        ADD_PARTC_FIELD(Uploader.HttpFailures5xx);
        ADD_PARTC_FIELD(Uploader.HttpEventsUploaded);
        ADD_PARTC_FIELD(Uploader.HttpEventsRejectedCounter);
        ADD_PARTC_FIELD(Uploader.HttpCompressedBytesUploaded);
        ADD_PARTC_FIELD(Uploader.HttpRepeatedUploadFailureDropped);
        ADD_PARTC_FIELD(Storage.DbDroppedCounter);
        ADD_PARTC_FIELD(Storage.DbDroppedFullCounter);
        ADD_PARTC_FIELD(Storage.DbDroppedFailureCounter);
        ADD_PARTC_FIELD(Storage.DbCriticalDroppedCounter);
        ADD_PARTC_FIELD(Storage.DbLifetimeResetCounter);
        ADD_PARTC_FIELD(Storage.DbResetCounter);
        ADD_PARTC_FIELD(Storage.DbCriticalOverflowEntersCounter);
        ADD_PARTC_FIELD(Storage.DbEnteringCriticalOverflowDroppedCounter);
        ADD_PARTC_FIELD(Storage.DbCriticalDataDroppedCounter);
        ADD_PARTC_FIELD(Storage.DbResetSizeSum);
        ADD_PARTC_FIELD(Storage.DbEventsPersistedCounter);
        ADD_PARTC_FIELD(Container.ContainerBufferFullDropCounter);
        ADD_PARTC_FIELD(Container.EventsForwardedToHost);
        ADD_PARTC_FIELD(Container.HostConnectionTimeoutsCounter);
        ADD_PARTC_FIELD(Container.HostConnectionErrorsCounter);
        ADD_PARTC_FIELD(Container.LastHostConnectionError);
        ADD_PARTC_FIELD(Container.HalRegisterReceiverCount);
        ADD_PARTC_FIELD(Container.HalConnectionFailures);
        ADD_PARTC_FIELD(Container.HalBufferFullDrops);
        ADD_PARTC_FIELD(Container.HalBufferOtherDrops);
        ADD_PARTC_FIELD(Container.ProxyConnectionFailures);
        ADD_PARTC_FIELD(Container.ProxyConnectionSuccess);
        ADD_PARTC_FIELD(Container.ProxyMissingCallbackDrops);
        ADD_PARTC_FIELD(HeartBeatSequenceNumber);
        ADD_PARTC_FIELD(LastUploadTime);
        ADD_PARTC_FIELD(SystemUptimeSec);

#undef ADD_PARTC_FIELD

        return _serializer.WriteSyntheticEvent(std::move(event));
    }
    CATCH_LOG();

    return false;
}

/*++
    Increases the counter for server service http attempts.
--*/
void HeartBeat::RecordNetworkAttempt()
{
    std::unique_lock<std::recursive_mutex> lock(_countersMutex);
    _counters->Uploader.NetworkAttempts++;
}

/*++
    Handle std::error_code in the networking stack, increasing
    NetworkErrorCounter and optionally NetworkTimeoutCounter. Also sets
    NetweorkLastError.

Arguments:
    networkResult - std::error_code of unsuccessful networking call to server
--*/
void HeartBeat::RecordNetworkFailure(std::error_code networkResult)
{
    TraceLoggingWrite(
        g_tlgProv,
        "HeartBeat_RecordNetworkFailure",
        TraceLoggingErrorCode(networkResult, "ErrorCode"));

    std::unique_lock<std::recursive_mutex> lock(_countersMutex);
    _counters->Uploader.NetworkErrorCounter++;
    _counters->Uploader.NetworkLastError =
        static_cast<uint32_t>(networkResult.value());

    // TODO: Correctly match against network errors for timeout
    // if (networkResult == TelemetryError::E_WINHTTP_TIMEOUT)
    // {
    //     _counters->Uploader.NetworkTimeoutCounter++;
    // }
}

/*++
    Increases the counter for non-2XX/400 response codes from server

Arguments:
    httpStatusCode - [0, 5XX] HTTP status code from server
--*/
void HeartBeat::RecordHttpResponseFailure(uint32_t httpStatusCode)
{
    TraceLoggingWrite(
        g_tlgProv,
        "HeartBeat_RecordHttpResponseFailure",
        TraceLoggingValue(httpStatusCode, "HttpStatusCode"));

    std::unique_lock<std::recursive_mutex> lock(_countersMutex);

    _counters->Uploader.HttpResponseFailures++;

    if (httpStatusCode / 100 == 4)
    {
        _counters->Uploader.HttpFailures4xx++;
    }
    else if (httpStatusCode / 100 == 5)
    {
        _counters->Uploader.HttpFailures5xx++;
    }
}

/*++
    Increments uploader counters in response to accepted/rejected event counts
    from a server response.

Arguments:
    acceptedEvents          - Count of events accepted
    rejectedEvents          - Count of events rejected
    compressedBytesUploaded - Count of compressed payload bytes in the request
--*/
void HeartBeat::RecordHttpAcceptedAndRejectedEvents(
    uint32_t acceptedEvents,
    uint32_t rejectedEvents,
    uint32_t compressedBytesUploaded)
{
    std::unique_lock<std::recursive_mutex> lock(_countersMutex);

    _counters->Uploader.HttpEventsUploaded += acceptedEvents;
    _counters->Uploader.HttpEventsRejectedCounter += rejectedEvents;
    _counters->Uploader.HttpCompressedBytesUploaded += compressedBytesUploaded;

    if (rejectedEvents > 0)
    {
        _counters->Uploader.HttpResponsesWithRejectedEvents++;
    }
}

/*++
    Increments Events lost due to repeated upload failure

Arguments:
    eventsLost - Number of events lost
--*/
void HeartBeat::RecordHttpRepeatedUploadFailureDroppedEvents(uint32_t eventsLost)
{
    std::unique_lock<std::recursive_mutex> lock(_countersMutex);
    _counters->Uploader.HttpRepeatedUploadFailureDropped += eventsLost;
}

/*++
    Increases the counter for events dropped by the event store.
    Events not added to the db or events deleted from the db when its maximum
    size is hit would be counted as dropped.

Arguments:
    eventCount - Number of events dropped
    criticalDataEventCount - Number of critical data events dropped
    persistence - The TriggerPersistence of the events dropped
    isBecauseFull - true if the events were dropped because storage was full
                    (throttling), false otherwise
--*/
void HeartBeat::RecordDbDropped(
    uint32_t eventCount,
    TriggerPersistence persistence,
    bool isBecauseFull)
{
    std::unique_lock<std::recursive_mutex> lock(_countersMutex);

    _counters->Storage.DbDroppedCounter += eventCount;

    if (isBecauseFull)
    {
        _counters->Storage.DbDroppedFullCounter += eventCount;
    }
    else
    {
        _counters->Storage.DbDroppedFailureCounter += eventCount;
    }

    if (persistence == TriggerPersistence::Critical)
    {
        _counters->Storage.DbCriticalDroppedCounter += eventCount;
    }
}

/*++
    Increases the counter upon reset of an event store due to error detection.

Arguments:
    eventStoreSize - Specifies the size of the event store being reset.
--*/
void HeartBeat::RecordDbReset(uint32_t eventStoreSize)
{
    std::unique_lock<std::recursive_mutex> lock(_countersMutex);
    _counters->Storage.DbResetCounter += 1;
    _counters->Storage.DbLifetimeResetCounter += 1;
    _counters->Storage.DbResetSizeSum += eventStoreSize;
}

/*++
    Increases the counter for bytes dropped when a normal buffer enters critical
    overflow mode, and all its payload needs to be thrown away.  Also increments
    a counter that tracks how many times critical overflow mode has been
    entered, independent of the number of bytes that have been dropped because
    of it.

Arguments:
    bytes - Number of bytes dropped
--*/
void HeartBeat::RecordDbEnteringCriticalOverflow(uint32_t bytes)
{
    std::unique_lock<std::recursive_mutex> lock(_countersMutex);
    _counters->Storage.DbEnteringCriticalOverflowDroppedCounter += bytes;
    _counters->Storage.DbCriticalOverflowEntersCounter++;
}

/*++
    Increases the counter for events ingested.
--*/
void HeartBeat::RecordDbEventPersisted()
{
    std::unique_lock<std::recursive_mutex> lock(_countersMutex);
    _counters->Storage.DbEventsPersistedCounter++;
}

/*++
    Sets the text for the last event to violate the maximum event size list.

    NOTE: Right now this is not strictly the name, but it is a big enough chunk
    of the event to contain the name.

Arguments:
    eventSizeOffender - The full event, or the first 4K characters of the
                                event JSON, whichever comes first.
--*/
void HeartBeat::RecordDbEventSizeOffender(std::string_view eventSizeOffender)
{
    std::unique_lock<std::recursive_mutex> lock(_countersMutex);
    _counters->LastEventSizeOffender = eventSizeOffender;
}

/*++
    Increases the counter for number of events dropped due to container buffer
    being full
--*/
void HeartBeat::RecordContainerBufferFullDropped()
{
    std::unique_lock<std::recursive_mutex> lock(_countersMutex);
    _counters->Container.ContainerBufferFullDropCounter++;
}

/*++
    Increases the counter for events forwarded to host
--*/
void HeartBeat::RecordEventsForwardedToHost()
{
    std::unique_lock<std::recursive_mutex> lock(_countersMutex);
    _counters->Container.EventsForwardedToHost++;
}

/*++
    Increases the counter for failed host connections due to timeout
--*/
void HeartBeat::RecordHostConnectionTimeouts()
{
    std::unique_lock<std::recursive_mutex> lock(_countersMutex);
    _counters->Container.HostConnectionTimeoutsCounter++;
}

/*++
    Increases the counter for failed host connections due to issues other than
    timeout and records the last error seen.

Arguments:
    error - The last error seen as part of a non-timeout connection issue
--*/
void HeartBeat::RecordHostConnectionErrors(std::error_code error)
{
    std::unique_lock<std::recursive_mutex> lock(_countersMutex);
    _counters->Container.HostConnectionErrorsCounter++;
    _counters->Container.LastHostConnectionError = error.value();
}

/*++
    Increases the counter for failed host connections due to issues other than
    timeout and records the last error seen.

Arguments:
    error - The last error seen as part of a non-timeout connection issue
--*/
void HeartBeat::SetContainerHeartbeatFields(
    uint64_t halRegisterReceiverCount,
    uint64_t halConnectionFailures,
    uint64_t halBufferFullDrops,
    uint64_t halBufferOtherDrops,
    uint64_t proxyConnectionFailures,
    uint64_t proxyConnectionSuccess,
    uint64_t proxyMissingCallbackDrops)
{
    std::unique_lock<std::recursive_mutex> lock(_countersMutex);
    _counters->Container.HalRegisterReceiverCount = halRegisterReceiverCount;
    _counters->Container.HalConnectionFailures = halConnectionFailures;
    _counters->Container.HalBufferFullDrops = halBufferFullDrops;
    _counters->Container.HalBufferOtherDrops = halBufferOtherDrops;
    _counters->Container.ProxyConnectionFailures = proxyConnectionFailures;
    _counters->Container.ProxyConnectionSuccess = proxyConnectionSuccess;
    _counters->Container.ProxyMissingCallbackDrops = proxyMissingCallbackDrops;
}

}}}
