// Copyright (c) Microsoft Corporation. All rights reserved.

#include "HttpSessionManager.h"

#include <memory>

#include <azure-device-util/ResultException.h>
#include <azure-device-util/String.h>
#include <cpprest/containerstream.h>
#include <cpprest/http_client.h>

#include "TlgProvider.h"

using namespace web;                   // Common features like URIs.
using namespace web::http;             // Common HTTP functionality
using namespace web::http::client;     // HTTP client features
using namespace concurrency::streams;  // Asynchronous streams

namespace Azure { namespace Device { namespace Telemetry {

/*++
    Transforms each header name/value pair into "{Name}: {Value}\n"
    appended together into a single string.

Return value:
    Result of appending transformed values together.
--*/
std::string FlattenHttpHeaderMap(const HttpHeaderMap& headerMap)
{
    std::string result;
    for (const auto& [name, value] : headerMap)
    {
        result += name;
        result += ": ";
        result += value;
        result += "\n";
    }
    return result;
}

HttpSessionManager::HttpSessionManager() = default;
HttpSessionManager::~HttpSessionManager() = default;

void HttpSessionManager::SetServerUrl(std::string_view serverUrl)
{
    THROW_HR_IF(make_error_code(std::errc::operation_not_permitted), _httpClient);

    http_client_config config;

    if (!_httpProxy.empty())
    {
        web_proxy proxy(_httpProxy);
        config.set_proxy(proxy);
        std::cout << "Proxy set to " << _httpProxy << std::endl;
    }

    _httpClient = std::make_unique<http_client>(std::string{ serverUrl }, config);
}

bool HttpSessionManager::SetProxy(std::string_view proxy) 
{
    if (web::uri::validate(std::string{proxy}) && !proxy.empty())
    {
       _httpProxy = proxy;
       return true;
    }
    
    return false;
}

/*++
    This is a wrapper to do a single HTTP request on the object's httpClient.

Arguments:
    path                - URL path to request
    httpMethod          - HTTP verb to use (GET, POST)
    headers             - Header name/value vector to attach to the request
    payload             - Data payload to be sent with the request for allowed
                            verbs (POST)
    cancellationToken   - Token that will observe a cancellation request to
                            abort the request before completion
    response            - Out parameter containing the response headers/payload
                            in the case of network success
--*/
std::error_code HttpSessionManager::DoSynchronousHttpRequest(
    std::string_view path,
    const HttpQueryMap& queryParams,
    HttpMethod httpMethod,
    const HttpHeaderMap& headers,
    const std::vector<std::byte>& payload,
    const pplx::cancellation_token& cancellationToken,
    HttpResponse& response)
{
    try
    {
        FAIL_FAST_IF(!_httpClient);
        FAIL_FAST_IF(httpMethod == HttpMethod::GET && !payload.empty());

        method method;
        switch (httpMethod)
        {
        case HttpMethod::GET:
            method = methods::GET;
            break;
        case HttpMethod::POST:
            method = methods::POST;
            break;
        default:
            FAIL_FAST();
        }

        const std::vector<unsigned char>& payloadCorrectType =
            *reinterpret_cast<const std::vector<unsigned char>*>(&payload);

        // Build request URI and start the request.
        uri_builder builder;
        builder.set_path(std::string{ path });

        for (const auto& [name, value] : queryParams)
        {
            builder.append_query(name, value, true);
        }

        http_request request{ method };
        request.set_request_uri(builder.to_uri());

        // Add headers
        for (auto& [name, value] : headers)
        {
            request.headers().add(name, value);
        }

        // Now set the body, to prevent additional Content-Type field
        request.set_body(payloadCorrectType);

        std::cout << std::endl
                  << "--Sending request with headers--" << std::endl;

        std::string flatHeaders = FlattenHttpHeaderMap(headers);
        std::cout << flatHeaders << std::endl;

        TraceLoggingWrite(
            g_tlgProv,
            "HttpSessionManager_Request",
            TraceLoggingValue(flatHeaders.c_str(), "RequestHeaders"));

        http_response httpResponse =
            _httpClient->request(request, cancellationToken).get();

        container_buffer<std::string> bodyBuffer;
        httpResponse.body().read_to_end(bodyBuffer).get();

        HttpHeaderMap headerMap;
        for (const auto& [name, value] : httpResponse.headers())
        {
            headerMap.emplace(name, value);
        }

        std::string etag;
        {
            auto it = headerMap.find("Etag");
            if (it != headerMap.end())
            {
                etag = it->second;
            }
        }

        std::string retryAfter;
        {
            auto it = headerMap.find("Retry-After");
            if (it != headerMap.end())
            {
                retryAfter = it->second;
            }
        }

        response.Data = bodyBuffer.collection();
        response.ETag = etag;
        response.Headers = std::move(headerMap);
        response.RetryAfterSeconds = Utils::String::ToUIntT<uint32_t>(retryAfter);
        response.StatusCode = httpResponse.status_code();

        TraceLoggingWrite(
            g_tlgProv,
            "HttpSessionManager_Response",
            TraceLoggingValue(response.Data.c_str(), "ReponseString"),
            TraceLoggingValue(
                FlattenHttpHeaderMap(response.Headers).c_str(), "ResponseHeaders"),
            TraceLoggingValue(response.RetryAfterSeconds, "RetryAfterSeconds"),
            TraceLoggingValue(response.StatusCode, "StatusCode"),
            TraceLoggingValue(response.ETag.c_str(), "ETag"));

        return {};
    }
    CATCH_RETURN();
}

}}}
