// Copyright (c) Microsoft Corporation. All rights reserved.

#include "KillSwitchManager.h"

#include "TimeService.h"
#include "TlgProvider.h"

namespace Azure { namespace Device { namespace Telemetry {

KillSwitchManager::KillSwitchManager(ITimeService& timeService)
    : _timeService(timeService)
{}

/*++
    Tests whether a particular tenantToken + persistence value has a kill
    effect in place.

Arguments:
    tenantToken - Token to test
    persistence - Persistence of data to test.

Return Value:
    true - kill is in place, false otherwise.
--*/
bool KillSwitchManager::IsKilled(
    std::string_view tenantToken,
    TriggerPersistence persistence)
{
    std::unique_lock<std::recursive_mutex> lock{ _mutex };

    // If no entry found, no kill
    auto itr = _killStates.find(tenantToken);
    if (itr == _killStates.end())
    {
        return false;
    }

    // If expired, no kill.  Remove expired entry.
    if (itr->second.UntilTime < _timeService.SteadyNow())
    {
        _killStates.erase(itr);
        return false;
    }

    // If kill request is for less persistent things that this event, no kill.
    if (itr->second.Persistence < persistence)
    {
        return false;
    }

    // Request is valid and applies to this event, kill.
    return true;
}

/*++
    Add a kill request for the given tenant token.  This will prevent persisting
    all events with this tenant token and persistence until the duration has
    expired.

    This will overwrite any existing request on the same tenant token.

Arguments:
    tenantToken     - Token to kill
    killMode        - Persistence of data to kill.  Anything equal or lower.
    killDuration    - Number of seconds to keep the kill request in place
--*/
void KillSwitchManager::SetKillRequest(
    std::string tenantToken,
    TriggerPersistence killMode,
    std::chrono::seconds killDuration)
{
    std::unique_lock<std::recursive_mutex> lock{ _mutex };

    TraceLoggingWrite(
        g_tlgProv,
        "KillSwitchManager_SetKillRequest",
        TraceLoggingValue(tenantToken.c_str(), "TenantToken"),
        TraceLoggingValue(static_cast<int8_t>(killMode), "Persistence"),
        TraceLoggingValue(killDuration.count(), "KillDurationSec"));

    KillState killState;
    killState.Persistence = killMode;
    killState.UntilTime = _timeService.SteadyNow() + killDuration;

    _killStates[std::move(tenantToken)] = killState;
};

}}}
