// Copyright (c) Microsoft Corporation. All rights reserved.

#include <azure-device-util/AndroidLogHelper.h>
#include <azure-device-util/ConversionFactors.h>
#include <azure-device-util/FileSystem.h>
#include <azure-device-util/ResultException.h>
#include <azure-device-util/SignalLoop.h>
#include <azure-device-util/TomlConfig.h>
#include <azure-device-util/Version.h>

#ifdef AZURE_DEVICE_TELEMETRY_USE_UPLOADER_HTTP
    #include "HttpSessionManager.h"
    #include "IAsimovUploadWorkItemFactory.h"
    #include "IHttpSessionManager.h"
#endif

#ifdef AZURE_DEVICE_TELEMETRY_USE_SYSTEMD_BASED_SIGNALLING
    #include <systemd/sd-daemon.h>
#endif

#include "AppState.h"
#include "AsimovEvent.h"
#include "AsimovEventBuffer.h"
#include "AsimovSyntheticEvent.h"
#include "ConfigPaths.h"
#include "ConsumerFactory.h"
#include "DirtyShutdownMonitor.h"
#include "EventStoreFactory.h"
#include "HeartBeat.h"
#include "IEventConsumer.h"
#include "IEventSerializer.h"
#include "IEventStore.h"
#include "IUploadScheduler.h"
#include "KillSwitchManager.h"
#include "LocalPolicyManager.h"
#include "RemotePolicyManager.h"
#include "SerializerFactory.h"

#ifdef AZURE_DEVICE_TELEMETRY_DO_SETTINGS_DOWNLOADS
    #include "SettingsManager.h"
#endif

#include "TimeService.h"
#include "TlgProvider.h"
#include "UploaderFactory.h"

using namespace Azure::Device::Utils;
using namespace Azure::Device::Policy;

namespace Azure { namespace Device { namespace Telemetry {

/*++
    Registered as our error handling callback with the ErrorHelper from
    azure-device-utils package. Logs to LTTNG.

Parameters:
    message - The message to log.
    file - The file name (in debug builds) which generated the error.
    line - The line number (in debug builds) which generated the error.
    err - The error code generated.
--*/
static void LogCallback(
    std::string_view message,
    std::string_view file,
    int lineNumber,
    std::error_code err)
{
    TraceLoggingWrite(
        g_tlgProv,
        "LogError",
        TraceLoggingStringView(file, "FileName"),
        TraceLoggingValue(lineNumber, "Line"),
        TraceLoggingStringView(message, "Message"),
        TraceLoggingErrorCode(err, "ErrorCode"));

    ADHS_LOG(
        "LogError: %.*s::%d, %.*s %d",
        static_cast<int>(file.size()),
        file.data(),
        lineNumber,
        static_cast<int>(message.size()),
        message.data(),
        err.value());

    std::cerr << "LogError: ";
    std::cerr << "loc='" << file << "@" << lineNumber << "' ";
    std::cerr << "msg='" << message << "' ";
    std::cerr << "err={msg='" << err.message() << "' ";
    std::cerr << "cat='" << err.category().name() << "' ";
    std::cerr << "val='" << err.value() << "'}" << std::endl;
}

#ifdef AZURE_DEVICE_TELEMETRY_USE_UPLOADER_HTTP
void InitializeStaticNetworkingThreadpool()
{
    // Set the threadpool to a more reasonable size on cpprest versions that
    // allow it
    #if (CPPREST_VERSION >= (2 * 100000 + 10 * 100 + 8))
    crossplat::threadpool::initialize_with_threads(10);
    #endif
}
#endif

//
// Creates the telemetry configuration, returns nullptr and logs any failure
// Also validates the existence of required fields:
// * Telemetry_TenantToken
//
std::unique_ptr<TomlConfig> CreateTelemetryConfig()
{
    std::filesystem::path runtimeDataPath = FileSystem::GetConfDirectory();
    runtimeDataPath /= "azure-device-telemetryd/config.toml";

    std::unique_ptr<TomlConfig> configService;
    try
    {
        configService = std::make_unique<TomlConfig>(runtimeDataPath.string());
    }
    catch (const std::exception& ex)
    {
        std::cout << "Failed to parse required telemetry conf file from: "
                  << runtimeDataPath.string() << std::endl;
        std::cout << "Details:" << std::endl;
        std::cout << "    " << ex.what() << std::endl;
        return nullptr;
    }

    auto validateResult = configService->ValidateConfigPathExistsString(
        Config::Telemetry_TenantToken);
    switch (validateResult)
    {
    case IConfig::ValidateStatus::PathNotFound: {
        std::cout << "Failed to find " << Config::Telemetry_TenantToken
                  << " in " << runtimeDataPath.string() << std::endl;
        return nullptr;
    }
    case IConfig::ValidateStatus::TypeError: {
        std::cout << "Config value for " << Config::Telemetry_TenantToken
                  << " was not of type string" << std::endl;
        return nullptr;
    }
    case IConfig::ValidateStatus::Ok:
        break;
    }

    return configService;
}

static std::vector<std::unique_ptr<IEventConsumer>>
CreateEventConsumers(IEventSerializer& serializer, IPolicyManager& policyManager)
{
    std::vector<std::unique_ptr<IEventConsumer>> consumers;

    (void) serializer;
    (void) policyManager;

#ifdef AZURE_DEVICE_TELEMETRY_USE_CONSUMER_LTTNG
    consumers.emplace_back(ConsumerFactory::CreateLttngConsumer(
        serializer, policyManager, std::chrono::milliseconds{ 50 }));
#endif

#ifdef AZURE_DEVICE_TELEMETRY_USE_CONSUMER_JOURNALD
    consumers.emplace_back(ConsumerFactory::CreateJournalDConsumer(
        serializer, policyManager, std::chrono::milliseconds{ 100 }));
#endif

#ifdef AZURE_DEVICE_TELEMETRY_USE_CONSUMER_HWBINDER
    consumers.emplace_back(
        ConsumerFactory::CreateHwBinderConsumer(serializer, policyManager));
#endif

#ifdef AZURE_DEVICE_TELEMETRY_USE_CONSUMER_LOGCAT
    consumers.emplace_back(
        ConsumerFactory::CreateLogcatConsumer(serializer, policyManager));
#endif

    return consumers;
}

/*++
    Wire up all telemetry subsystems in the correct order.
--*/
int RunTelemetryClient()
{
    TraceLoggingRegister(g_tlgProv);

    Azure::Device::Utils::SetResultLoggingCallback(
        Azure::Device::Telemetry::LogCallback);

    SignalLoop signalLoop{ SIGTERM, SIGINT, SIGHUP };

    std::string_view version = Version::GetString();
    TraceLoggingWrite(
        g_tlgProv, "Main_Startup", TraceLoggingStringView(version, "Version"));

    std::cout << "Staring azure-device-telemetryd with version: " << version
              << std::endl;

    std::filesystem::path dataDirectory = AppState::GetRuntimeDataDirectory();
    FAIL_FAST_IF_FAILED(
        Utils::FileSystem::CreateDirectoryAndParents(dataDirectory.string()));

    // Now that we have a directory to work in, check for a running flag
    bool isDirtyShutdown = false;
    DirtyShutdownMonitor monitor;
    if (monitor.RunningFlagExists())
    {
        ADHS_LOG("Service did not shutdown cleanly");
        std::cout << "Service did not shutdown cleanly" << std::endl;
        isDirtyShutdown = true;
    }
    else
    {
        monitor.SetRunningFlag();
    }

    auto configService = CreateTelemetryConfig();
    if (!configService)
    {
        return 1;
    }

    TimeService timeService;
    KillSwitchManager killSwitchManager{ timeService };

    std::unique_ptr<IPolicyManager> policyManager;
    std::unique_ptr<IEventSerializer> serializer;

#ifdef AZURE_DEVICE_TELEMETRY_USE_SERIALIZER_ASIMOV
    policyManager = std::make_unique<LocalPolicyManager>();
    serializer = SerializerFactory::CreateAsimovSerializer(
        *policyManager, killSwitchManager, *configService);
#elif AZURE_DEVICE_TELEMETRY_USE_SERIALIZER_BONDI
    policyManager = std::make_unique<LocalPolicyManager>();
    serializer = SerializerFactory::CreateBondiSerializer(
        *policyManager, killSwitchManager, *configService);
#else  // AZURE_DEVICE_TELEMETRY_USE_SERIALIZER_HVSOCKET
    auto remotePolicyManager = std::make_unique<RemotePolicyManager>();
    serializer = SerializerFactory::CreateHvSocketSerializer(
        *remotePolicyManager, killSwitchManager, *configService);
    policyManager = std::move(remotePolicyManager);
#endif

    policyManager->Update();

    FAIL_FAST_IF(serializer == nullptr);
    HeartBeat heartBeat{ "Linux", *serializer };
    serializer->SetHeartBeat(heartBeat);

#ifdef AZURE_DEVICE_TELEMETRY_DO_SETTINGS_DOWNLOADS
    SettingsManager settingsManager{ heartBeat, *serializer, *configService };
    settingsManager.StartTimers();
#endif

    std::unique_ptr<IEventStore> eventStore;

#ifdef AZURE_DEVICE_TELEMETRY_USE_STORAGE_SQLITE
    uint32_t sqliteStoreBytes = configService->GetInt(
        Config::Telemetry_EventStoreSizeBytes, 32 * ConversionFactors::mbToBytes);

    AsimovEventBufferFactory eventBufferFactory;
    eventStore = EventStoreFactory::CreateSqliteEventStore(
        heartBeat,
        *serializer,
        eventBufferFactory,
        dataDirectory.string(),
        sqliteStoreBytes,
        AsimovEventBuffer::GetNormalEventBufferMaxEvents(),
        AsimovEventBuffer::GetCriticalEventBufferMaxEvents(),
        AsimovEventBuffer::GetEventBufferMaxBytes());
#endif

    // Some serializers do not require an underlying event store, so passing
    // null here is ok.
    serializer->BeginSerializing(eventStore.get());

    //
    // This is the earliest we can log an event here. Unfortunately it means we
    // can lose some dirty shutdown metrics if the service is failing to startup
    // very early in the boot process.
    //
    if (isDirtyShutdown)
    {
        auto event =
            std::make_unique<AsimovSyntheticEvent>("FailedToShutdownCorrectly");
        serializer->WriteEvent(std::move(event));
    }

    std::unique_ptr<IUploadScheduler> uploadScheduler;
#ifdef AZURE_DEVICE_TELEMETRY_USE_UPLOADER_HTTP
    InitializeStaticNetworkingThreadpool();

    HttpSessionManager uploaderHttpSessionManager;
    AsimovUploadWorkItemFactory asimovUploadWorkItemFactory;

    using namespace std::string_view_literals;
    constexpr std::string_view c_uploadUrl = "https://"
                                             "adhs.events.data.microsoft.com/"
                                             "OneCollector/1.0/"sv;

    // Using HTTP Uploader requires an event store to buffer until upload
    // happens.
    FAIL_FAST_IF(!eventStore);

    const char* httpsProxy = std::getenv("HTTPS_PROXY");
    if (httpsProxy == nullptr)
    {
        httpsProxy = std::getenv("https_proxy");
    }

    if (httpsProxy != nullptr && !uploaderHttpSessionManager.SetProxy(httpsProxy))
    {
        std::cout << "Invalid URI " << httpsProxy
                  << " used for proxy, proxy was not set" << std::endl;
    }

    uploadScheduler = UploaderFactory::CreateHttpUploader(
        c_uploadUrl,
        heartBeat,
        *serializer,
        *eventStore,
        uploaderHttpSessionManager,
        asimovUploadWorkItemFactory);

    FAIL_FAST_IF(uploadScheduler == nullptr);
    uploadScheduler->StartTimers();
#endif

    // Note that uploadScheduler may be nullptr here if not using an uploader
    serializer->BeginSettingUploadTimers(uploadScheduler.get());

    heartBeat.StartTimer();

    auto eventConsumers = CreateEventConsumers(*serializer, *policyManager);
    for (const auto& consumer : eventConsumers)
    {
        consumer->StartConsuming();
    }

#ifdef AZURE_DEVICE_TELEMETRY_USE_SYSTEMD_BASED_SIGNALLING
    sd_notify(0, "READY=1");
    TraceLoggingWrite(g_tlgProv, "Main_NotifiedSystemd");
#endif  // AZURE_DEVICE_TELEMETRY_USE_SYSTEMD_BASED_SIGNALLING

    auto shutdownHandler = [](int signal) {
        std::cout << "Shutting down due to signal " << signal << std::endl;
        return false;
    };

    auto reloadHandler = [&](int signal) {
        std::cout << "Reloading configuration because of signal " << signal
                  << std::endl;

        //
        // Update azure-device-health-services config
        // Only do this if we support local update by signal
        //
        if (policyManager != nullptr)
        {
            policyManager->Update();
        }

        // Update telemetry config
        configService = CreateTelemetryConfig();
        if (!configService)
        {
            return true;
        }
        serializer->ReloadConfiguration(*configService);

        // Update derived consumer configs
        for (auto& consumer : eventConsumers)
        {
            consumer->ReloadConfiguration();
        }

        return true;
    };

    signalLoop.RegisterSignalHandler(SIGTERM, shutdownHandler);
    signalLoop.RegisterSignalHandler(SIGINT, shutdownHandler);
    signalLoop.RegisterSignalHandler(SIGHUP, reloadHandler);

    // This will return gracefully on SIGINT or SIGTERM
    signalLoop.Run();

    // Shutdown/Cleanup block
    for (const auto& consumer : eventConsumers)
    {
        consumer->StopConsuming();
    }

    for (const auto& consumer : eventConsumers)
    {
        consumer->WaitForCompletion();
    }

    if (uploadScheduler != nullptr)
    {
        uploadScheduler->EndTimers();
    }

    heartBeat.StopTimer();

#ifdef AZURE_DEVICE_TELEMETRY_DO_SETTINGS_DOWNLOADS
    settingsManager.EndTimers();
#endif

    monitor.ClearRunningFlag();

    std::cout << "Exiting gracefully!" << std::endl;

    TraceLoggingUnregister(g_tlgProv);

    return 0;
}
}}}

int main()
{
    int rc = 0;
    try
    {
        rc = Azure::Device::Telemetry::RunTelemetryClient();
    }
    catch (std::exception& ex)
    {
        std::cout << "Telemetry daemon exiting with uncaught exception:"
                  << std::endl;
        std::cout << ex.what() << std::endl;
        rc = 1;
    }
    catch (...)
    {
        std::cout
            << "Telemetry daemon exiting with uncaught non-exception object."
            << std::endl;
        rc = 1;
    }

    return rc;
}