// Copyright (c) Microsoft Corporation. All rights reserved.

#include "AsimovEventSerializer.h"

#include <iostream>
#include <memory>

#include <azure-device-util/ResultException.h>

#include "AsimovEvent.h"
#include "AsimovEventBuffer.h"
#include "AsimovSharedPartAState.h"
#include "AsimovSyntheticEvent.h"
#include "HeartBeat.h"
#include "IEventStore.h"
#include "IPolicyManager.h"
#include "IUploadScheduler.h"
#include "KillSwitchManager.h"
#include "TlgProvider.h"

using namespace Azure::Device::Policy;

namespace Azure { namespace Device { namespace Telemetry {

AsimovEventSerializer::AsimovEventSerializer(
    IPolicyManager& policyManager,
    IKillSwitchManager& killSwitchManager,
    Utils::IConfig& config)
    : _policyManager(policyManager)
    , _killSwitchManager(killSwitchManager)
    , _partAState(std::make_unique<AsimovSharedPartAState>(config))
{}

AsimovEventSerializer::~AsimovEventSerializer() {}

/*++
    Signal that serialization of events to storage can begin.  Saves a reference
    to the passed in IEventStore and expects that reference to remain good for
    the lifetime of this object.

    Writes all queued synthetic events that happened before this call to storage
    and clears the queue, which is no longer needed.

Arguments:
    eventStore - IEventStore to serialize events to
--*/
void AsimovEventSerializer::BeginSerializing(IEventStore* eventStore)
{
    std::lock_guard<std::recursive_mutex> lock{ _queuedEventMutex };

    // This should only be set once
    FAIL_FAST_IF(_eventStore != nullptr);

    // The passed event store should never be null
    FAIL_FAST_IF(eventStore == nullptr);
    _eventStore = eventStore;

    for (auto& synthEvent : _queuedEvents)
    {
        WriteEvent(std::move(synthEvent));
    }

    _queuedEvents.clear();
}

/*++
    Signal that upload timers are now safe to set when new events come in.

Arguments:
    uploadQueue - AsimovUploadQueue to upload
--*/
void AsimovEventSerializer::BeginSettingUploadTimers(IUploadScheduler* uploadScheduler)
{
    _uploadScheduler = uploadScheduler;
}

/*++
    Stores a synthetic event to the event database.  If no database has yet
    been provided, it will instead store to an internal queue to be drained
    when the event store is ready.

Arguments:
    event - AsimovSyntheticEvent to store
    options -  Additional flags governing event storage options

Return Value:
    true if event was stored successfully, false otherwise
--*/
bool AsimovEventSerializer::WriteSyntheticEvent(
    std::unique_ptr<AsimovSyntheticEvent> event,
    AsimovEventSerializer::WriteEventOptions options)
{
    {
        std::lock_guard<std::recursive_mutex> lock{ _queuedEventMutex };

        if (_eventStore == nullptr)
        {
            _queuedEvents.emplace_back(std::move(event));
            return true;
        }
    }

    return WriteEvent(std::move(event), options);
}

/*++
    Stores an event to the event database.  Also signals the upload
    timer to begin so an upload job will eventually upload the event.

Arguments:
    event - AsimovSyntheticEvent to store
    options -  Additional flags governing event storage options

Return Value:
    true if event was stored successfully, false otherwise
--*/
bool AsimovEventSerializer::WriteEvent(
    std::unique_ptr<AsimovEvent> event,
    AsimovEventSerializer::WriteEventOptions options)
{
    TelemetryPermission telemetryPermissionSnapshot =
        _policyManager.GetTelemetryPermission();
    event->SetTelemetryPermissionAtCreation(telemetryPermissionSnapshot);

    if (event->GetMinimumNeededTelemetryPermission() > telemetryPermissionSnapshot)
    {
        TraceLoggingWrite(
            g_tlgProv,
            "AsimovEventSerializer_EventRejectedForNotMeetingTelemetryPermissionPolicy",
            TraceLoggingCountedString(
                event->GetName().data(), event->GetName().size(), "EventName"),
            TraceLoggingValue(
                static_cast<int8_t>(event->GetMinimumNeededTelemetryPermission()),
                "MinimumNeededTelemetryPermission"),
            TraceLoggingValue(
                static_cast<int8_t>(telemetryPermissionSnapshot),
                "CurrentTelemetryPermission"));

        return false;
    }

    if (!IsSampledIn(*event))
    {
        // Logging happens in child function
        return false;
    }

    auto partAState = IncrementAndCopyPartAState();
    std::string_view tenantToken = partAState.GetTenantToken();

    if (_killSwitchManager.IsKilled(tenantToken, event->GetPersistence()))
    {
        TraceLoggingWrite(
            g_tlgProv,
            "AsimovEventSerializer_EventRejectedForBeingKillSwitched",
            TraceLoggingCountedString(
                event->GetName().data(), event->GetName().size(), "EventName"),
            TraceLoggingCountedString(
                tenantToken.data(), tenantToken.size(), "TenantToken"),
            TraceLoggingValue(
                static_cast<int8_t>(event->GetPersistence()), "Persistence"));

        return false;
    }

    //
    // Prepare a lock that will be taken part-way through Render and be held
    // through WriteEventToDatabase.  This is because some fields in an
    // event JSON depend on the exact buffer the event is written in to.  To
    // make sure that is consistent, we must own the DB through both.
    //
    {
        std::unique_lock<std::recursive_mutex> eventStoreLock(
            _eventStore->GetEventStoreMutex(), std::defer_lock);

        std::string fullEventPayload =
            event->Render(partAState, *_eventStore, eventStoreLock);

        std::cout << std::endl
                  << "--Persisting event--" << std::endl
                  << fullEventPayload << std::endl;

        TraceLoggingWrite(
            g_tlgProv,
            "AsimovEventSerializer_PersistEvent",
            TraceLoggingValue(fullEventPayload.c_str(), "EventPayload"),
            TraceLoggingValue(static_cast<int8_t>(event->GetLatency()), "Latency"),
            TraceLoggingValue(
                static_cast<int8_t>(event->GetPersistence()), "Persistence"));

        if (_uploadScheduler != nullptr && !options.SkipArmTimer)
        {
            _uploadScheduler->ArmTimer(event->GetLatency());
        }

        return WriteEventToDatabase(
            std::move(fullEventPayload),
            event->GetLatency(),
            event->GetPersistence());
    }
}

AsimovSharedPartAState AsimovEventSerializer::IncrementAndCopyPartAState()
{
    // Here we increment our member PartAState to account for another event
    // arriving We also make a copy of PartAState, also under the lock, and
    // return that to the caller to use safely, even if another thread updates
    // PartAState.
    std::lock_guard<std::recursive_mutex> lock{ _partAStateMutex };
    auto partAStateCopy = *_partAState;
    _partAState->IncrementEventCount();
    return partAStateCopy;
}

void AsimovEventSerializer::ReloadConfiguration(Utils::IConfig& config)
{
    std::lock_guard<std::recursive_mutex> lock{ _partAStateMutex };
    _partAState->ReloadConfiguration(config);
}

void AsimovEventSerializer::UpdateTelemetrySettings(
    std::unique_ptr<ITelemetrySettings> telemetrySettings)
{
    std::lock_guard lock{ _telemetrySettingsMutex };
    _telemetrySettings = std::move(telemetrySettings);
}

bool AsimovEventSerializer::IsSampledIn(const AsimovEvent& event)
{
    // TODO 31296666: Need way to pull provider name and event name so this
    // doesn't break for people who put dots in their event name
    auto fullEventName = event.GetName();
    size_t lastDotIndex = fullEventName.rfind('.');
    if (lastDotIndex == std::string::npos)
    {
        TraceLoggingWrite(
            g_tlgProv,
            "AsimovEventSerializer_RejectedEventWithNoPeriod",
            TraceLoggingStringView(fullEventName, "FullEventName"));

        return false;
    }

    auto providerName = fullEventName.substr(0, lastDotIndex);
    auto eventOnlyName = fullEventName.substr(lastDotIndex + 1);

    TelemetrySettingOverrides telemetryOverrides = [&]() {
        std::lock_guard lock{ _telemetrySettingsMutex };
        FAIL_FAST_IF(_telemetrySettings == nullptr);

        return _telemetrySettings->GetTelemetryEventSettings(
            event.GetMinimumNeededTelemetryPermission(),
            providerName,
            eventOnlyName);
    }();

    // If there is no explicit rule, the event is rejected.
    // This should be very unusual, as we expect default rules
    // at least to exist for every event.
    if (!telemetryOverrides.SampleRate)
    {
        TraceLoggingWrite(
            g_tlgProv,
            "AsimovEventSerializer_RejectedEventWithNoApplicableRules",
            TraceLoggingStringView(fullEventName, "FullEventName"));

        return false;
    }

    double deviceSampleRate = _partAState->GetDeviceSampleRate();
    double eventSampleRate = *telemetryOverrides.SampleRate;
    if (deviceSampleRate + eventSampleRate <= 100.0)
    {
        TraceLoggingWrite(
            g_tlgProv,
            "AsimovEventSerializer_EventRejectedForBeingSampledOut",
            TraceLoggingStringView(event.GetName(), "EventName"),
            TraceLoggingValue(eventSampleRate, "EventSampleRate"),
            TraceLoggingValue(deviceSampleRate, "DeviceSampleRate"));

        return false;
    }

    return true;
}

/*++
    Saves a pointer to HeartBeat for the serializer to write stats to.

Arguments:
    heartBeat - HeartBeat for the serializer to write stats to
--*/
void AsimovEventSerializer::SetHeartBeat(IHeartBeat& heartBeat)
{
    FAIL_FAST_IF(_heartBeat != nullptr);
    _heartBeat = &heartBeat;
}

/*++
    Writes an event to the event store.
    Events with illegal size are rejected.

Arguments:
    event       - event to write
    latency     - latency of event
    persistence - persistence of event

Return Value:
    true if successful, false otherwise
--*/
bool AsimovEventSerializer::WriteEventToDatabase(
    std::string&& event,
    TriggerLatency latency,
    TriggerPersistence persistence)
{
    UploadData uploadData;
    uploadData.Latency = latency;
    uploadData.Persistence = persistence;
    uploadData.Payload = std::move(event);

    //
    // If this event's payload exceeds k_maxSingleEventSizeInBytes, it cannot be
    // stored. If this ever becomes common, we can do a bit more work to make
    // the reported string more easily traceable to a given owner.
    //
    if (uploadData.Payload.length() > AsimovEventBuffer::k_maxEventSizeInBytes)
    {
        // Send up to 4KB of the event string.
        std::string portionOfEventString = uploadData.Payload.substr(
            0, std::min<size_t>(uploadData.Payload.length(), 4 * 1024));

        if (_heartBeat != nullptr)
        {
            _heartBeat->RecordDbEventSizeOffender(portionOfEventString);
        }

        TraceLoggingWrite(
            g_tlgProv,
            "AsimovUploader_RejectEventForExceedingSizeLimit",
            TraceLoggingValue(
                portionOfEventString.c_str(), "PortionOfEventString"),
            TraceLoggingValue(uploadData.Payload.length(), "Length"));

        return false;
    }

    return _eventStore->AddData(uploadData);
}

}}}
