// Copyright (c) Microsoft Corporation. All rights reserved.

#pragma once

#include <memory>
#include <mutex>
#include <string>
#include <vector>

#include <azure-device-util/IConfig.h>

#include "CommonDefs.h"
#include "IEventSerializer.h"
#include "IPolicyManager.h"
#include "ITelemetrySettings.h"

namespace Azure { namespace Device { namespace Telemetry {

//
// This class is responsible for serializing individual Asimov events and
// completed scenarios, and writing them to storage.
//
class AsimovEventSerializer : public IEventSerializer
{
  public:
    AsimovEventSerializer(
        Policy::IPolicyManager& policyManager,
        IKillSwitchManager& killSwitchManager,
        Utils::IConfig& config);
    ~AsimovEventSerializer();

    void BeginSerializing(IEventStore* eventStore) override;
    void BeginSettingUploadTimers(IUploadScheduler* uploadQueue) override;

    void SetHeartBeat(IHeartBeat& heartBeat) override;

    void ReloadConfiguration(Utils::IConfig& config) override;

    void UpdateTelemetrySettings(
        std::unique_ptr<ITelemetrySettings> telemetrySettings) override;

    // IAsimovEventSerializer

    IKillSwitchManager& GetKillSwitchManager() override
    {
        return _killSwitchManager;
    }

    bool WriteSyntheticEvent(
        std::unique_ptr<AsimovSyntheticEvent> event,
        WriteEventOptions options = {}) override;

    bool WriteEvent(
        std::unique_ptr<AsimovEvent> event,
        WriteEventOptions options = {}) override;

  private:
    bool WriteEventToDatabase(
        std::string&& event,
        TriggerLatency latency,
        TriggerPersistence persistence);

    AsimovSharedPartAState IncrementAndCopyPartAState();

    bool IsSampledIn(const AsimovEvent& event);

  private:
    Policy::IPolicyManager& _policyManager;
    IKillSwitchManager& _killSwitchManager;

    std::recursive_mutex _partAStateMutex;
    std::unique_ptr<AsimovSharedPartAState> _partAState;

    IHeartBeat* _heartBeat = nullptr;

    // Guard _eventStore and _queuedEvents with a mutex
    std::recursive_mutex _queuedEventMutex;
    IEventStore* _eventStore = nullptr;
    std::vector<std::unique_ptr<AsimovSyntheticEvent>> _queuedEvents;

    IUploadScheduler* _uploadScheduler = nullptr;

    std::recursive_mutex _telemetrySettingsMutex;
    std::unique_ptr<ITelemetrySettings> _telemetrySettings;
};

}}}
