// Copyright (c) Microsoft Corporation. All rights reserved.

#include "OneSettingsHttpClient.h"

#include <sstream>

#include <azure-device-util/Version.h>

#include "AsimovSharedPartAState.h"
#include "TlgProvider.h"

namespace Azure { namespace Device { namespace Telemetry {

OneSettingsHttpClient::OneSettingsHttpClient(IHttpSessionManager& httpSessionManager)
    : _httpSessionManager(httpSessionManager)
{}

/*++
    Downloads a OneSettings namespace and stores the result in HttpResponse.

    ETag is supplied to prevent downloading the full paload when the content
    of the request has not been modified since our last pull.

Arguments:
    partner     - OneSettings partner namespace to download [e.g. telemetry]
    feature     - OneSettings feature namespace to download [e.g. ASM-Linux]
    etag        - Identifier of previous content to prevent redownloading
    queryParams - Extra metadata about the device to help OneSettings targeting
    response    - Stores Http response code and payload data
--*/
std::error_code OneSettingsHttpClient::DownloadNamespace(
    std::string_view partner,
    std::string_view feature,
    std::string_view etag,
    const HttpQueryMap& queryParams,
    const pplx::cancellation_token& cancellationToken,
    HttpResponse& response)
{
    std::string flattenedQueryParams = FlattenHttpHeaderMap(queryParams);
    TraceLoggingWrite(
        g_tlgProv,
        "OneSettingsHttpClient_DownloadNamespace",
        TraceLoggingStringView(etag, "ETag"),
        TraceLoggingStringView(partner, "Partner"),
        TraceLoggingStringView(feature, "Feature"),
        TraceLoggingValue(flattenedQueryParams.c_str(), "QueryParams"));

    HttpHeaderMap requestHeaders;
    if (!etag.empty())
    {
        requestHeaders.emplace("If-None-Match", etag);
    }

    std::string oneSettingsUriPath = "settings/v3.0/";
    oneSettingsUriPath += partner;
    oneSettingsUriPath += "/";
    oneSettingsUriPath += feature;

    std::error_code downloadResult = _httpSessionManager.DoSynchronousHttpRequest(
        oneSettingsUriPath,
        queryParams,
        HttpMethod::GET,
        requestHeaders,
        {},
        cancellationToken,
        response);

    TraceLoggingWrite(
        g_tlgProv,
        "OneSettingsHttpClient_ResponseLog",
        TraceLoggingStringView(etag, "ETag"),
        TraceLoggingStringView(partner, "Partner"),
        TraceLoggingStringView(feature, "Feature"),
        TraceLoggingValue(
            FlattenHttpHeaderMap(requestHeaders).c_str(), "RequestHeaders"),
        TraceLoggingValue(
            FlattenHttpHeaderMap(queryParams).c_str(), "QueryParams"),
        TraceLoggingValue(response.StatusCode, "StatusCode"),
        TraceLoggingValue(
            FlattenHttpHeaderMap(response.Headers).c_str(), "ResponseHeaders"),
        TraceLoggingValue(response.Data.c_str(), "ResponseText"),
        TraceLoggingErrorCode(downloadResult, "ErrorCode"));

    return downloadResult;
}

/*++
    Builds a full set of OneSettings HTTP query parameters from supplied
    data.

Arguments:
    osReleaseData   - OsReleaseData population with OS version information

Return Value:
    HttpQueryMap suitable to hand to OneSettings
--*/
HttpQueryMap OneSettingsHttpClient::CreateOneSettingsQueryMap(
    const AsimovSharedPartAState::OsReleaseData& osReleaseData)
{
    HttpQueryMap queryMap;

    queryMap.emplace("os", osReleaseData.Name);
    queryMap.emplace("osVer", osReleaseData.Version);
    queryMap.emplace("app", "azure-device-health-services");
    queryMap.emplace("appVer", Utils::Version::GetString());

    return queryMap;
}
}}}
