// Copyright (c) Microsoft Corporation. All rights reserved.

#include "SettingsConfigReader.h"

#include <iostream>
#include <optional>

#include <azure-device-util/FileSystem.h>
#include <azure-device-util/IConfig.h>
#include <azure-device-util/String.h>

#include "ConfigPaths.h"
#include "TlgProvider.h"

namespace Azure { namespace Device { namespace Telemetry {

bool operator<(const SettingsNamespaceConfig& left, const SettingsNamespaceConfig& right)
{
    return std::tie(left.Partner, left.Feature) <
        std::tie(right.Partner, right.Feature);
}

std::string SettingsNamespaceTypeToString(SettingsNamespaceType type)
{
    switch (type)
    {
    case SettingsNamespaceType::Telemetry:
        return "Telemetry";
    default:
        FAIL_FAST();
    }
}

std::string SettingsNamespaceConfig::ToString() const
{
    std::string output;

    output += "{Partner='";
    output += Partner;
    output += "', Feature='";
    output += Feature;
    output += "', Type='";
    output += SettingsNamespaceTypeToString(Type);
    output += "'}";

    return output;
}

/*++
    Looks in config object for correctly specified OneSettings namespaces
    and returns them.

Arguments:
    config   - Telemetry service configuration object

Return Value:
    Vector of all parseable namespaces found on config object
--*/
std::vector<SettingsNamespaceConfig>
SettingsConfigReader::LoadSettingsFromIConfig(Utils::IConfig& config)
{
    std::vector<SettingsNamespaceConfig> result;

    Utils::ConfigPath arrayPath = Config::SettingsNamespaces;
    size_t arraySize = config.GetArraySize(arrayPath);
    for (size_t i = 0; i < arraySize; i++)
    {
        auto optSettingsNamespace =
            ParseSettingsNamespace(config, arrayPath.CreateWithChildIndex(i));

        if (optSettingsNamespace)
        {
            result.emplace_back(*std::move(optSettingsNamespace));
        }
    }

    return result;
}

/*++
    Parses the addressed config section to create a well-formed
    SettingsNamespaceConfig

Arguments:
    config      - Telemetry service configuration object
    elementPath - Indexer that points to desired config section to parse

Return Value:
    SettingsNamespaceConfig if parsed, empty if misconfigured
--*/
std::optional<SettingsNamespaceConfig>
SettingsConfigReader::ParseSettingsNamespace(
    Utils::IConfig& config,
    const Utils::ConfigPath& elementPath)
{
    std::vector<std::string> brokenFields;

    auto loadRequiredStringField = [&](std::string& structFieldToSet,
                                       std::string_view fieldName) {
        structFieldToSet = config.GetString(
            elementPath.CreateWithChildName(std::string{ fieldName }), "");

        if (structFieldToSet.empty())
        {
            brokenFields.emplace_back(fieldName);
        }
    };

    SettingsNamespaceConfig settingsNamespace;
    loadRequiredStringField(settingsNamespace.Partner, "Partner");
    loadRequiredStringField(settingsNamespace.Feature, "Feature");

    if (!brokenFields.empty())
    {
        LogBrokenSettingsNamespace(elementPath, brokenFields);
        return {};
    }

    LogSuccessfullyParsedSettingsNamespace(settingsNamespace);
    return settingsNamespace;
}

void SettingsConfigReader::LogSuccessfullyParsedSettingsNamespace(
    const SettingsNamespaceConfig& settingsNamespace)
{
    std::cout << "Loaded configuration for settingsNamespace: "
              << settingsNamespace.ToString() << std::endl;

    TraceLoggingWrite(
        g_tlgProv,
        "SettingsConfigReader_ParsedSettingsNamespace",
        TraceLoggingValue(settingsNamespace.Partner.c_str(), "Partner"),
        TraceLoggingValue(settingsNamespace.Feature.c_str(), "Feature"),
        TraceLoggingValue(
            SettingsNamespaceTypeToString(settingsNamespace.Type).c_str(),
            "Type"));
}

void SettingsConfigReader::LogBrokenSettingsNamespace(
    const Utils::ConfigPath& elementPath,
    const std::vector<std::string>& brokenFields)
{
    std::string fields = Utils::String::Join(brokenFields, ",");
    std::string elementPathString = elementPath.ToString();

    std::cout
        << "[ERROR] Failed to parse settings namespace at: " << elementPathString
        << "| brokenFields = {" << fields << "}" << std::endl;

    TraceLoggingWrite(
        g_tlgProv,
        "SettingsConfigReader_ParsedSettingsNamespace_Failed",
        TraceLoggingValue(elementPathString.c_str(), "ElementPath"),
        TraceLoggingValue(fields.c_str(), "BrokenFields"));
}

}}}
