// Copyright (c) Microsoft Corporation. All rights reserved.

#include "SettingsDownloader.h"

#include <memory>
#include <optional>

#include "AsimovSyntheticEvent.h"
#include "HeartBeat.h"
#include "IEventSerializer.h"
#include "ISettingsRepository.h"
#include "OneSettingsHttpClient.h"
#include "TlgProvider.h"

namespace Azure { namespace Device { namespace Telemetry {

SettingsDownloader::SettingsDownloader(
    IHeartBeat& heartBeat,
    IEventSerializer& serializer,
    IHttpSessionManager& httpSessionManager,
    ISettingsRepository& repository,
    HttpQueryMap queryMap)
    : _heartBeat(heartBeat)
    , _serializer(serializer)
    , _httpSessionManager(httpSessionManager)
    , _settingsRepository(repository)
    , _queryParamMap(std::move(queryMap))
{}

/*++
    For each specified namespace, try to download it and write to the
    repository.

Arguments:
    oldNamespaceStates  - config/metadata pairs specifying what to download and
                            any ETag to describe content ew already have
    cancelToken         - token to cancel any async download in progress
--*/
void SettingsDownloader::DownloadAllNamespacesToRepository(
    const std::vector<SettingsNamespaceState>& oldNamespaceStates,
    pplx::cancellation_token cancelToken)
{
    for (const auto& state : oldNamespaceStates)
    {
        try
        {
            auto optDownloadResult = DownloadNamespace(
                state.Config, state.Metadata.ETag, cancelToken);
            if (optDownloadResult)
            {
                // Only update content if ETag doesn't match
                if (optDownloadResult->ETag != state.Metadata.ETag)
                {
                    _settingsRepository.WriteContent(
                        state.Config, optDownloadResult->Content);
                }

                SettingsNamespaceMetadata metadata;
                metadata.ETag = optDownloadResult->ETag;
                metadata.LastDownloadTime = std::chrono::system_clock::now();

                _settingsRepository.WriteMetadata(state.Config, metadata);
            }
        }
        CATCH_LOG();
    }
}

/*++
    HTTP request for specified namespace config.  Provide ETag to prevent
    downloading duplicate data.

Arguments:
    settingsNamespace   - config data to name the namespace
    etag                - ETag to prevent redownloading content
    cancelToken         - token to cancel any async download in progress

Return Value:
    DownloadResult in case of successful connection and response,
    std::nullopt for network failure or HTTP error code
--*/
std::optional<SettingsDownloader::SettingDownloadResult>
SettingsDownloader::DownloadNamespace(
    const SettingsNamespaceConfig& settingNamespace,
    std::string_view etag,
    pplx::cancellation_token cancelToken)
{
    OneSettingsHttpClient httpClient{ _httpSessionManager };

    HttpResponse response;
    std::error_code uploadResult = httpClient.DownloadNamespace(
        settingNamespace.Partner,
        settingNamespace.Feature,
        etag,
        _queryParamMap,
        cancelToken,
        response);

    // Increase connection attempts
    _heartBeat.RecordNetworkAttempt();

    if (FAILED(uploadResult))
    {
        HandleNetworkError(uploadResult);
        return {};
    }

    std::cout << "--Receieved response--:" << std::endl;
    std::cout << "Status code: " << response.StatusCode << std::endl
              << std::endl;
    std::cout << "Headers: " << FlattenHttpHeaderMap(response.Headers)
              << std::endl;
    std::cout << "Body: " << response.Data << std::endl;

    return HandleNetworkSuccess(settingNamespace, response);
}

/*++
    Log network failure.

Arguments:
    result - std::error_code from failed download
--*/
void SettingsDownloader::HandleNetworkError(std::error_code result)
{
    TraceLoggingWrite(
        g_tlgProv,
        "SettingsDownloader_HandleNetworkError",
        TraceLoggingErrorCode(result, "Result"));

    auto event = std::make_unique<AsimovSyntheticEvent>(
        "SettingsDownloader_NetworkError_0");
    event->AddPartCField("ErrorCodeValue", result.value());
    event->AddPartCField("ErrorCodeCategoryName", result.category().name());

    _serializer.WriteSyntheticEvent(std::move(event));

    _heartBeat.RecordNetworkFailure(result);
}

/*++
    Process HTTP response and return SettingDownloadResult in case
    of service success.

Arguments:
    settingsNamespace   - config data to name the namespace
    response            - response metadata/content from service

Return Value:
    DownloadResult in case of successful response,
    std::nullopt otherwise
--*/
std::optional<SettingsDownloader::SettingDownloadResult>
SettingsDownloader::HandleNetworkSuccess(
    const SettingsNamespaceConfig& settingNamespace,
    const HttpResponse& response)
{
    TraceLoggingWrite(
        g_tlgProv,
        "SettingsDownloader_NetworkSuccess",
        TraceLoggingValue(
            settingNamespace.Partner.c_str(), "SettingNamespacePartner"),
        TraceLoggingValue(
            settingNamespace.Feature.c_str(), "SettingNamespaceFeature"),
        TraceLoggingValue(response.StatusCode, "HttpStatusCode"),
        TraceLoggingValue(response.ETag.c_str(), "ETag"),
        TraceLoggingValue(response.Data.c_str(), "ResponseData"));

    if (response.StatusCode == 200 || response.StatusCode == 304)
    {
        SettingDownloadResult result;
        result.Content = response.Data;
        result.ETag = response.ETag;
        return result;
    }

    return {};
}

}}}
