// Copyright (c) Microsoft Corporation. All rights reserved.

#include "SettingsLoader.h"

#include "ISettingsRepository.h"
#include "TelemetrySettings.h"

namespace Azure { namespace Device { namespace Telemetry {

SettingsLoader::SettingsLoader(
    ISettingsRepository& settingsRepository,
    std::vector<SettingsNamespaceConfig> namespaceConfigs)
    : _settingsRepository(settingsRepository)
    , _configs(std::move(namespaceConfigs))
{}

/*++
    Loads all namespaces listed in _configs from repository
    into memory.

    Stores new states internally in _loadedStates.
--*/
void SettingsLoader::LoadAllNamespacesFromRepository()
{
    std::vector<SettingsNamespaceState> results;

    for (const auto& namespaceConfig : _configs)
    {
        SettingsNamespaceState state;
        state.Config = namespaceConfig;

        if (_settingsRepository.MetadataExists(namespaceConfig))
        {
            if (auto optMetadata =
                    _settingsRepository.ReadMetadata(namespaceConfig))
            {
                state.Metadata = std::move(*optMetadata);
            }
        }

        if (_settingsRepository.ContentExists(namespaceConfig))
        {
            if (auto optContent = _settingsRepository.ReadContent(namespaceConfig))
            {
                state.Content = std::move(*optContent);
            }
        }

        if (state.Content.empty() && state.Metadata.ETag != "\"\"")
        {
            // If Content is empty, then ETag must be empty or the '""' tag
            // sent by the server when it has no content.
            state.Metadata.ETag.clear();
        }

        results.push_back(std::move(state));
    }

    _loadedStates = std::move(results);
}

/*++
    Return a new TelemetrySettings based on all Telemetry namespace content.
--*/
std::unique_ptr<ITelemetrySettings>
SettingsLoader::BuildTelemetrySettingsFromCurrentState()
{
    auto telemetrySettings = std::make_unique<TelemetrySettings>();

    // These are the default rules, any cloud rule will override them
    telemetrySettings->AddSetting("REQUIRED:::SampleRate", "100");
    telemetrySettings->AddSetting("OPTIONAL:::SampleRate", "100");

    // Loop over all states, routing each namespace type to the correct
    // location
    for (const auto& namespaceState : _loadedStates)
    {
        if (namespaceState.Config.Type == SettingsNamespaceType::Telemetry)
        {
            for (const auto& [settingName, settingValue] : namespaceState.Content)
            {
                bool succeeded =
                    telemetrySettings->AddSetting(settingName, settingValue);
                if (!succeeded)
                {
                    std::cout << "Invalid setting: \"" << settingName
                              << "\" -> \"" << settingValue << "\"" << std::endl;
                }
            }
        }
    }

    return telemetrySettings;
}
}}}
