// Copyright (c) Microsoft Corporation. All rights reserved.

#include "SettingsManager.h"

#include <azure-device-util/FileSystem.h>
#include <azure-device-util/IConfig.h>
#include <azure-device-util/Version.h>

#include "AppState.h"
#include "AsimovSharedPartAState.h"
#include "AsimovSyntheticEvent.h"
#include "ConfigPaths.h"
#include "HeartBeat.h"
#include "HttpSessionManager.h"
#include "IEventSerializer.h"
#include "OneSettingsHttpClient.h"
#include "SettingsConfigReader.h"
#include "SettingsRepository.h"
#include "TelemetryError.h"
#include "TelemetrySettings.h"
#include "TlgProvider.h"

namespace Azure { namespace Device { namespace Telemetry {

SettingsManager::SettingsManager(
    IHeartBeat& heartBeat,
    IEventSerializer& serializer,
    Utils::IConfig& config)
    : _settingsRepository(
          AppState::GetRuntimeDataDirectory() /
          std::filesystem::path{ "settings" })
    , _eventSerializer(serializer)
{
    _httpSessionManager.SetServerUrl("https://settings.data.microsoft.com");

    auto osReleaseData = AsimovSharedPartAState::ReadOsReleaseData();
    HttpQueryMap httpQueryMap =
        OneSettingsHttpClient::CreateOneSettingsQueryMap(osReleaseData);

    SettingsConfigReader settingsConfigReader;
    std::vector<SettingsNamespaceConfig> settingsConfigs =
        settingsConfigReader.LoadSettingsFromIConfig(config);

    _loader = std::make_unique<SettingsLoader>(
        _settingsRepository, std::move(settingsConfigs));

    _downloader = std::make_unique<SettingsDownloader>(
        heartBeat,
        serializer,
        _httpSessionManager,
        _settingsRepository,
        std::move(httpQueryMap));

    LoadSettingsAndPublish();
}

SettingsManager::~SettingsManager()
{
    EndTimers();
}

/*++
    On each timer expiry:
        1. Download all namespaces again
        2. Load the full set of state from the repository into memory
        3. Arm timer to do it all again
--*/
void SettingsManager::OnTimerExpiry()
{
    _downloader->DownloadAllNamespacesToRepository(
        _loader->GetStates(), _shutdownTokenSource.get_token());

    LoadSettingsAndPublish();

    // Now rearm ourselves to run again
    ArmTimerFromNow(c_defaultDownloadFrequency);
}

/*++
    Load settings from repository again, then publish them to _eventSerializer
    so they can take effect for sampling.
--*/
void SettingsManager::LoadSettingsAndPublish()
{
    _loader->LoadAllNamespacesFromRepository();
    auto telemetrySettings = _loader->BuildTelemetrySettingsFromCurrentState();
    _eventSerializer.UpdateTelemetrySettings(std::move(telemetrySettings));
}

/*++
    Arms timer to run the expiry function again asynchronously after duration.
--*/
void SettingsManager::ArmTimerFromNow(std::chrono::steady_clock::duration duration)
{
    _downloadTimer.expires_from_now(duration);
    _downloadTimer.async_wait([this](const boost::system::error_code& ec) {
        if (!ec)
        {
            OnTimerExpiry();
        }
        else
        {
            TraceLoggingWrite(
                g_tlgProv,
                "SettingsManager_ArmTimerFromNow_Failed",
                TraceLoggingErrorCode(ec, "ErrorCode"));

            std::cout << "Timer error: " << ec;
        }
    });
}

/*++
    Start timer callback processing thread and arm timer.
--*/
void SettingsManager::StartTimers()
{
    FAIL_FAST_IF(_downloadThread.joinable());

    _downloadThread = std::thread{ [this] {
        // work is created to prevent run() from returning immediately
        boost::asio::io_service::work work{ _downloadTimerIoService };
        _downloadTimerIoService.run();
        std::cout << "DownloadThread exiting!" << std::endl;
    } };

    ArmTimerFromNow(std::chrono::milliseconds{ 0 });
}

/*++
    Cancel timers and stop callback thread.
--*/
void SettingsManager::EndTimers()
{
    _shutdownTokenSource.cancel();

    if (!_downloadTimerIoService.stopped())
    {
        _downloadTimerIoService.stop();
    }

    if (_downloadThread.joinable())
    {
        _downloadThread.join();
    }
}

}}}
