// Copyright (c) Microsoft Corporation. All rights reserved.

#include "SettingsRepository.h"
#include <optional>

#include <azure-device-util/FileSystem.h>

#include "SettingsSerializers.h"
#include "TlgProvider.h"

namespace Azure { namespace Device { namespace Telemetry {

SettingsRepository::SettingsRepository(std::filesystem::path settingsDirectory)
    : _settingsDirectory(std::move(settingsDirectory))
{
    std::filesystem::create_directories(_settingsDirectory);
}

bool SettingsRepository::ExistsHelper(
    const SettingsNamespaceConfig& settingsNamespace,
    std::string_view dataType)
{
    try
    {
        auto path = CreateFullFileSystemPath(
            settingsNamespace.Partner,
            settingsNamespace.Feature,
            dataType,
            FileType::Final);
        return std::filesystem::exists(path);
    }
    CATCH_LOG();

    return false;
}

bool SettingsRepository::ContentExists(const SettingsNamespaceConfig& settingsNamespace)
{
    return ExistsHelper(settingsNamespace, c_content);
}

bool SettingsRepository::MetadataExists(
    const SettingsNamespaceConfig& settingsNamespace)
{
    return ExistsHelper(settingsNamespace, c_metadata);
}

std::optional<SettingsNamespaceContent>
SettingsRepository::ReadContent(const SettingsNamespaceConfig& settingsNamespace)
{
    try
    {
        auto path = CreateFullFileSystemPath(
            settingsNamespace.Partner,
            settingsNamespace.Feature,
            c_content,
            FileType::Final);

        std::string fileContent;
        THROW_IF_FAILED(Utils::FileSystem::ReadUTF8StringFromFile(
            path.string(), fileContent));

        SettingsNamespaceContentSerializer ser;
        return ser.FromJsonString(fileContent);
    }
    CATCH_LOG();
    return {};
}

std::optional<SettingsNamespaceMetadata>
SettingsRepository::ReadMetadata(const SettingsNamespaceConfig& settingsNamespace)
{
    try
    {
        auto path = CreateFullFileSystemPath(
            settingsNamespace.Partner,
            settingsNamespace.Feature,
            c_metadata,
            FileType::Final);

        std::string fileContent;
        THROW_IF_FAILED(Utils::FileSystem::ReadUTF8StringFromFile(
            path.string(), fileContent));

        SettingsNamespaceMetadataSerializer ser;
        return ser.FromJsonString(fileContent);
    }
    CATCH_LOG();
    return {};
}

void SettingsRepository::WriteHelper(
    std::string_view partner,
    std::string_view feature,
    std::string_view dataType,
    std::string_view fileData)
{
    try
    {
        auto tempPath =
            CreateFullFileSystemPath(partner, feature, dataType, FileType::Temp);

        THROW_IF_FAILED(Utils::FileSystem::WriteUTF8StringToFile(
            tempPath.string(), fileData));

        auto finalPath = CreateFullFileSystemPath(
            partner, feature, dataType, FileType::Final);

        std::filesystem::rename(tempPath, finalPath);
    }
    CATCH_LOG();
}

void SettingsRepository::WriteContent(
    const SettingsNamespaceConfig& settingsNamespace,
    std::string_view content)
{
    WriteHelper(
        settingsNamespace.Partner, settingsNamespace.Feature, c_content, content);
}

void SettingsRepository::WriteMetadata(
    const SettingsNamespaceConfig& settingsNamespace,
    const SettingsNamespaceMetadata& metadata)
{
    SettingsNamespaceMetadataSerializer metadataSer;
    std::string metadataStr = metadataSer.ToJsonString(metadata);

    WriteHelper(
        settingsNamespace.Partner,
        settingsNamespace.Feature,
        c_metadata,
        metadataStr);
}

std::filesystem::path SettingsRepository::CreateFullFileSystemPath(
    std::string_view partner,
    std::string_view feature,
    std::string_view dataType,
    FileType fileType)
{
    std::filesystem::path outputPath = _settingsDirectory;

    outputPath /= std::string{ partner };
    outputPath += ".";
    outputPath += std::string{ feature };
    outputPath += ".";
    outputPath += std::string{ dataType };
    outputPath += ".json";

    if (fileType == FileType::Temp)
    {
        outputPath += ".temp";
    }

    return outputPath;
}

}}}
