// Copyright (c) Microsoft Corporation. All rights reserved.

#include "TelemetrySettings.h"
#include <optional>

#include "TlgProvider.h"

namespace Utils = Azure::Device::Utils;

namespace Azure { namespace Device { namespace Telemetry {

/*++
    Adds a telemetry setting name/value pair to the set.
    This will parse the name/value pair for correct formatting, and reject
    any illegal setting.

Arguments:
    settingName - e.g. "MyProvider:MyEvent::SampleRate"
    settingValue - e.g. "100"

Return Code:
    true if successfully added, false otherwise
--*/
bool TelemetrySettings::AddSetting(
    std::string_view settingName,
    std::string_view settingValue)
{
    // Setting names are of the form
    // "<PermissionLevelName|ProviderName>:<EventName>:<unused>:<SettingCommandName>"
    auto settingNameComponents = Utils::String::Split(settingName, ":");

    if (settingNameComponents.size() != 4)
    {
        LogSettingRejection(
            settingName,
            settingValue,
            "All 3 colons need to be present to create a valid setting");
        return false;
    }

    std::string_view permissionOrProviderName = settingNameComponents[0];
    std::string_view eventName = settingNameComponents[1];
    // unused settingNameComponents[2];
    std::string_view settingCommandName = settingNameComponents[3];

    if (permissionOrProviderName.empty())
    {
        LogSettingRejection(
            settingName,
            settingValue,
            "<PermissionLevelName|ProviderName> can never be empty");
        return false;
    }

    auto commandOverrides =
        BuildCommandOverrides(settingCommandName, settingValue);
    if (!commandOverrides)
    {
        LogSettingRejection(
            settingName, settingValue, "Command or value not parseable");
        return false;
    }

    // If name matches a permission name, target that object to mutate.
    // Otherwise, use event name (possibly empty) to target an object in
    // the _providerAndEventSettings map.
    TelemetrySettingOverrides& targetOverrides =
        GetTargetOverrides(permissionOrProviderName, eventName);

    targetOverrides.MergeSettingsOverrides(*commandOverrides);
    return true;
}

/*++
    Builds a new TelemetrySettingsOverride with the single 'commanded' field
    set to non-optional.

Arguments:
    settingCommandName - e.g. "SampleRate"
    settingValue - e.g. "100"

Return Code:
    Valid object with one non-optional field if parsed correctly,
    std::nullopt otherwise
--*/
std::optional<TelemetrySettingOverrides> TelemetrySettings::BuildCommandOverrides(
    std::string_view settingCommandName,
    std::string_view settingValue)
{
    TelemetrySettingOverrides overrides;

    // Only accept SampleRate as command name for now
    if (Utils::String::ICompare(settingCommandName, "SampleRate") == 0)
    {
        overrides.SampleRate = Utils::String::ToFloat<double>(settingValue);
        return overrides;
    }

    // If we didn't successfully parse a command and a value, return nullopt
    return {};
}

/*++
    Returns a reference to the correct TelemetrySettingsOverrides object
    to merge new settings into.

Arguments:
    permissionOrProviderName    - First section of setting,
                                    e.g. "Required" or "MyProvider"
    eventName                   - Second section of setting, or empty
                                    e.g. "" or "MyEventName"

Return Code:
    Reference to object to merge settings into
--*/
TelemetrySettingOverrides& TelemetrySettings::GetTargetOverrides(
    std::string_view permissionOrProviderName,
    std::string_view eventName)
{
    // Since permission and provider take the same slot, see if the name
    // matches a permission value.  If it does, reference one of those objects.
    // If not, create/reference an entry in the map.
    if (auto optTelPermission =
            Utils::Permission::TryConvertStringToTelemetryPermission(
                permissionOrProviderName))
    {
        if (*optTelPermission == Utils::TelemetryPermission::Required)
        {
            return _requiredPermissionOverrides;
        }
        else if (*optTelPermission == Utils::TelemetryPermission::Optional)
        {
            return _optionalPermissionOverrides;
        }
    }

    return _providerAndEventSettings[std::string{ permissionOrProviderName }]
                                    [std::string{ eventName }];
}

/*++
    Decides settings overrides for one particular event by applying the
    most broad-level settings first and then allowing more narrowly-scoped
    settings to override them.

Arguments:
    permissionLevel - TelemetryPermission of event
    providerName    - Provider name of event, e.g. "MyProvider"
    eventName       - Event name of event, e.g. "MyEventName"

Return Code:
    TelemtrySettingsOverrides object with potentially std::nullopt on
    any field if no applicable settings of that command type were found.
--*/
TelemetrySettingOverrides TelemetrySettings::GetTelemetryEventSettings(
    Utils::TelemetryPermission permissionLevel,
    std::string_view providerName,
    std::string_view eventName)
{
    TelemetrySettingOverrides overrides;

    // First, merge in any settings from the permission level
    if (permissionLevel == Utils::TelemetryPermission::Required)
    {
        overrides.MergeSettingsOverrides(_requiredPermissionOverrides);
    }
    else if (permissionLevel == Utils::TelemetryPermission::Optional)
    {
        overrides.MergeSettingsOverrides(_optionalPermissionOverrides);
    }
    else
    {
        THROW_HR(Utils::UtilsError::E_INVALID_PERMISSION_VALUE);
    }

    // Look for provider level settings
    auto providerItr = _providerAndEventSettings.find(providerName);
    if (providerItr != _providerAndEventSettings.end())
    {
        const auto& providerMap = providerItr->second;

        auto mergeNameOverrides = [&overrides,
                                   &providerMap](std::string_view eventName) {
            auto eventItr = providerMap.find(eventName);
            if (eventItr != providerMap.end())
            {
                overrides.MergeSettingsOverrides(eventItr->second);
            }
        };

        // Second, empty string for provider level overrides
        mergeNameOverrides("");

        // Third, real event name for event level overrides
        mergeNameOverrides(eventName);
    }

    return overrides;
}

void TelemetrySettings::LogSettingRejection(
    std::string_view settingName,
    std::string_view settingValue,
    std::string_view rejectionReason)
{
    TraceLoggingWrite(
        g_tlgProv,
        "TelemetrySettings_SettingRejected",
        TraceLoggingStringView(settingName, "SettingName"),
        TraceLoggingStringView(settingValue, "SettingValue"),
        TraceLoggingStringView(rejectionReason, "RejectionReason"));
}

}}}
