// Copyright (c) Microsoft Corporation. All rights reserved.

#include "Sqlite.h"

#include <azure-device-util/ConversionFactors.h>
#include <azure-device-util/ResultException.h>
#include <azure-device-util/ScopeExit.h>
#include <azure-device-util/String.h>

#include "TelemetryError.h"
#include "TlgProvider.h"

namespace Azure { namespace Device { namespace Telemetry {

class SqliteErrorCat : public std::error_category
{
    const char* name() const noexcept override { return "SqliteErrorCat"; }

    std::string message(int sqliteError) const override
    {
        std::string output = "SqliteErrorCat: ";
        output += sqlite3_errstr(sqliteError);
        return output;
    }
};

static const SqliteErrorCat c_sqliteErrorCat;

std::error_code make_sqlite_error(SqliteResult errCode)
{
    if (errCode == SQLITE_OK || errCode == SQLITE_DONE || errCode == SQLITE_ROW)
    {
        return {};
    }

    return std::error_code{ errCode, c_sqliteErrorCat };
}

bool IsErrorSqliteCorrupt(std::error_code ec)
{
    return ec.category() == c_sqliteErrorCat && ec.value() == SQLITE_CORRUPT;
}

bool IsErrorSqliteFull(std::error_code ec)
{
    return ec.category() == c_sqliteErrorCat && ec.value() == SQLITE_FULL;
}

SqlConnection::SqlConnection(
    std::string_view databaseDirectory,
    bool skipQuickIntegrityCheck)
{
    std::string fileName;
    //
    // If the filename is an empty string, then a private, temporary on-disk
    // database will be created. This private database will be automatically
    // deleted as soon as the database connection is closed.
    //
    if (!databaseDirectory.empty())
    {
        fileName.append(databaseDirectory.data(), databaseDirectory.size());
        fileName += "/EventStore.db";
    }

    sqlite3* sql = nullptr;
    THROW_IF_FAILED(make_sqlite_error(sqlite3_open(fileName.c_str(), &sql)));
    _dbConnectionHandle.reset(sql);

    if (!databaseDirectory.empty() && !skipQuickIntegrityCheck)
    {
        SqliteStatement quickCheck{ _dbConnectionHandle, "PRAGMA quick_check;" };
        std::string quickCheckResult{ quickCheck.Bind().Step()->NextString() };

        THROW_HR_IF(
            TelemetryError::E_DB_INTEGRITY_CHECK_FAIL, quickCheckResult != "ok");
    }

    // We demand a 4KB page size on all platforms.  This is a no-op
    // on newer Sqlite versions.
    SqliteStatement pageSizeUpdate(_dbConnectionHandle, "PRAGMA page_size=4096;");
    pageSizeUpdate.Bind().Step();

    SqliteStatement pageSizeQuery(_dbConnectionHandle, "PRAGMA page_size;");
    uint32_t pageSize = pageSizeQuery.Bind().Step()->NextInt<uint32_t>();

    // After the update, our new max should be equal to the requested new
    // max, or greater if it couldn't shrink
    FAIL_FAST_IF(pageSize != 4096);
}

/*++
    Reclaim disk space via a VACUUM command.
--*/
void SqlConnection::Vacuum()
{
    TraceLoggingWrite(g_tlgProv, "SqlConnection_Vacuum");

    //
    // Some versions of SQLite don't support doing a vacuum if mmap_size >
    // 0. Set mmap_size to 0 before doing the vacuum and restore it after
    // the vacuum is done.
    //

    // Query the current mmap size
    uint32_t currentMmapSize = 0;
    {
        SqliteStatement mmapPragmaQuery{ _dbConnectionHandle,
                                         "PRAGMA mmap_size;" };
        currentMmapSize = mmapPragmaQuery.Bind().Step()->NextInt<uint32_t>();
    }

    // On scope exit, restore the mmap size
    auto scopeExit = Utils::MakeScopeExit([currentMmapSize, this]() {
        std::string mmapStatementString = "PRAGMA mmap_size ";
        mmapStatementString += std::to_string(currentMmapSize);
        mmapStatementString += ";";

        SqliteStatement mmapPragmaSetToOriginal(
            _dbConnectionHandle, mmapStatementString);
        mmapPragmaSetToOriginal.Bind().Step();
    });

    // Set mmap size to zero
    {
        SqliteStatement mmapPragmaQuery{ _dbConnectionHandle,
                                         "PRAGMA mmap_size = 0;" };
        mmapPragmaQuery.Bind().Step();
    }

    // Perform the vacuum
    {
        SqliteStatement vacuumStatement{ _dbConnectionHandle, "VACUUM;" };
        vacuumStatement.Bind().Step();
    }

    // Do a DB checkpoint
    CheckpointDb();
}

/*++
    Begin a transaction if one is not already opened.
--*/
void SqlConnection::TryBeginTransaction()
{
    if (!IsTransactionStarted())  // Open transaction;
    {
        TraceLoggingWrite(g_tlgProv, "SqlConnection_BeginTransaction");

        THROW_IF_FAILED(make_sqlite_error(sqlite3_exec(
            _dbConnectionHandle.get(), "BEGIN TRANSACTION", 0, 0, nullptr)));
    }
}

/*++
    Commit a transaction if one is opened.
--*/
void SqlConnection::CommitIfTransactionStarted()
{
    if (IsTransactionStarted())  // Not in open transaction
    {
        TraceLoggingWrite(g_tlgProv, "SqlConnection_CommitTransaction");

        THROW_IF_FAILED(make_sqlite_error(
            sqlite3_exec(_dbConnectionHandle.get(), "COMMIT;", 0, 0, nullptr)));
    }
}

/*++
    Runs checkpoint operation on the database and the log file is written
    back into the database. This blocks until there is no other writer to the
    database and all readers are reading from the database only. We only do
    single db operation at once so it should not be a concern for us.
--*/
void SqlConnection::CheckpointDb()
{
    TraceLoggingWrite(g_tlgProv, "SqlConnection_CheckpointDb");

    int sizeOfWalLogInFrames, totalNumberOfFramesCheckpointed;
    THROW_IF_FAILED(make_sqlite_error(sqlite3_wal_checkpoint_v2(
        _dbConnectionHandle.get(),
        nullptr,
        SQLITE_CHECKPOINT_TRUNCATE,
        &sizeOfWalLogInFrames,
        &totalNumberOfFramesCheckpointed)));
}

/*++
    Calculates the percentage of pages used in the whole db and stores
    into _percentageFull.
--*/
void SqlConnection::UpdatePercentageFull()
{
    uint32_t usedPageCount = GetCurrentUsedPageCount();

    _percentageFull = static_cast<double>(usedPageCount) * 100 / _maxPageCount;
}

/*++
    Calculates the number of pages used in the whole DB.
--*/
uint32_t SqlConnection::GetCurrentUsedPageCount()
{
    int32_t pageCount = 0;
    int32_t freeListCount = 0;

    auto GetColumnAsInt32 = [](void* output, int argc, char** argv, char**) -> int {
        FAIL_FAST_IF(argc != 1);
        int32_t* value = reinterpret_cast<int32_t*>(output);
        *value = Utils::String::ToIntT<int32_t>(argv[0]);
        return 0;
    };

    LOG_IF_FAILED(make_sqlite_error(sqlite3_exec(
        _dbConnectionHandle.get(),
        "PRAGMA page_count;",
        GetColumnAsInt32,
        &pageCount,
        nullptr)));

    LOG_IF_FAILED(make_sqlite_error(sqlite3_exec(
        _dbConnectionHandle.get(),
        "PRAGMA freelist_count;",
        GetColumnAsInt32,
        &freeListCount,
        nullptr)));

    int32_t delta = pageCount - freeListCount;
    FAIL_FAST_IF(delta < 0);

    return static_cast<uint32_t>(delta);
}

/*++
    Sets the amount of pages our storage is allowed to use. This is how
    sqlite sets max database size and sets the memory mapped file size if this
    isn't a in-memory database.

Arguments:
    maxPageCount    - the max amount of pages allowed to be used by db.
    skipMmapPragma  - skip setting the pragma for mmap_size if this is true.
--*/
uint32_t
SqlConnection::SetMaxPageCount(uint32_t desiredMaxPageCount, bool skipMmapPragma)
{
    TraceLoggingWrite(
        g_tlgProv,
        "SqlConnection_SetMaxPageCount",
        TraceLoggingValue(desiredMaxPageCount, "DesiredMaxPageCount"));

    std::string maxPagePragmaQuery = "PRAGMA max_page_count=";
    maxPagePragmaQuery.append(std::to_string(desiredMaxPageCount));
    maxPagePragmaQuery += ";";

    SqliteStatement sizeLimitUpdate(_dbConnectionHandle, maxPagePragmaQuery);

    _maxPageCount = sizeLimitUpdate.Bind().Step()->NextInt<uint32_t>();

    // After the update, our new max should be equal to the requested new
    // max, or greater if it couldn't shrink
    FAIL_FAST_IF(_maxPageCount < desiredMaxPageCount);

    // If we set the max to the desired amount, then take extra action
    if (_maxPageCount == desiredMaxPageCount)
    {
        if (!skipMmapPragma)
        {
            //
            // The mmap_size is the max number of bytes of the database file
            // that sqlite will try to map into the process address space at
            // one time We set this equal to our max database size to have a
            // fully memory mapped database.
            //
            // We will cap this at 10 * the requested DB size or 50MB,
            // whichever is smaller
            //
            uint32_t maxByteSize = desiredMaxPageCount * 4096;
            uint32_t mmapSize = std::min<uint32_t>(
                maxByteSize * 10, 50U * ConversionFactors::mbToBytes);

            std::string mmapPragma = "PRAGMA mmap_size = ";
            mmapPragma += std::to_string(mmapSize);
            mmapPragma += ";";

            SqliteStatement mmapPragmaUpdate(_dbConnectionHandle, mmapPragma);
            uint32_t updatedMmapSize =
                mmapPragmaUpdate.Bind().Step()->NextInt<uint32_t>();

            FAIL_FAST_IF(updatedMmapSize != mmapSize);
        }

        UpdatePercentageFull();
    }

    return _maxPageCount;
}

/*++
    Updates the WAL mode journal size for this database connection.

Arguments:
    journalSizeBytes - The maximum size of the write-ahead log journal
    autoCheckpointNumPages
        - The maximum number of pages in the write-ahead log
            (multiply by page size to get size in bytes)
--*/
void SqlConnection::UpdateWalJournalSize(
    uint32_t journalSizeBytes,
    uint32_t autoCheckpointNumPages)
{
    std::string journalSizeStatement = "PRAGMA journal_size_limit = ";
    journalSizeStatement += std::to_string(journalSizeBytes);
    journalSizeStatement += ";";

    SqliteStatement journalSize(GetSqliteDbHandle(), journalSizeStatement);
    journalSize.Bind().Step();

    std::string autocheckpointStatement = "PRAGMA wal_autocheckpoint = ";
    autocheckpointStatement += std::to_string(autoCheckpointNumPages);
    autocheckpointStatement += ";";

    SqliteStatement checkpoint(GetSqliteDbHandle(), autocheckpointStatement);
    checkpoint.Bind().Step();
}

/*++
    Sets up WAL mode for this database connection.
    https://sqlite.org/wal.html

Arguments:
    journalSizeBytes - The maximum size of the write-ahead log journal
    autoCheckpointNumPages - The maximum number of pages in the write-ahead
                               log (multiply by page size to get size in bytes)
--*/
void SqlConnection::SetWalMode(uint32_t journalSizeBytes, uint32_t autoCheckpointNumPages)
{
    TraceLoggingWrite(
        g_tlgProv,
        "SqlConnection_SetWalMode",
        TraceLoggingValue(journalSizeBytes, "JournalSizeBytes"),
        TraceLoggingValue(autoCheckpointNumPages, "AuthocheckpointNumPages"));

    SqliteStatement journalingModeWal(
        GetSqliteDbHandle(), "PRAGMA journal_mode = WAL;");
    journalingModeWal.Bind().Step();

    UpdateWalJournalSize(journalSizeBytes, autoCheckpointNumPages);
}

bool SqlConnection::IsTransactionStarted()
{
    // Returns 0 if in open transaction
    return sqlite3_get_autocommit(_dbConnectionHandle.get()) == 0;
}

}}}
