// Copyright (c) Microsoft Corporation. All rights reserved.

#pragma once

#include <memory>
#include <sqlite3.h>
#include <vector>

#include <azure-device-util/ResultException.h>
#include <nonstd/span.hpp>
#include <uuid/uuid.h>

namespace Azure { namespace Device { namespace Telemetry {

using SqliteResult = int;

std::error_code make_sqlite_error(SqliteResult sqliteErrorCode);

bool IsErrorSqliteCorrupt(std::error_code ec);
bool IsErrorSqliteFull(std::error_code ec);

struct sqlite3_stmt_reset_deleter
{
    void operator()(sqlite3_stmt* stmt) { sqlite3_reset(stmt); }
};

using unique_statement_reset =
    std::unique_ptr<sqlite3_stmt, sqlite3_stmt_reset_deleter>;

using shared_statement_reset = std::shared_ptr<sqlite3_stmt>;

struct sqlite3_deleter
{
    void operator()(sqlite3* sql)
    {
        LOG_IF_FAILED(make_sqlite_error(sqlite3_close(sql)));
    }
};

struct sqlite3_stmt_finalize_deleter
{
    void operator()(sqlite3_stmt* stmt)
    {
        LOG_IF_FAILED(make_sqlite_error(sqlite3_finalize(stmt)));
    }
};

using unique_sqlite3_db = std::unique_ptr<sqlite3, sqlite3_deleter>;

using unique_sqlite3_stmt =
    std::unique_ptr<sqlite3_stmt, sqlite3_stmt_finalize_deleter>;

class SqliteQueryResult
{
  public:
    SqliteQueryResult(shared_statement_reset statementHandle, uint32_t colCount)
        : _statement(statementHandle), _columnCount(colCount)
    {}

    template<typename T>
    T NextInt()
    {
        FAIL_FAST_IF(!HasNextColumn());

        return static_cast<T>(sqlite3_column_int(
            _statement.get(), static_cast<int>(_columnIndex++)));
    }

    template<typename T>
    T NextInt64()
    {
        FAIL_FAST_IF(!HasNextColumn());

        return static_cast<T>(sqlite3_column_int64(
            _statement.get(), static_cast<int>(_columnIndex++)));
    }

    nonstd::span<const std::byte> NextBlob()
    {
        FAIL_FAST_IF(!HasNextColumn());

        uint32_t colIndex = _columnIndex++;
        const std::byte* blobPtr = static_cast<const std::byte*>(
            sqlite3_column_blob(_statement.get(), static_cast<int>(colIndex)));
        int32_t blobLen =
            sqlite3_column_bytes(_statement.get(), static_cast<int>(colIndex));

        return nonstd::span<const std::byte>{ blobPtr,
                                              static_cast<size_t>(blobLen) };
    }

    std::string_view NextString()
    {
        FAIL_FAST_IF(!HasNextColumn());

        uint32_t colIndex = _columnIndex++;
        const char* str = reinterpret_cast<const char*>(
            sqlite3_column_text(_statement.get(), static_cast<int>(colIndex)));
        int32_t strLen =
            sqlite3_column_bytes(_statement.get(), static_cast<int>(colIndex));

        return std::string_view{ str, static_cast<size_t>(strLen) };
    }

  private:
    bool HasNextColumn() { return _columnIndex < _columnCount; }

  private:
    shared_statement_reset _statement;
    uint32_t _columnIndex = 0;
    uint32_t _columnCount = 0;
};

class SqliteBoundStatement
{
  public:
    SqliteBoundStatement(sqlite3_stmt* statementHandle, uint32_t colCount)
        : _statement(statementHandle, sqlite3_stmt_reset_deleter{})
        , _columnCount(colCount)
    {}

    SqliteBoundStatement(const SqliteBoundStatement& other) = delete;
    SqliteBoundStatement& operator=(const SqliteBoundStatement& other) = delete;
    SqliteBoundStatement(SqliteBoundStatement&& other) = default;
    SqliteBoundStatement& operator=(SqliteBoundStatement&& other) = default;

    std::unique_ptr<SqliteQueryResult> Step()
    {
        SqliteResult stepResult = sqlite3_step(_statement.get());
        if (stepResult == SQLITE_ROW)
        {
            return std::make_unique<SqliteQueryResult>(_statement, _columnCount);
        }
        else if (stepResult == SQLITE_DONE)
        {
            return nullptr;
        }
        else
        {
            THROW_HR(make_sqlite_error(stepResult));
        }
    }

  private:
    shared_statement_reset _statement;
    uint32_t _columnCount;
};

/*++
    Helper class for wrapping sqlite statements. This class creates and destroys
statement handles so the user does not have to worry about closing the handles
and handles leaking.

    The operations that act on individual statements are here:
        - Step
        - Bind (implicit and explicit)
        - Reset statements
        - Prepare Statement
--*/
class SqliteStatement
{
  public:
    template<typename... Values>
    SqliteStatement(
        const unique_sqlite3_db& dbConnectionHandle,
        std::string_view statementText)
    {
        sqlite3_stmt* value = nullptr;
        THROW_IF_FAILED(make_sqlite_error(sqlite3_prepare_v2(
            dbConnectionHandle.get(),
            statementText.data(),
            static_cast<int32_t>(statementText.size()),
            &value,
            nullptr)));

        _statementHandle.reset(value);

        _columnCount =
            static_cast<uint32_t>(sqlite3_column_count(_statementHandle.get()));
    }

    template<typename... Values>
    SqliteBoundStatement Bind(Values&&... values)
    {
        BindImpl(0, std::forward<Values>(values)...);
        return SqliteBoundStatement{ _statementHandle.get(), _columnCount };
    }

  private:
    void BindImpl(int32_t) {}

    template<typename First, typename... Rest>
    void BindImpl(int32_t index, First&& first, Rest&&... rest)
    {
        int32_t newIndex = index + 1;

        BindCol(newIndex, std::forward<First>(first));
        BindImpl(newIndex, std::forward<Rest>(rest)...);
    }

    void BindCol(int32_t index, int64_t value) const
    {
        FAIL_FAST_IF_FAILED(make_sqlite_error(
            sqlite3_bind_int64(_statementHandle.get(), index, value)));
    }

    // Binds string values
    void BindCol(int32_t index, std::string_view value) const
    {
        int32_t size = static_cast<int32_t>(value.size());
        FAIL_FAST_IF_FAILED(make_sqlite_error(sqlite3_bind_text(
            _statementHandle.get(), index, value.data(), size, SQLITE_TRANSIENT)));
    }

    // Blob
    void BindCol(int32_t index, const std::vector<std::byte>& blob) const
    {
        FAIL_FAST_IF_FAILED(make_sqlite_error(sqlite3_bind_blob(
            _statementHandle.get(),
            index,
            blob.data(),
            static_cast<int32_t>(blob.size()),
            SQLITE_TRANSIENT)));
    }

    // Guid (as binary data / Blob)
    void BindCol(int32_t index, const uuid_t& uuid) const
    {
        FAIL_FAST_IF_FAILED(make_sqlite_error(sqlite3_bind_blob(
            _statementHandle.get(),
            index,
            reinterpret_cast<const void*>(&uuid),
            static_cast<int32_t>(sizeof(uuid)),
            SQLITE_TRANSIENT)));
    }

  private:
    unique_sqlite3_stmt _statementHandle;
    uint32_t _columnCount;
};

/*++
    Helper class for wrapping sqlite db connection. This class creates and
destroys the sqlite database handle so the user does not have to worry about
    closing the handles and handles leaking.

    The operations that act directly on the database are here.
        - Vacuum
        - Transaction Status
        - Open Transaciton
        - Commit Transaciton
        - Force checkpoint database
--*/
class SqlConnection
{
  public:
    SqlConnection(std::string_view databaseDirectory, bool skipQuickIntegrityCheck);

    void TryBeginTransaction();
    void CommitIfTransactionStarted();

    void Vacuum();
    void CheckpointDb();

    double GetPercentageFull() const { return _percentageFull; }
    void UpdatePercentageFull();

    uint32_t GetCurrentUsedPageCount();

    uint32_t GetMaxPageCount() const { return _maxPageCount; }
    uint32_t
    SetMaxPageCount(uint32_t desiredMaxPageCount, bool skipMmapPragma = false);

    void SetWalMode(uint32_t journalSizeBytes, uint32_t autoCheckpointNumPages);

    unique_sqlite3_db& GetSqliteDbHandle() { return _dbConnectionHandle; }

  private:
    bool IsTransactionStarted();

    void
    UpdateWalJournalSize(uint32_t journalSizeBytes, uint32_t autoCheckpointNumPages);

  private:
    unique_sqlite3_db _dbConnectionHandle;
    uint32_t _maxPageCount = 0;
    double _percentageFull = 0;
};

}}}
