// Copyright (c) Microsoft Corporation. All rights reserved.

#include "SqliteEventStore.h"

#include <memory>

#include <azure-device-util/ConversionFactors.h>
#include <azure-device-util/FileSystem.h>
#include <azure-device-util/ResultException.h>
#include <azure-device-util/String.h>

#include "AsimovSyntheticEvent.h"
#include "HeartBeat.h"
#include "IEventSerializer.h"
#include "Sqlite.h"
#include "SqliteTable.h"
#include "TelemetryError.h"
#include "TlgProvider.h"

namespace Azure { namespace Device { namespace Telemetry {

constexpr std::string_view SqliteEventStore::k_storageFilePattern;

/*++
    Creates an event store of a given total size.

Arguments:
    heartbeat
        - HeartBeat to log stats to
    serializer
        - Serializer to write events to
    databaseDirectory
        - Full directory path where eventstorage(db) file should be kept.
            If this is empty, event storage will be in memory instead of being
            backed by an actual persistent file on the disk.
    maxSizeInBytes
        - Total number of bytes to be used by the event store.
    preCompressBufferNormalEventCount
        - Cap on the number of events that can be stored in a single upload
            buffer (for BufferType::Normal)
    preCompressBufferCriticalEventCount
        - Cap on the number of events that can be stored in a single upload
            buffer (for all other BufferType)
    preCompressBufferByteLimit
        - Cap on the number of bytes that can be stored in a single upload
            buffer
--*/
SqliteEventStore::SqliteEventStore(
    HeartBeat& heartBeat,
    IEventSerializer& serializer,
    IAsimovEventBufferFactory& asimovEventBufferFactory,
    std::string_view databaseDirectory,
    uint32_t maxSizeInBytes,
    uint32_t preCompressBufferNormalEventCount,
    uint32_t preCompressBufferCriticalEventCount,
    uint32_t preCompressBufferByteLimit)
    : _heartBeat(heartBeat)
    , _serializer(serializer)
    , _eventBufferFactory(asimovEventBufferFactory)
    , _databaseDirectory(databaseDirectory)
    , _eventStoreSizeInBytes(maxSizeInBytes)
    , _preCompressBufferNormalEventCount(preCompressBufferNormalEventCount)
    , _preCompressBufferCriticalEventCount(preCompressBufferCriticalEventCount)
    , _preCompressBufferByteLimit(preCompressBufferByteLimit)
{
    try
    {
        Initialize();
    }
    catch (...)
    {
        std::error_code databaseCreationHR = Utils::ResultFromCaughtException();

        // If the database was corrupted at the beginning, report it now.
        if (databaseCreationHR != TelemetryError::E_DB_SCHEMA_CHANGED)
        {
            auto event = std::make_unique<AsimovSyntheticEvent>(
                "SqliteEventStore_StorageCreationFailed",
                TriggerPersistence::Critical);

            event->AddPartCField("Path", _databaseDirectory);
            event->AddPartCField("ErrorCode", databaseCreationHR.message());

            _serializer.WriteSyntheticEvent(std::move(event));
        }
        ResetDatabase(databaseCreationHR);
    }
}

/*++
    Initializes the event store based on parameters from the constructor.

 Arguments:
    skipQuickIntegrityCheck - Set to skip integrity check on empty databases.
--*/
void SqliteEventStore::Initialize(bool skipQuickIntegrityCheck)
{
    TraceLoggingWrite(
        g_tlgProv,
        "SqliteEventStore_Initializer",
        TraceLoggingValue(skipQuickIntegrityCheck, "SkipQuickIntegrityCheck"));

    _dbConnection = std::make_unique<SqlConnection>(
        _databaseDirectory, skipQuickIntegrityCheck);

    _sqliteTable = std::make_unique<SqliteTable>(
        *_dbConnection,
        _heartBeat,
        _eventBufferFactory,
        _preCompressBufferNormalEventCount,
        _preCompressBufferCriticalEventCount,
        _preCompressBufferByteLimit);

    // Set up write-ahead log mode for this connection (see documentation on
    // function for more info)
    _dbConnection->SetWalMode(1048576u, 256u);

    //
    // Each sqlite page is 4KB
    // Hence, m_maxPageCount should be set to desired max_db_size (kb) / 4.
    // Note: PRAGMA m_maxPageCount is how we control the max file size. This
    // value is not persistent. It is a per connection setting. This means that
    // we have to set it every connection and every sql call during this
    // connection will be obey the file limit. Subsequent db connections or
    // connection from another source such as external tools can push the file
    // over this limit if they do not set the max_page_count.
    //
    EnforceEventStoreMaxSize();
}

/*++
    Uses the current _eventStoreSizeInBytes to set max_page_count for the DB.
    If the currently used page count is higher than that amount, rows will be
    deleted in order to trim the datababse to the new limit.
--*/
void SqliteEventStore::EnforceEventStoreMaxSize()
{
    try
    {
        TraceLoggingWrite(g_tlgProv, "SqliteEventStore_EnforceEventStoreMaxSize");

        uint32_t desiredMaxPageCount =
            _eventStoreSizeInBytes / (ConversionFactors::kbToBytes * 4);

        bool skipMmapPragma = _databaseDirectory.empty();

        uint32_t actualMaxPageCount =
            _dbConnection->SetMaxPageCount(desiredMaxPageCount, skipMmapPragma);

        // There is a loop here to deal with timing conditions where the
        // database manages to grow between a delete and vacuum
        while (actualMaxPageCount > desiredMaxPageCount)
        {
            // Delete data until page count used is less than desired max
            DeleteRowsUntilPageCountIsMet(desiredMaxPageCount);

            // Then perform one Vacuum, because it's very expensive.
            _dbConnection->Vacuum();

            // Try and set the max page count again.
            actualMaxPageCount = _dbConnection->SetMaxPageCount(
                desiredMaxPageCount, skipMmapPragma);
        }
    }
    catch (const std::system_error& ex)
    {
        CheckAndHandleCorruption(ex);

        // Log any failure to shrink, but don't bother bubbling up to the caller
        // who has no remediation steps
        TraceLoggingWrite(
            g_tlgProv,
            "SqliteEventStore_FailedToShrinkPagesFurther",
            TraceLoggingErrorCode(ex.code(), "ErrorCode"));
    }

    // Update percentage full now that page_count and max_page_count have been
    // changed
    _dbConnection->UpdatePercentageFull();
}

/*++
    Deletes chunks of rows in the database (starting with lowest priority data)
    until the currently used page_count is less than or equal to the
    desiredPageCount.

Arguments:
    desiredPageCount - Page count to trim the database to
--*/
void SqliteEventStore::DeleteRowsUntilPageCountIsMet(uint32_t desiredPageCount)
{
    TraceLoggingWrite(
        g_tlgProv,
        "SqliteEventStore_DeleteRowsUntilPageCountIsMet",
        TraceLoggingValue(desiredPageCount, "DesiredPageCount"));

    uint32_t pageCount = 0;
    do
    {
        //
        // Delete a chunk of any type of data (in priority order) to get the
        // page count down. It should not be possible to fail to delete enough
        // rows to shrink, but throw rather than crashing. Catch this higher up
        // the stack and log a shrink failure
        //
        THROW_HR_IF(
            TelemetryError::E_DB_CANNOT_SHRINK,
            !_sqliteTable->DeleteLowerPriorityData(BufferType::Realtime));

        // Update page count to see if we got below the limit
        pageCount = _dbConnection->GetCurrentUsedPageCount();

    } while (pageCount > desiredPageCount);
}

/*++
    Resets the pointers holding resources to the sqlite database, discards the
    database file and initializes a new database with a new file.

    This is reported to the heartbeat.

    This function will fail fast in case of error since we cannot continue
    without a storage layer.

Arguments:
    storageError - std::error_code that triggered this reset
--*/
void SqliteEventStore::ResetDatabase(std::error_code storageError) noexcept
{
    try
    {
        TraceLoggingWrite(g_tlgProv, "SqliteEventStore_BeginReset");

        // Restart objects holding database resources
        _sqliteTable.reset();
        _dbConnection.reset();

        int64_t size = 0;

        // Get file size if possible.
        if (!_databaseDirectory.empty())
        {
            if (storageError != TelemetryError::E_DB_SCHEMA_CHANGED)
            {
                std::string fileName = _databaseDirectory;
                fileName += "/EventStore.db";

                if (FAILED(Utils::FileSystem::ReadFileSize(fileName, size)))
                {
                    size = 0;
                }

                // Report drop data to heartbeat
                _heartBeat.RecordDbReset(static_cast<uint32_t>(size));
            }

            FAIL_FAST_IF_FAILED(Utils::FileSystem::DeleteAllFilesWithPattern(
                _databaseDirectory, SqliteEventStore::k_storageFilePattern));
        }

        // Re-initialize the db
        Initialize(true);
    }
    CATCH_FAIL_FAST();
}

SqliteEventStore::~SqliteEventStore()
{
    for (TriggerLatency latency : c_triggerLatencies)
    {
        SqliteEventStore::FlushPreCompressBuffers(latency);
    }
    _dbConnection->CommitIfTransactionStarted();
}

/*++
    Adds a single metadata+payload entry to the event store.
    Higher priority data can overwrite lower priority data in the sqliteTables.
    sqliteTable priorities are defined as Normal = 0, Critical = 1, Realtime = 2

Arguments:
    uploadData  - A single metadata+payload to be added to the event store

Return Value:
    true    - The event was added succesfully to the event store
    false   - The event was dropped outright.  This only happens in the
                attempted case of Normal data being written to a full
                sqliteTable OR a high priority event thrown away because the
                sqliteTable was full with high priority events as well OR
                unknown corruption that prevented insertion.
--*/
bool SqliteEventStore::AddData(const UploadData& uploadData)
{
    std::lock_guard<std::recursive_mutex> lock{ _mutex };

    // If we are above k_percentageThresholdForBatchingTransactions we no longer
    // accept Normal/Normal data
    if (_dbConnection->GetPercentageFull() >
            SqliteEventStore::k_percentageThresholdForBatchingTransactions &&
        (uploadData.Latency == TriggerLatency::Normal) &&
        (uploadData.Persistence == TriggerPersistence::Normal))
    {
        // Database too full to accept normal events.
        _heartBeat.RecordDbDropped(1, uploadData.Persistence, true);

        return false;
    }

    //
    // If we are above k_percentageThresholdForBatchingTransactions we commit
    // transaction if open. This ensures no rollback errors due to fullness of
    // the db. If under k_percentageThresholdForBatchingTransactions we open a
    // transaction in an attempt to batch. The transaction will close if the
    // upload loop calls flush. No need to open transaction if we are doing
    // realtime events. Need those to go into the db asap.
    //
    if (_dbConnection->GetPercentageFull() >=
        SqliteEventStore::k_percentageThresholdForBatchingTransactions)
    {
        _dbConnection->CommitIfTransactionStarted();
    }
    else if (
        _dbConnection->GetPercentageFull() <
            SqliteEventStore::k_percentageThresholdForBatchingTransactions &&
        TriggerLatency::Realtime != uploadData.Latency)
    {
        _dbConnection->TryBeginTransaction();
    }

    // Write will handle failures and update heartbeat
    std::error_code writeResult = _sqliteTable->Write(uploadData);
    if (FAILED(writeResult))
    {
        CheckAndHandleCorruption(writeResult);
        return false;
    }

    return true;
}

/*++
    Deletes an entry from a substore that was fetched from a read.

Arguments:
    bufferSeqNum        - ID of row to delete
    commitTransaction   - Should commit transaction if open after the delete.
                          Set to false as an optimization when deleting in bulk.
--*/
void SqliteEventStore::DeleteData(uint32_t bufferSeqNum, bool commitTransaction)
{
    std::lock_guard<std::recursive_mutex> lock{ _mutex };

    try
    {
        _sqliteTable->Delete(static_cast<int32_t>(bufferSeqNum));
    }
    catch (const std::system_error& ex)
    {
        CheckAndHandleCorruption(ex);
    }

    if (commitTransaction)
    {
        _dbConnection->CommitIfTransactionStarted();
    }
}

/*++
    Increment the read row's UploadFailCount for how many times this buffer has
    failed to upload.

Arguments:
    readResult - Data and metadata fetched from a read
--*/
void SqliteEventStore::IncrementUploadFailCount(const ReadResult& readResult)
{
    std::lock_guard<std::recursive_mutex> lock{ _mutex };

    TraceLoggingWrite(
        g_tlgProv,
        "SqliteEventStore_IncrementUploadFailCount",
        TraceLoggingValue(readResult.SequenceNumber, "SeqNum"));

    try
    {
        _sqliteTable->SetUploadFailureCounter(
            static_cast<int32_t>(readResult.SequenceNumber),
            static_cast<int32_t>(readResult.UploadFailCount + 1));
    }
    catch (const std::system_error& ex)
    {
        CheckAndHandleCorruption(ex);
    }

    _dbConnection->CommitIfTransactionStarted();
}

/*++
    Commits all sqlite db's open transctions and writes them to disk.

    Note: This function does not flush out the contents of the pre-compress
    buffer to disk.
--*/
void SqliteEventStore::FlushToFile()
{
    std::lock_guard<std::recursive_mutex> lock{ _mutex };

    TraceLoggingWrite(g_tlgProv, "SqliteEventStore_FlushToFile");

    _dbConnection->CommitIfTransactionStarted();
}

/*++
    Tries to compress, write, and empty the pre-compress buffers for a given
    latency.

Arguments:
    latency - PreCompressBuffer matching this latency will be flushed.
--*/
void SqliteEventStore::FlushPreCompressBuffers(TriggerLatency latency)
{
    std::lock_guard<std::recursive_mutex> lock{ _mutex };

    TraceLoggingWrite(
        g_tlgProv,
        "EventStore_FlushingPreCompressBuffers",
        TraceLoggingValue(static_cast<int8_t>(latency), "Latency"));

    _sqliteTable->FlushPreCompressBuffers(latency);

    _dbConnection->CommitIfTransactionStarted();
}

/*++
    Determines whether all of pre-compress buffers for the given latency are
    empty

Arguments:
    latency - Latency value

Return Value:
    true  - latency has at least one event sitting in a preCompress buffer
    false - latency has completely empty preCompress buffers
--*/
bool SqliteEventStore::ArePreCompressBuffersEmpty(TriggerLatency latency)
{
    std::lock_guard<std::recursive_mutex> lock{ _mutex };
    return _sqliteTable->ArePreCompressBuffersEmpty(latency);
}

/*++
    Determines whether there is at least one row of the given persistence
    latency combination.

Arguments:
    persistence - Persistence value - std::nullopt_t means to aggregate stores
                    of all persistence values.
    latency     - Latency value - std::nullopt_t means to aggregate stores of
                    all latency values.

Return Value:
    true  - The selected combination of persistence + latency has at least one
                upload payload remaining
    false - The selected group of combination of persistence + latency is
                completely empty
--*/
bool SqliteEventStore::HasRemainingData(
    const TriggerPersistence* persistence,
    const TriggerLatency* latency)
{
    std::lock_guard<std::recursive_mutex> lock{ _mutex };

    _dbConnection->CommitIfTransactionStarted();

    try
    {
        return _sqliteTable->HasRemainingData(persistence, latency);
    }
    catch (const std::system_error& ex)
    {
        CheckAndHandleCorruption(ex);
        return false;
    }
}

/*++
    Fetches one entry from a sqlite db matching the specified latency.

Arguments:
    latency - Latency value of events to pull out of the event store.
                Only events matching this latency value will be fetched.
    result  - ReadResult containing the data fetched as well as metadata needed
                to erase it.

Return Value:
    ReadResult if success, or empty
--*/
std::unique_ptr<ReadResult>
SqliteEventStore::GetDataByLatency(TriggerLatency latency)
{
    std::lock_guard<std::recursive_mutex> lock{ _mutex };

    // No need to commit transaction here since the upload loop flushes before
    // doing a get.
    try
    {
        return _sqliteTable->Read(latency);
    }
    catch (const std::system_error& ex)
    {
        CheckAndHandleCorruption(ex);
        return {};
    }
}

/*++
    Calculates the percentage of pages used in the whole db.

    We use 3 different variables to calculate this:
        freelist_count
            - Number of pages free in the file allocated by sqlite so far.
        page_count
            - Number of pages used by sqlite
        _eventStoreSizeInBytes
            - Max number of pages allowed to be allocated by the sqlite file.

Return Value:
    [0, 100.0] value representing the percentage of pages used versus the total
    number of pages
--*/
double SqliteEventStore::GetPercentageFull() const
{
    std::lock_guard<std::recursive_mutex> lock{ _mutex };
    return _dbConnection->GetPercentageFull();
}

/*++
    Stores a tenant token in the buffer.

Arguments:
    persistence - The persistence this event has.
    latency     - The latency this event has.
    tenantToken - Tenant token (no O: prefix) that will be added to the
                    collection for the buffer.
--*/
void SqliteEventStore::StoreTenantToken(
    TriggerPersistence persistence,
    TriggerLatency latency,
    std::string&& tenantToken)
{
    FAIL_FAST_IF(
        tenantToken.empty() || Utils::String::IStartsWith(tenantToken, "o:"));

    std::lock_guard<std::recursive_mutex> lock{ _mutex };

    return _sqliteTable->StoreTenantToken(
        persistence, latency, std::move(tenantToken));
}

uint32_t SqliteEventStore::GetEventStoreSizeBytes() const
{
    return _dbConnection->GetMaxPageCount() * 4096;
}

/*++
    Sets event store size in bytes.  Will throw in case of failure shrinking to
    needed size.

Arguments:
    bytes - size in bytes
--*/
void SqliteEventStore::SetEventStoreSizeBytes(uint32_t bytes)
{
    std::lock_guard<std::recursive_mutex> lock{ _mutex };

    _eventStoreSizeInBytes = bytes;
    EnforceEventStoreMaxSize();
}

/*++
    Fetches every entry from every substore.  Deletes data as it pulls it out
    because there's currently no way to have multiple consecutive reads without
    erasing to move the read pointer forward.

    Only used for tests!

Return Value:
    Vector of ReadResults that contain entry compressed payload+metadata for
    everything that was in the event store.
--*/
std::vector<ReadResult> SqliteEventStore::GetAndDeleteAllData()
{
    std::lock_guard<std::recursive_mutex> lock{ _mutex };

    TraceLoggingWrite(g_tlgProv, "SqliteEventStore_GetAndDeleteAllData");

    _dbConnection->TryBeginTransaction();

    std::vector<ReadResult> results;

    for (TriggerLatency latency : c_triggerLatencies)
    {
        GetAndDeleteAllDataByLatency(latency, results);
    }

    _dbConnection->CommitIfTransactionStarted();
    return results;
}

/*++
    Fetches and deletes all events from the substores of a particular latency.

    Only used for tests!

Arguments:
    latency - Latency value of events to pull out of the event store.
                Only events matching this latency value will be fetched.
    result  - ReadResult containing the data fetched as well as metadata needed
                to erase it.
--*/
void SqliteEventStore::GetAndDeleteAllDataByLatency(
    TriggerLatency latency,
    std::vector<ReadResult>& results)
{
    std::lock_guard<std::recursive_mutex> lock{ _mutex };

    TraceLoggingWrite(
        g_tlgProv,
        "SqliteEventStore_GetAndDeleteAllDataByLatency",
        TraceLoggingValue(static_cast<int8_t>(latency), "Latency"));

    while (auto readResult = GetDataByLatency(latency))
    {
        DeleteData(readResult->SequenceNumber, false);
        results.emplace_back(std::move(*readResult));
    }
}

void SqliteEventStore::CheckAndHandleCorruption(const std::system_error& ex) noexcept
{
    if (IsErrorSqliteCorrupt(ex.code()))
    {
        ResetDatabase(ex.code());
    }
}

}}}
