// Copyright (c) Microsoft Corporation. All rights reserved.

#pragma once

#include <string_view>

#include "IEventStore.h"

namespace Azure { namespace Device { namespace Telemetry {

using namespace std::string_view_literals;

class IEventSerializer;
class IAsimovEventBufferFactory;
class HeartBeat;
class SqlConnection;
class SqliteTable;

class SqliteEventStore : public IEventStore
{
  public:
    SqliteEventStore(
        HeartBeat& heartBeat,
        IEventSerializer& serializer,
        IAsimovEventBufferFactory& asimovEventBufferFactory,
        std::string_view databaseDirectory,
        uint32_t maxSizeInBytes,
        uint32_t preCompressBufferNormalEventCount,
        uint32_t preCompressBufferCriticalEventCount,
        uint32_t preCompressBufferByteLimit);

    ~SqliteEventStore();

    std::unique_ptr<ReadResult> GetDataByLatency(TriggerLatency latency) override;

    bool AddData(const UploadData& uploadData) override;

    void
    DeleteData(uint32_t bufferSeqNum, bool commitTransaction = true) override;

    void IncrementUploadFailCount(const ReadResult& readResult) override;

    void FlushToFile() override;

    void FlushPreCompressBuffers(TriggerLatency latency) override;

    double GetPercentageFull() const override;

    bool HasRemainingData(
        const TriggerPersistence* persistence,
        const TriggerLatency* latency) override;

    bool ArePreCompressBuffersEmpty(TriggerLatency latency) override;

    void StoreTenantToken(
        TriggerPersistence persistence,
        TriggerLatency latency,
        std::string&& tenantToken) override;

    std::recursive_mutex& GetEventStoreMutex() const override
    {
        return _mutex;
    };

    uint32_t GetEventStoreSizeBytes() const override;
    void SetEventStoreSizeBytes(uint32_t bytes) override;

  public:
    static constexpr std::string_view k_storageFilePattern = ".*\\.db"sv;
    static constexpr double k_percentageThresholdForBatchingTransactions = 90.0;

  private:
    std::vector<ReadResult> GetAndDeleteAllData() override;

    void GetAndDeleteAllDataByLatency(
        TriggerLatency latency,
        std::vector<ReadResult>& results);

    void Initialize(bool skipQuickIntegrityCheck = false);

    void ResetDatabase(std::error_code storageError) noexcept;

    void EnforceEventStoreMaxSize();

    void DeleteRowsUntilPageCountIsMet(uint32_t desiredPageCount);

    void CheckAndHandleCorruption(const std::system_error& ex) noexcept;

  private:
    mutable std::recursive_mutex _mutex;

    HeartBeat& _heartBeat;
    IEventSerializer& _serializer;
    IAsimovEventBufferFactory& _eventBufferFactory;

    const std::string _databaseDirectory;
    uint32_t _eventStoreSizeInBytes;
    const uint32_t _preCompressBufferNormalEventCount;
    const uint32_t _preCompressBufferCriticalEventCount;
    const uint32_t _preCompressBufferByteLimit;

    std::unique_ptr<SqlConnection> _dbConnection;
    std::unique_ptr<SqliteTable> _sqliteTable;
};

}}}
