// Copyright (c) Microsoft Corporation. All rights reserved.

#include "SqliteTable.h"

#include <memory>
#include <sstream>

#include <azure-device-util/ResultException.h>
#include <azure-device-util/ScopeExit.h>

#include "AsimovEventBuffer.h"
#include "HeartBeat.h"
#include "Sqlite.h"
#include "SqliteEventStore.h"
#include "TelemetryError.h"
#include "TlgProvider.h"

namespace Azure { namespace Device { namespace Telemetry {

/*++
    Creates a table in the sqlite database for holding compressed payload and
    related metadata. Also creates the corresponding pre-compress buffers.

Arguments:
    dbConnection
        - DB Connection to write to
    heartBeat
        - HeartBeat to write stats to
    preCompressBufferNormalEventCount
        - Cap on the number of events that can be stored in a single upload
            buffer (for BufferType::Normal)
    preCompressBufferCriticalEventCount
        - Cap on the number of events that can be stored in a single upload
            buffer (for all other BufferType)
    preCompressBufferByteLimit
        - Cap on the number of bytes that can be stored in a single upload
            buffer
--*/
SqliteTable::SqliteTable(
    SqlConnection& dbConnection,
    IHeartBeat& heartBeat,
    IAsimovEventBufferFactory& asimovEventBufferFactory,
    uint32_t preCompressBufferNormalEventCount,
    uint32_t preCompressBufferCriticalEventCount,
    uint32_t preCompressBufferByteLimit)
    : _dbConnection(dbConnection), _heartBeat(heartBeat)
{
    // Read user_version out of the database, if it exists
    SqliteStatement userVersion(
        _dbConnection.GetSqliteDbHandle(), "PRAGMA user_version;");

    // Expect a single row result with a single column, so use this compact
    // syntax
    uint32_t savedVersion = userVersion.Bind().Step()->NextInt<uint32_t>();

    if (savedVersion == 0)
    {
        std::string userVersionUpdate = "PRAGMA user_version=";
        userVersionUpdate += std::to_string(k_currentEventStoreVersion);
        userVersionUpdate += ";";

        SqliteStatement userVersionUpdateStatement(
            _dbConnection.GetSqliteDbHandle(), userVersionUpdate);
        userVersionUpdateStatement.Bind().Step();
    }
    else if (savedVersion != k_currentEventStoreVersion)
    {
        THROW_HR(TelemetryError::E_DB_SCHEMA_CHANGED);
    }

    auto executeSql = [this](std::string_view sqlStr) {
        SqliteStatement statement{ _dbConnection.GetSqliteDbHandle(), sqlStr };
        statement.Bind().Step();
    };

    //
    // We must manually create a table with RowId as integer primary key or
    // VACUUM will re-assign rows. This rowId will be a quick/performant way to
    // access data when iterating and deleting. Note: If you change this schema
    // plesae go update k_currentEventStoreVersion.
    //
    // Note: Update changes to table schema in SqlTableDumper tool
    //     Inconsistent table schema will cause dumper tool to fail.
    executeSql("CREATE TABLE IF NOT EXISTS AsimovEvents ("
               "RowId                          INTEGER PRIMARY KEY NOT NULL,"
               "UTF8ApiKeyHeader               TEXT,"
               "Payload                        BLOB,"
               "NumberOfEvents                 INTEGER,"
               "Latency                        INTEGER,"
               "Persistence                    INTEGER,"
               "UploadFailCount                INTEGER,"
               "RowSize                        INTEGER)");

    executeSql("CREATE INDEX IF NOT EXISTS Latency "
               "ON AsimovEvents (Latency)");

    executeSql("CREATE INDEX IF NOT EXISTS Persistence "
               "ON AsimovEvents (Persistence)");

    executeSql("CREATE INDEX IF NOT EXISTS LatencyAndPersistence "
               "ON AsimovEvents (Latency, Persistence)");

    PrepareStatements();

    // Create precompressbuffers
    for (BufferType bufferType : c_bufferTypes)
    {
        uint32_t eventCount = (bufferType == BufferType::Normal) ?
            preCompressBufferNormalEventCount :
            preCompressBufferCriticalEventCount;

        auto eventBuffer = asimovEventBufferFactory.Create(
            preCompressBufferByteLimit, eventCount);
        _preCompressBuffers.emplace_back(eventBuffer);
    }
}

SqliteTable::~SqliteTable() = default;

/*++
    Prepares statements that are used more than once. Run during construction.
    If any of these statements fail it is expected to throw.
--*/
void SqliteTable::PrepareStatements()
{
    auto prepStatement = [this](std::string_view sqlStr) {
        return std::make_unique<SqliteStatement>(
            _dbConnection.GetSqliteDbHandle(), sqlStr);
    };

    // Select a row
    _getDataRow = prepStatement("SELECT RowId,"
                                "UTF8ApiKeyHeader,"
                                "Payload,"
                                "NumberOfEvents,"
                                "UploadFailCount "
                                "FROM AsimovEvents "
                                "WHERE Latency = ?1 "
                                "ORDER BY RowId ASC "
                                "LIMIT 1");

    // Delete a row
    _deleteRow = prepStatement("DELETE FROM AsimovEvents "
                               "WHERE RowId = ?1;");

    // Insert a row
    _insertRow = prepStatement("INSERT INTO AsimovEvents ("
                               "UTF8ApiKeyHeader,"
                               "Payload,"
                               "NumberOfEvents,"
                               "Latency,"
                               "Persistence,"
                               "UploadFailCount,"
                               "RowSize) "
                               "VALUES "
                               "(?1,?2,?3,?4,?5,?6,?7)");

    // Update the 'UploadFailCount' column on ROW
    _updateUploadFailCount = prepStatement("UPDATE AsimovEvents "
                                           "SET UploadFailCount = ?1 "
                                           "WHERE RowId = ?2");

    // Check if data exists
    _dataExists = prepStatement("SELECT EXISTS "
                                "(SELECT * FROM AsimovEvents LIMIT 1);");

    // Check if Persistence and Latncy specific data exists
    _latencyAndPersistenceDataExists =
        prepStatement("SELECT EXISTS "
                      "(SELECT * FROM AsimovEvents "
                      "WHERE Latency = ?1 AND Persistence = ?2 LIMIT 1);");

    // Check if Latency specific data exists
    _latencyDataExists = prepStatement("SELECT EXISTS "
                                       "(SELECT * FROM AsimovEvents "
                                       "WHERE Latency = ?1 LIMIT 1);");

    // Check if Persistence specific data exists
    _persistenceDataExists = prepStatement("SELECT EXISTS "
                                           "(SELECT * FROM AsimovEvents "
                                           "WHERE Persistence = ?1 LIMIT 1);");

    // Get first 50 RowIds by sorted by least useful and oldest first
    _getRowIdSortedForDelete = prepStatement("SELECT Latency, Persistence, "
                                             "RowId "
                                             "FROM AsimovEvents "
                                             "ORDER BY "
                                             "Latency ASC, "
                                             "Persistence ASC, "
                                             "RowId ASC "
                                             "LIMIT 50;");
    _getSizeForRowId = prepStatement("SELECT RowSize FROM AsimovEvents WHERE "
                                     "RowId = ?1");
}

/*++
    Writes a single entry into the preCompressBuffers

Arguments:
    event - event to write
--*/
std::error_code SqliteTable::Write(const UploadData& event)
{
    try
    {
        IAsimovEventBuffer& buffer =
            GetPreCompressBuffer(event.Persistence, event.Latency);

        // If we are writing a real event and not just trying to flush the
        // buffer, try to add it to the  pre-compress buffer. Add event to the
        // pre-compress buffer.
        bool canAddAnotherEvent = buffer.AddEvent(event.Payload);

        TraceLoggingWrite(
            g_tlgProv,
            "SqliteTable_EventAddedSuccessfully",
            TraceLoggingValue(event.Payload.c_str(), "Payload"));

        // If the precompress buffer says there's still more room, there's
        // nothing left to do, this write has succeeded.
        if (canAddAnotherEvent)
        {
            return {};
        }

        //
        // If we got to this point, we know that the next event add will fail,
        // so we need to compress the buffer, write out to disk, reset buffer so
        // the next addition can succeed.
        //
        auto clearBufferOnExit =
            Utils::MakeScopeExit([&buffer]() { buffer.Clear(); });

        AsimovConstantEventBuffer constantEventBuffer = buffer.Commit();

        std::error_code insertSuccess =
            TryInsert(constantEventBuffer, event.Persistence, event.Latency);

        if (SUCCEEDED(insertSuccess))
        {
            TraceLoggingWrite(
                g_tlgProv,
                "EventStore_CommittedPreCompressBuffer",
                TraceLoggingValue(
                    constantEventBuffer.CompressedUtf8Json.size(), "NumBytes"),
                TraceLoggingValue(
                    constantEventBuffer.NumberOfEvents, "NumEvents"));
        }
        else
        {
            _heartBeat.RecordDbDropped(
                constantEventBuffer.NumberOfEvents, event.Persistence, true);
        }
        RETURN_IF_FAILED(insertSuccess);

        return {};
    }
    CATCH_RETURN();
}

/*++
    Reads a single event of the given latency.

Arguments:
    event - event to write

Return Value:
    ReadResult containing payload and metadata.  Empty if read failed.
--*/
std::unique_ptr<ReadResult> SqliteTable::Read(TriggerLatency latency)
{
    auto boundStatement = _getDataRow->Bind(static_cast<int>(latency));
    if (auto result = boundStatement.Step())
    {
        auto readResult = std::make_unique<ReadResult>();
        readResult->SequenceNumber = result->NextInt<uint32_t>();
        readResult->ApiKeyHttpHeader = result->NextString();

        auto payloadSpan = result->NextBlob();
        readResult->Payload.assign(payloadSpan.begin(), payloadSpan.end());

        readResult->EventCount = result->NextInt<uint32_t>();
        readResult->UploadFailCount = result->NextInt<uint32_t>();

        return readResult;
    }
    else
    {
        TraceLoggingWrite(g_tlgProv, "SqliteTable_Read_NoResult");

        return {};
    }
}

/*++
    Tries to compress, write, and empty the pre-compress buffers for a given
    latency.

Arguments:
    latency - PreCompressBuffers matching this latency will be flushed.
--*/
void SqliteTable::FlushPreCompressBuffers(TriggerLatency latency) noexcept
{
    TraceLoggingWrite(
        g_tlgProv,
        "SqliteTable_FlushPreCompressBuffers",
        TraceLoggingValue(static_cast<int8_t>(latency), "Latency"));

    // Get the precompress buffers for each storage buffer type (if there is
    // more than 1 for the current tenant. Ex: Windows Telemetry & Functional
    // Data)
    std::map<IAsimovEventBuffer*, TriggerPersistence> buffers;
    if (latency == TriggerLatency::Normal)
    {
        buffers.emplace(
            _preCompressBuffers[static_cast<int32_t>(BufferType::Normal)].get(),
            TriggerPersistence::Normal);
        buffers.emplace(
            _preCompressBuffers[static_cast<int32_t>(BufferType::Critical)].get(),
            TriggerPersistence::Critical);
    }
    else if (latency == TriggerLatency::Realtime)
    {
        buffers.emplace(
            _preCompressBuffers[static_cast<int32_t>(BufferType::Realtime)].get(),
            TriggerPersistence::Normal);
    }

    // Insert each buffer into sqlite database
    for (const auto& pair : buffers)
    {
        if (pair.first->IsEmpty())
        {
            continue;
        }

        AsimovConstantEventBuffer constantEventBuffer = pair.first->Commit();
        std::error_code result =
            TryInsert(constantEventBuffer, pair.second, latency);

        if (FAILED(result))
        {
            TraceLoggingWrite(
                g_tlgProv,
                "SqliteTable_TooFullToFlushPreCompressBuffer",
                TraceLoggingValue(
                    constantEventBuffer.CompressedUtf8Json.size(), "NumBytes"),
                TraceLoggingValue(
                    constantEventBuffer.NumberOfEvents, "NumEvents"));

            _heartBeat.RecordDbDropped(
                constantEventBuffer.NumberOfEvents, pair.second, true);
        }
        else
        {
            TraceLoggingWrite(
                g_tlgProv,
                "EventStore_CommittedPreCompressBuffer",
                TraceLoggingValue(
                    constantEventBuffer.CompressedUtf8Json.size(), "NumBytes"),
                TraceLoggingValue(
                    constantEventBuffer.NumberOfEvents, "NumEvents"));
        }
        pair.first->Clear();
    }
}

/*++
    Takes a vector of rowIds and deletes those rows from the sqlite table

Arguments:
    rowIds - vector of rowIds to delete from sqlite tale

Return Value:
    Returns number of estimated bytes deleted.
--*/
uint32_t SqliteTable::DeleteMultiple(const std::vector<int32_t>& rowIds)
{
    uint32_t sumDeleted = 0;
    for (int32_t rowId : rowIds)
    {
        uint32_t rowSize =
            _getSizeForRowId->Bind(rowId).Step()->NextInt<uint32_t>();
        sumDeleted += rowSize;

        _deleteRow->Bind(rowId).Step();
    }

    _dbConnection.UpdatePercentageFull();

    TraceLoggingWrite(
        g_tlgProv,
        "SqliteTable_DeletedMultiple",
        TraceLoggingValue(rowIds.size(), "NumRows"),
        TraceLoggingValue(sumDeleted, "BytesDeleted"));

    return sumDeleted;
}

/*++
    Deletes an entry from a substore that was fetched from a read.

Arguments:
    rowId             - rowId of the data to delete
--*/
void SqliteTable::Delete(int32_t rowId)
{
    _deleteRow->Bind(rowId).Step();

    _dbConnection.UpdatePercentageFull();
}

/*++
    Sets the upload failure counter for the given row ID

Arguments:
    rowId             - rowId of the data to update
    newValue          - Value for the UploadFailCount column
--*/
void SqliteTable::SetUploadFailureCounter(int32_t rowId, int32_t newValue)
{
    TraceLoggingWrite(
        g_tlgProv,
        "SqliteTable_SetUploadFailureCounter",
        TraceLoggingValue(rowId, "RowId"),
        TraceLoggingValue(newValue, "NumFailures"));

    _updateUploadFailCount->Bind(newValue, rowId).Step();
}

/*++
    Tries to write a AsimovConstantEventBuffer directly into the sqlite
    database. This is used when a Write call generates gets a full precompress
    buffer or we flush a precompress buffer. This call will delete lower
    priority data from the sqlite database if necessary.

Arguments:
    constantEventBuffer  - a commited/compressed buffer
    persistence          - persistence
    latency              - latency
--*/
std::error_code SqliteTable::TryInsert(
    const AsimovConstantEventBuffer& constantEventBuffer,
    TriggerPersistence persistence,
    TriggerLatency latency)
{
    //
    // This loop continues if the db was full and data was deleted.
    // If insert was successful or there was an error we return.
    //
    while (true)
    {
        uint64_t rowSize = (sizeof(int8_t) * 2)  // latency + persistence
            + sizeof(int32_t)                    // NumberOfEvents
            + constantEventBuffer.Utf8ApiKeyHeader.size()  // API header length
            + constantEventBuffer.CompressedUtf8Json.size();

        try
        {
            _insertRow
                ->Bind(
                    constantEventBuffer.Utf8ApiKeyHeader,
                    constantEventBuffer.CompressedUtf8Json,
                    static_cast<int>(constantEventBuffer.NumberOfEvents),
                    static_cast<int>(latency),
                    static_cast<int>(persistence),
                    0,
                    static_cast<int>(rowSize))
                .Step();

            _dbConnection.UpdatePercentageFull();

            // If we get here, we were able to insert into sqliteTable
            // succesfully
            TraceLoggingWrite(
                g_tlgProv,
                "SqliteTable_SuccessfullyWroteEvent",
                TraceLoggingValue(
                    constantEventBuffer.CompressedUtf8Json.size(), "NumBytes"),
                TraceLoggingValue(
                    constantEventBuffer.NumberOfEvents, "NumEvents"));

            return {};
        }
        catch (const std::system_error& ex)
        {
            if (IsErrorSqliteFull(ex.code()))
            {
                if (latency != TriggerLatency::Normal ||
                    persistence != TriggerPersistence::Normal)
                {
                    BufferType allowedToDelete = BufferType::Normal;
                    if (latency == TriggerLatency::Realtime)
                    {
                        allowedToDelete = BufferType::Critical;
                    }

                    if (DeleteLowerPriorityData(allowedToDelete))
                    {
                        continue;
                    }
                }

                TraceLoggingWrite(
                    g_tlgProv,
                    "SqliteTable_TooFullToCommitPreCompressBuffer",
                    TraceLoggingValue(
                        constantEventBuffer.CompressedUtf8Json.size(), "NumBytes"),
                    TraceLoggingValue(
                        constantEventBuffer.NumberOfEvents, "NumEvents"));
            }
            else
            {
                TraceLoggingWrite(
                    g_tlgProv,
                    "SqliteTable_UnexpectedErrorDuringTryInsert",
                    TraceLoggingErrorCode(ex.code(), "ErrorCode"));
            }

            // Unknown error or a full that wasn't able to recover enough to try
            // again
            RETURN_HR(ex.code());
        }
    }
}

/*++
    Deletes up to k_numOfRowsToDelete rows of data less than or equal to
    priority of the passed in BufferType

Arguments:
    highestAllowedBufferType - Highest priority of data allowed to be deleted
--*/
bool SqliteTable::DeleteLowerPriorityData(BufferType highestAllowedBufferType)
{
    auto convertToBufferType = [](TriggerLatency latency,
                                  TriggerPersistence persistence) {
        if (latency == TriggerLatency::Realtime)
        {
            return BufferType::Realtime;
        }
        else if (persistence == TriggerPersistence::Critical)
        {
            return BufferType::Critical;
        }
        else
        {
            return BufferType::Normal;
        }
    };

    std::vector<int32_t> rowsToDelete;  // RowIds start at 1

    // Need to scope boundStatement so it is destroyed before we try to
    // checkpoint
    {
        SqliteBoundStatement boundStatement = _getRowIdSortedForDelete->Bind();
        while (auto result = boundStatement.Step())
        {
            // Stop once we've hit 50 rows
            if (rowsToDelete.size() >= k_numOfRowsToDelete)
            {
                break;
            }

            TriggerLatency latency =
                static_cast<TriggerLatency>(result->NextInt<int8_t>());
            TriggerPersistence persistence =
                static_cast<TriggerPersistence>(result->NextInt<int8_t>());
            int32_t rowId = result->NextInt<int32_t>();

            BufferType bufferType = convertToBufferType(latency, persistence);

            // Stop if the only remaining data is too high priority
            if (bufferType > highestAllowedBufferType)
            {
                break;
            }

            rowsToDelete.push_back(rowId);
        }
    }

    if (!rowsToDelete.empty())
    {
        uint32_t bytesDeleted = DeleteMultiple(rowsToDelete);
        if (bytesDeleted > 0)
        {
            _dbConnection.CheckpointDb();
            _heartBeat.RecordDbEnteringCriticalOverflow(bytesDeleted);
            return true;
        }
    }
    return false;
}

/*++
    Gets the precompress buffer that maps to the given persistence and latency
    combination
--*/
IAsimovEventBuffer&
SqliteTable::GetPreCompressBuffer(TriggerPersistence persistence, TriggerLatency latency)
{
    if (latency == TriggerLatency::Normal)
    {
        if (persistence == TriggerPersistence::Normal)
        {
            return *_preCompressBuffers[static_cast<int>(BufferType::Normal)];
        }
        else if (persistence == TriggerPersistence::Critical)
        {
            return *_preCompressBuffers[static_cast<int>(BufferType::Critical)];
        }
    }
    else if (latency == TriggerLatency::Realtime)
    {
        return *_preCompressBuffers[static_cast<int>(BufferType::Realtime)];
    }
    FAIL_FAST();
}

/*++
    Determines whether there is at least one row of the given persistence
    latency combination.

Arguments:
    persistence - Persistence value - std::nullopt_t means to aggregate stores
                    of all persistence values.
    latency     - Latency value - std::nullopt_t means to aggregate stores of
                    all latency values.

Return Value:
    true  - The selected combination of persistence + latency has at least one
                upload payload remaining
    false - The selected group of combination of persistence + latency is
                completely empty
--*/
bool SqliteTable::HasRemainingData(
    const TriggerPersistence* persistence,
    const TriggerLatency* latency)
{
    if (!persistence && !latency)
    {
        return _dataExists->Bind().Step()->NextInt<bool>();
    }
    else if (latency && persistence)
    {
        return _latencyAndPersistenceDataExists
            ->Bind(static_cast<int>(*latency), static_cast<int>(*persistence))
            .Step()
            ->NextInt<bool>();
    }
    else if (latency)
    {
        return _latencyDataExists->Bind(static_cast<int>(*latency))
            .Step()
            ->NextInt<bool>();
    }
    else
    {
        return _persistenceDataExists->Bind(static_cast<int>(*persistence))
            .Step()
            ->NextInt<bool>();
    }
}

/*++
    Determines whether all of pre-compress buffers for the given latency are
    empty

Arguments:
    latency - Latency value

Return Value:
    true  - latency has at least one event sitting in a preCompress buffer
    false - The selected latency has completely empty preCompress buffers
--*/
bool SqliteTable::ArePreCompressBuffersEmpty(TriggerLatency latency)
{
    bool isEmpty = true;
    if (latency == TriggerLatency::Normal)
    {
        isEmpty &=
            _preCompressBuffers[static_cast<int32_t>(BufferType::Normal)]->IsEmpty();

        isEmpty &= _preCompressBuffers[static_cast<int32_t>(BufferType::Critical)]
                       ->IsEmpty();
    }
    else if (latency == TriggerLatency::Realtime)
    {
        isEmpty &= _preCompressBuffers[static_cast<int32_t>(BufferType::Realtime)]
                       ->IsEmpty();
    }

    return isEmpty;
}

/*++
    Stores a tenant token in the buffer. It will land in the ApiKey header on
    upload.

Arguments:
    persistence - Peristence this event has
    latency     - Latency this event has
    tenantToken - Tenant token (no O: prefix) that will be added to the
                    collection for the buffer.
--*/
void SqliteTable::StoreTenantToken(
    TriggerPersistence persistence,
    TriggerLatency latency,
    std::string&& tenantToken)
{
    IAsimovEventBuffer& buffer = GetPreCompressBuffer(persistence, latency);
    return buffer.StoreTenantToken(std::move(tenantToken));
}

}}}
