// Copyright (c) Microsoft Corporation. All rights reserved.

#pragma once

#include <vector>

#include "CommonDefs.h"
#include "IEventStore.h"

namespace Azure { namespace Device { namespace Telemetry {

struct AsimovConstantEventBuffer;
class IAsimovEventBuffer;
class IHeartBeat;
class IAsimovEventBufferFactory;
class SqlConnection;
class SqliteEventStore;
class SqliteStatement;

class SqliteTable
{
  public:
    SqliteTable(
        SqlConnection& dbConnection,
        IHeartBeat& heartBeat,
        IAsimovEventBufferFactory& asimovEventBufferFactory,
        uint32_t preCompressBufferNormalEventCount,
        uint32_t preCompressBufferCriticalEventCount,
        uint32_t preCompressBufferByteLimit);

    ~SqliteTable();

    std::error_code Write(const UploadData& event);

    std::unique_ptr<ReadResult> Read(TriggerLatency latency);

    void Delete(int32_t rowId);

    void SetUploadFailureCounter(int32_t rowId, int32_t newValue);

    void FlushPreCompressBuffers(TriggerLatency latency) noexcept;

    bool HasRemainingData(
        const TriggerPersistence* persistence,
        const TriggerLatency* latency);

    bool ArePreCompressBuffersEmpty(TriggerLatency latency);

    void StoreTenantToken(
        TriggerPersistence persistence,
        TriggerLatency latency,
        std::string&& tenantToken);

    bool DeleteLowerPriorityData(BufferType highestAllowedBufferType);

  public:
    static constexpr uint32_t k_numOfRowsToDelete = 50;

  private:
    void PrepareStatements();

    uint32_t DeleteMultiple(const std::vector<int32_t>& rowIds);

    std::error_code TryInsert(
        const AsimovConstantEventBuffer& constantEventBuffer,
        TriggerPersistence persistence,
        TriggerLatency latency);

    IAsimovEventBuffer&
    GetPreCompressBuffer(TriggerPersistence persistence, TriggerLatency latency);

  private:
    // This needs to be incremented every time we want to change store schema.
    static constexpr uint32_t k_currentEventStoreVersion = 1;

    SqlConnection& _dbConnection;
    IHeartBeat& _heartBeat;
    std::vector<std::shared_ptr<IAsimovEventBuffer>> _preCompressBuffers;

    // prepared statements
    std::unique_ptr<SqliteStatement> _getDataRow;
    std::unique_ptr<SqliteStatement> _deleteRow;
    std::unique_ptr<SqliteStatement> _insertRow;
    std::unique_ptr<SqliteStatement> _updateUploadFailCount;
    std::unique_ptr<SqliteStatement> _dataExists;
    std::unique_ptr<SqliteStatement> _latencyAndPersistenceDataExists;
    std::unique_ptr<SqliteStatement> _latencyDataExists;
    std::unique_ptr<SqliteStatement> _persistenceDataExists;
    std::unique_ptr<SqliteStatement> _getRowIdSortedForDelete;
    std::unique_ptr<SqliteStatement> _getSizeForRowId;
};

}}}
