// Copyright (c) Microsoft Corporation. All rights reserved.

#include "AsimovHttpClient.h"

#include <memory>

#include <azure-device-util/ConversionFactors.h>
#include <azure-device-util/ResultException.h>
#include <azure-device-util/String.h>
#include <azure-device-util/Time.h>

#include "CommonDefs.h"
#include "TelemetryError.h"
#include "TlgProvider.h"

namespace Azure { namespace Device { namespace Telemetry {

AsimovHttpClient::AsimovHttpClient(
    std::string_view endpointUrl,
    IHttpSessionManager& httpSessionManager)
    : _sessionManager(httpSessionManager), _endpointFullUrl(endpointUrl)
{
    using namespace std::string_view_literals;
    constexpr std::string_view k_httpsPrefix = "https://"sv;
    THROW_HR_IF(
        make_error_code(std::errc::invalid_argument),
        !Utils::String::IStartsWith(endpointUrl, k_httpsPrefix));

    auto slashIndex = endpointUrl.find(L'/', k_httpsPrefix.size());
    if (slashIndex != std::string::npos)
    {
        _endpointServer = endpointUrl.substr(0, slashIndex);

        // Is this safe if the provided url is 'server.com/'   ?
        _endpointPath = endpointUrl.substr(slashIndex + 1);
    }
    else
    {
        _endpointServer = endpointUrl;
    }

    _sessionManager.SetServerUrl(_endpointServer);

    // Construct the common header set that will be shared by all requests.
    _requestHeaders.emplace("Accept", "application/json");
    _requestHeaders.emplace("Accept-Encoding", "gzip, deflate");
    _requestHeaders.emplace(
        "Content-Type", "application/x-json-stream; charset=utf-8");
    _requestHeaders.emplace("Cache-Control", "no-cache");
    _requestHeaders.emplace("Content-Encoding", "deflate");
}

/*++
    POSTs the specified payload/headers to the server.
    Also increments backoff in case of network failure or a specific
    RetryAfter header from the server.

Arguments:
    headerData        - Values for various http headers
    payload           - The event payload to upload
    numEvents         - The number of events contained in the payload
    cancellationToken - Token used to cancel blocking request
    response          - Will contain a populated Response in the case of success
--*/
std::error_code AsimovHttpClient::UploadEventBuffer(
    const UploadEventBufferHeaders& headerData,
    const std::vector<std::byte>& payload,
    uint32_t numEvents,
    const pplx::cancellation_token& cancellationToken,
    HttpResponse& response)
{
    try
    {
        TraceLoggingWrite(g_tlgProv, "AsimovUploader_UploadEventBuffer");

        RETURN_HR_IF(
            make_error_code(std::errc::operation_in_progress),
            _backoffConfig.GetBackoffUntilTime() >
                std::chrono::steady_clock::now());

        //
        // Write the client's belief about the current time as a header.  The
        // server will use this to reason about how wrong the client's clock is.
        //
        HttpHeaderMap headers = _requestHeaders;

        std::chrono::system_clock::time_point currentTime =
            std::chrono::system_clock::now();
        headers.emplace(
            "Upload-Time",
            Utils::Time::SystemTimeTo8601MillisecondString(currentTime));

        headers.emplace("Reliability-Mode", "standard");
        headers.emplace("Client-Version", GenerateClientVersionString());
        headers.emplace("ApiKey", headerData.ApiKeyHeader);

        std::error_code uploadResult = _sessionManager.DoSynchronousHttpRequest(
            _endpointPath,
            {},
            HttpMethod::POST,
            headers,
            payload,
            cancellationToken,
            response);

        TraceLoggingWrite(
            g_tlgProv,
            "AsimovHttpClient_ResponseLog",
            TraceLoggingValue(
                FlattenHttpHeaderMap(headers).c_str(), "RequestHeaders"),
            TraceLoggingValue(response.StatusCode, "StatusCode"),
            TraceLoggingValue(
                FlattenHttpHeaderMap(response.Headers).c_str(), "ResponseHeaders"),
            TraceLoggingValue(response.Data.c_str(), "ResponseText"),
            TraceLoggingValue(numEvents, "NumEvents"),
            TraceLoggingErrorCode(uploadResult, "ErrorCode"));

        if (FAILED(uploadResult))
        {
            _backoffConfig.Backoff();

            // Now leave this function if the upload failed
            return uploadResult;
        }

        if (response.RetryAfterSeconds != 0)
        {
            // The server has sent us an explicit RetryAfterSeconds directive,
            // which applies to all backoffs.
            _backoffConfig.Pause(response.RetryAfterSeconds);
        }

        return {};
    }
    CATCH_RETURN();
}

/*++
    Returns client version string for use in header with server.
--*/
std::string AsimovHttpClient::GenerateClientVersionString()
{
    return "UTC-Linux-C++-no-experiement-no";
}

/*++
    Increment backoff due to client being unhappy with server response.
    Typically this will happen in reaction to HTTP status codes that are
    indicative of temporary failures that would benefit from reducing traffic
    briefly.

    NOTE: This will do nothing if the httpResponse has a RetryAfter header set.
    It takes priority, and that backoff has already occured.

Arguments:
    httpResponse - HttpResponse details
--*/
void AsimovHttpClient::BackoffDueToBadServerResponse(const HttpResponse& httpResponse)
{
    if (httpResponse.RetryAfterSeconds == 0)
    {
        _backoffConfig.Backoff();
    }
}

}}}
