// Copyright (c) Microsoft Corporation. All rights reserved.

#pragma once

#include <memory>
#include <string>
#include <string_view>
#include <vector>

#include "BackoffConfig.h"
#include "HttpSessionManager.h"

namespace pplx {
class cancellation_token;
}

namespace Azure { namespace Device { namespace Telemetry {

class IAsimovHttpClient
{
  public:
    struct UploadEventBufferHeaders
    {
        std::string ApiKeyHeader;
    };

    virtual ~IAsimovHttpClient() = default;

    virtual std::error_code UploadEventBuffer(
        const UploadEventBufferHeaders& headerData,
        const std::vector<std::byte>& payload,
        uint32_t numEvents,
        const pplx::cancellation_token& cancellationToken,
        HttpResponse& response) = 0;

    virtual void
    BackoffDueToBadServerResponse(const HttpResponse& httpResponse) = 0;

    virtual void ResetBackoffs() = 0;

    virtual std::chrono::steady_clock::time_point GetBackoffUntilTime() const = 0;
};

//
// This class is responsible for opening a connection to the server and
// POST'ing HTTP requests.
//
class AsimovHttpClient : public IAsimovHttpClient
{
  public:
    AsimovHttpClient(
        std::string_view endpointUrl,
        IHttpSessionManager& httpSessionManager);

    std::error_code UploadEventBuffer(
        const UploadEventBufferHeaders& headerData,
        const std::vector<std::byte>& payload,
        uint32_t numEvents,
        const pplx::cancellation_token& cancellationToken,
        HttpResponse& response) override;

    void BackoffDueToBadServerResponse(const HttpResponse& httpResponse) override;

    void ResetBackoffs() override { _backoffConfig.Reset(); }

    std::chrono::steady_clock::time_point GetBackoffUntilTime() const override
    {
        return _backoffConfig.GetBackoffUntilTime();
    }

  public:
    static std::string GenerateClientVersionString();

  private:
    IHttpSessionManager& _sessionManager;

    // Setup once at construction time
    std::string _endpointFullUrl;
    std::string _endpointPath;
    std::string _endpointServer;
    HttpHeaderMap _requestHeaders;

    BackoffConfig _backoffConfig;
};
}}}
