// Copyright (c) Microsoft Corporation. All rights reserved.

#include "AsimovUploadQueue.h"

#include <memory>

#include "AsimovUploadWorkItem.h"
#include "HttpSessionManager.h"
#include "IAsimovUploadWorkItemFactory.h"
#include "IEventStore.h"
#include "IHttpSessionManager.h"
#include "TlgProvider.h"

namespace Azure { namespace Device { namespace Telemetry {

constexpr std::chrono::seconds AsimovUploadQueue::c_realtimePeriod;
constexpr std::chrono::seconds AsimovUploadQueue::c_normalPeriod;

AsimovUploadQueue::AsimovUploadQueue(
    std::string_view uploadUrl,
    IHeartBeat& heartBeat,
    IEventSerializer& serializer,
    IEventStore& eventStore,
    IHttpSessionManager& httpSessionManager,
    IAsimovUploadWorkItemFactory& uploadWorkItemFactory)
    : _heartBeat(heartBeat)
    , _serializer(serializer)
    , _eventStore(eventStore)
    , _uploadWorkItemFactory(uploadWorkItemFactory)
    , _httpClient(uploadUrl, httpSessionManager)
{}

/*++
    Create threads to process io_service job queues.  Each job queue
    is single threaded, so only one scheduler-related action and one upload
    work item-related action can be occuring at once.

    _schedulerThread: Services _schedulerIoService
        - Runs timer completion lambdas
        - If there is work to do, creates a work item and posts a message to
            _workItemExecuteIoService for execution.
        - Handles work item completion messages to potentially start more work.

    _workItemExecuteThread: Services _workItemExecuteIoService
        - This thread runs the blocking execution of an AsimovUploadWorkItem
        - Only on execution runs at a time.
        - On completion, posts notification back to

    After this method is called, timer expirations will actually cause
    actions to happen.
--*/
void AsimovUploadQueue::StartTimers()
{
    _schedulerThread = std::thread{ [this] {
        // work is created to prevent run() from returning immediately
        boost::asio::io_service::work work{ _schedulerIoService };
        _schedulerIoService.run();
        std::cout << "SchedulerThread exiting!" << std::endl;
    } };

    _workItemExecuteThread = std::thread{ [this] {
        // work is created to prevent run() from returning immediately
        boost::asio::io_service::work work{ _workItemExecuteIoService };
        _workItemExecuteIoService.run();
        std::cout << "WorkItemExecuteThread exiting!" << std::endl;
    } };

    ArmTimersIfStorageBuffersHaveData(nullptr);
}

/*++
    Stops currently running io_services and waits for their threads
    to exit to clean up safely.
--*/
void AsimovUploadQueue::EndTimers()
{
    _shutdownCancellationTokenSource.cancel();

    if (!_schedulerIoService.stopped())
    {
        _schedulerIoService.stop();
    }
    if (!_workItemExecuteIoService.stopped())
    {
        _workItemExecuteIoService.stop();
    }

    if (_schedulerThread.joinable())
    {
        _schedulerThread.join();
    }
    if (_workItemExecuteThread.joinable())
    {
        _workItemExecuteThread.join();
    }
}

/*++
    Adds a work request for a particular latency, then runs the service queue.

Arguments:
    latency - Latency of job to create

Environment:
    _schedulerThread ONLY
--*/
void AsimovUploadQueue::AddWork(TriggerLatency latency)
{
    TraceLoggingWrite(
        g_tlgProv,
        "AsimovUploadQueue_AddWork",
        TraceLoggingValue(static_cast<int8_t>(latency), "Latency"));

    _workQueue.emplace(UploadWorkDefinition{ latency });

    ServiceQueue();
}

/*++
    Runs the main scheduler logic.

    If there is no work to do, or we are shutting down, then exit.

    If there is outstanding work, and it is of lower priority,
        send a preemption signal and exit.

    Otherwise if there is no outstanding work, create a new work item
        and send a message to _workItemExecuteIoService to run it.

Environment:
    _schedulerThread ONLY
--*/
void AsimovUploadQueue::ServiceQueue()
{
    if (_shutdownCancellationTokenSource.get_token().is_canceled())
    {
        TraceLoggingWrite(
            g_tlgProv, "AsimovUploadQueue_ServiceQueue_Exiting_Shutdown");

        return;
    }

    // If queue is empty, nothing to do.
    if (_workQueue.empty())
    {
        TraceLoggingWrite(
            g_tlgProv, "AsimovUploadQueue_ServiceQueue_Exiting_EmptyQueue");

        return;
    }

    // If there is outstanding work, consider asking it to sto if priority is
    // low, then leave
    if (_outstandingWork != nullptr)
    {
        TryPreempt();

        TraceLoggingWrite(
            g_tlgProv, "AsimovUploadQueue_ServiceQueue_Exiting_AttemptedPreempt");
        return;
    }

    // Now we are guaranteed to have new work to do, go create it
    UploadWorkDefinition workDef = _workQueue.top();
    _workQueue.pop();

    _outstandingWork = _uploadWorkItemFactory.Create(
        _httpClient,
        _heartBeat,
        _eventStore,
        _serializer,
        UploadWorkDefinition{ workDef.Latency });

    TraceLoggingWrite(
        g_tlgProv,
        "AsimovUploadQueue_ServiceQueue_ExecutingWork",
        TraceLoggingInt8(
            static_cast<int8_t>(_outstandingWork->GetWorkDefinition().Latency),
            "WorkDefLatency"));

    // Execute the task on the work item thread pool
    _workItemExecuteIoService.post([this]() { ExecuteWorkItem(); });
}

/*++
    Runs Execute() on the _outstandingWork item.

    On completion, posts a message back to _schedulerIoService so it
    can perform any bookkeeping and potentially schedule more work.

Environment:
    _workItemExecuteThread ONLY
--*/
void AsimovUploadQueue::ExecuteWorkItem()
{
    auto cancelToken = _shutdownCancellationTokenSource.get_token();
    _outstandingWork->Execute(cancelToken);

    TraceLoggingWrite(
        g_tlgProv,
        "AsimovUploadQueue_Execute",
        TraceLoggingInt8(
            static_cast<int8_t>(_outstandingWork->GetWorkDefinition().Latency),
            "WorkDefLatency"));

    // Handle completion of the task on main timer callback ioService
    _schedulerIoService.post([this]() { HandleWorkCompletion(); });
}

/*++
    Does scheduler bookkeeping on the completion of a work item to clear
    _outstandingWorkItem and then runs the service queue.

Environment:
    _schedulerThread ONLY
--*/
void AsimovUploadQueue::HandleWorkCompletion()
{
    TraceLoggingWrite(
        g_tlgProv,
        "AsimovUploadQueue_WorkItemCompleted",
        TraceLoggingInt8(
            static_cast<int8_t>(_outstandingWork->GetWorkDefinition().Latency),
            "WorkDefLatency"),
        TraceLoggingValue(
            static_cast<int8_t>(_outstandingWork->GetResult()), "Result"),
        TraceLoggingValue(
            _shutdownCancellationTokenSource.get_token().is_canceled(),
            "IsShuttingDown"));

    UploadWorkDefinition workDef = _outstandingWork->GetWorkDefinition();
    bool failedARealtimeUpload = workDef.Latency == TriggerLatency::Realtime &&
        _outstandingWork->GetResult() == UploadResult::Failed;

    // When a job completes, re-arm the timer if there is any data of that type
    // left, unless a realtime upload ended in failure.  Rearming here would
    // result in continuous short interval timer expiry.
    if (!failedARealtimeUpload)
    {
        TraceLoggingWrite(
            g_tlgProv,
            "AsimovUploadQueue_WorkCompleted_ArmingTimers",
            TraceLoggingValue(static_cast<int8_t>(workDef.Latency), "Latency"));

        ArmTimersIfStorageBuffersHaveData(&workDef.Latency);
    }

    _outstandingWork->LogCompletionEvent();

    // Now that we've handled completion, clear _outstandingWork
    _outstandingWork.reset();

    ServiceQueue();
}

/*++
    Sends a preempt signal to _outstandingWork if higher priority work
    has appeared in the queue.

Environment:
    _schedulerThread ONLY
--*/
void AsimovUploadQueue::TryPreempt()
{
    UploadWorkDefinition proposedWorkDef = _workQueue.top();
    if (_outstandingWork->GetWorkDefinition() < proposedWorkDef)
    {
        TraceLoggingWrite(
            g_tlgProv,
            "AsimovUploadQueue_Preempting",
            TraceLoggingValue(
                static_cast<int8_t>(proposedWorkDef.Latency), "ProposedLatency"),
            TraceLoggingValue(
                static_cast<int8_t>(_outstandingWork->GetWorkDefinition().Latency),
                "OutstandingLatency"));

        _outstandingWork->Preempt();
    }
}

/*++
    Arms the scheduler timer for respective latency.

Arguments:
    latency - Latency of timer to set
--*/
void AsimovUploadQueue::ArmTimer(TriggerLatency latency)
{
    auto handleTimerSet = [this](
                              TriggerLatency latency,
                              std::recursive_mutex& mutex,
                              boost::asio::steady_timer& timer,
                              std::chrono::seconds period) {
        std::lock_guard<std::recursive_mutex> lock{ mutex };
        if (timer.expires_at() < std::chrono::steady_clock::now())
        {
            timer.expires_from_now(period);
            timer.async_wait([this, latency](const boost::system::error_code& ec) {
                if (!ec)
                {
                    AddWork(latency);
                }
                else
                {
                    TraceLoggingWrite(
                        g_tlgProv,
                        "AsimovUploadQueue_SetTimer_TimerError",
                        TraceLoggingValue(static_cast<int8_t>(latency), "Latency"),
                        TraceLoggingErrorCode(ec, "ErrorCode"));

                    std::cout << "Timer error: " << ec;
                }
            });

            TraceLoggingWrite(
                g_tlgProv,
                "AsimovUploadQueue_SetTimer_NewTimerSet",
                TraceLoggingValue(static_cast<int8_t>(latency), "Latency"));

            std::cout << "New timer expiry set: " << static_cast<int8_t>(latency)
                      << std::endl;
        }
        else
        {
            TraceLoggingWrite(
                g_tlgProv,
                "AsimovUploadQueue_SetTimer_TimerAlreadySet",
                TraceLoggingValue(static_cast<int8_t>(latency), "Latency"));

            std::cout << "Old timer expiry remains: "
                      << static_cast<int8_t>(latency) << std::endl;
        }
    };

    switch (latency)
    {
    case TriggerLatency::Realtime: {
        handleTimerSet(
            TriggerLatency::Realtime,
            _realtimeTimerMutex,
            _realtimeTimer,
            _realtimePeriod);
        break;
    }
    case TriggerLatency::Normal: {
        handleTimerSet(
            TriggerLatency::Normal, _normalTimerMutex, _normalTimer, _normalPeriod);
        break;
    }
    default:
        break;
    }
}

void AsimovUploadQueue::SetTimerPeriod(
    TriggerLatency latency,
    std::chrono::seconds seconds)
{
    switch (latency)
    {
    case TriggerLatency::Realtime: {
        std::lock_guard<std::recursive_mutex> lock{ _normalTimerMutex };
        _normalPeriod = seconds;
        break;
    }
    case TriggerLatency::Normal: {
        std::lock_guard<std::recursive_mutex> lock{ _realtimeTimerMutex };
        _realtimePeriod = seconds;
        break;
    }
    default:
        break;
    }
}

/*++
    Arm corresponding latency's timer if pre-compress or storage buffers
    still have data sitting in them.

Arguments:
    latency - Latency of timer to set.  Or all if nullptr.
--*/
void AsimovUploadQueue::ArmTimersIfStorageBuffersHaveData(
    const TriggerLatency* desiredLatency)
{
    for (TriggerLatency latency : c_triggerLatencies)
    {
        if (!desiredLatency || latency == *desiredLatency)
        {
            if (!_eventStore.ArePreCompressBuffersEmpty(latency) ||
                _eventStore.HasRemainingData(nullptr, &latency))
            {
                ArmTimer(latency);

                TraceLoggingWrite(
                    g_tlgProv,
                    "AsimovUploadQueue_ArmTimersIfStorageBuffersHaveData",
                    TraceLoggingValue(static_cast<int8_t>(latency), "Latency"));
            }
        }
    }
}

}}}
