// Copyright (c) Microsoft Corporation. All rights reserved.

#pragma once

#include <queue>
#include <string_view>
#include <thread>

#include <boost/asio/io_service.hpp>
#include <boost/asio/steady_timer.hpp>
#include <pplx/pplx.h>

#include "AsimovHttpClient.h"
#include "CommonDefs.h"
#include "IUploadScheduler.h"

namespace Azure { namespace Device { namespace Telemetry {

class IEventSerializer;
class AsimovUploadWorkItem;
class IHeartBeat;
class IEventStore;
class IHttpSessionManager;
class IAsimovUploadWorkItemFactory;
class IAsimovUploadWorkItem;

//
// This class is responsible for prioritizing requests for upload for different
// bundles of Asimov data and kicking off tasks to upload that data.
//
class AsimovUploadQueue : public IUploadScheduler
{
  public:
    AsimovUploadQueue(
        std::string_view uploadUrl,
        IHeartBeat& heartBeat,
        IEventSerializer& serializer,
        IEventStore& eventStore,
        IHttpSessionManager& httpSessionManager,
        IAsimovUploadWorkItemFactory& uploadWorkItemFactory);

    void SetTimerPeriod(TriggerLatency latency, std::chrono::seconds seconds);

    // IUploadScheduler
    void StartTimers() override;
    void EndTimers() override;
    void ArmTimer(TriggerLatency latency) override;

    // For tests
    boost::asio::io_service& _GetSchedulerIoService()
    {
        return _schedulerIoService;
    }
    boost::asio::io_service& _GetWorkItemExecuteIoService()
    {
        return _workItemExecuteIoService;
    }

  private:
    void AddWork(TriggerLatency latency);
    void HandleWorkCompletion();

    void ServiceQueue();
    void TryPreempt();
    void ExecuteWorkItem();

    void ArmTimersIfStorageBuffersHaveData(const TriggerLatency* desiredLatency);

  private:
    IHeartBeat& _heartBeat;
    IEventSerializer& _serializer;
    IEventStore& _eventStore;
    IAsimovUploadWorkItemFactory& _uploadWorkItemFactory;

    boost::asio::io_service _schedulerIoService;
    std::thread _schedulerThread;

    std::recursive_mutex _realtimeTimerMutex;
    boost::asio::steady_timer _realtimeTimer{ _schedulerIoService };
    static constexpr std::chrono::seconds c_realtimePeriod{ 1 };
    std::chrono::seconds _realtimePeriod = c_realtimePeriod;

    std::recursive_mutex _normalTimerMutex;
    boost::asio::steady_timer _normalTimer{ _schedulerIoService };
    static constexpr std::chrono::seconds c_normalPeriod{ 30 };
    std::chrono::seconds _normalPeriod = c_normalPeriod;

    boost::asio::io_service _workItemExecuteIoService;
    std::thread _workItemExecuteThread;

    pplx::cancellation_token_source _shutdownCancellationTokenSource;

    AsimovHttpClient _httpClient;

    std::priority_queue<UploadWorkDefinition> _workQueue;
    std::shared_ptr<IAsimovUploadWorkItem> _outstandingWork;
};

}}}
