// Copyright (c) Microsoft Corporation. All rights reserved.

#include "AsimovUploadWorkItem.h"

#include <memory>

#include <azure-device-util/ConversionFactors.h>
#include <azure-device-util/Math.h>
#include <azure-device-util/ResultException.h>
#include <azure-device-util/ScopeExit.h>
#include <azure-device-util/Time.h>
#include <nlohmann/json.hpp>

#include "AsimovHttpClient.h"
#include "AsimovSyntheticEvent.h"
#include "HeartBeat.h"
#include "IEventSerializer.h"
#include "IEventStore.h"
#include "KillSwitchManager.h"
#include "TelemetryError.h"
#include "TlgProvider.h"

namespace Azure { namespace Device { namespace Telemetry {

constexpr std::chrono::minutes AsimovUploadWorkItem::k_maxUploadLoopTime;

std::shared_ptr<IAsimovUploadWorkItem> AsimovUploadWorkItemFactory::Create(
    IAsimovHttpClient& httpClient,
    IHeartBeat& heartBeat,
    IEventStore& eventStore,
    IEventSerializer& serializer,
    UploadWorkDefinition workDefinition)
{
    return std::make_shared<AsimovUploadWorkItem>(
        httpClient, heartBeat, eventStore, serializer, workDefinition);
}

AsimovUploadWorkItem::AsimovUploadWorkItem(
    IAsimovHttpClient& httpClient,
    IHeartBeat& heartBeat,
    IEventStore& eventStore,
    IEventSerializer& serializer,
    UploadWorkDefinition workDefinition)
    : _httpClient(httpClient)
    , _heartBeat(heartBeat)
    , _eventStore(eventStore)
    , _serializer(serializer)
    , _workDefinition(workDefinition)
    , _result(UploadResult::Succeeded)
{
    // Controls whether upload should actually hit the server or fake out a
    // 200 response.
    _mockUploadCallsWithSuccessResponse = false;
}

/*++
    Executes the upload loop for the given UploadWorkDefinition.  Will continue
    in the face of temporary backoffs until the limit is reached.  Responsible
    for setting the return state of either:

        1. Succeeded (ran out of data)
        2. Failed (hit too many network failures)
        3. Preempted (cancelled by the AsimovUploadQueue while it was still
            working)

Arguments:
    cancelToken - Signalled when entire job should exit with preemption
--*/
void AsimovUploadWorkItem::Execute(pplx::cancellation_token cancelToken)
{
    try
    {
        // If cancelToken is canceled, fire the preempty event.  Also make sure
        // we tear down the registration on function exit.
        auto cancelCallbackRegistration = cancelToken.register_callback(  // NOLINT
            [this]() { _preemptEvent.set(); });
        auto deregisterToken =
            Utils::MakeScopeExit([&cancelToken, &cancelCallbackRegistration]() {
                cancelToken.deregister_callback(cancelCallbackRegistration);
            });

        OnWorkItemStart();

        while (true)
        {
            std::unique_ptr<ReadResult> readResult = ReadData();

            if (!readResult)
            {
                // Nothing to upload, done.
                _result = UploadResult::Succeeded;
                break;
            }

            //
            // We might not be allowed to upload quite yet. We need to possibly
            // sleep until we are next allowed to upload this readResult.
            //
            auto backoffUntilTime = _httpClient.GetBackoffUntilTime();

            auto durationUntilBackoffFinished =
                backoffUntilTime - std::chrono::steady_clock::now();

            constexpr std::chrono::seconds c_maxUploadDelay{ 15 };
            if (durationUntilBackoffFinished > c_maxUploadDelay)
            {
                // Upload has backed off for good.
                _result = UploadResult::Failed;
                break;
            }

            // Note: In the case of millisUntilUploadAllowed=0, we don't really
            // need to sleep, but we do need to check for preempted
            if (_preemptEvent.wait_until(backoffUntilTime))
            {
                // We have been preempted by Preempt() or cancelToken.
                _result = UploadResult::Preempted;
                break;
            }

            if (durationUntilBackoffFinished > std::chrono::milliseconds{ 0 })
            {
                // We just waited. Now we're ready to try uploading again, but
                // we need to start over with a fresh ReadResult
                continue;
            }

            //
            // Now that we've (maybe) waited, it's time to upload the ReadResult
            // and deal with any remaining work and then start over the loop
            // with a new ReadResult.
            //
            auto uploadStartTime = std::chrono::steady_clock::now();

            bool uploadOccurred = UploadBuffer(*readResult, cancelToken);

            if (uploadOccurred)
            {
                HandleSuccessfulUpload(*readResult);
            }
            else
            {
                // Require at least some kind of backoff is set if we're not
                // uploading now, to prevent any tight-loop retrying.  For
                // testing purposes, allow a time_point of zero to be supplied.
                FAIL_FAST_IF(_httpClient.GetBackoffUntilTime() <= uploadStartTime);
            }

            if (IsTotalWorkItemTimeExceeded())
            {
                _result = UploadResult::Failed;
                break;
            }

            // There may be more to upload. We need to loop back and check.
        }

        OnWorkItemCompletion();
    }
    catch (...)
    {
        LOG_HR(Utils::ResultFromCaughtException());
        _result = UploadResult::Failed;
    }
}

void AsimovUploadWorkItem::Preempt()
{
    _result = UploadResult::Preempted;
    _preemptEvent.set();
}

void AsimovUploadWorkItem::LogCompletionEvent()
{
    // Log stats event
    auto event = std::make_unique<AsimovSyntheticEvent>(
        "AsimovUploadWorkItem_Completion_0");
    event->AddPartCField(
        "WorkDefLatency", static_cast<uint32_t>(_workDefinition.Latency));
    event->AddPartCField("Result", static_cast<uint32_t>(_result));
    event->AddPartCField("RuntimeMillis", _stats.Runtime.count());
    event->AddPartCField("EventsAccepted", _stats.EventsAccepted);
    event->AddPartCField("EventsRejected", _stats.EventsRejected);
    event->AddPartCField("SuccessfulUploads", _stats.SuccessfulUploads);
    event->AddPartCField("FailedUploads", _stats.FailedUploads);
    event->AddPartCField("BytesUploaded", _stats.BytesUploaded);

    IEventSerializer::WriteEventOptions skipArmTimerOptions;
    skipArmTimerOptions.SkipArmTimer = true;

    _serializer.WriteEvent(std::move(event), skipArmTimerOptions);
}

/*++
    Deal with totally successful buffer upload.  The key things to do here:
        1. Delete data
        2. Reset backoff
        3. Save success in bandwidth monitors / ETW / regkeys

Arguments:
    readResult - Populated ReadResult data to record stats about
--*/
void AsimovUploadWorkItem::HandleSuccessfulUpload(const ReadResult& readResult)
{
    // Now that upload has succeeded, delete the data
    _eventStore.DeleteData(readResult.SequenceNumber);

    if (_workDefinition.Latency == TriggerLatency::Normal)
    {
        TraceLoggingWrite(g_tlgProv, "AsimovUploader_NormalUploadSucceeded");
    }
    else if (_workDefinition.Latency == TriggerLatency::Realtime)
    {
        TraceLoggingWrite(g_tlgProv, "AsimovUploader_RealtimeUploadSucceeded");
    }
    else
    {
        FAIL_FAST();
    }
}

/*++
    Returns true if this is not a force upload job and the 20 minute maximum
    runtime for a given upload work item has been exceeded.
--*/
bool AsimovUploadWorkItem::IsTotalWorkItemTimeExceeded()
{
    return (std::chrono::steady_clock::now() - _workItemStartTime) >=
        _maxUploadLoopTime;
}

/*++
    Take care of once-per-work-item tasks that need to happen before work
    begins.
--*/
void AsimovUploadWorkItem::OnWorkItemStart()
{
    _workItemStartTime = std::chrono::steady_clock::now();

    _eventStore.FlushPreCompressBuffers(_workDefinition.Latency);

    TraceLoggingWrite(
        g_tlgProv,
        "AsimovUploadWorkItem_OnWorkItemStarted",
        TraceLoggingInt8(
            static_cast<int8_t>(_workDefinition.Latency), "WorkDef_Latency"));
}

/*++
    Take care of once-per-work-item tasks that need to happen after work has
    finished.
--*/
void AsimovUploadWorkItem::OnWorkItemCompletion()
{
    // Flush the database all the way to disk after we are finished with the
    // upload loops, only in the case of normal success.
    if (_workDefinition.Latency == TriggerLatency::Normal)
    {
        _eventStore.FlushToFile();
    }

    _stats.Runtime = std::chrono::duration_cast<std::chrono::milliseconds>(
        std::chrono::steady_clock::now() - _workItemStartTime);

    TraceLoggingWrite(
        g_tlgProv,
        "AsimovUploadWorkItem_OnWorkItemCompletion",
        TraceLoggingInt8(
            static_cast<int8_t>(_workDefinition.Latency), "WorkDef_Latency"),
        TraceLoggingValue(static_cast<int8_t>(_result), "Result"));
}

/*++
    Upload the eventBuffer, handling network success or failure with backoffs or
    response processing.

Arguments:
    readResult  - Buffer to upload
    cancelToken - Token that will be signalled to cancel the blocking upload job

Return Value:
    Returns true if buffer was accepted and should not be resent, false
    otherwise
--*/
bool AsimovUploadWorkItem::UploadBuffer(
    const ReadResult& readResult,
    pplx::cancellation_token cancelToken)
{
    // If the test hook DisableAsimovUploader is set, fake out a 200 response
    // without hitting the server.
    if (_mockUploadCallsWithSuccessResponse)
    {
        return true;
    }

    AsimovHttpClient::UploadEventBufferHeaders headers{
        readResult.ApiKeyHttpHeader
    };

    HttpResponse response;
    std::error_code uploadResult = _httpClient.UploadEventBuffer(
        headers, readResult.Payload, readResult.EventCount, cancelToken, response);

    if (uploadResult == TelemetryError::E_NETWORK_BACKOFF_IN_PROGRESS)
    {
        // No upload was attempted
        return false;
    }

    // Increase connection attempts
    _heartBeat.RecordNetworkAttempt();

    if (FAILED(uploadResult))
    {
        HandleNetworkError(_heartBeat, uploadResult);

        return false;
    }

    std::cout << "--Receieved response--:" << std::endl;
    std::cout << "Status code: " << response.StatusCode << std::endl
              << std::endl;
    std::cout << "Headers: " << FlattenHttpHeaderMap(response.Headers)
              << std::endl;
    std::cout << "Body: " << response.Data << std::endl;

    return HandleNetworkSuccess(readResult, response);
}

/*++
    Deal with error in the networking stack (typically due to just lack of
    internet connection). This means we did not get a reply from the server.

Arguments:
    client      - Client which we failed to connect to
    heartBeat   - HeartBeat object to update with failure information
    result      - std::error_code from networking stack
--*/
void AsimovUploadWorkItem::HandleNetworkError(
    IHeartBeat& heartBeat,
    std::error_code result)
{
    TraceLoggingWrite(
        g_tlgProv,
        "AsimovUploadWorkItem_HandleNetworkError",
        TraceLoggingErrorCode(result, "Result"));

    auto event = std::make_unique<AsimovSyntheticEvent>(
        "AsimovUploadWorkItem_NetworkError_0");
    event->AddPartCField("ErrorCodeValue", result.value());
    event->AddPartCField("ErrorCodeCategoryName", result.category().name());

    _serializer.WriteSyntheticEvent(std::move(event));
    heartBeat.RecordNetworkFailure(result);
    _stats.FailedUploads++;
}

/*++
    Deal with success in the networking stack and determine whether we must
    resend the buffer. At this point the response from the server must be parsed
    to check for service unavailability, accepted/rejected event counts, custom
    retry-afters, etc.

    NOTE: A buffer can still be accepted and have every event within it
    rejected.  In this case the data should still be deleted and not resent, as
    there's nothing we can do to fix up these events.

Arguments:
    readResult  - Event store contents
    resposne    - HttpResponse from server

Return Value:
    true if the buffer was accepted, false otherwise
--*/
bool AsimovUploadWorkItem::HandleNetworkSuccess(
    const ReadResult& readResult,
    const HttpResponse& response)
{
    TraceLoggingWrite(
        g_tlgProv,
        "AsimovUploadWorKItem_NetworkSuccess_0",
        TraceLoggingValue(response.StatusCode, "HttpStatus"));

    ProcessResponseHeaders(response.Headers);

    //
    // 2XX : Succeeded
    // 4XX : Non-retriable failure (except 408, 429)
    // 5XX : Retriable failure (except 501, 505)
    //
    bool uploadSucceededOrShouldNotRetry =
        (response.StatusCode >= 200 && response.StatusCode < 300) ||
        (response.StatusCode >= 400 && response.StatusCode < 500 &&
         response.StatusCode != 408 && response.StatusCode != 429) ||
        response.StatusCode == 501 || response.StatusCode == 505;

    if (uploadSucceededOrShouldNotRetry && response.StatusCode >= 300)
    {
        TraceLoggingWrite(
            g_tlgProv,
            "AsimovUploadWorkItem_NonRetriableFailure",
            TraceLoggingValue(response.StatusCode, "HttpStatus"),
            TraceLoggingValue(readResult.EventCount, "RejectedNumberOfEvents"));
    }

    if (uploadSucceededOrShouldNotRetry)
    {
        _httpClient.ResetBackoffs();

        // Assuming empty or malformed reponse, count all events as lost
        uint32_t acceptedEvents = 0, rejectedEvents = readResult.EventCount;
        ParseEndpointResponse(response.Data, acceptedEvents, rejectedEvents);

        uint32_t payloadSizeBytes =
            static_cast<uint32_t>(readResult.Payload.size());

        _stats.BytesUploaded += payloadSizeBytes;
        _stats.EventsAccepted += acceptedEvents;
        _stats.EventsRejected += rejectedEvents;
        _stats.SuccessfulUploads++;

        TraceLoggingWrite(
            g_tlgProv,
            "AsimovUploader_AcceptedAndRejectedEvents",
            TraceLoggingValue(acceptedEvents, "AcceptedNumberOfEvents"),
            TraceLoggingValue(rejectedEvents, "RejecetedNumberOfEvents"));

        _heartBeat.RecordHttpAcceptedAndRejectedEvents(
            acceptedEvents, rejectedEvents, payloadSizeBytes);

        return true;
    }
    else
    {
        constexpr uint32_t k_maxUploadAttemptsPerBuffer = 12;

        if (readResult.UploadFailCount > k_maxUploadAttemptsPerBuffer)
        {
            TraceLoggingWrite(
                g_tlgProv,
                "AsimovUploadWorkItem_DeletingBufferForTooManyUploadFailures",
                TraceLoggingValue(readResult.EventCount, "EventCount"),
                TraceLoggingValue(readResult.Payload.size(), "BufferSize"));

            _heartBeat.RecordHttpRepeatedUploadFailureDroppedEvents(
                readResult.EventCount);

            _eventStore.DeleteData(readResult.SequenceNumber);
        }
        else
        {
            _eventStore.IncrementUploadFailCount(readResult);
        }

        _httpClient.BackoffDueToBadServerResponse(response);
        _heartBeat.RecordHttpResponseFailure(response.StatusCode);

        auto event = std::make_unique<AsimovSyntheticEvent>(
            "AsimovUploadWorkItem_ServerErrorResponse_0");
        event->AddPartCField("ErrorCode", response.StatusCode);

        _serializer.WriteSyntheticEvent(std::move(event));

        return false;
    }
}

/*++
    Parses interesting information from the HTTP response body.

    The number of accepted events should always be around in any well-formed
    response, but the out-params will not be filled out if the response is
    malformed for some reason.

Arguments:
    responseData    - String of UTF-8 encoded HTTP response data
    acceptedEvents  - Count of accepted events parsed from response.
    rejectedEvents  - Count of rejected events parsed from response. Only
                        populated if present.
--*/
void AsimovUploadWorkItem::ParseEndpointResponse(
    std::string_view responseData,
    /*OUT*/ uint32_t& acceptedEvents,
    /*OUT*/ uint32_t& rejectedEvents) noexcept
{
    try
    {
        nlohmann::json responseJson = nlohmann::json::parse(responseData);

        acceptedEvents = responseJson.at("acc").get<uint32_t>();
        auto rejectItr = responseJson.find("rej");
        rejectedEvents =
            (rejectItr == responseJson.end()) ? 0u : rejectItr->get<uint32_t>();
    }
    catch (nlohmann::json::exception& jsonEx)
    {
        TraceLoggingWrite(
            g_tlgProv,
            "AsimovUploadWorkItem_ParseEndpointResponse_JsonParseError",
            TraceLoggingValue(jsonEx.what(), "Message"));
    }
    CATCH_LOG();
}

/*++
    Attempts to read a single buffer of data for the work item's latency.

Return Value:
    optional populated with ReadResult on success, empty on failure
--*/
std::unique_ptr<ReadResult> AsimovUploadWorkItem::ReadData() const
{
    auto readResult = _eventStore.GetDataByLatency(_workDefinition.Latency);
    if (readResult)
    {
        TraceLoggingWrite(
            g_tlgProv,
            "AsimovUploader_SelectedDataByLatency",
            TraceLoggingValue(
                static_cast<int8_t>(_workDefinition.Latency), "Latency"),
            TraceLoggingValue(readResult->EventCount, "EventCount"));
    }

    return readResult;
}

/*++
    Looks through response headers for Shed requests or tenant kill requests,
    and handles them if they exist

Arguments:
    responseHeaders - Map of name->value pairs in the headers of the HTTP
                      response from the server
--*/
void AsimovUploadWorkItem::ProcessResponseHeaders(const HttpHeaderMap& responseHeaders)
{
    auto killTokensIt = responseHeaders.find("kill-tokens");
    if (killTokensIt != responseHeaders.end())
    {
        std::vector<std::string> parsedKillTokenValue =
            Utils::String::Split(killTokensIt->second, ":");

        if (parsedKillTokenValue.size() == 2)
        {
            // First, find the kill duration (if it is set)
            std::chrono::seconds killDurationSeconds = std::chrono::minutes{ 15 };
            {
                // This default value will always be overwritten in all legal
                // cases (except in the case that it is never read anyway)
                auto killDurationIt = responseHeaders.find("kill-duration");
                if (killDurationIt != responseHeaders.end())
                {
                    uint32_t parseSeconds = 0;
                    if (SUCCEEDED(Utils::String::ToUIntTStrict(
                            killDurationIt->second, parseSeconds)))
                    {
                        killDurationSeconds =
                            std::chrono::seconds{ parseSeconds };
                    }
                }
            }

            TriggerPersistence killMode = TriggerPersistence::Normal;
            if (Utils::String::ICompare(parsedKillTokenValue[1], "all") == 0)
            {
                killMode = TriggerPersistence::Critical;
            }

            _serializer.GetKillSwitchManager().SetKillRequest(
                std::move(parsedKillTokenValue[0]), killMode, killDurationSeconds);
        }
    }
}

}}}
