// Copyright (c) Microsoft Corporation. All rights reserved.

#pragma once

#include <string_view>

#include <azure-device-util/SyncEvent.h>
#include <pplx/pplx.h>

#include "AsimovUploadQueue.h"
#include "IAsimovUploadWorkItemFactory.h"

namespace Azure { namespace Device { namespace Telemetry {

//
// This class is handed a bundle of parameters to determine what to upload,
// then spins away uploading data until it is completed or pre-empted.
//
class AsimovUploadWorkItem : public IAsimovUploadWorkItem
{
  public:
    AsimovUploadWorkItem(
        IAsimovHttpClient& httpClient,
        IHeartBeat& heartBeat,
        IEventStore& eventStore,
        IEventSerializer& serializer,
        UploadWorkDefinition workDefinition);

    struct Stats
    {
        uint32_t EventsAccepted = 0;
        uint32_t EventsRejected = 0;
        uint32_t SuccessfulUploads = 0;
        uint32_t FailedUploads = 0;
        uint64_t BytesUploaded = 0;
        std::chrono::milliseconds Runtime{ 0 };
    };
    const Stats& GetStats() { return _stats; };

    void SetMaxUploadLoopTime(std::chrono::seconds maxLoopTime)
    {
        _maxUploadLoopTime = maxLoopTime;
    }

    // IAsimovUploadWorkItem
    void Execute(pplx::cancellation_token cancelToken) override;

    void Preempt() override;

    UploadWorkDefinition GetWorkDefinition() const override
    {
        return _workDefinition;
    }
    UploadResult GetResult() const override { return _result; }

    void LogCompletionEvent() override;

  private:
    void HandleSuccessfulUpload(const ReadResult& readResult);

    void OnWorkItemStart();
    void OnWorkItemCompletion();

    bool IsTotalWorkItemTimeExceeded();

    bool
    UploadBuffer(const ReadResult& readResult, pplx::cancellation_token cancelToken);

    void HandleNetworkError(IHeartBeat& heartBeat, std::error_code result);

    bool
    HandleNetworkSuccess(const ReadResult& readResult, const HttpResponse& response);

    std::unique_ptr<ReadResult> ReadData() const;

    void ProcessResponseHeaders(const HttpHeaderMap& responseHeaders);

    static void ParseEndpointResponse(
        std::string_view responseData,
        /*OUT*/ uint32_t& acceptedEvents,
        /*OUT*/ uint32_t& rejectedEvents) noexcept;

  private:
    IAsimovHttpClient& _httpClient;
    IHeartBeat& _heartBeat;
    IEventStore& _eventStore;
    IEventSerializer& _serializer;

    UploadWorkDefinition _workDefinition;
    UploadResult _result;
    Utils::ManualResetEvent _preemptEvent;
    std::chrono::steady_clock::time_point _workItemStartTime;
    Stats _stats;

    constexpr static std::chrono::minutes k_maxUploadLoopTime{ 20 };
    std::chrono::seconds _maxUploadLoopTime = k_maxUploadLoopTime;

    // Test hooks
    bool _mockUploadCallsWithSuccessResponse;
};

}}}
