// Copyright (c) Microsoft Corporation. All rights reserved.

#include "BackoffConfig.h"

#include <chrono>
#include <cmath>

#include <azure-device-util/Math.h>
#include <azure-device-util/ResultException.h>

#include "TelemetryError.h"
#include "TlgProvider.h"

namespace Azure { namespace Device { namespace Telemetry {

/*++
    Clear backoff data.
--*/
void BackoffConfig::Reset()
{
    _currentBackoffAttempts = 0;
    _backoffUntilTime = {};
}

/*++
    Increment backoff counter and set backoff time based on an exponential,
    randomized backoff amount based on that counter.

    Backoffs max out at after a while to set a maximum per-backoff delay.
--*/
void BackoffConfig::Backoff()
{
    _currentBackoffAttempts++;

    uint32_t currentAttemptValue = std::min(_currentBackoffAttempts, 6u);

    uint32_t initialBackoffSeconds = 3;
    uint32_t baseBackoffSeconds = initialBackoffSeconds *
        static_cast<uint32_t>(std::lround(std::pow(3, currentAttemptValue - 1)));
    uint32_t randomizedBackoffSeconds =
        CalculateRandomizedBackoff(baseBackoffSeconds, initialBackoffSeconds);

    _backoffUntilTime =
        NowTime() + std::chrono::seconds(randomizedBackoffSeconds);

    TraceLoggingWrite(
        g_tlgProv,
        "BackoffConfig_Backoff",
        TraceLoggingValue(_currentBackoffAttempts, "CurrentBackoffAttempts"),
        TraceLoggingValue(randomizedBackoffSeconds, "RandomizedBackoffSeconds"));
}

/*++
    Increase the current backoff time without incrementing the backoff
    attempts counter.

Arguments:
    customBackoffSeconds  - How long to backoff
--*/
void BackoffConfig::Pause(uint32_t customBackoffSeconds)
{
    _backoffUntilTime = NowTime() + std::chrono::seconds(customBackoffSeconds);

    TraceLoggingWrite(
        g_tlgProv,
        "BackoffConfig_Pause",
        TraceLoggingValue(customBackoffSeconds, "CustomBackoffSeconds"));
}

/*++
    Calculate a randomized backoff +/- randomizedBackoffPercentage of
    of baseBackoffSeconds, bounded by minBackoffSeconds.

Arguments:
    baseBackoffSeconds  - The median value to randomize around
    minBackoffSeconds   - Lowest possible value of randomized value

Return Value:
    The calculated randomized backoff value.
--*/
uint32_t BackoffConfig::CalculateRandomizedBackoff(
    uint32_t baseBackoffSeconds,
    uint32_t minBackoffSeconds)
{
    // Calculate upper and lower bounds for an escalated backoff
    constexpr uint32_t c_randomizedBackoffPercentage = 15;
    uint32_t lowerBound = static_cast<uint32_t>(
        baseBackoffSeconds * ((100.0 - c_randomizedBackoffPercentage) / 100.0));
    uint32_t upperBound = static_cast<uint32_t>(
        baseBackoffSeconds * ((100.0 + c_randomizedBackoffPercentage) / 100.0));

    // Bind the lower bound and upper bound to the global std::min
    lowerBound = std::max(lowerBound, minBackoffSeconds);
    upperBound = std::max(upperBound, minBackoffSeconds);

    // Lower should always be <= upper
    if (lowerBound > upperBound)
    {
        THROW_HR(make_error_code(std::errc::invalid_argument));
    }

    // Pick a random value between upper and lower bound
    uint32_t resultSeconds =
        Utils::Math::GenerateUniformDistributionRandomIntInRange(
            lowerBound, upperBound);
    return resultSeconds;
}

std::chrono::steady_clock::time_point BackoffConfig::NowTime()
{
    return _useZeroBaseTime ? std::chrono::steady_clock::time_point{} :
                              std::chrono::steady_clock::now();
}

}}}
