// Copyright (c) Microsoft Corporation. All rights reserved.

#include <iostream>
#include <string>

#include <azure-device-util/Permission.h>
#include <azure-device-util/Process.h>

using namespace Azure::Device::Utils;

enum class FailureReason
{
    NONE = 0,
    USAGE_ERROR = 1,
    FILE_ERROR = 2
};

void PrintUsage()
{
    std::cout << "Usage:" << std::endl
              << "azure-device-telemetry-level level [--reload]" << std::endl
              << "    level       \"Optional\" or \"Required\"" << std::endl
              << "    --reload    Instruct ADHS service to reload configuration."
              << std::endl;
    exit(static_cast<int>(FailureReason::USAGE_ERROR));
}

void ReloadDaemons()
{
    auto result = Process::RunCommand(
        "systemctl",
        { "kill", "-s", "SIGHUP", "azure-device-telemetryd.service" });

    std::cout
        << (result.ExitCode == 0 ?
                "Reload signal sent to azure-device-telemetryd!" :
                "ERROR: could not send reload signal to azure-device-telemetryd!")
        << std::endl;

    result = Process::RunCommand(
        "systemctl",
        { "kill", "-s", "SIGHUP", "azure-device-errorreporting-uploaderd.service" });

    std::cout
        << (result.ExitCode == 0 ?
                "Reload signal sent to azure-device-errorreporting-uploaderd!" :
                "ERROR: could not send reload signal to azure-device-errorreporting-uploaderd!")
        << std::endl;
}

int main(int argc, char** argv)
{
    if (argc < 2)
    {
        PrintUsage();
    }

    try
    {
        auto error = Permission::WritePermissionToDefaultFile(
            Permission::ConvertStringToTelemetryPermission(argv[1]));

        if (error)
        {
            std::cerr << "ERROR: could not write level to file:" << std::endl
                      << error << std::endl;
            exit(static_cast<int>(FailureReason::FILE_ERROR));
        }

        std::cout << "Telemetry level updated!" << std::endl;
    }
    catch (const std::exception& exc)
    {
        std::cerr << "Invalid telemetry level: \"" << argv[1] << "\""
                  << std::endl;
        PrintUsage();
    }

    // No options
    if (argc == 2)
    {
        return static_cast<int>(FailureReason::NONE);
    }

    std::string opt = argv[2];

    // Reload using systemd
    if (opt.compare("--reload") == 0)
    {
        ReloadDaemons();
    }
    else
    {
        std::cerr << "Unknown option: \"" << opt << "\"" << std::endl;
        PrintUsage();
    }

    return static_cast<int>(FailureReason::NONE);
}