// Copyright (c) Microsoft Corporation. All rights reserved.

#pragma once

#include <bitset>
#include <cstddef>
#include <memory>
#include <string_view>

#include <azure-device-util/ResultException.h>
#include <nonstd/span.hpp>

// Small convenience additions to gsl
namespace Azure { namespace Device { namespace Utils {

class Bytes
{
  public:
    /*++
        Converts a container (e.g. vector) to a span of bytes
    --*/
    template<typename Container>
    static nonstd::span<const std::byte> ContainerToBytes(const Container& c)
    {
        nonstd::span<const typename Container::value_type> containerSpan{ c };

        return { reinterpret_cast<const std::byte*>(containerSpan.data()),
                 containerSpan.size_bytes() };
    }

    /*++
        Converts a container (e.g. vector) to a span of writable bytes
    --*/
    template<typename Container>
    static nonstd::span<std::byte> ContainerToWriteableBytes(Container& c)
    {
        nonstd::span<typename Container::value_type> containerSpan{ c };

        return { reinterpret_cast<std::byte*>(containerSpan.data()),
                 containerSpan.size_bytes() };
    }

    /*++
        Converts a span of bytes to a const value (e.g. int)
    --*/
    template<typename Value>
    static const Value& BytesToValue(nonstd::span<const std::byte> bytes)
    {
        FAIL_FAST_IF(bytes.size() != sizeof(Value));
        return *reinterpret_cast<const Value*>(bytes.data());
    }

    /*++
        Converts a span of bytes to a value (e.g. int)
    --*/
    template<typename Value>
    static Value& BytesToValue(nonstd::span<std::byte> bytes)
    {
        FAIL_FAST_IF(bytes.size() != sizeof(Value));
        return *reinterpret_cast<Value*>(bytes.data());
    }

    /*++
        Converts a value (e.g. int) to a span of bytes
    --*/
    template<typename Value>
    static nonstd::span<std::byte> ValueToBytes(Value& c)
    {
        return { reinterpret_cast<std::byte*>(&c), sizeof(c) };
    }

    /*++
        Converts a value (e.g. int) to a span of bytes
    --*/
    template<typename Value>
    static nonstd::span<const std::byte> ValueToBytes(const Value& c)
    {
        return { reinterpret_cast<const std::byte*>(&c), sizeof(c) };
    }

    /*++
        Skips bytesToConsume bytes from 'bytes' argument
    --*/
    static void
    ConsumeBytesNoReturn(nonstd::span<const std::byte>& bytes, size_t bytesToConsume)
    {
        if (bytesToConsume > bytes.size())
        {
            THROW_HR(std::make_error_code(std::errc::value_too_large));
        }

        bytes = bytes.subspan(bytesToConsume);
    }

    /*++
        Consumes sizeof(ConsumeType) bytes from 'bytes' parameter
    --*/
    template<class ConsumeType>
    static ConsumeType ConsumeBytes(nonstd::span<const std::byte>& bytes)
    {
        size_t bytesToConsume = sizeof(ConsumeType);
        if (bytesToConsume > bytes.size())
        {
            THROW_HR(std::make_error_code(std::errc::value_too_large));
        }

        auto valueBytes = bytes.first(bytesToConsume);
        bytes = bytes.subspan(bytesToConsume);
        return Utils::Bytes::BytesToValue<ConsumeType>(valueBytes);
    }

    /*++
        Consumes a null terminated string from 'bytes' parameter
    --*/
    static std::string_view
    ConsumeNullTerminatedString(nonstd::span<const std::byte>& bytes)
    {
        for (size_t i = 0; i < bytes.size(); i++)
        {
            if (bytes[i] == std::byte{ 0 })
            {
                auto stringView = Utils::Bytes::BytesToStringView(bytes.first(i));
                bytes = bytes.subspan(i + 1);
                return stringView;
            }
        }

        THROW_HR(std::make_error_code(std::errc::value_too_large));
    }

    /*++
        Consumes a counted string from 'bytes' parameter.
    --*/
    static std::string_view
    ConsumeCountedString(nonstd::span<const std::byte>& bytes, size_t stringLengthBytes)
    {
        if (stringLengthBytes > bytes.size())
        {
            THROW_HR(std::make_error_code(std::errc::value_too_large));
        }

        auto valueBytes = bytes.first(stringLengthBytes);
        bytes = bytes.subspan(stringLengthBytes);
        auto stringView = BytesToStringView(valueBytes);
        return stringView;
    }

    /*++
        Creates a string_view from a span of bytes
    --*/
    inline static std::string_view
    BytesToStringView(nonstd::span<const std::byte> bytes)
    {
        return { reinterpret_cast<const char*>(bytes.data()),
                 static_cast<size_t>(bytes.size()) };
    }

    /*++
        Converts a container (e.g. vector) to a span of bytes
    --*/
    static inline nonstd::span<const std::byte>
    StringViewToBytes(std::string_view str)
    {
        nonstd::span<const char> containerSpan{ str };

        return { reinterpret_cast<const std::byte*>(containerSpan.data()),
                 containerSpan.size_bytes() };
    }

    /*++
        Creates a bit set of 8 from a byte.
    --*/
    inline static std::bitset<8> ByteToBitset(std::byte val)
    {
        return std::bitset<8>(std::to_integer<uint8_t>(val));
    }

    /*++
        Creates a byte from 8 bits.
    --*/
    inline static std::byte BitsetToByte(std::bitset<8> bits)
    {
        return std::byte{ static_cast<uint8_t>(bits.to_ulong()) };
    }
};

}}}
