// Copyright (c) Microsoft Corporation. All rights reserved.

#pragma once

#include <list>
#include <map>

namespace Azure { namespace Device { namespace Utils {

template<class TKey, class TValue>
class LruCache
{
  private:
    struct LruEntry
    {
        TKey Key;
        std::shared_ptr<TValue> Value;

        // Needed for emplace_front to work until C++20 changes
        // standard library constructions to use {} instead of ()
        LruEntry(TKey key, std::shared_ptr<TValue> value)
            : Key(std::move(key)), Value(std::move(value))
        {}
    };
    using LruList = std::list<LruEntry>;
    using LruIterator = typename LruList::iterator;

  public:
    explicit LruCache(size_t maxLruSize = 100) : _maxLruSize(maxLruSize) {}

    // Copy possible to implement later, but not done yet
    LruCache(const LruCache&) = delete;
    LruCache& operator=(const LruCache&) = delete;

    LruCache(LruCache&&) = default;
    LruCache& operator=(LruCache&&) = default;

    void SetMaxLruSize(size_t maxLruSize)
    {
        while (_lruValues.size() > maxLruSize)
        {
            PopBack();
        }

        _maxLruSize = maxLruSize;
    }

    std::shared_ptr<TValue> Get(const TKey& keyType)
    {
        // If element is not found, return nullptr
        auto itr = _keysToValues.find(keyType);
        if (itr == _keysToValues.end())
        {
            return nullptr;
        }

        // Move LRU list entry to the front
        _lruValues.splice(_lruValues.begin(), _lruValues, itr->second);

        // Return found value
        return itr->second->Value;
    }

    void Insert(const TKey& key, const std::shared_ptr<TValue>& value)
    {
        // Add to the front of the values LRU list
        _lruValues.emplace_front(key, value);

        // Remove the oldest element if the maximum size is exceeded
        if (_lruValues.size() > _maxLruSize)
        {
            PopBack();
        }

        // Insert the new key value pointing at the front element of the LRU
        _keysToValues.emplace(key, _lruValues.begin());
    }

  private:
    void PopBack()
    {
        const auto& keyToDelete = _lruValues.back().Key;
        _keysToValues.erase(keyToDelete);
        _lruValues.pop_back();
    }

  private:
    LruList _lruValues;
    std::map<TKey, LruIterator> _keysToValues;

    size_t _maxLruSize;
};

}}}
