// Copyright (c) Microsoft Corporation. All rights reserved.

#pragma once

#include <cstdlib>
#include <random>

#include <azure-device-util/Template.h>
#include <nonstd/span.hpp>

namespace Azure { namespace Device { namespace Utils {

class Math
{
  public:
    // Get the difference between a and b, even if unsigned
    template<typename T>
    static T Delta(T a, T b)
    {
        return static_cast<T>(
            labs(static_cast<typename std::make_signed<T>::type>(a - b)));
    }

    /*++
        Rounds a value up to nearest multiple of alignment

    Arguments:
        a - value to align
        b - aligment

    Return Value:
        if, a is evenly divisible by alignment -> a
        else -> a + (alignment = a % alignment)
    --*/
    template<typename T>
    constexpr static T AlignUp(T value, typename identity<T>::type alignment)
    {
        T remainder = value % alignment;
        if (remainder == 0)
        {
            return value;
        }
        else
        {
            return value + (alignment - remainder);
        }
    }

    /*++
        Perform addition on operand 'a', i.e. a + b. If 'a' is not large enough,
        then the result is the most significant possible value that fits into T.

    Arguments:
        a - first operand
        b - second operand

    Return Value:
        a + b truncated to the most significant possible value instead of
        rolling over.
    --*/
    template<class T>
    static T SafeAdd(T a, typename identity<T>::type b)
    {
        if (b >= 0 && a + b < a)
        {
            return std::numeric_limits<T>::max();
        }

        if (b < 0 && a + b > a)
        {
            return std::numeric_limits<T>::min();
        }

        return a + b;
    }

    // Return a - b, will return 0 in the case of unsigned underflow
    template<typename T>
    static T SafeUnsignedSubtract(T a, typename identity<T>::type b)
    {
        return a > b ? a - b : 0;
    }

    // Return a - b, will failfast in the case of unsigned underflow
    template<typename T>
    static T AssertSafeUnsignedSubtract(T a, typename identity<T>::type b)
    {
        FAIL_FAST_IF(a < b);
        return a - b;
    }

    /*++
        Perform addition inplace on operand 'a', i.e. a += b. If 'a' is not
        large enough, then the result is the most significant possible value
        that fits in a.

    Arguments:
        a - first operand and receives the result of this operation
        b - second operand

    Return Value:
        a += b truncated to the most significant possible value instead of
        rolling over.
    --*/
    template<class T>
    static T SafeAddInplace(T& a, typename identity<T>::type b)
    {
        if (b >= 0 && a + b < a)
        {
            return a = std::numeric_limits<T>::max();
        }

        if (b < 0 && a + b > a)
        {
            return a = std::numeric_limits<T>::min();
        }

        return a += b;
    }

    /*++
        Reduce an integer value to a smaller type. If the new type is too small
        to fit the original value, then the most significant expressable value
        that fits in the smaller type is used.

    Arguments:
        value - value to truncate

    Return Value:
        truncated v. It is the same value as v, or the most significant possible
        value that fits in T.
    --*/
    template<class T, class U>
    static T BoundedNarrowCast(const U& value)
    {
        static_assert(sizeof(T) <= sizeof(U), "Invalid truncation");
        static_assert(
            std::is_unsigned<T>::value == std::is_unsigned<U>::value,
            "Inconsistent signage");

        if (value >= 0)
        {
            return static_cast<T>(
                std::min(value, static_cast<U>(std::numeric_limits<T>::max())));
        }
        else
        {
            return static_cast<T>(
                std::max(value, static_cast<U>(std::numeric_limits<T>::min())));
        }
    }

    /*++
        Returns the lowest set 1 bit from a given value.

    Arguments:
        val - Any unigned integer

    Return Value:
        val with all 1 bits cleared except the lowest one. Example: 0b110101000
        -> 0b1000
    --*/
    template<typename T>
    static T LeastSetBit(T val)
    {
        return (val & -val);
    }

    /*++
        Generates a random uniformly distributed uint64_t in [low, high] range.

    ARguments:
        low - The lowest value to return, inclusive.
        high - The highest value to return, inclusive.
        device - The random device to use for generating the number. If you are
            doing multiple generations, you should pass your own and re-use it.

    Return Value:
        Random 64-bit integer
    --*/
    template<typename T>
    static T
    GenerateUniformDistributionRandomIntInRange(T low, T high, std::random_device& device)
    {
        static_assert(std::is_integral<T>::value, "T must be an integral type");
        std::uniform_int_distribution<T> distribution(low, high);
        return distribution(device);
    }

    /*++
        Generates a random uniformly distributed uint64_t in [low, high] range
        using the random_device algorithm.

    ARguments:
        low - The lowest value to return, inclusive.
        high - The highest value to return, inclusive.

    Return Value:
        Random 64-bit integer
    --*/
    template<typename T>
    static T GenerateUniformDistributionRandomIntInRange(T low, T high)
    {
        static_assert(std::is_integral<T>::value, "T must be an integral type");
        std::random_device device;
        std::uniform_int_distribution<T> distribution(low, high);
        return distribution(device);
    }

    /*++
        Generates a random uniformly distributed float in [low, high] range.

    ARguments:
        low - The lowest value to return, inclusive.
        high - The highest value to return, inclusive.

    Return Value:
        Random 64-bit float
    --*/
    template<typename T>
    static T GenerateUniformDistributionRandomFloatInRange(T low, T high)
    {
        static_assert(
            std::is_floating_point<T>::value, "T must be an floating point type");
        std::random_device device;
        std::uniform_real_distribution<T> distribution(low, high);
        return distribution(device);
    }

    // PRECONDITION: Non-empty span
    template<class T>
    static T Min(nonstd::span<const T> span)
    {
        return *std::min_element(span.cbegin(), span.cend());
    }

    // PRECONDITION: Non-empty span
    template<class T>
    static T Max(nonstd::span<const T> span)
    {
        return *std::max_element(span.cbegin(), span.cend());
    }

    // PRECONDITION: Non-empty span
    template<class T>
    static T Mean(nonstd::span<const T> span)
    {
        T total = std::accumulate(span.begin(), span.end(), 0);

        return total / span.size();
    }

    // PRECONDITION: Non-empty, sorted span
    template<class T>
    static T Median(nonstd::span<const T> span)
    {
        size_t middleIndex = span.size() / 2;
        if (span.size() % 2 == 0)
        {
            return (span[middleIndex] + span[middleIndex + 1]) / 2;
        }
        else
        {
            return span[middleIndex];
        }
    }
};

}}}
