// Copyright (c) Microsoft Corporation. All rights reserved.

#pragma once

#include <cstring>
#include <iostream>
#include <string_view>

#include <azure-device-util/ErrorHelper.h>

#define FAIL_FAST std::terminate

#define UNREFERENCED_PARAMETER(x) (void) x

// Parameters are: Error message, file name, line number, error code
using log_callback_t = std::add_pointer<
    void(std::string_view, std::string_view, int, std::error_code)>::type;

namespace Azure { namespace Device { namespace Utils {

// Defined in the CPP file
extern log_callback_t g_pfnLoggingCallback;

/*++
    Sets the error callback for the module.
--*/
inline void SetResultLoggingCallback(log_callback_t callbackFunction)
{
    g_pfnLoggingCallback = callbackFunction;
}

std::error_code ResultFromCaughtException();

std::error_code
ResultFromCaughtExceptionInternal(char* debugString, size_t debugStringSize);

std::error_code make_errno_error_code();

}}}

//*****************************************************************************
// Macros for logging failures (ignore or pass-through)
//*****************************************************************************
#define LOG_IF_FAILED(hr)                                              \
    do                                                                 \
    {                                                                  \
        std::error_code _hr = (hr);                                    \
        if (FAILED(_hr))                                               \
        {                                                              \
            if (Azure::Device::Utils::g_pfnLoggingCallback != nullptr) \
            {                                                          \
                Azure::Device::Utils::g_pfnLoggingCallback(            \
                    "LOG_HR: ", __FILE__, __LINE__, _hr);              \
            }                                                          \
            else                                                       \
            {                                                          \
                /* Set a logging callback to use LOG* macro! */        \
                std::terminate();                                      \
            }                                                          \
        }                                                              \
    } while (0)

#define LOG_HR(hr)                                                 \
    do                                                             \
    {                                                              \
        std::error_code _hr = (hr);                                \
        if (Azure::Device::Utils::g_pfnLoggingCallback != nullptr) \
        {                                                          \
            Azure::Device::Utils::g_pfnLoggingCallback(            \
                "LOG_HR: ", __FILE__, __LINE__, _hr);              \
        }                                                          \
        else                                                       \
        {                                                          \
            /* Set a logging callback to use LOG* macro! */        \
            std::terminate();                                      \
        }                                                          \
    } while (0)

// Use these macros in place of a catch block to handle exceptions
#define CATCH_RETURN() \
    catch (...) { return Azure::Device::Utils::ResultFromCaughtException(); }
#define CATCH_RETURN_EXPECTED() \
    catch (...) { return Azure::Device::Utils::ResultFromCaughtException(); }
#define CATCH_FAIL_FAST() \
    catch (...) { FAIL_FAST(); }
#define CATCH_LOG()                                                           \
    catch (...)                                                               \
    {                                                                         \
        if (Azure::Device::Utils::g_pfnLoggingCallback != nullptr)            \
        {                                                                     \
            char _DebugString[128] = "LOG_HR: ";                              \
            size_t _LogMessageSize = sizeof("LOG_HR: ") - 1;                  \
            std::error_code _hr =                                             \
                Azure::Device::Utils::ResultFromCaughtExceptionInternal(      \
                    &(_DebugString[_LogMessageSize]), 128 - _LogMessageSize); \
            Azure::Device::Utils::g_pfnLoggingCallback(                       \
                _DebugString, __FILE__, __LINE__, _hr);                       \
        }                                                                     \
        else                                                                  \
        {                                                                     \
            /* Set a logging callback to use LOG* macro! */                   \
            std::terminate();                                                 \
        }                                                                     \
    }

//*****************************************************************************
// Macros for returning failures as std::error_codes
//*****************************************************************************
#define RETURN_HR(hr)                                              \
    do                                                             \
    {                                                              \
        std::error_code _hr = (hr);                                \
        if (Azure::Device::Utils::g_pfnLoggingCallback != nullptr) \
        {                                                          \
            Azure::Device::Utils::g_pfnLoggingCallback(            \
                "RETURN_HR: ", __FILE__, __LINE__, _hr);           \
        }                                                          \
        return _hr;                                                \
    } while (0)

#define RETURN_IF_FAILED(hr)                                           \
    do                                                                 \
    {                                                                  \
        std::error_code _hr = (hr);                                    \
        if (FAILED(_hr))                                               \
        {                                                              \
            if (Azure::Device::Utils::g_pfnLoggingCallback != nullptr) \
            {                                                          \
                Azure::Device::Utils::g_pfnLoggingCallback(            \
                    "RETURN_HR: ", __FILE__, __LINE__, _hr);           \
            }                                                          \
            return _hr;                                                \
        }                                                              \
    } while (0)

#define RETURN_HR_IF(hr, condition)                                    \
    do                                                                 \
    {                                                                  \
        std::error_code _hr = (hr);                                    \
        if ((condition))                                               \
        {                                                              \
            if (Azure::Device::Utils::g_pfnLoggingCallback != nullptr) \
            {                                                          \
                Azure::Device::Utils::g_pfnLoggingCallback(            \
                    "RETURN_HR: ", __FILE__, __LINE__, _hr);           \
            }                                                          \
            return _hr;                                                \
        }                                                              \
    } while (0)

//*****************************************************************************
// Macros to fail fast the process on failures
//*****************************************************************************
#define FAIL_FAST_IF_FAILED(hr)                                        \
    do                                                                 \
    {                                                                  \
        std::error_code _hr = (hr);                                    \
        if (FAILED(_hr))                                               \
        {                                                              \
            if (Azure::Device::Utils::g_pfnLoggingCallback != nullptr) \
            {                                                          \
                Azure::Device::Utils::g_pfnLoggingCallback(            \
                    "FAILFAST_HR: ", __FILE__, __LINE__, _hr);         \
            }                                                          \
            std::terminate();                                          \
        }                                                              \
    } while (0)

#define FAIL_FAST_IF(condition)                                        \
    do                                                                 \
    {                                                                  \
        if ((condition))                                               \
        {                                                              \
            if (Azure::Device::Utils::g_pfnLoggingCallback != nullptr) \
            {                                                          \
                Azure::Device::Utils::g_pfnLoggingCallback(            \
                    "FAILFAST_HR: ", __FILE__, __LINE__, {});          \
            }                                                          \
            std::terminate();                                          \
        }                                                              \
    } while (0)

#define FAIL_FAST_HR(hr)                                           \
    do                                                             \
    {                                                              \
        if (Azure::Device::Utils::g_pfnLoggingCallback != nullptr) \
        {                                                          \
            Azure::Device::Utils::g_pfnLoggingCallback(            \
                "FAILFAST: ", __FILE__, __LINE__, hr);             \
        }                                                          \
        std::terminate();                                          \
    } while (0)

//*****************************************************************************
// Macros to throw exceptions on failure
//*****************************************************************************
#define THROW_HR(hr)                                               \
    do                                                             \
    {                                                              \
        std::error_code _hr = (hr);                                \
        if (Azure::Device::Utils::g_pfnLoggingCallback != nullptr) \
        {                                                          \
            Azure::Device::Utils::g_pfnLoggingCallback(            \
                "THROW_HR: ", __FILE__, __LINE__, _hr);            \
        }                                                          \
        throw std::system_error(_hr);                              \
    } while (0)

#define THROW_IF_FAILED(hr)                                            \
    do                                                                 \
    {                                                                  \
        std::error_code _hr = (hr);                                    \
        if (FAILED(_hr))                                               \
        {                                                              \
            if (Azure::Device::Utils::g_pfnLoggingCallback != nullptr) \
            {                                                          \
                Azure::Device::Utils::g_pfnLoggingCallback(            \
                    "THROW_HR: ", __FILE__, __LINE__, _hr);            \
            }                                                          \
            throw std::system_error(_hr);                              \
        }                                                              \
    } while (0)

#define THROW_HR_IF(hr, condition)                                     \
    do                                                                 \
    {                                                                  \
        std::error_code _hr = (hr);                                    \
        if ((condition))                                               \
        {                                                              \
            if (Azure::Device::Utils::g_pfnLoggingCallback != nullptr) \
            {                                                          \
                Azure::Device::Utils::g_pfnLoggingCallback(            \
                    "THROW_HR: ", __FILE__, __LINE__, _hr);            \
            }                                                          \
            throw std::system_error(_hr);                              \
        }                                                              \
    } while (0)
