// Copyright (c) Microsoft Corporation. All rights reserved.

#pragma once

// STL
#include <algorithm>
#include <iterator>
#include <limits>
#include <set>
#include <sstream>
#include <string>
#include <string_view>
#include <vector>

#include <azure-device-util/ResultException.h>
#include <nonstd/span.hpp>
#include <uuid/uuid.h>

#include "UuidBridge.h"

namespace Azure { namespace Device { namespace Utils {

class String
{
  public:
    // String conversion functions
    template<class T>
    static T ToUIntT(std::string_view str)
    {
        return static_cast<T>(strtoull(std::string{ str }.c_str(), nullptr, 0));
    }
    template<class T>
    static T ToIntT(std::string_view str)
    {
        return static_cast<T>(strtoll(std::string{ str }.c_str(), nullptr, 0));
    }

    template<class T>
    static T ToFloat(std::string_view str)
    {
        return static_cast<T>(atof(std::string{ str }.c_str()));
    }

    // This returns an error when it fails to parse exactly correctly.
    template<class T>
    static std::error_code ToUIntTStrict(
        std::string_view str,
        T& value,
        uint8_t radix = 0,
        bool allowNegative = false)
    {
        try
        {
            const std::string strNullTerminated{ str };
            if (!allowNegative)
            {
                const char* itStr = strNullTerminated.c_str();
                for (; isspace(*itStr); itStr++)
                    ;
                RETURN_HR_IF(
                    make_error_code(std::errc::invalid_argument), *itStr == '-');
            }

            errno = 0;
            char* strEnd;
            uint64_t v = strtoull(strNullTerminated.c_str(), &strEnd, radix);

            if (*strEnd != 0)
            {
                RETURN_HR(make_error_code(std::errc::invalid_argument));
            }

            switch (errno)
            {
            case ERANGE:
                RETURN_HR(make_error_code(std::errc::value_too_large));
            }

            errno = 0;

            if (v > std::numeric_limits<T>::max())
            {
                RETURN_HR(make_error_code(std::errc::value_too_large));
            }

            value = static_cast<T>(v);
            return {};
        }
        CATCH_RETURN();
    }

    static bool ToBool(std::string_view str);
    static std::string FromBool(bool b);

    static int32_t ICompare(std::string_view str1, std::string_view str2);
    static bool IContains(std::string_view str1, std::string_view str2);

    static bool IStartsWith(std::string_view str, std::string_view prefix);
    static bool IEndsWith(std::string_view str, std::string_view suffix);

    static std::string
    StringFromGuid(const uuid_t& uuid, bool includeBraces = true);

    static std::vector<std::string> Split(
        std::string_view str,
        std::string_view delimiter,
        bool return_empty_tokens = true);

    static std::string
    Join(nonstd::span<const std::string> strings, std::string_view joinSequence);

    static void
    Trim(std::string& trimString, const char* additionalChars = nullptr);

    struct CaseInsensitiveEqualToPredicate
    {
        template<typename T1, typename T2>
        bool operator()(const T1& x, const T2& y) const
        {
            return ICompare(x, y) == 0;
        }
    };

    struct CaseInsensitiveLessThanPredicate
    {
        // This allows .find(const T& t) on types that are comparable but not
        // the same as a set/map's KeyType
        typedef int is_transparent;

        template<typename T1, typename T2>
        bool operator()(const T1& x, const T2& y) const
        {
            return ICompare(x, y) < 0;
        }
    };

    static std::string GenerateGuidString(bool normalizeGuidString = true);

    static void ToUppercase(std::string& str);

    static size_t IFindSequence(
        std::string_view str,
        std::string_view sequence,
        size_t position = 0);

    static bool ReplaceAll(
        std::string& original,
        std::string_view toReplace,
        std::string_view replacement,
        bool caseSensitiveMatch = true);
};

}}}
