// Copyright (c) Microsoft Corporation. All rights reserved.

#pragma once

#include <chrono>
#include <condition_variable>
#include <mutex>

namespace Azure { namespace Device { namespace Utils {

enum class SyncEventType
{
    AutoReset,
    ManualReset,
};

template<SyncEventType EnumVal>
class SyncEvent
{
  public:
    explicit SyncEvent(bool initiallySignaled = false)
        : _signaled(initiallySignaled)
    {}

    void set()
    {
        std::lock_guard<std::mutex> lock{ _mutex };
        _signaled = true;

        // For auto-reset, try to signal just one waiter.  They will set
        // _signaled back to false. For manual reset, notify all waiters
        if (EnumVal == SyncEventType::AutoReset)
        {
            _cv.notify_one();
        }
        else
        {
            _cv.notify_all();
        }
    }

    template<
        SyncEventType Q = EnumVal,
        typename std::enable_if_t<Q == SyncEventType::ManualReset, bool> = true>
    void reset()
    {
        std::lock_guard<std::mutex> lock{ _mutex };
        _signaled = false;
    }

    void wait()
    {
        std::unique_lock<std::mutex> lock{ _mutex };
        _cv.wait(lock, [this]() { return _signaled; });

        // For an auto-reset event, we need to set signaled back to false while
        // we hold the lock.
        if (EnumVal == SyncEventType::AutoReset)
        {
            _signaled = false;
        }
    }

    // Returns true if signaled, false if timed out
    template<class Rep, class Period>
    bool wait_for(const std::chrono::duration<Rep, Period>& duration)
    {
        std::unique_lock<std::mutex> lock{ _mutex };

        auto isSignaledCheck = [this]() { return _signaled; };

        bool finishedSignaled = _cv.wait_for(lock, duration, isSignaledCheck);

        // For an auto-reset event, we need to set signaled back to false while
        // we hold the lock.
        if (EnumVal == SyncEventType::AutoReset)
        {
            _signaled = false;
        }

        return finishedSignaled;
    }

    // Returns true if signaled, false if timed out
    template<class Rep, class Period>
    bool wait_until(const std::chrono::time_point<Rep, Period>& timePoint)
    {
        std::unique_lock<std::mutex> lock{ _mutex };

        auto isSignaledCheck = [this]() { return _signaled; };

        bool finishedSignaled = _cv.wait_until(lock, timePoint, isSignaledCheck);

        // For an auto-reset event, we need to set signaled back to false while
        // we hold the lock.
        if (EnumVal == SyncEventType::AutoReset)
        {
            _signaled = false;
        }

        return finishedSignaled;
    }

  private:
    std::mutex _mutex;
    std::condition_variable _cv;
    bool _signaled;
};

using AutoResetEvent = SyncEvent<SyncEventType::AutoReset>;
using ManualResetEvent = SyncEvent<SyncEventType::ManualReset>;

}}}
