// Copyright (c) Microsoft Corporation. All rights reserved.

#include <vector>
#include <zlib.h>

#include <azure-device-util/DeflateStream.h>
#include <azure-device-util/ResultException.h>

namespace Azure { namespace Device { namespace Utils {

/*++
    Constructor
--*/
DeflateStream::DeflateStream() : m_zStream(), m_finished(false)
{
    //
    // By using deflateInit2 we can set windowBits to negative
    // which will then generate raw deflate data with no zlib header or
    // trailer, and will not compute an adler32 check value
    //
    std::error_code err = make_zlib_error_code(deflateInit2(
        &m_zStream,
        Z_DEFAULT_COMPRESSION,  // level
        Z_DEFLATED,             // method -- default
        -k_windowBits,          // windowBits -- negative default
        k_memLevel,             // memLevel -- default
        Z_DEFAULT_STRATEGY));   // strategy -- default

    THROW_HR_IF(err, err != make_zlib_error_code(Z_OK));
}

/*++
    Destructor
--*/
DeflateStream::~DeflateStream()
{
    deflateEnd(&m_zStream);
}

/*++
    Adds data to the stream.

Arguments:
    data - The data to compress.
--*/
void DeflateStream::AddData(nonstd::span<const std::byte> data)
{
    bool finishing = (data.size() == 0);

    m_zStream.next_in =
        reinterpret_cast<Byte*>(const_cast<std::byte*>(data.data()));
    m_zStream.avail_in = static_cast<uint32_t>(data.size());

    // Keep looping until we run out of input, on in the finishing case,
    // when we hit Z_STREAM_END
    std::error_code zerr = make_zlib_error_code(Z_OK);
    while (zerr == make_zlib_error_code(Z_OK) &&
           (m_zStream.avail_in > 0 || finishing))
    {
        std::byte outputChunk[4096];
        m_zStream.next_out = reinterpret_cast<Bytef*>(outputChunk);
        m_zStream.avail_out = sizeof(outputChunk);

        zerr = make_zlib_error_code(
            deflate(&m_zStream, finishing ? Z_FINISH : Z_NO_FLUSH));

        uint32_t bytesToWriteToOutput = sizeof(outputChunk) - m_zStream.avail_out;
        m_outputBytes.insert(
            m_outputBytes.end(), outputChunk, outputChunk + bytesToWriteToOutput);
    }

    int32_t expectedZerr = finishing ? Z_STREAM_END : Z_OK;

    THROW_HR_IF(zerr, zerr != make_zlib_error_code(expectedZerr));
}

/*++
    Finishes the stream and returns the compressed output

Return Value:
    Vector of compressed data
--*/
std::vector<std::byte> DeflateStream::Finish()
{
    // Flush out the last bits of buffered data to m_outputBytes
    AddData({});

    m_finished = true;

    return std::move(m_outputBytes);
}

/*++
    Returns the size of current output bytes + the size of the window in the
    compressor. This is effectively the maximum number of output bytes that
    would be produced if the stream were Finish'd right now.
--*/
size_t DeflateStream::GetMaxCurrentOutputSize() const
{
    //
    // Current size of compressed output + maximum size of window in zlib
    // based on our current setting for windowBits
    //
    // zlib window size is based on the following formula from their design
    // docs (ONLY WINDOW SIZE MATTERS FOR THIS FUNCTION): ' The memory
    // requirements for compression depend on two parameters, windowBits and
    // memLevel:
    //
    //       deflate memory usage(bytes) = (1 << (windowBits + 2)) + (1 <<
    //       (memLevel + 9)) '
    //
    return m_outputBytes.size() + (1 << (k_windowBits + 2));
}

/*++
    Returns the memory size in bytes of having this stream open.
--*/
size_t DeflateStream::GetCompressorMemoryCost()
{
    return (1 << (k_windowBits + 2)) + (1 << (k_memLevel + 9));
}

}}}