// Copyright (c) Microsoft Corporation. All rights reserved.

#include <iostream>
#include <optional>

#include <azure-device-util/Docker.h>
#include <azure-device-util/Process.h>
#include <azure-device-util/String.h>

namespace Azure { namespace Device { namespace Utils {

std::vector<Docker::Container> Docker::GetAllContainers()
{
    auto containerIds = Docker::GetAllContainerIDs();
    return Docker::GetContainers(containerIds);
}

std::vector<Docker::Container>
Docker::GetContainers(const std::vector<std::string>& containerIds)
{
    if (containerIds.empty())
    {
        return {};
    }

    std::vector<std::string> arguments = {
        "inspect", "--size", "-f", c_DockerInspectFormat
    };

    arguments.insert(arguments.end(), containerIds.begin(), containerIds.end());
    auto output = Process::RunCommand("docker", arguments);
    if (output.ExitCode)
    {
        return {};
    }

    std::vector<Container> containers;
    for (auto&& line : output.Output)
    {
        auto tokens = String::Split(line, c_ContainerFormatDelimiter);
        if (tokens.size() != static_cast<int>(ContainerAttributes::NUM_VALUES))
        {
            continue;
        }

        Container temp;

        // Extract docker image in format "<name>:<tag>", split it and populate
        // the corresponding fields.
        // ":<tag>" is not guaranteed to be present !
        std::string fullImageName =
            std::move(tokens[static_cast<int>(ContainerAttributes::Image)]);
        auto tagPos = fullImageName.rfind(':');
        if (tagPos != std::string::npos)
        {
            temp.ImageName = fullImageName.substr(0, tagPos);
            temp.ImageTag = fullImageName.substr(tagPos + 1, std::string::npos);
        }
        else
        {
            temp.ImageName = std::move(fullImageName);
            temp.ImageTag.clear();
        }

        temp.ID = std::move(tokens[static_cast<int>(ContainerAttributes::ID)]);
        temp.Name =
            std::move(tokens[static_cast<int>(ContainerAttributes::Name)]);
        temp.ImageId =
            std::move(tokens[static_cast<int>(ContainerAttributes::ImageID)]);
        temp.CreatedTime =
            std::move(tokens[static_cast<int>(ContainerAttributes::CreatedTime)]);
        temp.Status =
            std::move(tokens[static_cast<int>(ContainerAttributes::Status)]);
        temp.StartedTime =
            std::move(tokens[static_cast<int>(ContainerAttributes::StartedTime)]);
        temp.FinishedTime = std::move(
            tokens[static_cast<int>(ContainerAttributes::FinishedTime)]);
        temp.SizeRw = Utils::String::ToUIntT<std::uint64_t>(
            tokens[static_cast<int>(ContainerAttributes::SizeRw)]);
        temp.SizeRootFs = Utils::String::ToUIntT<std::uint64_t>(
            tokens[static_cast<int>(ContainerAttributes::SizeRootFs)]);
        containers.push_back(std::move(temp));
    }

    return containers;
}

bool Docker::AddContainerStats(std::vector<Container>& containers)
{
    if (containers.empty())
    {
        return true;
    }

    std::vector<std::string> arguments = {
        "stats", "--no-stream", "--format", c_DockerStatsFormat
    };

    for (auto&& container : containers)
    {
        arguments.push_back(container.ID);
    }

    auto output = Process::RunCommand("docker", arguments);
    if (output.ExitCode)
    {
        return false;
    }

    if (output.Output.size() != containers.size())
    {
        return false;
    }

    for (size_t i = 0; i < containers.size(); i++)
    {
        auto tokens = String::Split(output.Output[i], c_ContainerFormatDelimiter);
        if (tokens.size() != static_cast<int>(ContainerStatsFields::NUM_VALUES))
        {
            continue;
        }

        containers[i].MemUsage =
            std::move(tokens[static_cast<int>(ContainerStatsFields::MemUsage)]);
        containers[i].CpuUsage = std::move(
            tokens[static_cast<int>(ContainerStatsFields::CpuPercentage)]);
    }

    return true;
}

std::optional<Docker::Container>
Docker::GetContainer(std::string_view containerId)
{
    auto results = Docker::GetContainers({ std::string{ containerId } });
    if (results.size() == 1)
    {
        return results[0];
    }

    return {};
}

std::vector<std::string> Docker::GetAllContainerIDs()
{
    auto output =
        Process::RunCommand("docker", { "ps", "-a", "--format", "{{.ID}}" });

    if (output.ExitCode)
    {
        return {};
    }

    return output.Output;
}
}}}