// Copyright (c) Microsoft Corporation. All rights reserved.

#include <filesystem>
#include <fstream>
#include <map>
#include <regex>
#include <sstream>

#include <azure-device-util/ErrorHelper.h>
#include <azure-device-util/FileSystem.h>
#include <azure-device-util/String.h>

#include "Config.h"

namespace Azure { namespace Device { namespace Utils {

/*++
   Deletes all files matching a specific pattern (as defined by
   by the given regex pattern) in a specific directory.

Arguments:
   directory - Full path to the directory we will be working in.
   pattern - The regex pattern to search for.
--*/
std::error_code FileSystem::DeleteAllFilesWithPattern(
    std::string_view directory,
    std::string_view pattern)
{
    namespace fs = std::filesystem;
    RETURN_IF_FAILED(DoForEachFileWithPattern(
        directory, pattern, [](const std::filesystem::path& fullPath) {
            std::filesystem::remove(fullPath);
        }));

    return {};
}

/*++
    Creates a directory including any needed parents

Arguments:
    directory - The full path to the directory to create
--*/
std::error_code FileSystem::CreateDirectoryAndParents(std::string_view directory)
{
    std::error_code result;
    std::filesystem::create_directories(std::string{ directory }, result);

    if (result)
    {
        std::error_code stdError{ result.value(), std::generic_category() };
        return stdError;
    }

    return {};
}

/*++
    Executes the actionOnFile lambda for each file found with a given regex
    pattern.

Arguments:
    directory - The directory to search
    pattern - The regex to apply to any found file
    actionOnFile - The labmda to execute on each found file
--*/
std::error_code FileSystem::DoForEachFileWithPattern(
    std::string_view directory,
    std::string_view pattern,
    const std::function<void(const std::filesystem::path&)>& actionOnFile)
{
    std::string patternString{ pattern };

    try
    {
        namespace fs = std::filesystem;
        for (const std::filesystem::path& path :
             std::filesystem::directory_iterator{ std::string{ directory } })
        {
            if (path.has_filename())
            {
                std::string fileName = path.filename().native();

                std::regex patternRegex{ patternString };
                if (std::regex_match(fileName, patternRegex))
                {
                    try
                    {
                        actionOnFile(path);
                    }
                    CATCH_LOG();
                }
            }
        }

        return {};
    }
    CATCH_RETURN_EXPECTED();
}

/*++
    Writes a string to a file in UTF-8 encoding.

Arguments:
    fileName - The name of the file to write to
    data - The string contents to be written to the file as UTF8
--*/
std::error_code
FileSystem::WriteUTF8StringToFile(std::string_view fileName, std::string_view data)
{
    try
    {
        std::ofstream fileOutStream{ std::string{ fileName } };
        fileOutStream.exceptions(std::ofstream::failbit | std::ofstream::badbit);

        fileOutStream.write(data.data(), static_cast<ptrdiff_t>(data.size()));

        // Manually close so any exception gets thrown here instead of being
        // thrown in the destructor.
        fileOutStream.close();

        return {};
    }
    CATCH_RETURN();
}

/*++
    Reads a string from a UTF-8 encoded file.

Arguments:
    fileName - The name of the file to read.
    data - The string contents from the file.
--*/
std::error_code
FileSystem::ReadUTF8StringFromFile(std::string_view fileName, std::string& data)
{
    try
    {
        std::ifstream fileInStream{ std::string{ fileName } };

        if (!fileInStream.is_open())
        {
            return make_error_code(std::errc::no_such_file_or_directory);
        }

        std::stringstream dataStream;
        dataStream << fileInStream.rdbuf();
        data = dataStream.str();

        return {};
    }
    CATCH_RETURN();
}

/*++
    Reads a file size.

Arguments:
    fileName - The full path of the file we are getting the size to.
--*/
std::error_code FileSystem::ReadFileSize(std::string_view fileName, int64_t& size)
{
    try
    {
        size = static_cast<int64_t>(
            std::filesystem::file_size(std::string{ fileName }));

        return {};
    }
    CATCH_RETURN();
}

/*++
    Reads a dictionary of (key,value) pairs from a file, where each pair is
separated by a newline, and the key and value are separated by a delimiter
string.

Arguments:
    fileName - The name of the file to read.
    delimiter - The string that separates each key and value.
    data - Receives the (key,value) pairs from the file.
--*/
std::error_code FileSystem::ReadDictionaryFromFile(
    std::string_view fileName,
    std::string_view delimiter,
    std::map<std::string, std::string, std::less<>>& data)
{
    try
    {
        if (delimiter.empty())
        {
            RETURN_HR(make_error_code(std::errc::invalid_argument));
        }

        std::ifstream fileInStream{ std::string{ fileName } };

        if (!fileInStream.is_open())
        {
            return make_error_code(std::errc::no_such_file_or_directory);
        }

        std::string line;
        while (std::getline(fileInStream, line))
        {
            size_t delimiterPos =
                line.find(delimiter.data(), 0, delimiter.size());

            if (delimiterPos == std::string::npos || delimiterPos == 0 ||
                delimiterPos == (line.size() - 1))
            {
                break;
            }

            std::string key = line.substr(0, delimiterPos);
            std::string value = line.substr(delimiterPos + 1);

            data[key] = value;
        }

        return {};
    }
    CATCH_RETURN();
}

bool FileSystem::FileExists(std::string_view fileName)
{
    try
    {
        std::filesystem::path path(fileName.begin(), fileName.end());
        std::error_code errorCode;
        bool exists = std::filesystem::exists(path, errorCode);
        if (FAILED(errorCode))
        {
            return false;
        }

        return exists;
    }
    CATCH_LOG();

    return false;
}

std::filesystem::path FileSystem::GetConfDirectory()
{
    return AZURE_DEVICE_UTILS_SYSCONFDIR;
}

std::filesystem::path FileSystem::GetBinaryDirectory()
{
    return AZURE_DEVICE_UTILS_BINDIR;
}

std::filesystem::path FileSystem::GetLibraryDirectory()
{
#ifdef AZURE_DEVICE_UTILS_USE_LIB64_VARIABLE
    return AZURE_DEVICE_UTILS_LIB64DIR;
#else
    return AZURE_DEVICE_UTILS_LIBDIR;
#endif
}

std::filesystem::path FileSystem::GetLocalStateDirectory()
{
    return AZURE_DEVICE_UTILS_LOCALSTATEDIR;
}

}}}