// Copyright (c) Microsoft Corporation. All rights reserved.

#include <filesystem>
#include <iostream>
#include <optional>

#include <azure-device-util/PackageManager.h>
#include <azure-device-util/Process.h>
#include <azure-device-util/String.h>

namespace Azure { namespace Device { namespace Utils {

bool PackageManager::PackageManagerSupported()
{
    switch (GetSystemPackageManager())
    {
    case PackageManagerFamily::UNKNOWN:
    case PackageManagerFamily::OTHER:
        return false;
    default:
        return true;
    }
}

PackageManager::PackageManagerFamily PackageManager::GetSystemPackageManager()
{
#if defined(PACKAGE_MANAGER_RPM)
    return PackageManagerFamily::RPM;
#elif defined(PACKAGE_MANAGER_DPKG)
    return PackageManagerFamily::DPKG;
#else
    return PackageManagerFamily::UNKNOWN;
#endif
}

std::optional<PackageManager::Package>
PackageManager::GetFileOwner(std::string_view filePath)
{
    try
    {
        std::filesystem::path bPath(filePath.begin(), filePath.end());
        if (!std::filesystem::exists(bPath))
        {
            return {};
        }

        // Construct an absolute path to avoid entry duplication
        auto absolutePath = std::filesystem::absolute(bPath);

        switch (GetSystemPackageManager())
        {
        case PackageManagerFamily::RPM:
            return GetFileOwnerRPM(absolutePath.string());
        case PackageManagerFamily::DPKG:
            return GetFileOwnerDPKG(absolutePath.string());
        default:
            return {};
        }
    }
    CATCH_LOG()

    return {};
}

std::optional<PackageManager::Package>
PackageManager::GetProcessOwnerByPid(uint32_t processId)
{
    std::filesystem::path exeFilePath = "/proc";
    exeFilePath /= std::to_string(processId);
    exeFilePath /= "exe";

    if (!std::filesystem::exists(exeFilePath))
    {
        return {};
    }

    auto processFilePath = std::filesystem::read_symlink(exeFilePath);

    return GetFileOwner(processFilePath.string());
}

std::vector<PackageManager::Package> PackageManager::GetInstalledPackages()
{
    switch (GetSystemPackageManager())
    {
    case PackageManagerFamily::RPM:
        return GetInstalledPackagesRPM();
    case PackageManagerFamily::DPKG:
        return GetInstalledPackagesDPKG();
    default:
        return {};
    }
}

std::optional<PackageManager::Package>
PackageManager::GetFileOwnerRPM(std::string_view filePath)
{
    Process::CommandResult result = Process::RunCommand(
        "rpm",
        { "--queryformat", c_RPMPackageFormat, "-qf", std::string{ filePath } });

    if (result.ExitCode != 0 || result.Output.size() != 1)
    {
        return {};
    }

    return ParsePackage(result.Output[0]);
}

std::string PackageManager::GetDPKGPackageName(std::string_view filePath)
{
    Process::CommandResult result =
        Process::RunCommand("dpkg", { "-S", std::string{ filePath } });

    if (result.ExitCode != 0 || result.Output.size() != 1)
    {
        return {};
    }

    std::string dpkgResult = std::move(result.Output[0]);

    if (dpkgResult.empty())
    {
        return {};
    }

    // dpkg returns the owner's name and the file path separated by a colon.
    // Grab only the package name.
    return dpkgResult.substr(0, dpkgResult.find_first_of(':'));
}

std::optional<PackageManager::Package>
PackageManager::GetFileOwnerDPKG(std::string_view filePath)
{
    std::string packageName = GetDPKGPackageName(filePath);
    if (packageName.empty())
    {
        return {};
    }

    Process::CommandResult result = Process::RunCommand(
        "dpkg-query", { c_DPKGPackageFormat, "--show", packageName });

    if (result.ExitCode != 0 || result.Output.size() != 1)
    {
        return {};
    }

    return ParsePackage(result.Output[0]);
}

std::vector<PackageManager::Package> PackageManager::GetInstalledPackagesRPM()
{
    Process::CommandResult result = Process::RunCommand(
        "rpm", { "--queryformat", c_RPMPackageFormat, "-qa" });

    if (result.ExitCode != 0)
    {
        return {};
    }

    return ParsePackageList(result.Output);
}

std::vector<PackageManager::Package> PackageManager::GetInstalledPackagesDPKG()
{
    Process::CommandResult result =
        Process::RunCommand("dpkg-query", { c_DPKGPackageFormat, "-W" });

    if (result.ExitCode != 0)
    {
        return {};
    }

    return ParsePackageList(result.Output);
}

std::optional<PackageManager::Package>
PackageManager::ParsePackage(std::string_view entry)
{
    auto tokens = Utils::String::Split(entry, c_PackageFormatDelimiter);
    if (tokens.size() != static_cast<int>(PackageAttributes::AttributeCount))
    {
        return {};
    }

    Package package;
    package.Name = std::move(tokens[static_cast<int>(PackageAttributes::Name)]);
    package.Version =
        std::move(tokens[static_cast<int>(PackageAttributes::Version)]);
    package.Arch =
        std::move(tokens[static_cast<int>(PackageAttributes::Architecture)]);

    if (package.Name.empty() || package.Version.empty() || package.Arch.empty())
    {
        return {};
    }

    return package;
}

std::vector<PackageManager::Package>
PackageManager::ParsePackageList(const std::vector<std::string>& list)
{
    std::vector<PackageManager::Package> packages;

    for (const auto& entry : list)
    {
        if (auto package = ParsePackage(entry))
        {
            packages.emplace_back(*package);
        }
    }

    return packages;
}

}}}