// Copyright (c) Microsoft Corporation. All rights reserved.

#include <string_view>
#include <optional>

#include <azure-device-util/ErrorHelper.h>
#include <azure-device-util/FileSystem.h>
#include <azure-device-util/Permission.h>
#include <azure-device-util/ResultException.h>
#include <azure-device-util/String.h>
#include <azure-device-util/Toml.h>

namespace Azure { namespace Device { namespace Utils {

TelemetryPermission Permission::ReadPermissionFromDefaultFile()
{
    std::filesystem::path confFilePath = Utils::FileSystem::GetConfDirectory();
    confFilePath /= "azure-device-health-services/config.toml";

    toml::value tomlValue = toml::parse(confFilePath.string());
    const auto& permissionString =
        toml::find<std::string>(tomlValue, "Permission");

    return ConvertStringToTelemetryPermission(permissionString);
}

std::error_code
Permission::WritePermissionToDefaultFile(TelemetryPermission permission)
{
    std::filesystem::path confFilePath = Utils::FileSystem::GetConfDirectory();
    confFilePath /= "azure-device-health-services/config.toml";

    toml::value data;
    auto permissionString = ConvertTelemetryPermissionToString(permission);

    try
    {
        // If file exists and is valid, read it to preserve all other setting.
        data = toml::parse(confFilePath.string());
        toml::find<toml::value>(data, "Permission") = permissionString;
    }
    catch (...)
    {
        // If that failed, create a new and clean toml value to write to the
        // file.
        data = toml::value{ { "Permission", permissionString } };
    }

    return Utils::FileSystem::WriteUTF8StringToFile(
        confFilePath.string(), toml::format(data));
}

TelemetryPermission
Permission::ConvertStringToTelemetryPermission(std::string_view permissionString)
{
    if (auto optPermission =
            TryConvertStringToTelemetryPermission(permissionString))
    {
        return *optPermission;
    }
    else
    {
        THROW_HR(UtilsError::E_INVALID_PERMISSION_VALUE);
    }
}

std::optional<TelemetryPermission>
Permission::TryConvertStringToTelemetryPermission(std::string_view permissionString)
{
    if (Utils::String::ICompare(permissionString, "None") == 0)
    {
        return TelemetryPermission::None;
    }
    else if (Utils::String::ICompare(permissionString, "Required") == 0)
    {
        return TelemetryPermission::Required;
    }
    else if (Utils::String::ICompare(permissionString, "Optional") == 0)
    {
        return TelemetryPermission::Optional;
    }
    else
    {
        return {};
    }
}

std::string_view
Permission::ConvertTelemetryPermissionToString(TelemetryPermission permission)
{
    if (permission == TelemetryPermission::None)
    {
        return "None";
    }
    else if (permission == TelemetryPermission::Required)
    {
        return "Required";
    }
    else if (permission == TelemetryPermission::Optional)
    {
        return "Optional";
    }
    else
    {
        THROW_HR(UtilsError::E_INVALID_PERMISSION_VALUE);
    }
}

}}}