// Copyright (c) Microsoft Corporation. All rights reserved.

#include <filesystem>
#include <optional>

#include <azure-device-util/Process.h>
#include <azure-device-util/String.h>
#include <boost/process.hpp>
#include <boost/process/extend.hpp>

namespace Azure { namespace Device { namespace Utils {

std::string Process::GetPathForBinary(std::string_view binaryName)
{
    std::vector<std::string> pathsToTry;
    const char* pathPtr = std::getenv("PATH");
    if (pathPtr)
    {
        pathsToTry = Utils::String::Split(pathPtr, ":");
    }
    else
    {
        pathsToTry.push_back("/usr/local/bin");
        pathsToTry.push_back("/usr/bin");
        pathsToTry.push_back("/bin");
        pathsToTry.push_back("/vendor/bin");
        pathsToTry.push_back("/system/bin");
    }

    for (std::filesystem::path pathToTry : pathsToTry)
    {
        std::filesystem::path fullPathToTry = pathToTry / binaryName;
        if (std::filesystem::exists(fullPathToTry))
        {
            return fullPathToTry;
        }
    }

    return "";
}

Process::CommandResult Process::RunCommand(
    std::string_view binaryName,
    const std::vector<std::string>& args,
    std::optional<std::filesystem::path> inputPath)
{
    CommandResult result;

    std::string fullBinaryPath = GetPathForBinary(binaryName);
    if (fullBinaryPath.empty())
    {
        result.ExitCode = static_cast<int>(std::errc::no_such_file_or_directory);
        return result;
    }

    boost::process::ipstream outStream, errStream;

    if (inputPath.has_value())
    {
        result.ExitCode = boost::process::system(
            fullBinaryPath,
            boost::process::args = args,
            boost::process::std_in<inputPath.value(), boost::process::std_out> outStream,
            boost::process::std_err > errStream);
    }
    else
    {
        result.ExitCode = boost::process::system(
            fullBinaryPath,
            boost::process::args = args,
            boost::process::std_in.null(),
            boost::process::std_out > outStream,
            boost::process::std_err > errStream);
    }

    auto readLines = [](std::vector<std::string>& lines, std::istream& stream) {
        for (std::string line; std::getline(stream, line);)
        {
            lines.push_back(std::move(line));
        }
    };

    readLines(result.Output, outStream);
    readLines(result.Error, errStream);

    return result;
}

struct reset_sigmask_in_child : boost::process::extend::handler
{
    template<class Executor>
    void on_exec_setup(Executor&) const
    {
        sigset_t emptySet;
        sigemptyset(&emptySet);
        pthread_sigmask(SIG_SETMASK, &emptySet, nullptr);
    }
};

struct set_sid_in_child : boost::process::extend::handler
{
    template<class Executor>
    void on_exec_setup(Executor&) const
    {
        setsid();
    }
};

void Process::SpawnDetachedProcess(
    std::string_view binaryName,
    const std::vector<std::string>& args)
{
    boost::process::spawn(
        GetPathForBinary(binaryName),
        boost::process::args = args,
        boost::process::std_in.null(),
        boost::process::std_out.null(),
        boost::process::std_err.null(),
        set_sid_in_child{},
        reset_sigmask_in_child{});
}

}}}
