// Copyright (c) Microsoft Corporation. All rights reserved.

#include <azure-device-util/ResultException.h>

namespace Azure { namespace Device { namespace Utils {

// Per module (SO) callback for logging errors.
log_callback_t g_pfnLoggingCallback = nullptr;

/*++
    Returns a std::error_code based on a given caught exception.

Return Value:
    std::error_code based on caught exception.
--*/
std::error_code ResultFromCaughtException()
{
    return ResultFromCaughtExceptionInternal(nullptr, 0);
}

/*++
    Helper for ResultFromCaughtException. Attempts to convert
    the exception into a std::error_code or give back a generic
    exception error. If a non-std::exception is caught, will
    terminate.

Arguments:
    DebugString - If given and the caught exception is a std::exception,
        will be populated with the exception's .what().
    DebugStringSize - If DebugString is non-null, this must contain the
        buffer size including null terminator.

Return Value:
    std::error_code based on caught exception.
--*/
std::error_code
ResultFromCaughtExceptionInternal(char* DebugString, size_t DebugStringSize)
{
    try
    {
        throw;
    }
    catch (std::system_error& ex)
    {
        return ex.code();
    }
    catch (std::exception& ex)
    {
        if (DebugString != nullptr)
        {
            size_t WhatSize = strlen(ex.what());
            strncpy(
                DebugString, ex.what(), std::min(DebugStringSize - 1, WhatSize));
            DebugString[DebugStringSize - 1] = '\0';
        }

        return UtilsError::E_STD_EXCEPTION;
    }
    catch (...)
    {
        FAIL_FAST();
    }
}

std::error_code make_errno_error_code()
{
    return std::error_code{ errno, std::generic_category() };
}

}}}