// Copyright (c) Microsoft Corporation. All rights reserved.

#include <azure-device-util/ResultException.h>
#include <azure-device-util/SignalLoop.h>

namespace Azure { namespace Device { namespace Utils {

SignalLoop::SignalLoop(std::initializer_list<int> signals)
{
    THROW_HR_IF(make_error_code(std::errc::invalid_argument), signals.size() == 0);

    sigemptyset(&_signals);

    for (int signal : signals)
    {
        auto [itr, inserted] = _signalToHandler.emplace(signal, SignalHandler{});
        THROW_HR_IF(make_error_code(std::errc::invalid_argument), !inserted);

        sigaddset(&_signals, signal);
    }

    pthread_sigmask(SIG_BLOCK, &_signals, nullptr);
}

void SignalLoop::RegisterSignalHandler(int signal, SignalHandler&& handler)
{
    THROW_HR_IF(make_error_code(std::errc::invalid_argument), !handler);

    auto itr = _signalToHandler.find(signal);
    THROW_HR_IF(
        make_error_code(std::errc::invalid_argument),
        itr == _signalToHandler.end());

    itr->second = std::move(handler);
}

void SignalLoop::Run()
{
    // Ensure each constructed signal has a real handler
    for (const auto& [signal, handler] : _signalToHandler)
    {
        THROW_HR_IF(make_error_code(std::errc::invalid_argument), !handler);
    }

    // Loop until a handler says we should stop
    while (true)
    {
        siginfo_t sigInfo;
        int signalValue = sigwaitinfo(&_signals, &sigInfo);
        if (signalValue < 0)
        {
            if (errno == EINTR)
            {
                continue;
            }
            THROW_HR(make_errno_error_code());
        }

        auto itr = _signalToHandler.find(signalValue);
        FAIL_FAST_IF(itr == _signalToHandler.end());

        const auto& handler = itr->second;
        if (!handler(signalValue))
        {
            break;
        }
    }
}

}}}
