// Copyright (c) Microsoft Corporation. All rights reserved.

#include <azure-device-util/String.h>

namespace Azure { namespace Device { namespace Utils {

/*++
    Converts a string to a boolean.

Return Value:
    Strings that start with 1, T, or t return true. All others return false.
--*/
bool String::ToBool(std::string_view str)
{
    return !str.empty() && (str[0] == L'1' || str[0] == L'T' || str[0] == L't');
}

/*++
    Converts a boolean to a "true" or "false" string.

Return Value:
    "true" or "false"
--*/
std::string String::FromBool(bool b)
{
    return b ? "true" : "false";
}

/*++
    Compares two strings in a case-insensitive fashion.

Return Value:
    Mirrors strcmp.
--*/
int32_t String::ICompare(std::string_view str1, std::string_view str2)
{
    int32_t result = strncasecmp(
        str1.data(), str2.data(), std::min(str1.size(), str2.size()));

    if (result == 0)
    {
        // This will intentionally overflow to a negative int32_t if str2 >
        // str1.
        return static_cast<int32_t>(str1.size() - str2.size());
    }
    else
    {
        return result;
    }
}

/*++
    Checks if str1 contains str2.

Return Value:
    true if str2 is a substring of str1, false otherwise.
--*/
bool String::IContains(std::string_view str1, std::string_view str2)
{
    if (str1.size() < str2.size())
    {
        return false;
    }

    return str1.find(str2) != std::string_view::npos;
}

/*++
    Checks if str starts with prefix.

Return Value:
    true if prefix is str's prefix, false otherwise.
--*/
bool String::IStartsWith(std::string_view str, std::string_view prefix)
{
    if (str.size() < prefix.size())
    {
        return false;
    }

    str.remove_suffix(str.size() - prefix.size());
    return ICompare(str, prefix) == 0;
}

/*++
    Checks if str1 ends with suffix.

Return Value:
    true if suffix is str's suffix, false otherwise.
--*/
bool String::IEndsWith(std::string_view str, std::string_view suffix)
{
    if (str.size() < suffix.size())
    {
        return false;
    }

    str.remove_prefix(str.size() - suffix.size());
    return ICompare(str, suffix) == 0;
}

/*++
    Write the GUID to a string

Arguments:
    guid - GUID to convert

Return Value:
    String representation of GUID
--*/
std::string String::StringFromGuid(const uuid_t& uuid, bool includeBraces)
{
    std::string outputString(includeBraces ? 39 : 37, '\0');

    if (includeBraces)
    {
        outputString[0] = '{';
        uuid_unparse(uuid, &outputString[1]);
        ToUppercase(outputString);
        outputString[37] = '}';
        outputString[38] = 0;
    }
    else
    {
        uuid_unparse(uuid, &outputString[0]);
        ToUppercase(outputString);
    }

    // Remove the null terminator
    outputString.pop_back();

    return outputString;
}

/*++
    Splits a string into a vector of strings by a delimiter.

Arguments:
    str - The string to split.  May be an empty string, which will
        result in a 1-element result with an empty string.
    delimiter - The string of characters used to split str into tokens. This
        string will never appear in the set of result strings.
    return_empty_tokens - If true, result will include tokens that are empty.

Return Value:
    Set of substrings of str separated by the delimiter.
    If str ends with the delimiter, there will be an extra empty string
    element in the return vector.

    Example:
        't,foo,apple,' will result in an array of:{"t", "foo", "apple",
""}
--*/
std::vector<std::string>
String::Split(std::string_view str, std::string_view delimiter, bool return_empty_tokens)
{
    std::vector<std::string> tokens;

    THROW_HR_IF(make_error_code(std::errc::invalid_argument), delimiter.empty());

    while (true)
    {
        size_t delimiterBeginning = str.find(delimiter);
        if (delimiterBeginning == std::string_view::npos)
        {
            // Append the string of characters from the beginning of the
            // current token to the null-terminating char
            tokens.emplace_back(str);
            break;
        }

        // Append the string of characters from the beginning of the token
        // to the beginning of the delimiter. Ensure to respect
        // return_empty_tokens
        if (return_empty_tokens || delimiterBeginning != 0)
        {
            tokens.emplace_back(str.data(), delimiterBeginning);
        }

        // Move the view forward past the end of the delimiter and then loop
        // again
        str.remove_prefix(delimiterBeginning + delimiter.size());
    }

    return tokens;
}

/*++
    Appends all the string values in 'strings' together, with the 'joinSequence'
    layered in between.

Arguments:
    strings         - String list to join together
    joinSequence    - Character sequence to insert in between values in strings

Return Value:
    string[0] + joinSeq + string[1] + joinSeq ... + string[n-1]
--*/
std::string
String::Join(nonstd::span<const std::string> strings, std::string_view joinSequence)
{
    std::string result;
    for (const auto& str : strings)
    {
        result += str;
        result += joinSequence;
    }

    // Chop off the trailing joinSequence
    size_t sizeWithoutTrailingJoinSeq = result.size() - joinSequence.size();
    result.resize(sizeWithoutTrailingJoinSeq);

    return result;
}

/*++
    Trim the whitespace off of an STL string. Unfortunately, the standard
    does not include a function (or <algorithm>) for this. This will trim both
    the left and the right side of the string (e.g. "   abc   " becomes "abc",
    and "  a b c  " becomes "a b c".

Arguments:
    trimString - An inout parameter that starts as the string to be trimmed
    and ends as a trimmed string.
--*/
void String::Trim(std::string& trimString, const char* additionalChars)
{
    const char* k_whitespaceDelimiters = " \n\r\t";

    const char* charsToTrim = k_whitespaceDelimiters;
    std::string storage;
    if (additionalChars != nullptr)
    {
        storage += k_whitespaceDelimiters;
        storage += additionalChars;
        charsToTrim = storage.c_str();
    }

    size_t start = trimString.find_first_not_of(charsToTrim);
    if (start != std::string::npos)
    {
        trimString = trimString.substr(start);
    }
    else
    {
        trimString = "";
    }

    size_t end = trimString.find_last_not_of(charsToTrim);
    if (end != std::string::npos)
    {
        trimString = trimString.substr(0, end + 1);
    }
}

/*++
    Generates a new GUID and Return Value: it in string form

Arguments:
    normalizeGuidString - true if the output should be normalized to
all-caps, braceless string form (i.e. 00000000-AB12-0000-0000-000000000000)
                            false if the output should be the unmodified
return value of StringFromGUID2 (i.e.
{00000000-AB12-0000-0000-000000000000}) Defaults to true if not specified.

Return Value:
    std::string representation of GUID
--*/
std::string String::GenerateGuidString(bool normalizeGuidString)
{
    constexpr uint32_t c_uuidStrLen = 36;
    uuid_t uuid;
    uuid_generate_time_safe(uuid);
    std::string g;
    g.resize(c_uuidStrLen);
    uuid_unparse_upper(uuid, &g[0]);

    std::string finalGuidString;
    if (!normalizeGuidString)
    {
        finalGuidString += "{";
        finalGuidString += g;
        finalGuidString += "}";
    }
    else
    {
        finalGuidString = g;
    }

    return finalGuidString;
}

/*++
    Converts a string to uppercase.

Arguments:
    str - The string to convert to uppercase.  Converted in-place.
--*/
void String::ToUppercase(std::string& str)
{
    std::transform(str.cbegin(), str.cend(), str.begin(), toupper);
}

/*++
    Performs a case-insensitive string find operation. Includes a templated
implmentation and specializations for string_view and wstring_view.

Arguments:
    str - The subject string to examine
    sequence - The sequence to find in the subject string
    position - The position in the subject string to begin the search

Return Value:
    The position of the first character in the first match, or npos if the
sequence was not found
--*/
size_t
String::IFindSequence(std::string_view str, std::string_view sequence, size_t position)
{
    if (position >= str.size())
    {
        return str.npos;
    }

    auto searchResult = std::search(
        str.begin() + position,
        str.end(),
        sequence.begin(),
        sequence.end(),
        [](char a, char b) { return ::tolower(a) == ::tolower(b); });

    return (searchResult == str.end()) ? str.npos : (searchResult - str.begin());
}

/*++
    Replaces all instances of a substring with another string.

Arguments:
    original - The string to parse and replace within.
    toReplace - The substring we want to replace.
    replacement - The replacement to insert for every toReplace instance.
    caseSensitiveMatch - Specifies whether or not the case of characters in
toReplace must be matched.

Return Value:
    Returns false if the original string is empty, true otherwise
--*/
bool String::ReplaceAll(
    std::string& original,
    std::string_view toReplace,
    std::string_view replacement,
    bool caseSensitiveMatch)
{
    if (toReplace.empty())
    {
        return false;
    }

    size_t currentPosition = 0;
    size_t newPosition = caseSensitiveMatch ?
        original.find(toReplace, currentPosition) :
        IFindSequence(original, toReplace, currentPosition);

    while (newPosition != original.npos)
    {
        original.replace(newPosition, toReplace.size(), replacement);

        // Update positions
        currentPosition = newPosition + replacement.size();
        newPosition = caseSensitiveMatch ?
            original.find(toReplace, currentPosition) :
            IFindSequence(original, toReplace, currentPosition);
    }
    return true;
}

}}}
