// Copyright (c) Microsoft Corporation. All rights reserved.

#include <azure-device-util/Time.h>
#include <sys/sysinfo.h>

namespace Azure { namespace Device { namespace Utils {

/*++
    Converts a filetime string to a millisecond-precise ISO 8601 time
    string.

    This version of the function is precise to the millisecond instead of
    the hundred nanosecond, which is nice if you want to save some bytes because
    you don't care about the precision.

Arguments:
    time - system_clock::time_point to translate to string

Return Value:
    std::string - A millisecond-precise ISO 8601 representation of time.
--*/
std::string
Time::SystemTimeTo8601MillisecondString(std::chrono::system_clock::time_point time)
{
    // This will get us a nubmer [0, 999] that is the remainder millis on the
    // timestamp.
    std::chrono::nanoseconds nanosSinceEpoch = time.time_since_epoch();
    std::chrono::milliseconds remainderMillis =
        std::chrono::duration_cast<std::chrono::milliseconds>(
            nanosSinceEpoch % std::chrono::seconds{ 1 });

    // Convert the filetime to a tm, to break up into individual components
    // (year, month, etc.)
    std::time_t printableTime = std::chrono::system_clock::to_time_t(time);
    tm timeStruct;
    gmtime_r(&printableTime, &timeStruct);

    static constexpr uint32_t k_8601StringSize = 24;
    std::string isoTime(k_8601StringSize, L'\0');

    std::snprintf(
        &isoTime[0],
        k_8601StringSize + 1,
        "%04u-%02u-%02uT%02u:%02u:%02u.%03uZ",
        timeStruct.tm_year + 1900,
        timeStruct.tm_mon + 1,
        timeStruct.tm_mday,
        timeStruct.tm_hour,
        timeStruct.tm_min,
        timeStruct.tm_sec,
        static_cast<int32_t>(remainderMillis.count()));

    return isoTime;
}

/*++
    Converts from unix timestamp to system time.

Arguments:
    unixTime - unix timestamp

Return Value:
    system_clock::time_point - system time that corresponds to unixTime
--*/
std::chrono::system_clock::time_point
Time::UNIXTimeToSystemTime(uint64_t unixTime)
{
    std::chrono::milliseconds duration(unixTime);
    std::chrono::system_clock::time_point systemTime(duration);
    return systemTime;
}

/*++
    Gets system uptime from sysinfo

Return Value:
    std::chrono::seconds - seconds since the system started
--*/
std::chrono::seconds Time::GetSystemUptime()
{
    struct sysinfo sysinfoData;
    if (sysinfo(&sysinfoData) == 0)
    {
        return std::chrono::seconds{ sysinfoData.uptime };
    }
    else
    {
        return std::chrono::seconds{ 0 };
    }
}

}}}
