// Copyright (c) Microsoft Corporation. All rights reserved.

#include <azure-device-util/TomlConfig.h>

namespace Azure { namespace Device { namespace Utils {

TomlConfig::TomlConfig(std::string_view tomlFilePath)
{
    _toml = toml::parse(std::string{ tomlFilePath });
}

// overloaded is unfortunate boilerplate needed to make
// std::visit implement one lambda callback per variant-type
template<class... Ts>
struct overloaded : Ts...
{
    using Ts::operator()...;
};
template<class... Ts>
overloaded(Ts...) -> overloaded<Ts...>;

const toml::value* TomlConfig::FindPath(const ConfigPath& key)
{
    if (key.GetComponents().empty())
    {
        return nullptr;
    }

    const toml::value* descendingValue = &_toml;
    for (const auto& keyPart : key.GetComponents())
    {
        std::visit(
            overloaded{ [&](const std::string& name) {
                           if (descendingValue && descendingValue->is_table() &&
                               descendingValue->contains(name))
                           {
                               descendingValue =
                                   &toml::find(*descendingValue, name);
                           }
                           else
                           {
                               descendingValue = nullptr;
                           }
                       },
                        [&](size_t index) {
                            if (descendingValue && descendingValue->is_array() &&
                                index < descendingValue->size())
                            {
                                descendingValue = &descendingValue->at(index);
                            }
                            else
                            {
                                descendingValue = nullptr;
                            }
                        } },
            keyPart);
    }

    return descendingValue;
}

IConfig::ValidateStatus TomlConfig::ValidateHelper(
    const ConfigPath& key,
    const std::function<bool(const toml::value&)>& typeTest)
{
    const toml::value* value = FindPath(key);
    if (value == nullptr)
    {
        return ValidateStatus::PathNotFound;
    }
    else if (!typeTest(*value))
    {
        return ValidateStatus::TypeError;
    }
    else
    {
        return ValidateStatus::Ok;
    }
}

IConfig::ValidateStatus
TomlConfig::ValidateConfigPathExistsBool(const ConfigPath& key)
{
    return ValidateHelper(
        key, [](const toml::value& v) { return v.is_boolean(); });
}
IConfig::ValidateStatus
TomlConfig::ValidateConfigPathExistsInt(const ConfigPath& key)
{
    return ValidateHelper(
        key, [](const toml::value& v) { return v.is_integer(); });
}
IConfig::ValidateStatus
TomlConfig::ValidateConfigPathExistsDouble(const ConfigPath& key)
{
    return ValidateHelper(
        key, [](const toml::value& v) { return v.is_floating(); });
}
IConfig::ValidateStatus
TomlConfig::ValidateConfigPathExistsString(const ConfigPath& key)
{
    return ValidateHelper(key, [](const toml::value& v) { return v.is_string(); });
}

template<class ReturnType, class DefaultValType>
ReturnType
TomlConfig::GetHelper(const ConfigPath& key, DefaultValType defaultVal)
{
    const toml::value* value = FindPath(key);
    if (value == nullptr)
    {
        return ReturnType{ defaultVal };
    }
    else
    {
        return toml::get<ReturnType>(*value);
    }
}

bool TomlConfig::GetBool(const ConfigPath& key, bool defaultVal)
{
    return GetHelper<bool>(key, defaultVal);
}

int64_t TomlConfig::GetInt(const ConfigPath& key, int64_t defaultVal)
{
    return GetHelper<int64_t>(key, defaultVal);
}

double TomlConfig::GetDouble(const ConfigPath& key, double defaultVal)
{
    return GetHelper<double>(key, defaultVal);
}

std::string
TomlConfig::GetString(const ConfigPath& key, std::string_view defaultVal)
{
    return GetHelper<std::string>(key, defaultVal);
}

size_t TomlConfig::GetArraySize(const ConfigPath& key)
{
    return GetHelper<toml::array>(key, toml::array{}).size();
}

}}}
